/* node.c --- implementation of the phylogenetic tree data structure
   Jim Blandy <jimb@gnu.ai.mit.edu> --- July 1994 */

#include <ctype.h>
#include <string.h>

extern int atoi ();

#include "xmalloc.h"
#include "lenstring.h"
#include "node.h"
#include "hash.h"



/* Allocate and return a new node.
   NAME and NAME_LEN are the node's name.
   DESCRIPTION and DESCRIPTION_LEN are the node's description.
   The node is created with no children, and no parent.  */

struct node *
new_node (char *name, size_t name_len,
          char *description, size_t description_len)
{
  struct node *n = (struct node *) xmalloc (sizeof (*n));

  make_lenstring (&n->name, name, name_len);
  make_lenstring (&n->description, description, description_len);
  n->depth = 0;
  n->num_children = 0;
  n->children = 0;
  n->child_index = 0;
  n->parent = 0;
  n->marks = 0;
  n->prev_marks = 0;

  return n;
}


/* Make CHILD a child-node of PARENT.  */
void
add_child (struct node *parent,
           struct node *child)
{
  /* You have to add nodes root-down, or else the depth won't be
     computed correctly.  */
  if (child->num_children != 0)
    abort ();

  /* Make sure PARENT's array of children is big enough to accomodate
     the new child.

     Let P be the least power of two >= num_children; there are always
     at least P elements allocated to the children array.  Thus, if
     num_children is currently a power of two, then we know we need to
     reallocate children to be double its current size.

     Fact: foo is a power of two iff ((foo - 1) & foo) == 0.  */
  if (((parent->num_children - 1) & parent->num_children) == 0)
    {
      if (parent->children)
        parent->children
          = (struct node **) xrealloc (parent->children,
				       (2 * parent->num_children
					* sizeof (*parent->children)));
      else
        parent->children
          = (struct node **) xmalloc (sizeof (*parent->children));
    }

  child->child_index = parent->num_children;
  parent->children[parent->num_children++] = child;
  child->parent = parent;
  child->depth = parent->depth + 1;
}


/* tree traversal */

/* Post-order traverse the tree rooted at NODE, applying FUNC as we go.  */
void
traverse_post (struct node *node,
	       node_function_t func)
{
  {
    int i;

    for (i = 0; i < node->num_children; i++)
      traverse_post (node->children[i], func);
  }

  (*func) (node);
}



/* Building node name tables for the tree */

static void hash_subtree (struct node *TREE,
			  struct string_hash *TABLE);

/* Build and return a string hash table mapping TREE's node names onto
   nodes.  If two nodes have the same name, rename them to be unique.
   Set the special node name "!root" to the root of the tree.  */
struct string_hash *
hash_node_names (struct node *tree)
{
  struct string_hash *table = new_hash_table ();

  /* Set up the "!root" name.  */
  *LOOKUP_NODE (table, "!root", 5) = tree;

  /* Hash the rest of the nodes into place.  */
  hash_subtree (tree, table);

  return table;
}

static void
hash_subtree (struct node *tree, 
	      struct string_hash *table)
{
  {
    struct node **cell
      = LOOKUP_NODE (table, tree->name.text, tree->name.len);

    /* Is there already another node in the tree by this name?  */
    if (*cell)
      {
	/* Try suffixing the node name with ":2", ":3", ... until the
	   hash table says we've found a name not already in use.  */
	int suffix = 2;
	char *buf = (char *) xmalloc (tree->name.len + 30);

	memcpy (buf, tree->name.text, tree->name.len);

	for (;;)
	  {
	    sprintf (buf + tree->name.len, ":%d", suffix);
	    cell = LOOKUP_NODE (table, buf, (tree->name.len
					     + strlen (buf + tree->name.len)));

	    if (! *cell) break;

	    suffix++;
	  }

	/* Rename the tree node.  */
	free (tree->name.text);
	make_lenstring (&tree->name, buf,
			(tree->name.len + strlen (buf + tree->name.len)));
      }

    /* Now *cell is an empty cell.  */
    *cell = tree;
  }

  /* Do children.  */
  {
    int i;

    for (i = 0; i < tree->num_children; i++)
      hash_subtree (tree->children[i], table);
  }
}


/* marks */

/* Given the name of a mark as a lenstring, return the number for that
   mark.  */
int
mark_by_name (struct lenstring *name)
{
  struct mark_name
    {
      char *name;
      int num;
    };
  static struct mark_name marks[] =
    {
      { "open", OPEN_N },
      { "selected", SELECTED_N },
      { "children-selected", CHILDREN_SELECTED_N },
      { "in-file", IN_FILE_N },
    };

  int i;

  for (i = 0; i < (sizeof (marks)/sizeof (marks[0])); i++)
    if (name->len == strlen (marks[i].name)
	&& ! memcmp (name->text, marks[i].name, name->len))
      return marks[i].num;
  
  return 0;
}


/* linear traversal */

/* Return t iff NODE1 appears strictly before NODE2 in the usual
   pre-order presentation of TREE.  */
int
node_before_p (struct node *node1, struct node *node2)
{
  /* Find the deepest common level of the two nodes.  */
  while (node1->depth > node2->depth)
    node1 = node1->parent;
  if (node1 == node2)		/* is node2 a parent of or equal to node1?  */
    return 0;
  while (node2->depth > node1->depth)
    node2 = node2->parent;
  if (node1 == node2)		/* is node1 a parent of node2?  */
    return 1;

  {
    int node1_before = 0;

    /* Walk from each node up to the parent, comparing child indices.
       The last child index that differs is the most significant.  */
    for (;;)
      {
	if (node1 == node2)
	  break;

	/* Since they're not equal, and they're in the same tree, and
	   they're at the same depth, neither of them can be the root.  */
	if (! node1->parent || ! node2->parent)
	  abort ();

	if (node1->child_index < node2->child_index)
	  node1_before = 1;
	else if (node1->child_index > node2->child_index)
	  node1_before = 0;
	/* If they're the same, keep the same value for node1_before.  */

	node1 = node1->parent;
	node2 = node2->parent;
      }

    return node1_before;
  }
}


/* Initialize the `next' pointers in TREE.
   Assume NEXT is the successor of the last node in tree.
   Return the first node in TREE.  */
struct node *
thread_tree (struct node *tree, struct node *next)
{
  int i;

  for (i = tree->num_children - 1; i >= 0; i--)
    next = thread_tree (tree->children[i], next);
      
  tree->next = next;
  return tree;
}



/* Return the node immediately following NODE, walking only visible nodes. 
   Return 0 iff NODE is the last node in the tree.  */
struct node *
next_visible_node (struct node *node)
{
  /* If we have any visible children, our first is our successor.  */
  if (node->num_children > 0
      && (node->marks & OPEN_M))
    return node->children[0];

  /* Go up until we have a successor, or have no parent.  */
  while (node->parent && node->child_index + 1 >= node->parent->num_children)
    node = node->parent;

  /* If this has no parent, then we have no successor.  */
  if (! node->parent)
    return 0;

  /* Who is our sibling at this level?  */
  node = node->parent->children[node->child_index + 1];

  return node;
}


/* Return non-zero iff NODE is visible.  */
int
node_visible_p (struct node *node)
{
  while (node->parent)
    {
      node = node->parent;
      if (! (node->marks & OPEN_M))
	return 0;
    }
  return 1;
}
