/* phylip.c --- converting alignments between Phylip and tabl formats
   Jim Blandy <jimb@cyclic.com> --- Jan 1995 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "xmalloc.h"
#include "getopt.h"
#include "lenstring.h"
#include "hash.h"
#include "careful.h"

/* Phylip files are kind of a pain.  In general, they don't contain
   enough information on their own to parse them; you may have to
   prompt the user to see which program they're intended for, and how
   the sequence data is arranged.  Furthermore, there are no
   per-sequence annotations in Phylip files, but there are things
   which might be called ``per-file annotations'' which the user would
   probably like preserved.

   Since it was a pain in the neck for me to figure this stuff out, I
   present it here in gathered form, so that you, O worthy maintainer,
   may better understand my motivations for doing such awful things.
   It also helps me understand the issues myself to write them up like
   this.

   Here's a simple Phylip file (remove three spaces of indentation):

    4 20
   E.coli       UCAGAU-UGA ACGCUGGCGG
   Aquifex_py   UCAGCG-CGA ACGCUGGCGG
   Thermotoga   UCAGGG-UGA ACGCUGGCGG
   Thermomicr   UCAGGG-GGA ACGCUGGCGG

   The "4" is the number of organisms.  The "20" is the number of
   sites given for each organism.  The first ten columns of the
   sequence lines contain the name of the organism; names may not be
   longer than ten characters, and if they are shorter, they must be
   padded to the tenth column with spaces.  The sequence data starts
   in the eleventh column, and can go as long as you like on a single
   line (although long lines make some programs unhappy).  The
   sequence data may use the characters:
        ABCDEFGHIJKLMNOPQRSTUVWXYZ*?-
   All other characters in the sequence area are ignored.  Upper- and
   lower-case letters are treated as equivalent.  Since that rule gets
   rid of spaces and numbers, it's easy to paste in data from files in
   other formats.  All very straightforward.
   
   However, what happens when you want to wrap a line?  Well, Phylip
   allows you two options: you can present all the sequence for a
   given organism at once, so that each organism's data appears
   entirely in a consecutive group of lines, or you can interleave the
   sequences, so that it's easy to see the alignment, even on
   eighty-column screens.  The Phylip documentation calls the former
   arrangement "sequential", and the latter arrangement "interleaved."

   Here is a dataset in sequential form:

    4 100
   E.coli       UCAGAU-UGAACGCU
   GGCGGCA-GGC-CUA A-CACAUGCA AGUCACGUGG             CGGACGGGUG AGUAAUGUCU
   GGGAACUGCC UGAUGGGGGG GAUAACUACU
   Aquifex_py   UCAGCG-CGA ACGCUGGCGG
             CG-UGC-CUA A-CACAUGCA AGUCGCGCGG
      CAAACGGGUG AGUAACACGU GGGUCCUACC CCCAGGGGGG GAUAACCCCG
   Thermotoga
         1 UCAGGG-UGAACGCUGGCGGCG-UGC-CUAA
        32 -CACAUGCAAGUCGCGCGGCGGACGGGUGAG
        63 UAACACGU GGGUCCUGCC CUCCGGGG GG
        94 GAUAACCAGG
   Thermomicr   UCAGGG-GGA ACGCUGGCGG CG-UGC-CUA A-UGCAUGCA AGUCACGUGG
                CGGACGGGUG CGUAACACGU GGGGCCCUCC CGGGUGGGGG GAUAACCCGG

   That's messier than what you usually see, but it demonstrates the
   possibilities.  The site count tells you how far to read before
   expecting the next organism.  Each organism starts on a new line.

   Here's the same dataset in interleaved form:

    4 100
   E.coli       UCAGAU-UG AACGCUGGCGGCA-GGC-CUAA-CACAUGCA
   Aquifex_py   UCAGCG-CGAAC GCUGGCGGCG-UGC-CUA
   Thermotoga   UCAGGG-UG AACGCUGGCGGCG-UGC-CUAA-CACAUGCAAGUCGCGC
   Thermomicr   UCAGGG-GGA ACGCUGGCGGCG-UGC-CUAA-UGCAUGC

   AGUCACGUGGCGGAC GGGUGAGUAAUGUCUGGGAACUGCCUGAUGGGGGGGAUAACUACU
   A-CACAUGCA AGUCGCGCGGCAAACGGGUGAGUAACACGUGGGUCCUACCCCCAGGGGGGGAUAACCCCG
   GGCGGACGGGUGAGUAACACGUGGG UCCUGCCCUCCGGGGGGGAUAACCAGG
   AAGUCACGUGGCGGACGGGUGCGUAACACGUGGGGCCCUC CCGGGUGGGGGGAUAACCCGG

   The seventh through tenth lines contain data for the organisms
   named on the second through fifth lines, respectively.  That is,
   the second linesworth of data for "Thermotoga" is on line nine.
   You're allowed to leave a blank line between the first bunch of
   lines, but you don't have to.  You may not repeat the organism
   names on lines after the first bunch.  The Phylip documentation
   doesn't say this, but I think that the interleaved format requires
   that all organisms use the same number of lines for their sequence
   data; in the above sample, all organisms used two lines.


   So, here's Problem #1: is the file below in interleaved or
   sequential format?

    2 9
   Bug          BUG
   Dog          dog
   HOG          hog
   Cat          CAT

   This is a valid sequential file containing two organisms: Bug, with
   sequence BUGDOGDOG, and HOG, with sequence HOGCATCAT.  This is also
   a valid interleaved file containing two organisms: BUG, with
   sequence BUGHOGHOG and DOG, with sequence DOGCATCAT.  This example
   is contrived, and golly, people usually indent all sequence lines
   but the first, and you can't indent organism names, but it seems to
   me that a good program doesn't rely on such things.  The clean way
   out seems to be to require the user to indicate if the file is
   interleaved or not.


   Pressing onward: Phylip files can include weight information, like
   the following file:

    2 20 W
   weights    0111110011
              1010101011
   big.parrot CGGACGGGUG
   bonehead   CGGAAGGGUU
              AGUAACGCGU
	      AGUAAUACAU

   The "W" on the first line indicates that category information is
   present.  The "W" at the beginning of the second line marks the
   start of the weight data, and the digits following the "W" are the
   weights.  The weight data comes after the file's first line, but
   before the sequences.  Note that, even though the sequences are in
   interleaved format, the weight data is *not* interleaved.  The
   Phylip documentation guarantees that the nine characters following
   the W are ignored, and in fact suggests that you write out the full
   word "WEIGHTS", to remind yourself what they are.

   I think some of the Phylip packages allow other sorts of per-site
   data, but it's probably handled pretty much the same way.

   Finally, Phylip input files may contain multiple data sets.  This
   feature is used in bootstrapping --- one program takes a data set
   and generates a bunch of variants, an analysis program maps over
   all the variants producing a result for each, and then a consensus
   program looks over the results and computes consensus and certainty
   information.  Here's a file containing two data sets:

	5    6
   Alpha     CCACCA
   Beta      CCAAAA
   Gamma     CAACCA
   Delta     AACAAC
   Epsilon   AACCCA
	5    6
   Alpha     CACACA
   Beta      CCAACC
   Gamma     CAACAC
   Delta     GCCTGG
   Epsilon   TGCAAT

   I don't think we need to bother supporting this format, since the
   user can always (and usually wants to!) edit the original file from
   which the bootstrap data was produced.

   In summary: Phylip's is a pretty straightforward format to parse,
   except that you can't reliably tell the difference between
   interleaved and sequential files.  You need to ask the user which
   format the file follows before reading it.


   Now, Gary Olsen's fastDNAml program takes its input following the
   Phylip format, but adds a few new tricks.  Ordinary Phylip packages
   usually present you a menu of "options" to frob and tweak before
   they start grinding away at your data.  For example, one option
   specifies whether the sequence data is interleaved or not.
   However, in fastDNAml, all the options are specified on the first
   line of the file.  Thus, you might see a fastDNAml file that looks
   like this:

    4 100 F G 4
   pCLO28       ACUGGAGAGU UUGAUUCUGG CUCAGAUUGA ACGCUGGCGG CGUGCUUAAC
   D.desulf_2   ACUGGAGAGU NUGAUUCUGG CUCAGAUUGA ACGCUGGCGG CGUGCUUAAC
   Dsm.pigra    ACUGGAGAGU UUGAUUCUGG CUCAGAUUGA ACGCUGGCGG CGUGCUUAAC
   Dsv.sp.PT-   ACUGGAGAGU UUGAUUCUGG CUGAGAUUGA ACGCUGGCGG CGUGCCUAAC

		ACAUGCAAGU CGAACGUGAA AUAAAGUGGC GCCGGGUGAG UAACCGUGGA
		ACAUGCAAGU CGAACGCGAA AUAAAGUGGC GCCGGGUGAG UAACCGUGGA
		ACAUGCAAGU CGUACGCGAA AUAAAGUGGC GGCGGGUGAG UAACCGUGGA
		ACAUGCAAGU CGUGCGUGAA AUAAAGCGGC GCCGGGUGAG UAACCGUGGA

   The "F", "G", and "4" on the first line all select options.

   In fastDNAml files, if the "I" option is present, then the sequence
   portion of the file should *not* be interleaved, i.e. it should be
   follow the sequential format.  If the "I" option is not present,
   the data may be assumed interleaved.

   Some fastDNAml options require further information, provided like
   this:

    3 100 F G J 4 T 
   J 135
   T 2.0
   pCLO28       ACUGGAGAGU UUGAUUCUGG CUCAGAUUGA ACGCUGGCGG CGUGCUUAAC
   D.desulf_2   ACUGGAGAGU NUGAUUCUGG CUCAGAUUGA ACGCUGGCGG CGUGCUUAAC
   Dsm.pigra    ACUGGAGAGU UUGAUUCUGG CUCAGAUUGA ACGCUGGCGG CGUGCUUAAC

		ACAUGCAAGU CGAACGUGAA AUAAAGUGGC GCCGGGUGAG UAACCGUGGA
		ACAUGCAAGU CGAACGCGAA AUAAAGUGGC GCCGGGUGAG UAACCGUGGA
		ACAUGCAAGU CGUACGCGAA AUAAAGUGGC GGCGGGUGAG UAACCGUGGA

   In this file, the first line selects the "J" and "T" options, which
   require numeric parameters; these are supplied on the file's second
   and third lines.  The fastDNAml documentations calls the lines that
   carry option values "auxiliary data lines".  The following options
   take auxiliary data lines:
       BCGJOTWY

   Most options require only one parameter line, but the W option may
   require more, and the C option requires at least two!  The W option
   works like its Phylip counterpart.  The C option requires one
   auxiliary data line, followed by per-site categorical information.
   Here's a fastDNAml file with C and W info:

    3 100 F G J 4 T C W
   W            0101010101 1111111000 1111100000
                1111100000 1111100000 1111100000
                0000000111 0000011111 0000000111
                1111100000
   C 5 1.0 0.9 1.1 1.2 1.3
   CATEGORIES   4444444444 4444441231 2312312312 3123123123 1231231231
                2312312312 3123123123 1231555555 5555523123 1231444444
   J 135
   T 2.0
   pCLO28       ACUGGAGAGU UUGAUUCUGG CUCAGAUUGA ACGCUGGCGG CGUGCUUAAC
   D.desulf_2   ACUGGAGAGU NUGAUUCUGG CUCAGAUUGA ACGCUGGCGG CGUGCUUAAC
   Dsm.pigra    ACUGGAGAGU UUGAUUCUGG CUCAGAUUGA ACGCUGGCGG CGUGCUUAAC

		ACAUGCAAGU CGAACGUGAA AUAAAGUGGC GCCGGGUGAG UAACCGUGGA
		ACAUGCAAGU CGAACGCGAA AUAAAGUGGC GCCGGGUGAG UAACCGUGGA
		ACAUGCAAGU CGUACGCGAA AUAAAGUGGC GGCGGGUGAG UAACCGUGGA
   
   Which brings us to Problem #2: How would fastDNAml and Phylip parse
   the following file?

    2 100 C
   C            5 1.0 0.9 1.1 1.2 1.3
   CATEGORIES   4444444444 4444441231 2312312312 3123123123 1231231231
                2312312312 3123123123 1231555555 5555523123 1231444444
   pCLO28       ACUGGAGAGU UUGAUUCUGG CUCAGAUUGA ACGCUGGCGG CGUGCUUAAC
   D.desulf_2   ACUGGAGAGU NUGAUUCUGG CUCAGAUUGA ACGCUGGCGG CGUGCUUAAC

		ACAUGCAAGU CGAACGUGAA AUAAAGUGGC GCCGGGUGAG UAACCGUGGA
		ACAUGCAAGU CGAACGCGAA AUAAAGUGGC GCCGGGUGAG UAACCGUGGA

   This is clearly a file containing fastDNAml data; it has the two
   category lines, which must not be confused with sequence data.  If
   we tried to read this as Phylip data, we'd barf.

   One might argue that, if a program can print an error message, it
   obviously knows that something is wrong, and ought to try a
   different interpretation of the data before complaining to the
   user.  That is, the program could initially assume that the file
   was in Phylip format, notice that a line contained too much data,
   and then abort the parse and re-start assuming fastDNAml's format.
   
   However, I think this could lead to truly mysterious behavior if
   sequence lines happened to end at just the wrong place.  I think
   the clean, reliable way out is simply to ask the user whether the
   file is intended for consumption by fastDNAml or Phylip.


   Finally, the fastDNAml "U" option allows the user to specify trees
   in Newick format or using Prolog syntax; the number of trees and
   the trees themselves come after all the sequence data.  Here is (I
   think) a legal fastDNAml file with two trees.

    4 100 F G 4 U
   pCLO28       ACUGGAGAGU UUGAUUCUGG CUCAGAUUGA ACGCUGGCGG CGUGCUUAAC
   D.desulf_2   ACUGGAGAGU NUGAUUCUGG CUCAGAUUGA ACGCUGGCGG CGUGCUUAAC
   Dsm.pigra    ACUGGAGAGU UUGAUUCUGG CUCAGAUUGA ACGCUGGCGG CGUGCUUAAC
   Dsv.sp.PT-   ACUGGAGAGU UUGAUUCUGG CUGAGAUUGA ACGCUGGCGG CGUGCCUAAC

		ACAUGCAAGU CGAACGUGAA AUAAAGUGGC GCCGGGUGAG UAACCGUGGA
		ACAUGCAAGU CGAACGCGAA AUAAAGUGGC GCCGGGUGAG UAACCGUGGA
		ACAUGCAAGU CGUACGCGAA AUAAAGUGGC GGCGGGUGAG UAACCGUGGA
		ACAUGCAAGU CGUGCGUGAA AUAAAGCGGC GCCGGGUGAG UAACCGUGGA
   2
   (('D.desulf_2':0.0722,
     'Dsm.pigra':0.1128
    ):0.0894,
    'Dsv.sp.PT-':0.1293
   ):0.0366;
   ('D.desulf_2':0.0874,
    ('Dsm.pigra':0.0864,
     'Dsv.sp.PT-':0.0721
    ):0.0222
   ):0.0381


   So, the program needs to ask the user two questions:

      First, is the file a Phylip or fastDNAml file?

	 If it's a fastDNAml file, we need look no further --- fastDNAml
	 files contain all the information necessary to parse them.

	 If it's a Phylip file, then we need to ask if it is in
	 interleaved or sequential format.

   Given those data, we can successfully parse the file.


   Which brings us to the next big step --- what do we write to the
   tabl files?

   The obvious thing to do is to focus on the sequence data, and write
   one tabl record per organism in the Phylip file.  However, that
   would mean that, if you open a Phylip file in Ale, make some
   changes, and then write it out, all your masks, categories,
   options, and tree data at the bottom will be deleted!  This is not
   okay.

   So, when converting Phylip files to tabl format, we should encode
   all the options and stuff in the tabl file in such a way that Ale
   will write them back out, and then, when we convert the tabl files
   to Phylip format, we can decode the options and put them back in
   their place.

   Here's the approach we'll take:

   Remember that there are two types of auxiliary data lines: those
   that contain just a few numbers (like J or T), and those that
   contain per-site data (like C or W).

   The first line of the file we'll write out as an organism named
   "headers", whose sequence is "see-annotations", and whose
   annotations are the contents of the first of the file line (with
   the organism and site counts omitted).

   Non-per-site auxiliary data lines we'll write out as 
   organisms named "aux:p:C" (where C is the parameter being set by
   that line), whose sequence is "see-annotations", and whose
   annotation is the entire contents of the data line.

   Auxiliary lines followed by per-site data we'll write as organisms
   named "aux:s:MUMBLE" (where MUMBLE is the name on the line), whose
   sequence is the per-site data.

   If there is extra data present at the end of the file, we'll make
   an organism called "tailers", whose sequence is "see-annotations",
   and whose annotations are the trailing garbage.

   Whew!  */


/* Main.  */

typedef struct
{
  lenstring *strings;
  int len;
} stringlist;

void give_usage (void);
void give_version (void);
void init_char_tables (void);
void read_tabl (char *name_filename,
		char *annotation_filename,
		char *sequence_filename,
		stringlist *names,
		stringlist *annotations,
		stringlist *sequences);
int maximum_string_len (stringlist *strings);
void free_stringlist (stringlist *);
void read_index_file (char *index_file_name);

void phylip_to_tabl (char *phylip, 
		     char *names, char *annotations, char *sequences,
		     char fastdnaml, char interleaved, int subset);
void tabl_to_phylip (char *phylip, 
		     char *names, char *annotations, char *sequences,
		     char interleaved);

char *progname;

enum long_only {
  opt_none = 2000,
  opt_include,
  opt_exclude,
};

/* An array of structures describing the arguments accepted by the
   program.  For use by getopt_long.  */
struct option longopts[] = {
  { "to-tabl",      no_argument, 0, 't' },
  { "to-phylip",    no_argument, 0, 'p' },

  { "fastdnaml",    no_argument, 0, 'f' },
  { "interleaved",  no_argument, 0, 'i' },
  
  { "names",        required_argument, 0, 'n' },
  { "annotations",  required_argument, 0, 'a' },
  { "sequences",    required_argument, 0, 's' },

  { "include",      required_argument, 0, opt_include },
  { "exclude",      required_argument, 0, opt_exclude },

  { "help",         no_argument, 0, 'h' },
  { "version",      no_argument, 0, 'v' },

  { 0, 0, 0, 0 }
};

int
main (int argc, char **argv)
{
  int option, longopts_ix;
  char to_tabl = 0;		/* converting Phylip -> tabl */
  char to_phylip = 0;		/* converting tabl -> Phylip */
  char fastdnaml = 0;		/* read fastDNAml data */
  char interleaved = 0;		/* read/write interleaved format;
				   if zero, assume sequential.  */
  char *name_filename = 0;
  char *annotation_filename = 0;
  char *sequence_filename = 0;
  char *phylip_filename = 0;

  char *include_filename = 0;
  char *exclude_filename = 0;

  progname = careful_prog_name (argv[0]);

  init_char_tables ();

  while ((option
	  = getopt_long (argc, argv, "tpfin:a:s:hv", longopts, &longopts_ix))
	 != -1)
    switch (option)
      {
      case 't': to_tabl   = 1; break;
      case 'p': to_phylip = 1; break;
      case 'f': fastdnaml = 1; break;
      case 'i': interleaved = 1; break;

      case 'n':
	name_filename = optarg;
	break;
      case 'a':
	annotation_filename = optarg;
	break;
      case 's':
	sequence_filename = optarg;
	break;

      case opt_include:
	include_filename = optarg;
	break;
      case opt_exclude:
	exclude_filename = optarg;
	break;

      case 'v':
	give_version ();
	break;
      case 'h': case '?':
	give_usage ();
	break;
      default:
	abort ();
      }
  
  if (optind + 1 < argc
      || to_tabl == to_phylip
      || (include_filename && exclude_filename))
    give_usage ();

  if (optind < argc)
    phylip_filename = argv[optind++];

  if (include_filename)
    read_index_file (include_filename);
  if (exclude_filename)
    read_index_file (exclude_filename);

  if (to_tabl)
    phylip_to_tabl (phylip_filename,
		    name_filename,
		    annotation_filename,
		    sequence_filename,
		    fastdnaml,
		    interleaved,
		    (include_filename ? opt_include
		     : exclude_filename ? opt_exclude
		     : opt_none));

  if (to_phylip)
    tabl_to_phylip (phylip_filename,
		    name_filename,
		    annotation_filename,
		    sequence_filename,
		    interleaved);

  return 0;
}


void
give_usage ()
{
  fprintf (stderr,
	   "\
Usage: %s [ PHYLIP-FILE ] OPTION ...\n\
\n\
   [--to-tabl]   or [-t]   convert from Phylip format to tabl format\n\
   [--to-phylip] or [-p]   convert from tabl format to Phylip format\n\
   (Specify exactly one of the above two options.)\n\
\n\
   [--fastdnaml]   or [-f] Read a file formatted for fastDNAml\n\
   [--interleaved] or [-i] Read/write Phylip's \"interleaved\" format\n\
                        (default is \"sequential\" format)\n\
\n\
If omitted, the following default to standard input/output:\n\
   [-n NAME-FILE]       or [--names NAME-FILE]\n\
   [-a ANNOTATION-FILE] or [--annotations ANNOTATION-FILE]\n\
   [-s SEQUENCE-FILE]   or [--sequences SEQUENCE-FILE]\n\
\n\
When reading Phylip files:\n\
   [--include FILE]     include only organisms listed in FILE\n\
   [--exclude FILE]     exclude all  organisms listed in FILE\n\
\n\
   [--help]    or [-h]\n\
   [--version] or [-v]\n\
", progname);

  exit (1);
}


void
give_version ()
{
  printf ("Ale readers version " VERSION "\n");
  printf ("RCS revision: $Id: phylip-tabl.c,v 1.14 1995/06/03 03:58:48 jimb Exp $\n");

  exit (0);
}


/* A hash table for use in selecting subsets, and functions to use it.  */

struct string_hash *index_table;

/* Read the contents of a file full of loci (one per line) into the
   hash table.  read_EMBL can then check each entry name field against the
   hash table to decide whether to exclude/include the entry.  */
void
read_index_file (char *index_file_name)
{
  FILE *index_file;
  lenstring buf;

  index_file = careful_open (index_file_name, "r", 0);
  index_table = new_hash_table ();

  while (read_delimited_lenstring (&buf, "\n", index_file) != EOF)
    {
      lookup_hash_table (index_table, buf.text, buf.len);
      free (buf.text);
    }

  careful_close (index_file, index_file_name);
}


/* Return non-zero iff INDEX[0..LEN-1] should be written, given SUBSET.  */
int
in_subset_p (lenstring *index, int subset)
{
  return (subset == opt_none
	  || (subset == opt_include
	      && lookup_hash_table_soft (index_table,
					 index->text, index->len))
	  || (subset == opt_exclude
	      && ! lookup_hash_table_soft (index_table,
					   index->text, index->len)));
}


/* Like in_subset_p, but operates on C strings.  */
int
str_in_subset_p (char *index, int subset)
{
  lenstring wrapper;

  wrapper.text = index;
  wrapper.len  = strlen (index);

  return in_subset_p (&wrapper, subset);
}



/* reading Phylip files */

#define NUM_CHARS (256)

/* Sequences --- amino and nucleic.  */
char is_sequence_table[NUM_CHARS];
#define is_sequence(c) (is_sequence_table[(unsigned) c])

/* Weights and categories --- alphanumeric, basically.  */
char is_weight_table[NUM_CHARS];

void
init_char_tables ()
{
  char *p;

  memset (is_sequence_table, 0, sizeof (is_sequence_table));
  for (p = "ABCDEFGHIJKLMNOPQRSTUVWXYZ*?-"; *p; p++)
    {
      is_sequence_table[(unsigned) *p] = 1;
      is_sequence_table[(unsigned) tolower (*p)] = 1;
    }

  memset (is_weight_table, 0, sizeof (is_weight_table));
  for (p = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"; *p; p++)
    {
      is_weight_table[(unsigned) *p] = 1;
      is_weight_table[(unsigned) tolower (*p)] = 1;
    }
}


/* STRING if STRING is non-zero, else "standard input".  */
#define CAREFUL_NAME(string) ((string) ? (string) : "standard input")

/* The length of an organism name in a Phylip file.  */
#define PHYLIP_NAME_LEN (10)


/* Set NAME to a sequence name read from IN.
   Use FILENAME in error messages.
   Don't read past a newline.  */
void
read_name (FILE *in, char *filename, lenstring *name)
{
  int i, c;

  name->text = (char *) xmalloc (PHYLIP_NAME_LEN * sizeof (name->text[0]));

  for (i = 0;
       (i < PHYLIP_NAME_LEN && (c = getc (in)) != EOF && c != '\n');
       i++)
    name->text[i] = c;
  name->len = i;

  /* Trim spaces from the end.  */
  while (name->len > 0 && isspace (name->text[name->len - 1]))
    name->len--;

  if (name->len == 0)
    {
      fprintf (stderr,
	       "%s: %s: apparently empty name; please check the input file\n",
	       progname, CAREFUL_NAME (filename));
      exit (2);
    }
  if (c == '\n')
    ungetc (c, in);
  else if (c == EOF)
    {
      fprintf (stderr, "%s: %s: file ends in midst of name: ",
	       progname, CAREFUL_NAME (filename));
      write_lenstring (name, stderr);
      fprintf (stderr, "\n%s: please check input file\n", progname);

      exit (2);
    }
}


void
incomplete_data_error (char *filename, lenstring *org_name)
{
  fprintf (stderr, "%s: %s: reached end of file in midst of sequence\n",
	   progname, CAREFUL_NAME (filename));
  fprintf (stderr, "%s: sequence is apparently named `", progname);
  write_lenstring (org_name, stderr);
  fprintf (stderr, "'\n%s: please check input file\n", progname);
  exit (2);
}


void
extra_data_error (char *filename, lenstring *org_name)
{
  fprintf (stderr, "%s: %s: found extra data at end of sequence\n",
	   progname, CAREFUL_NAME (filename));
  fprintf (stderr, "%s: sequence is apparently named `", progname);
  write_lenstring (org_name, stderr);
  fprintf (stderr, "'\n%s: please check input file\n", progname);
  exit (2);
}


/* Read a sequential-style per-site line from IN.  Read until we've
   found NUM_SITES characters of data; ignore the rest of the line.
   Ignore characters c for which CHARSET[c] is zero.
   
   Put the name of the organism (the first ten characters of the line)
   in NAME, and the data in DATA.  The caller should free these.  */
   
void
read_sequential_line (FILE *in, char *filename,
		      int num_sites, char *charset,
		      lenstring *name, lenstring *data)
{
  int i;

  read_name (in, filename, name);

  data->text = (char *) xmalloc (num_sites * sizeof (data->text[0]));

  for (i = 0; i < num_sites; i++)
    {
      int c;

      /* Skip non-sequence characters.  */
      while ((c = getc (in)) != EOF && ! charset[c])
	;

      if (c == EOF)
	incomplete_data_error (filename, name);

      data->text[i] = c;
    }
  data->len = i;

  /* Make sure the rest of the line doesn't have any sequence-ish data
     on it.  */
  {
    int c;

    while ((c = getc (in)) != EOF
	   && ! charset[c]
	   && c != '\n')
      ;
    
    if (charset[c])
      extra_data_error (filename, name);
  }
}


/* Read one line of an interleaved sequence.
   Read characters from IN, storing sequence chars in DATA, starting at POS.
   Use FILENAME and ORG_NAME in error messages.
   Assume data already has space allocated for its specified length.
   Throw away characters that don't fit in DATA.  
   Return the index of the next free position in DATA.  */
int
read_interleaved_line (FILE *in, char *filename, lenstring *org_name,
		       lenstring *data, int pos)
{
  char *p = data->text + pos;
  char *end = data->text + data->len;
  int c;

  while (p < end
	 && (c = getc (in)) != EOF
	 && c != '\n')
    if (is_sequence (c))
      *p++ = c;

  if (p >= end)
    {
      /* Make sure the rest of the line doesn't have any sequence-ish data
	 on it.  */
      int c;

      while ((c = getc (in)) != EOF
	     && !is_sequence (c)
	     && c != '\n')
	;
    
      if (is_sequence (c)) 
	extra_data_error (filename, org_name);
    }
    
  else if (c == EOF)
    incomplete_data_error (filename, org_name);

  return p - data->text;
}



void
header_line_error (char *name)
{
  fprintf (stderr, "%s: %s: first line is malformed;\n",
	   progname, CAREFUL_NAME (name));
  fprintf (stderr, "%s: please check input file\n", progname);
  exit (2);
}


/* Parse the first line of a Phylip file.
   F is the file; NAME is its name, for use in error messages.
   Set *NUM_ORGANISMS to the organism count,
       *NUM_SITES to the site count, and
       *FIRST_LINE_OPTIONS to the section of the line after the site count
           (which contains options describing the rest of the file).  */
void
phylip_first_line (FILE *f, char *name,
		   int *num_organisms, int *num_sites,
		   lenstring *first_line_options)
{
  lenstring line;

  if (read_delimited_lenstring (&line, "\n", f) != 0)
    header_line_error (name);

  {
    char *p = line.text;
    char *end = p + line.len;

    /* Skip initial whitespace.  */
    while (p < end && isspace (*p))
      p++;

    /* Read the organism count.  */
    {
      int n = 0;

      if (p >= end || !isdigit (*p))
	header_line_error (name);

      while (p < end && isdigit (*p))
	n = (n * 10) + (*p++ - '0');

      *num_organisms = n;
    }

    /* Skip some whitespace.  */
    if (p >= end || !isspace (*p))
      header_line_error (name);
    do
      p++;
    while (p < end && isspace (*p));
  
    /* Read the site count.  */
    {
      int n = 0;

      if (p >= end || !isdigit (*p))
	header_line_error (name);

      while (p < end && isdigit (*p))
	n = (n * 10) + (*p++ - '0');

      *num_sites = n;
    }

    /* Copy the remainder to *first_line_options.  */
    first_line_options->len = end - p;
    first_line_options->text 
      = (char *) xmalloc (first_line_options->len
			  * sizeof (*first_line_options->text));
    memcpy (first_line_options->text, p, first_line_options->len);
  }

  free (line.text);
}
   

void
aux_line_error (char *name, char letter)
{
  fprintf (stderr, "%s: %s: auxiliary `%c' data line is malformed;\n",
	   progname, CAREFUL_NAME (name), letter);
  fprintf (stderr, "%s: please check input file\n", progname);
  exit (2);
}


void
fastdnaml_headers (FILE *phylip, char *phylip_name,
		   FILE *names, FILE *annotations, FILE *sequences,
		   int num_sites, int subset, lenstring *options)
{
  /* Option letters for which fastDNAml expects a parameter line.  */
  static char parameters[] = "BCGJOTY";

  /* Option letters for which fastDNAml expects per-site data.  */
  static char per_site[] = "CW";

  /* Headers we've seen before; '\0'-terminated.  */
  char headers_seen[50];

  /* We store headers in the tabl file as an organism named "headers",
     with sequence "see-annotations"; the annotations contain the
     portion of the first line after the organism and site counts.
     
     Non-per-site auxiliary data lines we'll write out as 
     organisms named "aux:p:C" (where C is the parameter being set by
     that line), whose sequence is "see-annotations", and whose
     annotation is the entire contents of the data line.
     
     Auxiliary lines followed by per-site data we'll write as organisms
     named "aux:s:MUMBLE" (where MUMBLE is the name on the line), whose
     sequence is the per-site data.  */
  if (str_in_subset_p ("headers", subset))
    {
      fputs ("headers\n", names);
      write_lenstring (options, annotations);
      fputs ("\f", annotations);
      fputs ("see-annotations\n", sequences);
    }

  /* We haven't seen any headers yet.  */
  headers_seen[0] = '\0';

  /* Read header lines, and decide what to do with them.  */
  for (;;)
    {
      /* Get the first character of the line.  */
      int c = getc (phylip);
      ungetc (c, phylip);

      c = toupper (c);

      /* Does it start with a letter that appears on the first line?
	 Is it a letter we haven't seen before?  */
      if (c != EOF
	  && (memchr (options->text, c, options->len)
	      || memchr (options->text, tolower (c), options->len))
	  && ! strchr (headers_seen, c))
	{
	  /* If c appears in neither parameters nor per_site, we have
	     reached the end of the header lines, even if it does
	     appear on the first line.  */
	  int processed = 0;

	  if (strchr (parameters, c))
	    {
	      lenstring line;

	      processed = 1;

	      if (0 != read_delimited_lenstring (&line, "\n", phylip))
		aux_line_error (phylip_name, c);

	      {
		char buf[7];
		lenstring name;

		/* Build the name for this parameter, so we can see if
                   it's been included/excluded.  */
		name.text = buf;
		name.len  = sizeof (buf);
		memcpy (name.text, "aux:p:", 6);
		name.text[6] = c;

		if (in_subset_p (&name, subset))
		  {
		    write_lenstring (&name, names);
		    putc ('\n', names);
		    write_lenstring (&line, annotations);
		    fputs ("\f", annotations);
		    fputs ("see-annotations\n", sequences);
		  }
	      }

	      free (line.text);
	    }

	  if (strchr (per_site, c))
	    {
	      lenstring name, data;

	      processed = 1;

	      /* If c appears in both parameters and per_site, then
		 this could be a different line from the one we
		 originally read c from.  Make sure this line starts
		 with the proper letter.  */
	      {
		int c2 = getc (phylip);
		ungetc (c2, phylip);

		if (c != c2)
		  {
		    fprintf (stderr,
			     "%s: %s: `%c' auxiliary data line is missing\n",
			     progname, CAREFUL_NAME (phylip_name), c);
		    fprintf (stderr, "%s: its per-site data; please check input file\n", 
			     progname);
		    exit (2);
		  }
	      }

	      read_sequential_line (phylip, phylip_name,
				    num_sites, is_weight_table,
				    &name, &data);

	      /* Build the name for this parameter, so we can see if
		 it's been included/excluded.  */
	      {
		lenstring n;

		n.len = 6 + name.len;
		n.text = (char *) xmalloc (n.len * sizeof (n.text[0]));
		memcpy (n.text, "aux:s:", 6);
		memcpy (n.text + 6, name.text, name.len);

		if (in_subset_p (&n, subset))
		  {
		    write_lenstring (&n, names);
		    putc ('\n', names);
		    fputs ("\f", annotations);
		    write_lenstring (&data, sequences);
		    putc ('\n', sequences);
		  }

		free (n.text);
	      }

	      free (name.text);
	      free (data.text);
	    }

	  /* The character doesn't appear in either parameters or
             per_site, so it's probably the start of a real organism
             name.  Who knows why it appears on the first line.  */
	  if (! processed)
	    break;

	  {
	    int len = strlen (headers_seen);
	    headers_seen[len++] = c;
	    headers_seen[len++] = '\0';
	  }
	}
      else
	/* The character doesn't appear on the first line, so it's
           probably the start of a real organism name.  */
	break;
    }
}


void
phylip_headers (FILE *phylip, char *phylip_name,
		FILE *names, FILE *annotations, FILE *sequences,
		int num_sites, int subset, lenstring *options)
{
  /* Option letters for which Phylip expects per-site data.  */
  static char per_site[] = "W";

  /* We store headers in the tabl file as an organism named "headers",
     with sequence "see-annotations"; the annotations contain the
     portion of the first line after the organism and site counts,
     terminated by a newline.  */
  if (str_in_subset_p ("headers", subset))
    {
      fputs ("headers\n", names);
      write_lenstring (options, annotations);
      fputs ("\f", annotations);
      fputs ("see-annotations\n", sequences);
    }

  /* Read header lines, and decide what to do with them.  */
  for (;;)
    {
      /* Get the first character of the line.  */
      int c = getc (phylip);
      ungetc (c, phylip);

      c = toupper (c);

      /* Does it start with a letter that appears on the first line?  */
      if (c != EOF
	  && (memchr (options->text, c, options->len)
	      || memchr (options->text, tolower (c), options->len))
	  && strchr (per_site, c))
	{
	  lenstring name, data;

	  read_sequential_line (phylip, phylip_name,
				num_sites, is_weight_table,
				&name, &data);

	  /* Build the name for this parameter, so we can see if
	     it's been included/excluded.  */
	  {
	    lenstring n;
	    
	    n.len = 6 + name.len;
	    n.text = (char *) xmalloc (n.len * sizeof (n.text[0]));
	    memcpy (n.text, "aux:s:", 6);
	    memcpy (n.text + 6, name.text, name.len);

	    if (in_subset_p (&n, subset))
	      {
		write_lenstring (&n, names);
		putc ('\n', names);
		fputs ("\f", annotations);
		write_lenstring (&data, sequences);
		putc ('\n', sequences);
	      }

	    free (n.text);
	  }

	  free (name.text);
	  free (data.text);
	}
      else
	break;
    }
}


void
interleaved_sequences (FILE *phylip, char *phylip_name,
		       int num_organisms, int num_sites, int subset,
		       FILE *names, FILE *annotations, FILE *sequences)
{
  lenstring *name, *data;
  int *pos;

  name = (lenstring *) xmalloc (num_organisms * sizeof (*name));
  data = (lenstring *) xmalloc (num_organisms * sizeof (*data));
  pos  = (int *)       xmalloc (num_organisms * sizeof (*pos));

  /* Allocate the per-organism data.  */
  {
    int i;

    for (i = 0; i < num_organisms; i++)
      {
	data[i].len  = num_sites;
	data[i].text = (char *) xmalloc (data[i].len * sizeof (data[i].text));

	pos[i] = 0;
      }
  }

  /* Read the interleaved data.  */
  {
    int first_group = 1;

    for (;;)
      {
	int i;

	for (i = 0; i < num_organisms; i++)
	  {
	    if (first_group)
	      read_name (phylip, phylip_name, &name[i]);

	    pos[i] = read_interleaved_line (phylip, phylip_name, &name[i],
					    &data[i], pos[i]);
	  }

        first_group = 0;

	/* Are there any organisms that still need residues?  */
	for (i = 0; i < num_organisms; i++)
	  if (pos[i] < data[i].len)
	    break;

	if (i >= num_organisms)
	  /* No, we're done.  */
	  break;
	else
	  /* There may be more organisms.  Skip a blank line, if there
             is one.  */
	  {
	    int c;

	    do
	      c = getc (phylip);
	    while (isspace (c) && c != '\n');

	    if (c != '\n')
	      ungetc (c, phylip);
	  }
      }
  }

  /* Write the data out to the tabl file.  */
  {
    int i;

    for (i = 0; i < num_organisms; i++)
      {
	lenstring *n = &name[i];

	if (in_subset_p (n, subset))
	  {
	    write_lenstring (n, names);
	    putc ('\n', names);
	    fputs ("\f", annotations);
	    write_lenstring (&data[i], sequences);
	    putc ('\n', sequences);
	
	    free (name[i].text);
	    free (data[i].text);
	  }
      }
  }

  free (name);
  free (data);
  free (pos);
}


void
sequential_sequences (FILE *phylip, char *phylip_name,
		      int num_organisms, int num_sites, int subset,
		      FILE *names, FILE *annotations, FILE *sequences)
{
  int i;

  for (i = 0; i < num_organisms; i++)
    {
      lenstring name, data;

      read_sequential_line (phylip, phylip_name,
			    num_sites, is_sequence_table,
			    &name, &data);

      if (in_subset_p (&name, subset))
	{
	  write_lenstring (&name, names);
	  putc ('\n', names);

	  fputs ("\f", annotations);

	  write_lenstring (&data, sequences);
	  putc ('\n', sequences);
	}

      free (name.text);
      free (data.text);
    }
}


void
write_tailer (FILE *phylip,
	      FILE *names, FILE *annotations, FILE *sequences,
	      int subset)
{
  lenstring tailer;
  int i, c;

  if (! str_in_subset_p ("tailers", subset))
    return;

  tailer.len = 10;
  tailer.text = (char *) xmalloc (tailer.len * sizeof (tailer.text[0]));
  i = 0;

  while ((c = getc (phylip)) != EOF)
    {
      if (i >= tailer.len)
	{
	  tailer.len *= 2;
	  tailer.text
	    = (char *) xrealloc (tailer.text,
				 tailer.len * sizeof (tailer.text[0]));
	}

      tailer.text[i++] = c;
    }
  tailer.len = i;

  /* Did we get anything but spaces?  */
  for (i = 0; i < tailer.len; i++)
    if (! isspace (tailer.text[i]))
      break;

  /* If we got anything, write it.  */
  if (i < tailer.len)
    {
      fputs ("tailers\n", names);
      write_lenstring (&tailer, annotations);
      putc ('\f', annotations);
      fputs ("see-annotations\n", sequences);
    }

  free (tailer.text);
}


void
phylip_to_tabl (char *phylip_name, 
		char *names_name, char *annotations_name, char *sequences_name,
		char fastdnaml, char interleaved, int subset)
{
  FILE *phylip 	    = careful_open (phylip_name,      "r", stdin);
  FILE *names  	    = careful_open (names_name,       "w", stdout);
  FILE *annotations = careful_open (annotations_name, "w", stdout);
  FILE *sequences   = careful_open (sequences_name,   "w", stdout);

  int num_organisms, num_sites;
  lenstring first_line_options;

  phylip_first_line (phylip, phylip_name,
		     &num_organisms, &num_sites, &first_line_options);

  if (fastdnaml)
    {
      /* Get the interleaved/sequential info from the first line options.
	 The presence of an 'I' means that the file is *not* interleaved.  */
      interleaved
	= !(memchr (first_line_options.text, 'I', first_line_options.len)
	    || memchr (first_line_options.text, 'i', first_line_options.len));

      fastdnaml_headers (phylip, phylip_name,
			 names, annotations, sequences,
			 num_sites, subset, &first_line_options);
    }
  else
    phylip_headers (phylip, phylip_name,
		    names, annotations, sequences,
		    num_sites, subset, &first_line_options);

  check_file (names, names_name, "writing Phylip headers");
  check_file (annotations, annotations_name, "writing Phylip headers");
  check_file (sequences, sequences_name, "writing Phylip headers");

  /* And the data.  */
  if (interleaved)
    interleaved_sequences (phylip, phylip_name,
			   num_organisms, num_sites, subset,
			   names, annotations, sequences);
  else
    sequential_sequences (phylip, phylip_name,
			  num_organisms, num_sites, subset,
			  names, annotations, sequences);

  check_file (names, names_name, "writing Phylip sequence data");
  check_file (annotations, annotations_name, "writing Phylip sequence data");
  check_file (sequences, sequences_name, "writing Phylip sequence data");

  /* Put any trailing stuff in a "tailer" mock-organism.  */
  write_tailer (phylip, names, annotations, sequences, subset);

  check_file (names, names_name, "writing Phylip tailers");
  check_file (annotations, annotations_name, "writing Phylip tailers");
  check_file (sequences, sequences_name, "writing Phylip tailers");

  free (first_line_options.text);

  careful_close (phylip, phylip_name);
  careful_close (names, names_name);
  careful_close (annotations, annotations_name);
  careful_close (sequences, sequences_name);
}


/* reading tabl files */


/* Set the I'th string of STRINGS to point to the LEN bytes at TEXT.
   If STRINGS isn't long enough, realloc it.  */
void
set_strings (stringlist *strings, int i, char *text, int len)
{
  if (strings->len == 0)
    {
      strings->len = 10;
      strings->strings
	= (lenstring *) xmalloc (strings->len * sizeof (strings->strings[0]));
    }
  else
    {
      int len = strings->len;

      while (i >= len)
	len *= 2;

      if (len > strings->len)
	{
	  strings->len = len;
	  strings->strings
	    = (lenstring *) xrealloc (strings->strings,
				      len * sizeof (strings->strings[0]));
	}
    }

  strings->strings[i].text = text;
  strings->strings[i].len  = len;
}
   

/* Read a tabl file.  Get names, annotations, and sequences from
   NAME_FILENAME, ANNOTATION_FILENAME, and SEQUENCE_FILENAME.  Put
   them in *NAMES, *ANNOTATIONS, and *SEQUENCES.  */
void read_tabl (char *name_filename,
		char *annotation_filename,
		char *sequence_filename,
		stringlist *names,
		stringlist *annotations,
		stringlist *sequences)
{
  FILE *name_file       = careful_open (name_filename,       "r", stdin);
  FILE *annotation_file = careful_open (annotation_filename, "r", stdin);
  FILE *sequence_file   = careful_open (sequence_filename,   "r", stdin);

  lenstring n, a, s;

  int i;

  names->len = annotations->len = sequences->len = 0;

  for (i = 0; ; i++)
    {
      int n_status = read_delimited_lenstring (&n, "\n", name_file);
      int a_status = read_delimited_lenstring (&a, "\f", annotation_file);
      int s_status = read_delimited_lenstring (&s, "\n", sequence_file);

      if (n_status || a_status || s_status)
	{
	  if (! (n_status == EOF && a_status == EOF && s_status == EOF))
	    {
	      fprintf (stderr, "%s: %s, %s, %s:\n",
		       progname,
		       CAREFUL_NAME (name_filename),
		       CAREFUL_NAME (annotation_filename),
		       CAREFUL_NAME (sequence_filename));
	      fprintf (stderr,
		       "%s: Hallelujah!  Some of your tabl files are incomplete.\n",
		       progname);
	      exit (2);
	    }

	  
	  break;
	}

      set_strings (names,       i, n.text, n.len);
      set_strings (annotations, i, a.text, a.len);
      set_strings (sequences,   i, s.text, s.len);
    }

  names->len = i;
  annotations->len = i;
  sequences->len = i;

  careful_close (name_file, name_filename);
  careful_close (annotation_file, annotation_filename);
  careful_close (sequence_file, sequence_filename);
}


/* Display LEN bytes of SEQUENCE starting at element START, as
   if SEQUENCE were padded to infinite length with PAD_CHAR.
   Write to F.
   We should really fold this functionality into display_clipped_lenstring.  */
void
display_padded_sequence (lenstring *sequence, int start, int len,
			 char pad_char, 
			 FILE *f)
{
  int seq_start, seq_end;
  
  seq_start = start;
  seq_end = start + len;

  if (seq_start > sequence->len)
    seq_start = sequence->len;
  if (seq_end > sequence->len)
    seq_end = sequence->len;

  fwrite (sequence->text + seq_start, sizeof (sequence->text[0]),
	  seq_end - seq_start,
	  f);
  
  /* Set len to the number of pad characters we actually have to print.  */
  len = len - (seq_end - seq_start);
  while (len-- > 0)
    putc (pad_char, f);
}


/* Display LEN characters from the beginning of SEQUENCE to file F.

   If SEQUENCE is shorter than len, behave as if it were padded to
   infinity with PAD_CHAR.

   Indent all lines but the first with INDENT.  */
void
display_sequence (lenstring *sequence,
		  int len, char pad_char, char *indent,
		  FILE *f)
{
  int pos = 0;

  /* Display the text in five columns of ten residues each, separated
     by spaces.  */
  while (pos < len)
    {
      int line_end = pos + 50;

      if (line_end > len)
	line_end = len;

      if (pos > 0)
	fputs (indent, f);
	
      while (pos < line_end)
	{
	  int column_width = len - pos;

	  if (column_width > 10)
	    column_width = 10;

	  display_padded_sequence (sequence, pos, column_width, pad_char, f);

	  pos += column_width;

	  if (pos < line_end)
	    putc (' ', f);
	}

      putc ('\n', f);
    }
}


/* Display sequences in interleaved format.
   Take names from NAMES, and sequence data from SEQUENCES.
   Pretend that all the sequences are padded with gaps to NUM_SITES in length.
   Write to F.  */
void
display_interleaved (stringlist *names, stringlist *sequences, int num_sites,
		     FILE *f)
{
  int num_organisms = names->len;
  int page_pos;

  for (page_pos = 0; page_pos < num_sites; page_pos += 50)
    {
      int i;

      for (i = 0; i < num_organisms; i++)
	{
	  int pos = page_pos;
	  int line_end;

	  if (pos == 0)
	    {
	      display_clipped_lenstring (&names->strings[i], 10, f);
	      fputs ("   ", f); /* 3 spaces */
	    }
	  else
	    fputs ("             ", f); /* 13 spaces */

	  line_end = pos + 50;
	  if (line_end > num_sites)
	    line_end = num_sites;

	  while (pos < line_end)
	    {
	      int len = num_sites - pos;

	      if (len > 10)
		len = 10;

	      display_padded_sequence (&sequences->strings[i],
				       pos, len, '-', f);

	      pos += len;
	      if (pos < line_end)
		putc (' ', f);
	    }		    

	  /* Put a newline after each line.  */
	  putc ('\n', f);
	}

      /* Put a blank line after each page but the last.  */
      if (page_pos + 50 < num_sites)
	putc ('\n', f);
    }
}


void
display_sequential (stringlist *names, stringlist *sequences, int num_sites,
		    FILE *f)
{
  int i;

  for (i = 0; i < names->len; i++)
    {
      display_clipped_lenstring (&names->strings[i], 10, f);
      fputs ("   ", f);	/* 3 spaces */
      display_sequence (&sequences->strings[i], num_sites, '-',
			"             ", /* 13 spaces */
			f);
    }
}


/* Return non-zero iff NAME is equal to PATTERN (if ALLOW_TAIL is zero)
   or if PATTERN is a prefix of NAME (if ALLOW_TAIL is non-zero).  */
int
name_match (lenstring *name, char *pattern, int allow_tail)
{
  int len = strlen (pattern);

  return ((allow_tail ? name->len >= len : name->len == len)
	  && ! memcmp (name->text, pattern, len));
}


void
tabl_to_phylip (char *phylip_filename, 
		char *names_filename,
		char *annotations_filename,
		char *sequences_filename,
		char interleaved)
{
  stringlist names, annotations, sequences;
  stringlist organism_names, organism_sequences;

  /* This is annoying, but we have to read the entire tabl file before
     we can start writing the Phylip file, because we need to compute
     the organism and site counts.  */
  read_tabl (names_filename, annotations_filename, sequences_filename,
	     &names, &annotations, &sequences);

  /* Sort out the sequence data proper from the headers, tailers, and
     auxiliary data lines.  */
  {
    FILE *phylip = careful_open (phylip_filename, "w", stdout);

    int headers_entry = -1;
    int tailers_entry = -1;

    int num_organisms = 0;
    int num_sites = 0;

    organism_names.len = organism_sequences.len = 0;

    /* Search the alignment for headers and tailers; count the
       organisms, and determine the number of sites they have.  */
    {
      int i;

      for (i = 0; i < names.len; i++)
	{
	  lenstring *name = &names.strings[i];

	  if (name_match (name, "headers", 0))
	    headers_entry = i;
	  else if (name_match (name, "tailers", 0))
	    tailers_entry = i;
	  else if (name_match (name, "aux:", 1))
	    ;
	  else
	    {
	      set_strings (&organism_names, num_organisms,
			   name->text, name->len);
	      set_strings (&organism_sequences, num_organisms,
			   sequences.strings[i].text,
			   sequences.strings[i].len);

	      num_organisms++;
	      if (sequences.strings[i].len > num_sites)
		num_sites = sequences.strings[i].len;
	    }
	}
    }

    organism_names.len = organism_sequences.len = num_organisms;
    num_sites = maximum_string_len (&organism_sequences);

    /* Write the header line.   */
    fprintf (phylip, " %d %d", num_organisms, num_sites);
    if (headers_entry != -1)
      write_lenstring (&annotations.strings[headers_entry], phylip);
    putc ('\n', phylip);

    /* Write the auxiliary lines.  */
    {
      int i;

      for (i = 0; i < names.len; i++)
	{
	  lenstring *name = &names.strings[i];

	  /* Is this a per-site auxiliary line?  */
	  if (name_match (name, "aux:s:", 1))
	    {
	      lenstring original_name;

	      /* Strip the "aux:s:" from the name.  */
	      original_name.text = name->text + 6;
	      original_name.len  = name->len - 6;

	      display_clipped_lenstring (&original_name, 10, phylip);
	      fputs ("   ", phylip); /* 3 spaces */
	      /* I guess '0' is a good padding character for
		 per-site auxiliary data.  */
	      display_sequence (&sequences.strings[i], num_sites, '0',
				"             ", /* 13 spaces */
				phylip);
	    }

	  /* Is this a parameter line?  */
	  else if (name_match (name, "aux:p:", 1))
	    {
	      /* It's just a parameter line.  The annotation
		 contains the text that should go into the Phylip
		 file.  */
	      write_lenstring (&annotations.strings[i], phylip);
	      putc ('\n', phylip);
	    }
	}
    }

    check_file (phylip, phylip_filename, "writing Phylip headers");

    /* Write the data lines.  */
    if (interleaved)
      display_interleaved (&organism_names, &organism_sequences, num_sites, 
			   phylip);
    else
      display_sequential (&organism_names, &organism_sequences, num_sites,
			  phylip);

    check_file (phylip, phylip_filename,
		"writing Phylip sequence data");

    /* Write the tailers.  */
    if (tailers_entry != -1)
      write_lenstring (&annotations.strings[tailers_entry], phylip);

    careful_close (phylip, phylip_filename);
  }

  free (organism_names.strings);
  free (organism_sequences.strings);

  free_stringlist (&names);
  free_stringlist (&annotations);
  free_stringlist (&sequences);
}



/* general utilities */

/* Return the length of the longest string in STRINGS.  */
int
maximum_string_len (stringlist *strings)
{
  int max = 0;
  int i;

  for (i = 0; i < strings->len; i++)
    if (strings->strings[i].len > max)
      max = strings->strings[i].len;

  return max;
}


void
free_stringlist (stringlist *s)
{
  int i;

  for (i = 0; i < s->len; i++)
    free (s->strings[i].text);

  free (s->strings);
}
