;;;; gene-analyze.el --- calling analysis programs
;;;; Copyright (C) 1994 Indiana University Foundation
;;;;
;;;; This software is copyrighted under the terms of the GNU General
;;;; Public License (see terms below).  The copyright covers this file,
;;;; specifically, and the program of which it is a part.
;;;; 
;;;;      This program is free software; you can redistribute it and/or modify
;;;;      it under the terms of the GNU General Public License as published by
;;;;      the Free Software Foundation; either version 1, or (at your option)
;;;;      any later version.
;;;; 
;;;;      This program is distributed in the hope that it will be useful,
;;;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;;      GNU General Public License for more details.
;;;; 
;;;;      You should have received a copy of the GNU General Public License
;;;;      along with this program; if not, write to the Free Software
;;;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;;;;

(require 'cl)
(require 'shell)

(require 'gene-io)


;;;; Managing the analysis frame.

(defconst genalyze-buffer "*Analysis*")

(defun genalyze-run-command (command)
  "Run COMMAND in the *Analysis* buffer, and display that buffer.
Command is a list of strings, the first of which is the name of the
command to run."
  (let* ((buf (get-buffer-create genalyze-buffer))
	 (proc (get-buffer-process buf)))
    (set-buffer buf)
    (gene-menu-bar-make-hide-button)
    (gene-frame-show
     buf
     `((foreground-color . ,gene-foreground-color)
       (background-color . ,gene-background-color)
       (cursor-color     . ,gene-cursor-color)
       (mouse-color      . ,gene-cursor-color)
       (unsplittable     . t)
       (menu-bar-lines   . 1)))
    (if (and proc
	     (eq (process-status proc) 'run))
	(when (gene-y-or-n-p
	       "There is an analysis command running already; kill it? ")
	  (kill-process proc)))
    (goto-char (point-max))
    (insert "\n\n")
    (save-excursion
      (insert (mapconcat 'identity command " ") "\n"))
    (set-window-start (selected-window) (point))
    (set-window-point (selected-window) (point))
    (let ((new-proc (apply 'start-process "*Analysis*" (current-buffer) 
			   command)))
      (set-process-sentinel new-proc 'genalyze-sentinel))))

(defvar genalyze-temp-files nil
  "Files to be deleted when the next analyzer exits.")

(defun genalyze-delete-temp-files ()
  "Delete the temp files used by the last analysis program we ran."
  (while (consp genalyze-temp-files)
    (condition-case data
        (delete-file (car genalyze-temp-files))
      (file-error))
    (setq genalyze-temp-files (cdr genalyze-temp-files))))

(defun genalyze-sentinel (process string)
  "The sentinel function used for analysis subprocesses."
  (genalyze-delete-temp-files))


;;;; Running analyzers on the current group.

(defun genalyze-run-on-tabl-group (command)
  "Run COMMAND on the ids and sequences of the current group.
The ids and sequences are stored in two files, separated by newlines;
think of it as tabl format without the annotations.
COMMAND is a list of strings specifying the command to run.  It can
contain symbols `ids' and `sequences', which will be replaced by the
names of files containing the ids and sequences of the current group.
The final set of strings are executed by treating the first one as the
name of the program and the rest as its command-line arguments."
  (gene-with-watch-pointer
    (let (ids sequences)
      (let ((temp-files
	     (gene-io-with-not-so-temp-files
	      (let* ((group (gene-sgroup-displayed-group-here))
		     (organisms (if group (gene-sgroup-sequences group)
				  (list (gene-get-here-sym))))
		     (files (gene-io-write-organisms-no-ann organisms)))
		(setq ids (elt files 0))
		(setq sequences (elt files 2))))))
	(setq genalyze-temp-files
	      (append genalyze-temp-files temp-files)))
      (let ((command-string (mapcar 'eval command)))
	(genalyze-run-command command-string)))))
	 

;;;; Running count-tabl, the composition counter

(defun genalyze-composition (combination three-position)
  "Analyze the composition of the sequences in the current group."
  (interactive
   (list (gene-y-or-n-p "Compute percentages of combinations (i.e. A+C)? ")
	 (gene-y-or-n-p "Compute separate sums for each codon position? ")))
  (genalyze-run-on-tabl-group (append '("count-tabl")
				      (if combination '("-c"))
				      (if three-position '("-p"))
				      '(ids sequences))))


;;;; prologue

(provide 'gene-analyze)
