;; gene-data-input.el (part of ale).
;; Inputting and double-checking new data.
;; Copyright (C) 1995  Karl Fogel <kfogel@cyclic.com>
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;;
;; Search for instances of "todo:" to find notes about future plans.

(require 'gene-mode-tools)


;;;; Code:

;;; Creating new sequences

(defvar gene-new-sequence-count 0
  "Number of new sequences generated this session.
This variable is used in creating names for them.")

(defun gene-new-sequence (name)
  "Open a new, empty sequence named NAME above the line at point.
Place point in the new sequence.
NAME defaults to a concoction based on username.
The new sequence is assigned the same file as the sequence at point, and
its inititial lock state is Unlocked."
  (interactive
   (let* ((default-name
            (concat
             (user-login-name)
             "-"
             (number-to-string gene-new-sequence-count)))
          (input 
           (read-string
            (format "Name of new sequence (default \"%s\"): " default-name))))
     (list (if (string-equal "" input)
               default-name
             input))))

  (let* ((source (gene-get-here-source name))
         (file   (gene-source-filename source))
         (format (gene-source-file-format source))
         (new-sym
          (gene-open-sequence
           name
           ""
           (gene-io-make-annotation name format)
           file
           format
           gene-locking-unlocked-state)))
    (setq gene-new-sequence-count (1+ gene-new-sequence-count))
    (gene-seq-goto-organism new-sym)))



;;; Input from gels, and double-checking from gels.

(defsubst gene-seq-maybe-handle-data-input-command (char)
  "If in data-input mode, then deal with CHAR and return non-nil.
Else return nil."
  (and
   gene-data-input-mode
   (eq (gene-get-here-sym) gene-data-input-sequence)
   (if (eolp)
       ;; At end of line, we just accept new data for input.
       (gene-seq-insert-char (gene-data-input-translate-char char) parg)
     ;; Else do a double-check.
     (cond
      ((equal char ?p)         (gene-data-input-handle-replace))
      ((equal char ?u)         (gene-data-input-handle-delete))
      ((equal char ?)        (gene-data-input-handle-delete))
      ((equal char 'backspace) (gene-data-input-handle-delete))
      ((equal char 'delete)    (gene-data-input-handle-delete))
      ((equal char ?i)         (gene-data-input-handle-insert))
      ((equal char ?I)         (gene-data-input-handle-insert))
      ((equal char ??)         (gene-data-input-handle-help))
      (t
       (gene-data-input-double-check-char
        (gene-data-input-translate-char char)))))
   (gene-data-input-auto-gappify)
   ;; Return t for caller's sake.
   t))


;;; Entering new sequences and double-checking them:
;;
;; Two things need to be done here.
;;
;; 1. When entering a sequence for the first time, it's nice to be
;; able to type it in in tandem with a sequence already in the buffer.
;; This will allow gaps to be inserted for you, to a point, and also
;; allow the new sequence to be almost aligned by the time you're done
;; -- and you'll know exactly where the trouble spots are, since you
;; will have been forced to deal with them as you typed in.  Hmm, now
;; that I mention it, I'm not so sure that's a good thing...
;;
;; 2. Double-checking a newly entered sequence.
;;    (explanation here soon)

(defvar gene-data-input-translation-table
  '((?j ?U)
    (?k ?A)
    (?l ?C)
    (?\; ?G)
    (?  ?-) ; this gap is not in Carl's list
    ;; Would be nice if we had a way of handling context-gap...
    ;; Maybe we should do the keyboard-translation by hand...
    (?' ?n)
    (?m ?u)
    (?, ?a)
    (?. ?c)
    (?/ ?g)
    )
  "*Keyboard translation table for entering new sequences.")


(defun gene-data-input-translate-char (char)
  "Translate CHAR according to `gene-data-input-translation-table'.
Returns the new char."
  (or
   (nth 1 (assoc char gene-data-input-translation-table))
   char))


(defvar gene-data-input-mode-modeline-string ""
  "String for modeline, set by `gene-data-input-mode'.")


(defvar gene-data-input-sequence nil
  "The seq into which data is being entered in gene-data-input-mode.")


(defvar gene-data-input-mode nil
  "Mode variable for gene-data-input-mode.")

(defun gene-data-input-mode (&optional flag seq)
  "Minor mode for entering new sequence data.
Optional arg FLAG nil means toggle the mode.
Else if \(> \(prefix-numeric-value FLAG\) 0\), turn it on.
Else turn it off.

Optional second arg SEQ is the sequence in which the new data will be
entered; it defaults to the sequence at point."
  (interactive "P")

  ;; We don't want to actually set the mode var,
  ;; `gene-data-input-mode', until the end of the function, in
  ;; case of error, so we use a temp variable:
  (let (gene-data-input-mode-tmp)
    (setq gene-data-input-mode-tmp
          (if (null flag) (not gene-data-input-mode)
            (> (prefix-numeric-value flag) 0)))
    
    (if gene-data-input-mode-tmp
        ;; Mode is on.
        (let ((sequence (or seq (gene-get-here-sym))))
          (if sequence
              (setq gene-data-input-sequence sequence)
            (gene-error "Can't find sequence in which to enter data."))

          ;; todo: hmmm, how to handle multiple invocations?  We
          ;; certainly only want to be inputting into one sequence at
          ;; a time; so a repeated invokation should have the effect
          ;; of switching the target sequence.

          (let ((mode-str
                 (format
                  "(Entering Data: %s)" (gene-organism-name sequence))))
            (setq minor-mode-alist
                  (delete-if
                   (lambda (cell)
                     (eq (car cell) 'gene-data-input-mode))
                   minor-mode-alist)
                  minor-mode-alist
                  (cons (list 'gene-data-input-mode mode-str)
                        minor-mode-alist)
                  gene-data-input-mode-modeline-string
                  mode-str)))
      
      ;; Else mode is off.
      (setq gene-data-input-sequence nil
            minor-mode-alist
            (delete-if
             (lambda (cell)
               (eq (car cell) 'gene-data-input-mode))
             minor-mode-alist)
            gene-data-input-mode-modeline-string
            ""))
    
    ;; Finally, make the mode var equal its temporary stand-in.
    (setq gene-data-input-mode gene-data-input-mode-tmp)
    (force-mode-line-update)))


(defun gene-data-input-auto-gappify ()
  "Automatically insert or skip gaps after entering new sequence data.
If at end of line, inserts; else skips to next base or eol, whichever
comes first."
  (if (eolp)
      nil ; Do nothing right now.
    (skip-chars-forward gene-skip-gaps-regexp (gene-eol-point))))


(defun gene-data-input-handle-help ()
  "Give brief instructions for double-check mode."
   (gene-message "`p' replaces, `i' inserts, and `u' deletes."))


(defun gene-data-input-handle-replace ()
  "Promptingly replace the char at point with a new one.
This always prompts the user, so don't try to use it non-interactively."
  (let (buffer-read-only
        (o-char (char-after (point))))
    (let ((new-char
           (gene-data-input-translate-char
            (gene-read-char
             (format "Replace `%c' with: " o-char)
             (concat (format "Not confirmed! (%c --> " o-char) "%c)")
             'read-char-exclusive))))
      (gene-translate-point-to-char (point) new-char)
      (forward-char 1)
      (gene-message "Replaced `%c' with `%c'" o-char new-char))))
          

(defun gene-data-input-handle-insert ()
  "Promptingly insert a new char at point.
This always prompts the user, so don't try to use it non-interactively."
  (let (buffer-read-only
        (o-char (char-after (point))))
    (let ((new-char
           (gene-data-input-translate-char
            (gene-read-char
             (format "Insert before `%c': " o-char)
             (concat "Not confirmed! (insertion of `%c')")
             'read-char-exclusive))))
      (insert-char new-char 1 gene-sel-stretch-selections)
      (forward-char 1)
      (gene-message "Inserted `%c' before `%c'." new-char o-char))))


(defun gene-data-input-handle-delete ()
  "Promptingly delete the char at point.
This always prompts the user, so don't try to use it non-interactively."
  (let (buffer-read-only
        (cursor-in-echo-area t)
        (o-char (char-after (point))))
    (gene-message "Delete `%c'?" o-char)
    (let ((confirmation (read-char-exclusive)))
      (if (member confirmation gene-read-char-confirmation-chars)
          (progn
            (delete-char 1)
            (gene-message "Deleted `%c'." o-char))
        (gene-error
         "Not confirmed! (deletion of %c)" o-char)))))


(defun gene-data-input-double-check-char (char)
  "Double-check that char at point is CHAR, enter correction mode if not.  
In the correction mode, you can enter one-character correction commands:

     RETURN   replace character at point  (prompts for confirmation)
     DEL      deletes character at point  (prompts for confirmation)
     i        insert a char before point  (prompts for confirmation)
     ?        gets help"
  (let ((point-char (char-after (point))))

    (if (= (upcase point-char) (upcase char))
        ;; User typed same as what's in sequence, so move on.
        (forward-char 1)
      
      ;; Else mismatch, so sound the bell and don't move forward:
      (ding)
      (gene-message
       "\"%c\" doesn't match \"%c\" (type ? for help)" char point-char))))



(provide 'gene-data-input)
