;;;; gene-dialog.el --- popping up dialog boxes 'n' stuff.
;; Copyright (C) 1994 RDP, Indiana University Foundation
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

(require 'gene-mode-tools)

;;;; What goes in this file?
;;
;; By "dialog", I mean "dialog with the user" in the general sense,
;; not just "dialog boxes".  So any sort of question/answer interface
;; could go in here.  But most of the code probably will be about
;; dialog-boxes.
;;
;; We need to decide some of the standard interface objects in dialog
;; boxes.  There are:
;;
;; - radio buttons   (choose 1)
;; - toggle buttons  (choose N)
;; - sliders         (slide the box: "-------------[*]--------------")
;; - text fields     ("Enter % mismatch:__________________")
;; - 



;;;; Code.

;;; Reading characters.

(defconst gene-read-char-confirmation-chars
  '(?\r ? ?y ?Y)
  "Characters treated as \"affirmative\" in confirmation requests.")


(defun gene-read-char (prompt &optional require-confirm read-func)
  "Read a single character, prompting with PROMPT.
PROMPT should be arranged so that the new char looks appropriate
tacked onto its end, for example \"My char: X\".

If optional second arg REQUIRE-CONFIRM is non-nil, then require that
the user type one of `gene-read-char-confirmation-chars' to confirm
it.  If she does not confirm, then an error will be signalled.

If REQUIRE-CONFIRM is a string, it is also used as a message to
display on non-confirmation.  The message may contain one \"%c\"
format constructs; it will be replaced with the character that was
read but not confirmed.

Optional third arg READ-FUNC is a function to use for reading
characters.  It defaults to `read-char'.  One useful alternative
might be `read-char-exclusive'."
  (message prompt)
  (let* ((cursor-in-echo-area t)
         (ch (funcall (or read-func 'read-char)))
         (confirmation  (if require-confirm
                            (progn
                              (message "%s%c" prompt ch)
                              (member (funcall (or read-func 'read-char))
                                      gene-read-char-confirmation-chars))
                          ;; Else confirmation is assumed affirmative
                          t)))
    (if confirmation
        ch
      (gene-error
       (if (stringp require-confirm) require-confirm "Not confirmed: %c")
       ch))))


(defun gene-y-or-n-p (prompt &optional minibuffer-prompt)
  "Like `y-or-n-p', but more insistent about using a popup box if possible.
If optional second argument MINIBUFFER-PROMPT is non-nil, it is a
string to use if prompting takes place in the minibuffer instead of in
a menu."
  (if (eq window-system 'x)
      (let* ((prompt-len (max 3 (- (length prompt) 2)))
             (left-of-center     (max 0 (1- (/ prompt-len 2))))
             (yes-str (let ((str (make-string prompt-len ? )))
                        (aset str left-of-center       ?y)
                        (aset str (+ left-of-center 1) ?e)
                        (aset str (+ left-of-center 2) ?s)
                        str))
             (no-str  (let ((str (make-string prompt-len ? )))
                        (aset str left-of-center       ?n)
                        (aset str (+ left-of-center 1) ?o)
                        str))
             (ans
              (x-popup-menu
               (list
                (list (/ (frame-pixel-width) 2) (/ (frame-pixel-height) 2))
                (selected-window))
               (list prompt
                     (list prompt
                           ""
                           (cons yes-str t) (cons no-str nil) "")))))
        ans)

    ;; Else just use minibuffer
    (y-or-n-p (or minibuffer-prompt prompt))
    ))



;;; Popping up dialog boxes.
;; 
;; How is this going to work?
;;
;; The two option fields we will support first are TEXT FIELDS and
;; Y-OR-N QUESTIONS.  They look sort of like this:
;;
;;     Foo?  Yes   No
;;     Bar:  ____________
;;
;; respectively (but without the underscores, I think).
;;
;; Anyway, covering each field will be a property,
;; `gene-dialog-field'.  It is a function that, when
;; invoked, returns a string that is the option generated by that
;; field.
;;
;; In addition, the Y-OR-N fields have mouse-sensitive
;; `gene-dialog-here-command' (function) properties.  The property
;; is invoked by any mouse-click, and runs an anonymous function which
;; knows about markers delimiting the text field and what text the
;; option wants.  So, for example:
;;
;; "Auto Generate? Yes  No"
;;
;; Clicking on Yes highlights it and unhighlights No.
;; Then later, when the user "sends" the command (hits OKAY or
;; whatever), the result of evaluating the
;; `gene-dialog-option-generator' function there will yield the string
;; "-autogen" for the program's options.
;;
;; This will work because when you clicked Yes, it changed some
;; property's value from "" to "-autogen".
;;
;; THOUGHTS: dialog boxes should have <HIDE> buttons, and if hidden
;; and then reshown, they should preserve information.  So you can
;; fill one in partially, hide it, and then come back to it later by
;; choosing the same dialog-box-up-popping command.
;; Also, after sending a command, might as well leave the dialog box
;; filled in, since the user will likely use pretty much the same
;; options on the next invocation.
;;
;; Think some more, then write.

(defun gene-dialog-here-command (event)
  "If `gene-dialog-here-command' text property is non-nil at EVENT, run it."
  (interactive "e")
  (let* ((point (posn-point (event-start event)))
         (cmd   (get-text-property point 'gene-dialog-here-command))
         (ov    (get-text-property point 'gene-dialog-radio-overlay)))

    ;; If this is a radio button (possibly boolean), then...
    (if ov
        (let* ((sf  (1- (gene-dialog-beginning-of-field point)))
               (ef  (1- (gene-dialog-end-of-field point)))
               (beg (previous-single-property-change
                     (1+ point) 'gene-dialog-radio-overlay nil sf))
               (end (next-single-property-change
                     point 'gene-dialog-radio-overlay nil ef)))
          (move-overlay ov beg end)))
    (if cmd
        (funcall cmd)
      ;; Else no command property, so just go to point
      (mouse-set-point event))))


(defun gene-dialog-beginning-of-field (&optional location property)
  "Return the point at which this field begins.
Optional first arg LOCATION is a location to search from.
Optional second arg PROPERTY is a property to use instead of
`gene-dialog-field'."
  (previous-single-property-change
   (or location (point)) (or property 'gene-dialog-field) nil 1))


(defun gene-dialog-end-of-field (&optional location property)
  "Return the point at which this field ends.
Optional first arg LOCATION is a location to search from.
Optional second arg PROPERTY is a property to use instead of
`gene-dialog-field'."
  (next-single-property-change
   (or location (point)) (or property 'gene-dialog-field) nil (point-max)))


(defun gene-dialog-field-bounds (&optional location property)
  "Return the pair \(start . end\) indicating this option field's location.
Optional first arg LOCATION is a location to search from.
Optional second arg PROPERTY is a property to use instead of
`gene-dialog-field'."
  (cons
   (gene-dialog-beginning-of-field location property)
   (gene-dialog-end-of-field       location property)))


(defun gene-dialog-move-to-previous-field ()
  "Move to the previous field in the dialog box."
  (interactive)
  (let* ((one (previous-single-property-change (point) 'gene-dialog-start))
         (two (previous-single-property-change   one   'gene-dialog-start)))
    (if (or two one)
        ;; We count on intangible property to make this work.
        (goto-char (or two one))
      (gene-message "No previous field."))))


(defun gene-dialog-move-to-next-field ()
  "Move to the next field in the dialog box."
  (interactive)
  (goto-char (gene-dialog-next-field)))


(defvar gene-dialog-box-mode-map (make-sparse-keymap))


(defun gene-dialog-setup-keymap (&optional testing)
  "Set up a dialog-box keymap in the current buffer.
Optional arg TESTING means only set mouse-bindings, not keyboard
bindings, so we can test things in a scratch buffer."
  (use-local-map gene-dialog-box-mode-map)
  (gene-menu-bar-make-hide-button 'gene-dialog-hide-dialog-box)

  (define-key
    gene-dialog-box-mode-map [down-mouse-1] 'gene-dialog-here-command)
  (define-key
    gene-dialog-box-mode-map [mouse-1]      'gene-absorb-mouse-event)
  (define-key
    gene-dialog-box-mode-map [drag-mouse-1] 'gene-absorb-mouse-event)
  (define-key
    gene-dialog-box-mode-map [down-mouse-2] 'gene-dialog-here-command)
  (define-key
    gene-dialog-box-mode-map [mouse-2]      'gene-absorb-mouse-event)
  (define-key
    gene-dialog-box-mode-map [drag-mouse-2] 'gene-absorb-mouse-event)
  (define-key
    gene-dialog-box-mode-map [down-mouse-3] 'gene-dialog-here-command)
  (define-key
    gene-dialog-box-mode-map [mouse-3]      'gene-absorb-mouse-event)
  (define-key
    gene-dialog-box-mode-map [drag-mouse-3] 'gene-absorb-mouse-event)

  (if testing
      nil
    (define-key gene-dialog-box-mode-map "\n" (lambda ()
                                            (interactive)
                                            (funcall
                                             gene-dialog-local-command)))
    (define-key gene-dialog-box-mode-map "\r" (lambda ()
                                            (interactive)
                                            (funcall 
                                             gene-dialog-send-button-command)))
    
    (define-key
      gene-dialog-box-mode-map "\C-i" 'gene-dialog-move-to-next-field)
    (define-key
      gene-dialog-box-mode-map [up] 'gene-dialog-move-to-previous-field)
    (define-key
      gene-dialog-box-mode-map [down] 'gene-dialog-move-to-next-field)
    ;; (define-key gene-dialog-box-mode-map "\C-p" 'gene-dialog-previous-field)
    ;; (define-key gene-dialog-box-mode-map "\C-n" 'gene-dialog-next-field)
    ))


(defun gene-dialog-box-mode (&optional command)
  "Major Mode for dialog box buffers.
Optional argument COMMAND is a function to use as the value of the
local variable `gene-dialog-local-command'; it defaults to
`gene-dialog-issue-system-command'."
  (gene-dialog-setup-keymap)
  (setq major-mode 'gene-dialog-box-mode)
  (setq mode-name "Dialog Box")
  (setq mode-line-format "%b")
  (make-local-variable 'gene-dialog-local-command)
  (setq
   gene-dialog-local-command
   (or command 'gene-dialog-issue-system-command))
  )



;; Faces for dialog-box parts:

(defvar gene-dialog-boolean-color "rosybrown"
  "Color for boolean buttons.")

(defvar gene-dialog-label-color "blue"
  "Color for labels.")

(defun gene-dialog-make-faces ()
  (make-face 'gene-dialog-label-face)
  (set-face-background 'gene-dialog-label-face gene-dialog-label-color)
  (make-face 'gene-dialog-boolean-face)
  (set-face-background 'gene-dialog-boolean-face gene-dialog-boolean-color)
  )



;; Yes/No Questions.

(defun gene-dialog-insert-boolean-field (label yes-option no-option default)
  "Insert a yes/no question.
LABEL is the question; YES-OPTION is the option string if the user
chooses YES; NO-OPTION likewise; and DEFAULT is `yes'  or `no'.
The options will be padded with spaces on the right and left.
\(For DEFAULT, anything other than `yes' counts as `no'.\)"
  (gene-dialog-insert-radio-buttons
   label
   (if (eq default 'yes) "YES" "NO")
   (cons "YES" yes-option)
   (cons "NO" no-option)
   ))


(defun gene-dialog-insert-radio-buttons (label default &rest options)
  "Insert radio buttons.
LABEL is the question.
DEFAULT is a string identifying the radio-button that is on by default.
OPTIONS are the actual buttons.
Each OPTION is of the form

       \(NAME . OPTION\)

where OPTION is the option string obtained by clicking on button NAME.
OPTIONS will be padded with a space on both sides."
  (let* ((sym      (gensym "gene-dialog-boolean-"))
         (ov       (make-overlay 1 1)))

    ;; Color the overlay.
    (overlay-put ov 'face 'gene-dialog-boolean-face)

    ;; Insert label.
    (gene-dialog-insert-label
     label `(lambda () ,sym) 'gene-dialog-label-face t)

    ;; Insert choices.
    (mapcar
     (lambda (option-pair)
       (let* ((name  (car option-pair))
              (opt   (cdr option-pair))
              (func  `(lambda () 
                        (set (quote (, sym)) (concat " " ,opt " "))))
              (str   (concat " " name))
              (len   (1+ (length name)))
              (label (progn
                       (put-text-property
                        1 len 'gene-dialog-here-command func str)
                       (put-text-property
                        1 len 'mouse-face 'gene-mouse-face str)
                       (put-text-property
                        1 len 'gene-dialog-radio-overlay ov str)
                       str))
              opoint)
         (setq opoint (1+ (point)))
         (gene-dialog-insert-label label (lambda () "") nil t)
         ;; If this is the default option, make it so.
         (if (string-equal name default)
             (progn
               (set sym (concat " " opt " "))
               (move-overlay ov opoint (1- (+ opoint len)))))))
     options)))



;; Fields and labels.


(defun gene-dialog-text-field-text-end ()
  "Return the end of the text field at point.
We search for the beginning of another field, or a newline, whichever
comes first."
  (let* ((here-prop (get-text-property (point) 'gene-dialog-field))
         (eol       (gene-eol-point))
         (next      (text-property-not-all
                     (point) eol 'gene-dialog-field here-prop)))
    (or next eol)))


(defun gene-dialog-insert-text-field (label option &optional default-text)
  "Insert a text-field into this dialog box at point.
The text field's name will be LABEL, and it controls OPTION \(a string\).
LABEL will be padded on the right with a space.
It will control option OPTION thusly: when the dialog box's options
are being compiled, this text field will produce the option
`\(concat \" \" OPTION TEXT\)', where TEXT is the text in the field at the
moment of option compilation."
  (let* ((mkr (point-marker))
         (opt-func
          `(lambda ()
             (goto-char ,mkr)
             ;; We depend on the intangible property skipping us to
             ;; the right place automatically.
             (goto-char (gene-dialog-beginning-of-field))
             (let ((str (buffer-substring-no-properties
                         (point) (gene-dialog-text-field-text-end))))
               (concat " " (, option) str)))))
    (gene-dialog-insert-label label opt-func 'gene-dialog-label-face)
    (if default-text (insert default-text))
    ))


(defun gene-dialog-insert-label
  (name option-func &optional face no-landing)
  "Insert a field label NAME at point, with option function OPTION-FUNC.
NAME will be padded on the right with a space.
OPTION-FUNC is a function of no arguments, to be called when options
are being compiled.
The fieldname will be `intangible' and read-only.
Optional third arg FACE is a face for the name.
Optional fourth arg NO-LANDING means this is not a field as far as
movement commands are concerned -- they should skip over it."
  (if (zerop (length name))
      (gene-error "Cannot insert an empty label.")
    (let ((opoint (point)))
      (insert name " ")
      (unless no-landing
       (put-text-property opoint (1+ opoint) 'gene-dialog-start t))
      (put-text-property (1- (point)) (point) 'rear-nonsticky t)
      (put-text-property opoint (point) 'intangible t)
      (put-text-property opoint (point) 'gene-dialog-field option-func)
      (if face
          (put-text-property opoint (1- (point)) 'face face))
      ;; Put the read-only property last.  We could get around this by
      ;; binding `inhibit-read-only', but Oy Vey.
      ;; We use gensym for the read-only property because we want any
      ;; adjacent read-only regions to have _different_ non-nil values
      ;; for this property, thus allowing insertion between them.  See
      ;; `Special Properties' in Emacs Lisp Info.
      (put-text-property opoint (point) 'read-only (gensym)))))


;; Turned out that Jim's version of above had been sitting here all
;; along, duh.  Why I didn't look I don't know.
;;
;; (defun gene-insert-label (name)
;;   "Insert a label for a form.
;; The label has its read-only and intangible properties set up to make
;; it behave appropriately for a label in a form --- you can put the
;; cursor at its start or end, but you cannot put the cursor inside it.
;; It cannot be deleted, entirely or partially, and you can't insert in
;; it (because you can't put the cursor in it)."
;;   (if (zerop (length name))
;;       (error "insert-label: name can't be the empty string."))
;;   (let ((start (point)))
;;     (insert name)
;;     (put-text-property (1+ start) (point) 'intangible t)
;;     (put-text-property (1- (point)) (point) 'rear-nonsticky t)
;;     (put-text-property start (point) 'read-only t)))


(defun gene-dialog-popup-dialog-box (name &rest fields)
  "Popup a dialog box named NAME.
FIELDS are args indicating what this dialog box does, in some
as-yet-to-be-determined format."
  (setq name (buffer-name (generate-new-buffer name)))

  (save-excursion
    (set-buffer name)
    (gene-dialog-setup-keymap)
    

    ;; ... set up the dialog box here ...
    
    ;; Hmm, do we want some way to lock out everything but the dialog
    ;; box?  No, I can't see any reason to do that.  User can go ahead
    ;; and edit, ignoring the box, maybe filling in a field now and
    ;; then.  There will a button that "sends" it when done; they can
    ;; click that at the end.
    
    (gene-frame-show name)
    )
  )



;; Compiling the options.

;; why does this affect point in the test case???
(defun gene-dialog-compile-options ()
  "Return an option string compiled from this dialog-box buffer."
  (save-excursion
    (let ((spot (point-min))
          (end  (point-max))
          options)
      (while (setq 
              spot 
              (text-property-not-all spot end 'gene-dialog-field nil))
        (setq
         options
         (cons (funcall (get-text-property spot 'gene-dialog-field)) options))
        (setq spot
              (or
               (text-property-any spot end 'gene-dialog-field nil)
               end)))
      (let ((str (apply 'concat (nreverse options))))
        (set-text-properties 0 (length str) nil str)
        str))))




;; Sending the command

(defun gene-dialog-issue-system-command (command)
  "Issue COMMAND to the system.  It should include any options; i.e.:
`gene-dialog-compile-options' should have been called already."
  (ding)
  (gene-message command)
  (ding))


(defun gene-dialog-hide-dialog-box (&optional dialog-box-buffer)
  "Hide DIALOG-BOX-BUFFER \(defaults to the current buffer\)."
  (interactive)
  (or dialog-box-buffer (setq dialog-box-buffer (current-buffer)))
  (make-frame-invisible
   (window-frame (get-buffer-window dialog-box-buffer t))))


(defvar gene-dialog-send-button-command nil
  "The function that would be run if the user clicked on the send button.
`gene-dialog-insert-send-button' sets this after constructing the command.")


;; The local variable `gene-dialog-local-command' had better be set in
;; each dialog box:
(defun gene-dialog-insert-send-button (label command)
  "Insert button LABEL to send COMMAND \(a string\).
Options will be compiled and concatenated after COMMAND, and then
whatever function is the value of `gene-dialog-local-command' will be
passed that string as argument."
  (let ((len    (length label))
        (lbl    (copy-sequence label))
        (newcom `(lambda ()
                   (interactive)
                   (funcall
                    gene-dialog-local-command
                    (concat ,command " " (gene-dialog-compile-options)))
                   (gene-dialog-hide-dialog-box))))
    (setq gene-dialog-send-button-command newcom)
    (put-text-property 0 len 'gene-dialog-here-command newcom lbl)
    (put-text-property 0 len 'mouse-face 'gene-mouse-face lbl)
    (gene-dialog-insert-label lbl nil nil t)
   ))



;;; Slider bars.

;; The Plan:
;;
;; Define what a slider is, then when a slider is created, it is
;; covered with `gene-slider' text-property, which value is the
;; slider.
;; The Slider data structure will indicate START and END of slider,
;; RANGE of numbers it is to cover, and a NUMBER-FIELD which it is
;; controlling.
;;
;; Someday.

(defun gene-dialog-drag-horiz-slider (evt &optional func)
  "Drag the mouse horizontally in current frame, ignore vertical motion.
Specifically, start dragging at event EVT, and at each new position,
call FUNC with the new buffer position as argument.  Do not go past
bol or eol."
  (interactive "e")
  (save-excursion
    (let* ((posn      (event-start evt))
           (start-pos (posn-point posn)))
      (goto-char start-pos)
      (let ((bol      (save-excursion (beginning-of-line) (point)))
            (eol      (save-excursion (end-of-line) (point))))
        ;; Do the right stuff here.
        ))))


;; What happens to the POSN as the mouse pointer hits borders?
;; when go out left:   PT becomes nil, COL becomes negative
;;             right:  PT becomes nil, COL goes higher than window-width
;;             top:    PT becomes nil, ROW goes negative
;;             bottom: Ugh.  PT becomes `mode-line', then 1 (when enter
;;                     minibuffer, at which time ROW also becomes 0 of
;;                     course), then nil again.  ROW just keeps
;;                     increasing past bottom row, except for a brief
;;                     interlude in the minibuffer, when it's 0.


(defun gene-dialog-drag-with-no-scroll (event func)
  "Given a click EVENT, track the mouse, applying FUNC to each position.
Return when the mouse button pressed in EVENT is released.

FUNC should be a function of three arguments, PT, ROW, and COL,
which specify the current position of the mouse.

If the mouse moves out of the window, then do nothing."
  (save-excursion
    (let* ((o-window   (posn-window (event-start event))))
      (track-mouse
        (while (mouse-movement-p (setq event (read-event)))
          (let* ((posn     (event-start event))
                 (window   (posn-window posn))
                 (pt       (posn-point posn))
                 ;; What's the difference between this and `posn-x-y'?
                 (col-row  (posn-col-row posn))
                 (col      (car col-row))
                 (row      (cdr col-row)))
            (and
             (numberp pt)
             (funcall func pt col row)))))
      (message "Done dragging."))))

;; (define-key gene-dialog-box-mode-map [down-mouse-3] 'kf-test-drag-with)


(defun gene-popup-alternatives (prompt alternatives)
  "Pop up a menu labeled with PROMPT, offering ALTERNATIVES.
PROMPT should be a string.
ALTERNATIVES may be a list of strings or symbols.
Return the element of ALTERNATIVES chosen."
  (let ((pane (cons prompt
		    (mapcar (lambda (alternative)
			      (cons (if (symbolp alternative)
					(symbol-name alternative)
				      alternative)
				    alternative))
			    alternatives))))
    (x-popup-menu t (list "x-popup-alternatives" pane))))




(defun gene-popup-checkbox (position label items)
  "Let user select a subset of the items on a pop-up menu.
POSITION specifies the position of the menu, as for x-popup-menu.
LABEL is the label to place at the top of the item list.
ITEMS is an association list mapping names to appear in the menu onto
symbols to represent them in the return value.
The return value is an association list mapping the cdrs of the pairs
in ITEMS onto t or nil, indicating whether each one was selected."
  (let* ((done-cookie (list 'done))
	 (pane (append (list label)
		       (mapcar (lambda (item)
				 (let ((name (concat "  " (car item))))
				   (list name name (cdr item))))
			       items)
		       (list ""
			     (cons "  Done" done-cookie))))
	 (menu (list "a meaningless title" pane)))
    (while (let ((choice (x-popup-menu position menu)))
	     (cond
	      ((null choice) t)
	      ((eq choice done-cookie) nil)
	      (t (let ((label (car choice)))
		   (aset label 0 (if (= (aref label 0) ?*) ?\  ?*)))
		 t))))
    (delq done-cookie
	  (mapcar (lambda (item)
		    (if (and (consp item)
			     (= (length item) 3))
			(cons (elt item 2) (= (aref (car item) 0) ?*))
		      done-cookie))
		  pane))))


;; (gene-popup-checkbox t "dog food" '(("crunchy" . crunchy)
;;                                     ("smelly" . smelly)
;;                                     ("bright red" . bright-red)))



;; These three functions were suggested by Jim, and should probably
;; exist when a more complete dragging library is done.

;; (defun highlight-dragged-to (posn anchor overlay)
;;   "Highlight up to POSN, from ANCHOR, using OVERLAY for display.
;; If POSN is not visible in the current window, scroll until it is."
;;   ...)
;; 
;; (defun enclosed-words (start end)
;;   "Return the smallest region on word boundaries containing START ... END.
;; This is useful when dragging a word at a time."
;;   ...)
;; 
;; (defun enclosed-lines (start end)
;;   "Return the smallest region on line boundaries containing START ... END.
;; This is useful when dragging a line at a time."
;;   ...)



(provide 'gene-dialog)
