;; gene-colors.el (part of ale).
;; Copyright (C) 1993, 1994  Karl Fogel <kfogel@darwin.life.uiuc.edu>
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;;
;; Search for instances of "todo:" to find notes about future plans.



(require 'cl)
(require 'gene-mode-tools)
(require 'disp-table)


;;;; Code:
;;; Residue Colors and Display Table fiddling

(defstruct gene-color-set

  ;; This color set's name.
  name

  ;; The display table --- a vector made by make-display-table.
  display-table

  ;; Non-nil means the set's characters' backgrounds are colored,
  ;; instead of the foregrounds.
  inverted-p

  ;; The background and foreground text colors --- nil if the default
  ;; is fine.
  background foreground)

(defvar gene-color-sets nil
  "A list of the residue-coloring sets we support.")


(defun gene-make-color-set (name)
  "Create a color set named NAME, with our favorite characters installed."
  (let ((table (make-display-table)))
    ;; Get rid of those annoying $ signs, use a space instead.
    (set-display-table-slot table 'truncation ? )
    ;; Same with vertical window borders
    (set-display-table-slot table 'vertical-border ? )

    (make-gene-color-set :display-table table :name name)))


(defun gene-lookup-color-set (name &optional no-create)
  "Return the display table for the color set named NAME.
If no such color set is defined, and optional second arg NO-CREATE is
nil, make and return a new one named NAME."
  (let ((set (catch 'set
               (mapcar
                (lambda (cset)
                  (if (string-equal name (gene-color-set-name cset))
                      (throw 'set cset)))
                gene-color-sets)
               nil)))
    (or
     set
     ;; Else we're creating a new color set, unless NO-CREATE:
     (unless no-create
       (let ((color-set (gene-make-color-set name)))
         (push color-set gene-color-sets)
         color-set)))))


(defun gene-current-color-set ()
  "Return the currently active color-set."
  (catch 'set
    (mapcar
     (lambda (cset)
       (if (eq buffer-display-table (gene-color-set-display-table cset))
           (throw 'set cset)))
     gene-color-sets)
    nil))


(defun gene-face-from-id (face-id)
  "Return the face (a symbol) whose id number is FACE-ID, or nil if none.
Why doesn't this come with Emacs?"
  (let ((faces (mapcar 'internal-get-face (face-list))))
    (catch 'face
      (mapcar
       (lambda (face-thing)
         (let ((this-id (face-id face-thing)))
           (if (= face-id this-id)
               (throw 'face (face-name face-thing)))))
       faces)
      nil)))


;; Mmm, thanks for writing this, Jim.
;; I'm embarrassed to admit that I did.  You'd think Emacs would provide this
;;   in a library somewhere, but no, we'll just hard-code the fact that the
;;   lower nineteen bits of a glyph code are the character, while the upper...
(defun gene-make-glyph (character face)
  "Return a glyph code which displays CHARACTER using FACE."
  (+ character (lsh (face-id face) 19)))


(defun gene-color-character-internal
  (color-set character color &optional background)
  "Internal: in COLOR-SET, color CHARACTER with COLOR.
The difference between this and `gene-color-character' 
is that the latter is more relaxed about quoting.
Optional third arg BACKGROUND means color CHARACTER's background,
making the foreground black."
  (setq color (gene-check-color-allowable color))
  (let ((table (gene-color-set-display-table
		(gene-lookup-color-set color-set))))
    (setq character (string-to-char character))
    (let ((face (intern (format "gene-%s-face" color))))
      (make-face face)
      (if (not background)
          (set-face-foreground face (gene-check-color-allowable color))
        (set-face-background face (gene-check-color-allowable color))
        (set-face-foreground face (gene-check-color-allowable "black")))
      (aset table
	    character
	    (vector (gene-make-glyph character face))))))


(defun gene-color-background-internal (color-set color)
  "Internal: in COLOR-SET, use background color COLOR.
The difference between this and `gene-color-background' 
is that the latter is more relaxed about quoting."
  (setq color (gene-check-color-allowable color gene-background-color))
  (unless (eq window-system 'x)
    (error "Must be running under X Windows."))
  (let ((set (gene-lookup-color-set color-set)))
    (setf (gene-color-set-background set) color)))


(defun gene-color-foreground-internal (color-set color)
  "Internal: in COLOR-SET, use foreground color COLOR.
The difference between this and `gene-color-foreground' 
is that the latter is more relaxed about quoting."
  (setq color (gene-check-color-allowable color))
  (unless (eq window-system 'x)
    (error "Must be running under X Windows."))
  (let ((set (gene-lookup-color-set color-set)))
    (setf (gene-color-set-foreground set) color)))


(defun gene-reverse-color-set (&optional color-set)
  "\"Swap\" fore- and background colors of characters in COLOR-SET.
COLOR-SET defaults to the current color set.
Specifically, if the foreground of char X is colored blue, then make
X's background be blue and foreground be black.  Or vice-versa, make
X's foreground be its background \(blue\), and background be
COLOR-SET's background color."
  (or color-set (setq color-set (gene-current-color-set)))
  (gene-message "Reversing display...")
  (let ((foreground     (gene-color-set-foreground color-set))
        (background     (gene-color-set-background color-set))
        (d-table        (gene-color-set-display-table color-set))
        (inverted-p     (gene-color-set-inverted-p color-set))
        (visited-faces  nil)
        (tmp-color      nil)
        (i              0))
    (while (< i 256)
      (let ((entry (aref d-table i)))
        (if entry
            (let* ((val  (aref (aref d-table i) 0))
                   (id   (lsh val -8))
                   (face (gene-face-from-id id)))
              (if (memq face visited-faces)
                  nil
                ;; Else haven't seen this face before, so [re]invert.
                (if inverted-p
                    (progn
                      (set-face-foreground face (face-background face))
                      (set-face-background face "black")
                      ;; (setf tmp-color
                      ;; (gene-color-set-foreground color-set)
                      ;; (gene-color-set-foreground color-set)
                      ;; (gene-color-set-background color-set)
                      ;; (gene-color-set-background color-set)
                      ;; tmp-color)
                      (setf (gene-color-set-inverted-p color-set) nil))
                  ;; Else the set was normal, so invert it.
                  (set-face-background face (face-foreground face))
                  (set-face-foreground face "black")
                  ;; (setf tmp-color
                  ;; (gene-color-set-foreground color-set)
                  ;; (gene-color-set-foreground color-set)
                  ;; (gene-color-set-background color-set)
                  ;; (gene-color-set-background color-set)
                  ;; tmp-color)
                  (setf (gene-color-set-inverted-p color-set) t))
                (setq visited-faces (cons face visited-faces))))))
      (setq i (1+ i)))
    (gene-message "Reversing display... done.")))


(defun gene-copy-color-set-internal (source dest)
  "Internal: Copy color set SOURCE to color set DEST, defining DEST if needed.
SOURCE must be an existing color set name; DEST can be an existing
color set name or a new one.  If it is an existing color set, the
colors in SOURCE will override the ones already defined in DEST."
  (let ((src-set  (gene-lookup-color-set source t))) ; pass NO-CREATE flag
    (if src-set
        (let* ((dest-set     (gene-lookup-color-set dest))
               (src-table    (gene-color-set-display-table src-set))
               (dest-table   (gene-color-set-display-table dest-set)))
          (map-char-table (lambda (range value)
                            (set-char-table-range dest-table range value))
                          src-table))
      ;; Else SOURCE didn't exist.  No big deal, but user should know.
      (gene-err-accumulate-error
       "Cannot copy from non-existent color set \"%s\"." source)
      )))


(defmacro gene-use-color-set (color-set)
  "Use color set COLOR-SET."
  (list 'gene-use-color-set-internal
	(if (symbolp color-set) (symbol-name color-set) color-set)))


(defun gene-use-color-set-internal (color-set)
  "Use COLOR-SET to display the characters in the alignment."
  ;; Emptive Note: see Pre-emptive Note.  
  ;; Pre-emptive Note: Please never modify this function to check if
  ;; COLOR-SET is already being used and do nothing if so.  Sometimes
  ;; it's necessary to call it with the current color-set, if changes
  ;; have been made to that set.
  (let* ((color-set (gene-lookup-color-set color-set))
	 (table     (gene-color-set-display-table color-set))
	 (foreground (gene-color-set-foreground color-set))
	 (background (gene-color-set-background color-set)))
    (save-excursion
      (set-buffer gene-seq-buffer)
      (setq buffer-display-table table)
      (let (params1 params2)
	(when foreground
	  (push (cons 'foreground-color
		      (setq gene-foreground-color foreground))
		params1)
	  (push (cons 'cursor-color (setq gene-cursor-color foreground))
		params2)
	  (push (cons 'mouse-color foreground)
		params2))
	(when background
	  (push (cons 'background-color
		      (setq gene-background-color background))
		params1))
	(let ((window (get-buffer-window (current-buffer))))
	  (if window
	      (let ((frame (window-frame window)))
		;; For some reason, Emacs insists we set the cursor
		;; color separately.
		(modify-frame-parameters frame params1)
		(modify-frame-parameters frame params2)))))

      ;; This gets the display table change noticed.
      (force-mode-line-update))))


(defun gene-generate-color-set-menu ()
  "Generate a section of the Display menu listing the available color sets."
  
  (let*

      ;; What is the current color set?
      ((current (progn (set-buffer gene-seq-buffer) buffer-display-table))

       ;; Sort the list of color sets alphabetically.
       (sorted (sort (copy-list gene-color-sets)
		     (lambda (entry1 entry2)
		       (string-lessp
                        (gene-color-set-name entry1)
                        (gene-color-set-name entry2)))))

       ;; Generate a list of menu entries.
       (entries
	(mapcar (lambda (set)
		  (let* ((name (gene-color-set-name set))
			 (table (gene-color-set-display-table set)))
		    (let ((closure (list 'lambda nil
					 '(interactive)
					 (list 'gene-use-color-set
					       (list 'quote name)))))
		      (cons (format "%c %s" 
				    (if (eq table current) ?* ?\ ) 
				    name)
			    closure))))
		sorted)))

    ;; Stick a title at the top.
    (cons '"Color Sets:" entries)))


;;; Macros for users to invoke in their .ale files.
(defmacro gene-color-character (color-set character color)
  "In COLOR-SET, color CHARACTER with COLOR.
COLOR-SET can be one of the names appearing on the Display menu, or a
new name, in which case you'd be defining a new color set \(see also
`gene-copy-color-set'\).
Here are the color sets predefined by Ale:
    Nucleic
    Amino
    Uncolored
For example, here's how to make A's in RNA sequences green:
    (gene-color-character Nucleic A green)

See the manual for more details."
  (list 'gene-color-character-internal
	(if (symbolp color-set) (symbol-name color-set) color-set)
	(if (symbolp character) (symbol-name character) character)
	(if (symbolp color) (symbol-name color) color)))

(eval-when-compile
  (defalias 'gene-color-residue 'gene-color-character))


(defmacro gene-color-character-background (color-set character color)
  "In COLOR-SET, color CHARACTER's background with COLOR.
COLOR-SET can be one of the names appearing on the Display menu, or a
new name, in which case you'd be defining a new color set \(see also
`gene-copy-color-set'\).
Here are the color sets predefined by Ale:
    Nucleic
    Amino
    Uncolored
For example, here's how to make A's background in RNA sequences green:
    (gene-color-character Nucleic A green)

See the manual for more details."
  (list 'gene-color-character-internal
	(if (symbolp color-set) (symbol-name color-set) color-set)
	(if (symbolp character) (symbol-name character) character)
	(if (symbolp color) (symbol-name color) color)
        t))

(eval-when-compile
  (defalias 'gene-color-residue-background 'gene-color-character-background))


(defmacro gene-color-background (color-set color)
  "In COLOR-SET, use background color COLOR.
COLOR-SET can be one of the names appearing on the Display menu, or a
new name, in which case you'd be defining a new color set \(see also
`gene-copy-color-set'\).
Here are the color sets predefined by Ale:
    Nucleic
    Amino
    Uncolored
For example, here's how to use a white background in RNA sequences:
    (gene-color-background Nucleic white)"
  (list 'gene-color-background-internal
	(if (symbolp color-set) (symbol-name color-set) color-set)
	(if (symbolp color) (symbol-name color) color)))


(defmacro gene-color-foreground (color-set color)
  "In COLOR-SET, use foreground color COLOR.
COLOR-SET can be one of the names appearing on the Display menu, or a
new name, in which case you'd be defining a new color set \(see also
`gene-copy-color-set'\).
Here are the color sets predefined by Ale:
    Nucleic
    Amino
    Uncolored
For example, here's how to use a white foreground in RNA sequences:
    (gene-color-foreground Nucleic white)"
  (list 'gene-color-foreground-internal
	(if (symbolp color-set) (symbol-name color-set) color-set)
	(if (symbolp color) (symbol-name color) color)))


(defmacro gene-copy-color-set (source dest)
  "Copy color set SOURCE to color set DEST, defining DEST if needed.
SOURCE must be one of the names appearing in the Display menu; DEST
can be the name of an existing color set or a new name.  If it is an
existing color set, the colors in SOURCE will override the ones
already defined in DEST."
  (list 'gene-copy-color-set-internal
	(if (symbolp source) (symbol-name source) source)
	(if (symbolp dest) (symbol-name dest) dest)))


;;; Leave this around for backward compatibility.
(defmacro gene-define-color-set (&rest dummy)
  "Obsolete.  Remove uses of this."
  nil)


(defmacro gene-startup-with-color-set (color-set)
  "Come up using the colors in COLOR-SET.
COLOR-SET must be one of the default Ale color sets, or one you
defined yourself.  The default sets are:
    Nucleic
    Amino
    Uncolored"
  (list 'add-hook
        ''gene-seq-mode-hook
        (list
         'function
         (list
          'lambda '()
          (list
           'gene-use-color-set
           color-set)))))


;;; Some default color settings.

(defun gene-create-default-color-sets ()
  "Create the default color sets, Uncolored, Nucleic, and Amino."
  (gene-make-Uncolored-color-set)
  (gene-make-Nucleic-color-set)
  (gene-make-Amino-Acids-5-Group-color-set)
  (gene-make-Nucleic-GDE-color-set)

  ;; New sets from Niels
  (gene-make-Amino-Acids-Hydropathy-Scale-color-set)
  (gene-make-Amino-Acids-Default-color-set)
  (gene-make-Amino-Acids-Isoelectric-Point-color-set)
  (gene-make-DNA/RNA-Default-color-set)
  (gene-make-Purine/Pyrimidine-Highlight-color-set)
  )


(defun gene-make-Uncolored-color-set ()
  "Create the \"Uncolored\" color set."
  (gene-color-foreground "Uncolored" "white")
  (gene-color-background "Uncolored" "black")
  )


(defun gene-make-Nucleic-color-set ()
  "Create the \"Nucleic\" color set."
  ;; Nucleic acids:
  (gene-color-background Nucleic black)
  (gene-color-foreground Nucleic white)
  
  (gene-color-character Nucleic A gold)
  (gene-color-character Nucleic a gold)
  (gene-color-character Nucleic G yellow)
  (gene-color-character Nucleic g yellow)
  (gene-color-character Nucleic C cyan)
  (gene-color-character Nucleic c cyan)
  (gene-color-character Nucleic U "pale green")
  (gene-color-character Nucleic u "pale green")
  (gene-color-character Nucleic T "pale green")
  (gene-color-character Nucleic t "pale green")
  )


;;; GDE-style nucleic colors
(defun gene-make-Nucleic-GDE-color-set ()
  "Create the \"Nucleic, GDE-style\" color set."

  (gene-color-background "Nucleic, GDE-style" white)
  (gene-color-foreground "Nucleic, GDE-style" black)

  (gene-color-character "Nucleic, GDE-style" A red)
  (gene-color-character "Nucleic, GDE-style" C blue)
  (gene-color-character "Nucleic, GDE-style" T green)
  (gene-color-character "Nucleic, GDE-style" U green)
  (gene-color-character "Nucleic, GDE-style" G black)

  (gene-color-character "Nucleic, GDE-style" a red)
  (gene-color-character "Nucleic, GDE-style" c blue)
  (gene-color-character "Nucleic, GDE-style" t green)
  (gene-color-character "Nucleic, GDE-style" u green)
  (gene-color-character "Nucleic, GDE-style" g black))


(defun gene-make-Amino-color-set ()
  "Create the \"Amino\" color set."
  ;; Hydrophobic (nonpolar):
  (gene-color-background Amino black)
  (gene-color-foreground Amino white)
  
  (gene-color-character Amino G "gold")
  (gene-color-character Amino g "gold")
  (gene-color-character Amino A "gold")
  (gene-color-character Amino a "gold")
  (gene-color-character Amino V "gold")
  (gene-color-character Amino v "gold")
  (gene-color-character Amino L "gold")
  (gene-color-character Amino l "gold")
  (gene-color-character Amino I "gold")
  (gene-color-character Amino i "gold")
  (gene-color-character Amino P "gold")
  (gene-color-character Amino p "gold")
  (gene-color-character Amino F "gold")
  (gene-color-character Amino f "gold")
  (gene-color-character Amino M "gold")
  (gene-color-character Amino m "gold")
  (gene-color-character Amino W "gold")                    
  (gene-color-character Amino w "gold")                    
  (gene-color-character Amino C "gold")
  (gene-color-character Amino c "gold")
  ;; Basic
  (gene-color-character Amino K "yellow")
  (gene-color-character Amino k "yellow")
  (gene-color-character Amino R "yellow")
  (gene-color-character Amino r "yellow")
  (gene-color-character Amino H "yellow")
  (gene-color-character Amino h "yellow")
  ;; Acidic
  (gene-color-character Amino D "cyan")
  (gene-color-character Amino d "cyan")
  (gene-color-character Amino E "cyan")
  (gene-color-character Amino e "cyan")
  ;; Uncharged polar
  (gene-color-character Amino N "pale green")
  (gene-color-character Amino n "pale green")
  (gene-color-character Amino Q "pale green")
  (gene-color-character Amino q "pale green")
  (gene-color-character Amino S "pale green")
  (gene-color-character Amino s "pale green")
  (gene-color-character Amino T "pale green")
  (gene-color-character Amino t "pale green")
  (gene-color-character Amino Y "pale green")
  (gene-color-character Amino y "pale green")
  )


(defun gene-make-Amino-Acids-5-Group-color-set ()
  (gene-color-character "Amino Acids 5-Group" "-" "dim grey")
  (gene-color-character "Amino Acids 5-Group" "~" "grey")
  (gene-color-character "Amino Acids 5-Group" "." "grey51")
  (gene-color-character "Amino Acids 5-Group" G "cornsilk")
  (gene-color-character "Amino Acids 5-Group" g "cornsilk")
  (gene-color-character "Amino Acids 5-Group" A "cornsilk")
  (gene-color-character "Amino Acids 5-Group" a "cornsilk")
  (gene-color-character "Amino Acids 5-Group" V "cornsilk")
  (gene-color-character "Amino Acids 5-Group" v "cornsilk")
  (gene-color-character "Amino Acids 5-Group" L "cornsilk")
  (gene-color-character "Amino Acids 5-Group" l "cornsilk")
  (gene-color-character "Amino Acids 5-Group" I "cornsilk")
  (gene-color-character "Amino Acids 5-Group" i "cornsilk")
  (gene-color-character "Amino Acids 5-Group" P "cornsilk")
  (gene-color-character "Amino Acids 5-Group" p "cornsilk")
  (gene-color-character "Amino Acids 5-Group" F "PaleTurquoise1")
  (gene-color-character "Amino Acids 5-Group" f "PaleTurquoise1")
  (gene-color-character "Amino Acids 5-Group" Y "PaleTurquoise1")
  (gene-color-character "Amino Acids 5-Group" y "PaleTurquoise1")
  (gene-color-character "Amino Acids 5-Group" W "PaleTurquoise1")
  (gene-color-character "Amino Acids 5-Group" w "PaleTurquoise1")
  (gene-color-character "Amino Acids 5-Group" S "LightSteelBlue1")
  (gene-color-character "Amino Acids 5-Group" s "LightSteelBlue1")
  (gene-color-character "Amino Acids 5-Group" T "LightSteelBlue1")
  (gene-color-character "Amino Acids 5-Group" t "LightSteelBlue1")
  (gene-color-character "Amino Acids 5-Group" C "LightSteelBlue1")
  (gene-color-character "Amino Acids 5-Group" c "LightSteelBlue1")
  (gene-color-character "Amino Acids 5-Group" M "LightSteelBlue1")
  (gene-color-character "Amino Acids 5-Group" m "LightSteelBlue1")
  (gene-color-character "Amino Acids 5-Group" N "LightSteelBlue1")
  (gene-color-character "Amino Acids 5-Group" n "LightSteelBlue1")
  (gene-color-character "Amino Acids 5-Group" Q "LightSteelBlue1")
  (gene-color-character "Amino Acids 5-Group" q "LightSteelBlue1")
  (gene-color-character "Amino Acids 5-Group" D "light cyan")
  (gene-color-character "Amino Acids 5-Group" d "light cyan")
  (gene-color-character "Amino Acids 5-Group" E "light cyan")
  (gene-color-character "Amino Acids 5-Group" e "light cyan")
  (gene-color-character "Amino Acids 5-Group" K "RosyBrown1")
  (gene-color-character "Amino Acids 5-Group" k "RosyBrown1")
  (gene-color-character "Amino Acids 5-Group" R "RosyBrown1")
  (gene-color-character "Amino Acids 5-Group" r "RosyBrown1")
  (gene-color-character "Amino Acids 5-Group" H "RosyBrown1")
  (gene-color-character "Amino Acids 5-Group" h "RosyBrown1")
  )


(defun gene-make-Amino-Acids-Hydropathy-Scale-color-set ()
  "Create the Amino Acids Hydropathy Scale color set."
  ;; 
  ;; Shows hydrophobic residues in reddish, hydrophilic in blues
  ;;
  (gene-color-foreground "Amino Acids Hydropathy Scale" "white")
  (gene-color-background "Amino Acids Hydropathy Scale" "black")

  (gene-color-character "Amino Acids Hydropathy Scale" "-" "dim grey")
  (gene-color-character "Amino Acids Hydropathy Scale" "~" "grey")
  (gene-color-character "Amino Acids Hydropathy Scale" "." "grey51")
  (gene-color-character "Amino Acids Hydropathy Scale" "_" "grey51")
  (gene-color-character "Amino Acids Hydropathy Scale" G "#74aa8a")
  (gene-color-character "Amino Acids Hydropathy Scale" A "#b2964c")
  (gene-color-character "Amino Acids Hydropathy Scale" V "#f68108")
  (gene-color-character "Amino Acids Hydropathy Scale" L "#eb8513")
  (gene-color-character "Amino Acids Hydropathy Scale" I "#ff7f00")
  (gene-color-character "Amino Acids Hydropathy Scale" P "#52b4ac")
  (gene-color-character "Amino Acids Hydropathy Scale" F "#ce8d30")
  (gene-color-character "Amino Acids Hydropathy Scale" Y "#5ab1a4")
  (gene-color-character "Amino Acids Hydropathy Scale" W "#66ae99")
  (gene-color-character "Amino Acids Hydropathy Scale" S "#68ad96")
  (gene-color-character "Amino Acids Hydropathy Scale" T "#6bac93")
  (gene-color-character "Amino Acids Hydropathy Scale" C "#c69038")
  (gene-color-character "Amino Acids Hydropathy Scale" M "#b59549")
  (gene-color-character "Amino Acids Hydropathy Scale" N "#1cc5e2")
  (gene-color-character "Amino Acids Hydropathy Scale" Q "#1cc5e2")
  (gene-color-character "Amino Acids Hydropathy Scale" D "#1cc5e2")
  (gene-color-character "Amino Acids Hydropathy Scale" E "#1cc5e2")
  (gene-color-character "Amino Acids Hydropathy Scale" K "#11c8ee")
  (gene-color-character "Amino Acids Hydropathy Scale" R "#00ceff")
  (gene-color-character "Amino Acids Hydropathy Scale" H "#24c2da")
  )


(defun gene-make-Amino-Acids-Isoelectric-Point-color-set ()
  "Create the Amino Acids Isoelectric Point color set."
  ;;
  ;; Shows positive residues in reddish, negative in blues
  ;;
  (gene-color-foreground "Amino Acids Isoelectric Point" "white")
  (gene-color-background "Amino Acids Isoelectric Point" "black")

  (gene-color-character "Amino Acids Isoelectric Point" G "#66ae98")
  (gene-color-character "Amino Acids Isoelectric Point" A "#67ad97")
  (gene-color-character "Amino Acids Isoelectric Point" V "#66ae98")
  (gene-color-character "Amino Acids Isoelectric Point" L "#66ae98")
  (gene-color-character "Amino Acids Isoelectric Point" I "#67ad97")
  (gene-color-character "Amino Acids Isoelectric Point" P "#76a988")
  (gene-color-character "Amino Acids Isoelectric Point" F "#56b3a8")
  (gene-color-character "Amino Acids Isoelectric Point" Y "#5cb1a2")
  (gene-color-character "Amino Acids Isoelectric Point" W "#63af9b")
  (gene-color-character "Amino Acids Isoelectric Point" S "#5cb1a2")
  (gene-color-character "Amino Acids Isoelectric Point" T "#62af9c")
  (gene-color-character "Amino Acids Isoelectric Point" C "#49b7b5")
  (gene-color-character "Amino Acids Isoelectric Point" M "#5eb0a0")
  (gene-color-character "Amino Acids Isoelectric Point" N "#54b3aa")
  (gene-color-character "Amino Acids Isoelectric Point" Q "#5bb1a3")
  (gene-color-character "Amino Acids Isoelectric Point" D "#00ceff")
  (gene-color-character "Amino Acids Isoelectric Point" E "#0ec9f0")
  (gene-color-character "Amino Acids Isoelectric Point" K "#de8920")
  (gene-color-character "Amino Acids Isoelectric Point" R "#ff7f00")
  (gene-color-character "Amino Acids Isoelectric Point" H "#999e65")
  )


(defun gene-make-Amino-Acids-Default-color-set ()
  "Create the Amino Acids Default color set."
  (gene-color-foreground "Amino Acids Default" "white")
  (gene-color-background "Amino Acids Default" "black")

  (gene-color-character "Amino Acids Default" "-" "dim grey")
  (gene-color-character "Amino Acids Default" "~" "grey")
  (gene-color-character "Amino Acids Default" "." "grey51")

  ;; glycine  (stands alone)
  ;;
  (gene-color-character "Amino Acids Default" G "white")
  (gene-color-character "Amino Acids Default" g "white")
  ;;
  ;; Small hydrophobic:  (tend to be buried)
  ;;
  (gene-color-character "Amino Acids Default" A "light grey")
  (gene-color-character "Amino Acids Default" a "light grey")
  (gene-color-character "Amino Acids Default" V "light grey")
  (gene-color-character "Amino Acids Default" v "light grey")
  (gene-color-character "Amino Acids Default" L "light grey")
  (gene-color-character "Amino Acids Default" l "light grey")
  (gene-color-character "Amino Acids Default" I "light grey")
  (gene-color-character "Amino Acids Default" i "light grey")
  (gene-color-character "Amino Acids Default" M "light grey")
  (gene-color-character "Amino Acids Default" m "light grey")
  (gene-color-character "Amino Acids Default" C "light grey")
  (gene-color-character "Amino Acids Default" c "light grey")
  ;;
  ;; Large aromatic:  (tend to substitute only for each other)
  ;;
  (gene-color-character "Amino Acids Default" W "gold")                    
  (gene-color-character "Amino Acids Default" w "gold")                    
  (gene-color-character "Amino Acids Default" F "gold")
  (gene-color-character "Amino Acids Default" f "gold")
  (gene-color-character "Amino Acids Default" Y "gold")
  (gene-color-character "Amino Acids Default" y "gold")
  ;;
  ;; Positively charged: (basic; histidine is a bit different) 
  ;;
  (gene-color-character "Amino Acids Default" K "deep sky blue")
  (gene-color-character "Amino Acids Default" k "deep sky blue")
  (gene-color-character "Amino Acids Default" R "deep sky blue")
  (gene-color-character "Amino Acids Default" r "deep sky blue")
  (gene-color-character "Amino Acids Default" H "turquoise")
  (gene-color-character "Amino Acids Default" h "turquoise")
  ;;
  ;; Uncharged polar: (could substitute for +/- charged)
  ;;
  (gene-color-character "Amino Acids Default" A "pale green")
  (gene-color-character "Amino Acids Default" a "pale green")
  (gene-color-character "Amino Acids Default" N "pale green")
  (gene-color-character "Amino Acids Default" n "pale green")
  (gene-color-character "Amino Acids Default" Q "pale green")
  (gene-color-character "Amino Acids Default" q "pale green")
  (gene-color-character "Amino Acids Default" S "pale green")
  (gene-color-character "Amino Acids Default" s "pale green")
  (gene-color-character "Amino Acids Default" T "pale green")
  (gene-color-character "Amino Acids Default" t "pale green")
  (gene-color-character "Amino Acids Default" P "pale green")
  (gene-color-character "Amino Acids Default" p "pale green")
  ;;
  ;; Negatively charged: (acidic)
  ;;
  (gene-color-character "Amino Acids Default" D "salmon")
  (gene-color-character "Amino Acids Default" d "salmon")
  (gene-color-character "Amino Acids Default" E "salmon")
  (gene-color-character "Amino Acids Default" e "salmon")
  )


(defun gene-make-DNA/RNA-Default-color-set ()
  "Create the DNA/RNA Default color set."
  (gene-copy-color-set "Amino Acids Default" "DNA/RNA Default")

  (gene-color-foreground "DNA/RNA Default" "white")
  (gene-color-background "DNA/RNA Default" "black")

  (gene-color-character  "DNA/RNA Default" A "cornsilk")
  (gene-color-character  "DNA/RNA Default" a "cornsilk")
  (gene-color-character  "DNA/RNA Default" G "navajo white")
  (gene-color-character  "DNA/RNA Default" g "navajo white")
  (gene-color-character  "DNA/RNA Default" U "light cyan")
  (gene-color-character  "DNA/RNA Default" u "light cyan")
  (gene-color-character  "DNA/RNA Default" T "light cyan")
  (gene-color-character  "DNA/RNA Default" t "light cyan")
  (gene-color-character  "DNA/RNA Default" C "LightSteelBlue1")
  (gene-color-character  "DNA/RNA Default" c "LightSteelBlue1")
  )


(defun gene-make-Purine/Pyrimidine-Highlight-color-set ()
  "Create the Purine/Pyrimidine Highlight color set."
  (gene-copy-color-set
   "Amino Acids Default" "Purine/Pyrimidine Highlight")

  (gene-color-foreground "Purine/Pyrimidine Highlight" "white")
  (gene-color-background "Purine/Pyrimidine Highlight" "black")

  (gene-color-character  "Purine/Pyrimidine Highlight"  A "white")
  (gene-color-character  "Purine/Pyrimidine Highlight"  G "white")
  (gene-color-character  "Purine/Pyrimidine Highlight"  a "white")
  (gene-color-character  "Purine/Pyrimidine Highlight"  g "white")
  (gene-color-character  "Purine/Pyrimidine Highlight"  U "grey66")
  (gene-color-character  "Purine/Pyrimidine Highlight"  C "grey66")
  (gene-color-character  "Purine/Pyrimidine Highlight"  T "grey66")
  (gene-color-character  "Purine/Pyrimidine Highlight"  u "grey66")
  (gene-color-character  "Purine/Pyrimidine Highlight"  c "grey66")
  (gene-color-character  "Purine/Pyrimidine Highlight"  t "grey66")
  )



(provide 'gene-display)
