;;;; gene-error.el --- Displaying errors nicely, but noticeably.
;; Copyright (C) 1994 RDP, Indiana University Foundation
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


(require 'gene-mode-tools)


;;;; Code:

;; todo: perhaps this should pop up a dialog box later?
(defalias 'gene-error 'error)


(defvar gene-error-data nil
  "For safely debugging errors encountered in custom handlers.
If you are running the gene editor and get what looks like a custom
error message, then chances are you can examine this variable for more
information.")

;; todo: make a custom condition-case form that automates some of the
;; above?  See function `gene-sgroup-drag-select' for an example
;; of how condition-case is being used right now for cleanup.



;;; The accumulated error code.

(defconst gene-err-accumulated-error-buffer
  (get-buffer-create "*Warnings*")
  "Buffer in which non-fatal errors are accumulated and displayed.")


(defun gene-err-accumulate-error (error-str &rest format-args)
  "Add ERROR-STR to `gene-err-accumulated-error-buffer'.
A newline will be inserted after ERROR-STR.
\"Error\" is used here in a general sense; it does not necessarily
imply a non-local exit.  The error will not be displayed until you
call `gene-err-display-accumulated-errors'."
  (gene-save-excursion-in gene-err-accumulated-error-buffer
    (goto-char (point-max))
    (insert (apply 'format error-str format-args) "\n")))


(defun gene-err-clear-accumulated-errors ()
  "Clear the accumulated errors in `gene-err-accumulated-error-buffer'."
  (gene-save-excursion-in gene-err-accumulated-error-buffer
    (erase-buffer)))


(defun gene-err-accumulated-errors-p ()
  "Return non-nil iff there are accumulated errors to show."
  (gene-save-excursion-in gene-err-accumulated-error-buffer
    (/= 0 (buffer-size))))


(defun gene-err-display-accumulated-errors ()
  "If there are any accumulated errors, pop up a frame showing them.
If none, do nothing."
  (gene-save-excursion-in gene-err-accumulated-error-buffer
    (gene-menu-bar-make-hide-button
     'gene-err-hide-accumulated-errors))
  (if (gene-err-accumulated-errors-p)
      (gene-frame-show
       gene-err-accumulated-error-buffer
       `((foreground-color . ,gene-foreground-color)
         (background-color . ,gene-background-color)
         (cursor-color     . ,gene-cursor-color)
         (mouse-color      . ,gene-cursor-color)
         (unsplittable     . t)
         (menu-bar-lines   . 1)))))



(defun gene-err-hide-accumulated-errors (&optional noclear)
  "Hide the accumulated error frame, if it is showing.
Unless optional arg NOCLEAR is non-nil, clear out the accumulated error
list."
  (interactive)
  (unless noclear (gene-err-clear-accumulated-errors))    
  (gene-frame-hide gene-err-accumulated-error-buffer))



(provide 'gene-error)
