;;;; gene-io.el --- Reading and writing sequences (part of ale)
;; Copyright (C) 1993, 1994  Karl Fogel <kfogel@darwin.life.uiuc.edu>
;; Copyright (C) 1994 Indiana University Foundation
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;; needed: 
;; `gene-io-revert-sequence' -- revert this sequence to the version in
;; the database.
;;
;; What else?
;; Throw away the multi-source support.  It's a mess, confusing to the
;; user, and not really what's wanted anyway.


;;;; Requirements

(require 'cl)

(require 'gene-message)
(require 'gene-frame)
(require 'gene-mode-tools)


;;;; Self-erasing temporary files

(defvar gene-io-temp-file-dir
  (cond
   ((file-directory-p "/tmp")     "/tmp")
   ((file-directory-p "/usr/tmp") "/usr/tmp")
   (t (expand-file-name "~")))
  "*Directory in which gene-mode should put temporary files.")

(defvar gene-io-temp-files nil
  "A list of the temporary files we've created and not yet cleaned up.")

(defvar gene-io-temp-file-number-start 0
  "The smallest number to use in temp file names.
We use this to make sure that the names generated within a
`gene-io-with-not-so-temp-files' form never conflict with other temp
file names.  After generating a bunch of not-so-temporary filenames, we
set this to the largest number we used in their names, ensuring that
all temporary filenames created thereafter use larger numbers than the
not-so-temporary ones we returned.")

(defun gene-io-clean-temp-files (until)
  "Erase files from `gene-io-temp-files' until we hit a cdr eq to UNTIL.
If UNTIL is a tail of the list `gene-io-temp-files', the file names it
contains will not be touched."
  (while (and (consp gene-io-temp-files)
              (not (eq gene-io-temp-files until)))
    (condition-case data
        (delete-file (car gene-io-temp-files))
      (file-error))
    (setq gene-io-temp-files (cdr gene-io-temp-files))))

(defmacro gene-io-with-temp-files (&rest forms)
  "Evaluate FORMS, and then erase the temporary files they created.
This only erases those temporary files whose names were generated using
`gene-io-temp-file-name'; all others are left alone."
  ;; Hee hee.  The last sentence in the doc string sounds like it's
  ;; describing a very attractive feature.  We will quietly ignore the
  ;; fact that doing it any other way would have been very difficult.
  ;; Smile.  -kff

  ;; It only sounds like a very attractive feature if you *want* files
  ;; to be left alone.  -JimB

  ;; Well, that's not quite what I was thinking.  I say that it
  ;; "sounds like it's describing a very attractive feature", not
  ;; because it is or isn't an attractive feature, but because of the
  ;; language used in describing the feature.  In other words, the
  ;; keyword is "sounds", not "attractive feature".
  ;;
  ;; Usually (i.e.: in advertising), a meaningless phrase is only
  ;; added on if it is thought to make the product more attractive
  ;; than it was without the phrase.  For example "Less fattening!",
  ;; or "Enhanced Safety Features!".  They don't really tell the
  ;; consumer anything interesting, but they're designed to give off
  ;; an aura of Quality.  Of course, they all end up meaning
  ;; absolutely nothing, partly because they don't say anything and
  ;; partly because the stock phrases are so over-used anyway.
  ;;
  ;; Now, I'm not claiming that this is the spirit in which you wrote
  ;; "all others are left alone".  It's just the spirit in which I
  ;; read it.  If the sentence had ended without the phrase "all
  ;; others are left alone", it would have contained exactly the same
  ;; amount of information as it does now.  Since it was pretty much a
  ;; null-content tail, I was instinctively kicked into
  ;; "anti-advertising" mode, in which I am hypersensitive to
  ;; meaningless phrases.  I don't think all meaningless phrases
  ;; should be omitted from our writing, nor am I advocating that this
  ;; particular phrase be deleted from the doc string.  I'm just
  ;; saying it has a bit of the familiar advertising ring to it
  ;; (unintended, no doubt), and that's what spurred me to write my
  ;; original comment. 
  ;;
  ;; *Whew*!  Aren't you glad we've got that settled?  You can beat a
  ;; dead horse with water, but you can't make him drink... -kff

  (list 'let '((gene-io-temp-file-marker gene-io-temp-files))
        (list 'unwind-protect (cons 'progn forms)
              '(gene-io-clean-temp-files gene-io-temp-file-marker))))

;;; Tell lisp-mode how to indent expressions using gene-io-with-temp-files.
(put 'gene-io-with-temp-files 'lisp-indent-function 0)


(defmacro gene-io-with-not-so-temp-files (&rest forms)
  "Evaluate FORMS; return the names of the temporary files created.
Any temp files created within the scope of `gene-io-with-temp-files'
forms are not included in this list.  This form may not be used within
the dynamic scope of any `gene-io-with-temp-files' forms."
  (list 'progn
	'(if (boundp 'gene-io-temp-file-marker)
	     (error "gene-io-with-not-so-temp-files: don't nest me!"))
	(list 'let '((gene-io-temp-file-marker nil))
	      (cons 'progn forms)
	      '(prog1 gene-io-temp-files
		 (incf gene-io-temp-file-number-start
		       (length gene-io-temp-files))
		 (setq gene-io-temp-files nil)))))


(defun gene-io-register-temp-file (file)
  "Request that FILE be deleted along with any other temporary files.
FILE will be deleted upon exiting the innermost dynamically enclosing
`gene-io-with-temp-files' form.

This function signals an error if we are not within the dynamic scope
of any `gene-io-with-temp-files' form."

  ;; Make sure we're always evaluated within the scope of a
  ;; gene-io-with-temp-files form.
  (unless (boundp 'gene-io-temp-file-marker)
    (error "gene-io-register-temp-file: not inside a `gene-io-with-temp-files' form."))

  (push file gene-io-temp-files))


(defun gene-io-temp-file-name (prefix)
  "Return an unused file name using PREFIX, saving it in `gene-io-temp-files.'
Any file generated using this function will automatically be erased
upon exiting the nearest enclosing `gene-io-with-temp-files' form."
  (or (<= (length prefix) 6)
      (error "gene-io-temp-file-name: prefix too long for SysV: %s"
             prefix))
  (let ((name
         (expand-file-name
          (make-temp-name
           (concat prefix
                   ;; Make sure not to clash with other temp files
                   ;; created by this process.
                   (number-to-string (+ gene-io-temp-file-number-start
					(length gene-io-temp-files)))))
          gene-io-temp-file-dir)))
    (gene-io-register-temp-file name)
    name))



;;;; Temporary buffers

(defun gene-io-make-temp-buffer ()
  "Return a new, empty buffer."
  (generate-new-buffer " *gene-temp*"))

(defun gene-io-kill-temp-buffer (buffer)
  "Delete BUFFER, if it is a buffer."
  (if (and (bufferp buffer)
	   (buffer-name buffer))
      (kill-buffer buffer)))

(defmacro gene-io-with-temp-buffer (VAR &rest BODY-FORMS)
  "Macro: bind VAR to a fresh buffer, evaluate BODY-FORMS, delete the buffer.
The buffer will be deleted even if BODY-FORMS makes a non-local exit."
  (list 'let (list VAR)
	(list 'unwind-protect
	      (list 'progn
		    (list 'setq VAR '(gene-io-make-temp-buffer))
		    (cons 'progn BODY-FORMS))
	      (list 'gene-io-kill-temp-buffer VAR))))
(put 'gene-io-with-temp-buffer 'lisp-indent-function 1)


;;;; Processing buffer-or-filename arguments

;;; We'd like most of our functions to take either a buffer or a file
;;; as input.  Calling this function is an easy and as-quick-as-
;;; possible way to make sure that the rest of your code only has to
;;; deal with files.
(defun gene-io-coerce-to-file (object)
  "Make sure OBJECT's contents are in a file, and return the file's name.
If OBJECT is a buffer, this function writes it to a temporary file and
returns the name.  If OBJECT is a string, it is assumed to be a
filename, passed through `expand-file-name', and returned."
  (cond
   ((bufferp object)
    (let ((name (gene-io-temp-file-name "gene")))
      (save-excursion
        (set-buffer object)
        (gene-message "Saving %s..." (buffer-name object))
        (write-region (point-min) (point-max) name nil 'no-message))
      name))
   ((stringp object)
    (expand-file-name object))
   (t
    (error "gene-io-coerce-to-file: neither buffer nor filename: %s"
           object))))



;;;; Centralized information about file formats

;;; We choose a symbol to stand for each file format --- `genbank' for
;;; GenBank, `embl' for EMBL, etcetera.  Then we put various
;;; properties on that symbol to provide information about it ---
;;; whether it's indexed, which function to call to read it, or write
;;; it, etcetera.

;;; The `gene-format-define' function initializes the properties on
;;; symbols, and keeps a list of all formats; the `gene-format-mumble'
;;; functions retrieve info from the symbols' plists.

(defvar gene-format-list nil
  "A list of the symbols for the known file formats.
Use the function `gene-format-define' to add elements to it.")

(defun gene-format-define (symbol alist)
  "Set up a format symbol SYMBOL with the appropriate properties, from ALIST.
ALIST is a list of pairs of the following forms: 
  (header-regexp . REGEXP)
    REGEXP matches the initial characters of files in this format.
  (filename-regexp . REGEXP)
    REGEXP matches the names of files in this format.
  (indexed-p . BOOLEAN)
    BOOLEAN is non-nil iff this file format is indexed.  Indexed
    formats allow you to change some sequences and leave others alone
    without rewriting the entire file; GDBM is an indexed file format.
  (to-tabl-method . FUNCTION)
    A function to convert files in this format to tabl format.  The
    function should take the same arguments as, for example,
    `gene-io-genbank-to-tabl'.
  (from-tabl-method . FUNCTION)
    A function to convert tabl files into this format.  The function
    should take the same arguments as, for example,
    `gene-io-tabl-to-genbank'.  If this isn't set, we'll try to use
    the `update-from-tabl-method' property.
  (update-from-tabl-method . FUNCTION)
    A function to merge a named set of sequences with those already in
    the file.  FUNCTION should be called like
    `gene-io-tabl-into-gdbm'.  This property may be omitted, in which
    case we use `from-tabl-method' and `to-tabl-method'.
  (list-ids-method . FUNCTION)
    A function to list the short-ids used in the file.  The arguments
    should be as with `gene-io-gdbm-list-ids'.  This property may be
    omitted, in which case we use the `to-tabl-method', and throw away
    all but the short-id list.
  (make-annotation-method . FUNCTION)
    A function to construct an annotation for a new sequence.  The function
    takes one argument --- the short ID of the sequence."

  (push symbol gene-format-list)
  (dolist (binding alist)
    (put symbol (car binding) (cdr binding))))
(put 'gene-format-define 'lisp-indent-function 'defun)

(defsubst gene-format-header-regexp (format)
  (get format 'header-regexp))

(defsubst gene-format-filename-regexp (format)
  (get format 'filename-regexp))

(defsubst gene-format-indexed-p (format)
  (get format 'indexed-p))

(defsubst gene-format-to-tabl-method (format)
  (get format 'to-tabl-method))

(defsubst gene-format-from-tabl-method (format)
  (get format 'from-tabl-method))

(defsubst gene-format-update-from-tabl-method (format)
  (get format 'update-from-tabl-method))

(defsubst gene-format-list-ids-method (format)
  (get format 'list-ids-method))

(defsubst gene-format-make-annotation-method (format)
  (get format 'make-annotation-method))


;;;; Guessing file formats.

(defun gene-io-file-format (filename &optional no-error)
  "Return the symbol indentifying the format of FILENAME.
Signal an error if we can't figure it out, unless optional second arg
NO-ERROR is non-nil.  If no error, and no format, then return nil."
  (or (gene-io-file-format-from-contents filename)
      (gene-io-file-format-from-name filename)
      (or
       no-error
       (error "gene-io-file-format: can't determine format of %s" filename))))


(defun gene-io-file-format-from-name (file)
  "Return a symbol identifying the format of FILE, based on its name alone.
Return nil if we can't guess the format.
Possible format symbols are described in the docstring for
`gene-io-file-format'."
  (dolist (format gene-format-list)
    (if (and (gene-format-filename-regexp format)
	     (string-match (gene-format-filename-regexp format) file))
	(return format))))


(defun gene-io-file-format-from-contents (file)
  "Return a symbol identifying the format of FILE, based on its contents.
Return nil if we can't guess the format."

  (cond

   ;; If the file doesn't exist, use whatever the alignment has already.
   ((not (file-exists-p file))
    (or (gene-io-current-file-format)
	'fasta))

   ;; If the file exists, but isn't readable, try to guess it by name
   ;; (i.e. return nil and let gene-io-file-format-from-name take a
   ;; shot.
   ((not (file-readable-p file))
    nil)

   ;; Read in the first 300 bytes of the file, and do some regexp
   ;; matching on it.
   (t
    (save-excursion
      (gene-io-with-temp-buffer text
	(set-buffer text)
	(insert-file-contents file nil 0 300)
	(goto-char (point-min))
	(gene-io-header-format))))))


(defun gene-io-header-format ()
  "Return the format of the header starting at point.
Return nil if we can't figure it out."
  (save-excursion
    (save-restriction

      ;; Skip blank lines immediately after point.
      (if (looking-at "[ \t\n]*\n")
	  (goto-char (match-end 0)))

      (narrow-to-region (point) (point-max))

      (dolist (format gene-format-list)
	(if (and (gene-format-header-regexp format)
		 (looking-at (gene-format-header-regexp format)))
	    (return format))))))
    


;;;; Running synchronous subprocesses, and displaying resulting error messages.

(defun gene-io-call-process-carefully (program &rest args)
  "Call PROGRAM synchronously in separate process.
The program's input comes from `/dev/null'.
Remaining arguments are strings passed as command arguments to PROGRAM.
Waits for PROGRAM to terminate and returns a numeric exit status or a
signal description string.

If the program produces any output, this is assumed to contain error
messages; they are displayed, and an error is signalled.

If you quit, the process is killed with SIGINT, or SIGKILL if you quit again."
  (let ((error-buffer (get-buffer-create "*Errors*")))
    (save-excursion
      (set-buffer error-buffer)
      (erase-buffer)
      (let ((result
	     (apply 'call-process program nil error-buffer nil args)))
	(if (zerop (buffer-size))
	    (let ((f (gene-frame-get-frame (buffer-name error-buffer))))
	      (if (frame-live-p f)
		  (delete-frame f)))
	  (progn
	    (gene-frame-show error-buffer '((name . "*Errors*")
					    (width . 80)
					    (height . 10)
					    (menu-bar-lines . 0)
					    (vertical-scroll-bars . nil)))
	    (error "Error occurred while running %s..." program)))

	result))))
	  


;;;; Calling GenBank<->tabl (e.g. Emacs buffer format) converters

(gene-format-define 'genbank
  '((header-regexp . "^LOCUS[ \t]")
    (filename-regexp . "\\.gb\\'")
    (to-tabl-method . gene-io-genbank-to-tabl)
    (from-tabl-method . gene-io-tabl-to-genbank)
    (make-annotation-method . gene-io-make-genbank-annotation)))

(defvar gene-io-genbank-to-tabl-program "gb-tabl"
  "*The name of the GenBank->tabl conversion program.")

(defvar gene-io-tabl-to-genbank-program "tabl-gb"
  "*The name of the tabl->GenBank conversion program.")


(defun gene-io-genbank-to-tabl (source &optional subset-op indices)
  "Extract loci, headers, and sequences from SOURCE, a file of GenBank entries.
SOURCE may also be a buffer (but not a buffer name, to avoid ambiguity).
This returns a vector [LOCUSFILE HEADERFILE SEQUENCEFILE],
where each element is the name of a file containing the obvious data.
Loci in LOCUSFILE are separated by newlines.
Headers in HEADERFILE are separated by form feeds (?\f).
Sequences in SEQUENCEFILE are separated by newlines.

Second and third optional arguments SUBSET-OP and INDICES help extract
only a subset of the file.
INDICES should be a file/buffer of loci, separated by newlines.
If SUBSET-OP is `include', extract only entries whose loci INDICES lists.
If SUBSET-OP is `exclude', extract only entries whose loci INDICES does not."

  (let ((sourcefile   (gene-io-coerce-to-file source))
        (locusfile    (gene-io-temp-file-name "genelo"))
        (headerfile   (gene-io-temp-file-name "genehe"))
        (sequencefile (gene-io-temp-file-name "genese"))
        (subset-args  (cond ((eq subset-op 'include)
                             (list "--include-loci"
                                   (gene-io-coerce-to-file indices)))
                            ((eq subset-op 'exclude)
                             (list "--exclude-loci"
                                   (gene-io-coerce-to-file indices)))
                            (t nil)))
        (subset-message (cond ((eq subset-op 'include) "selected ")
                              ((eq subset-op 'exclude) "unselected ")
                              (t ""))))
    (gene-message "extracting %sheaders and sequences..." subset-message)
    (apply 'gene-io-call-process-carefully
	   gene-io-genbank-to-tabl-program
	   "--gb-file" sourcefile
	   "--locus-name-out" locusfile
	   "--annotation-out" headerfile
	   "--sequence-out" sequencefile
	   subset-args)
    (gene-message "extracting %sheaders and sequences... done."
                  subset-message)

    (vector locusfile headerfile sequencefile)))


(defun gene-io-tabl-to-genbank (loci headers sequences)
  "Combine LOCI, HEADERS and SEQUENCES into a GenBank file; return its name.
LOCI should contain organism short ID's, separated by newlines.  LOCI
  isn't actually used; we get the short ID's from HEADERS.
HEADERS should contain GenBank-format headers, separated by form feeds.
SEQUENCES should contain tabl-format sequences, separated by newlines."
  (let ((headerfile   (gene-io-coerce-to-file headers))
        (sequencefile (gene-io-coerce-to-file sequences))
        (genbankfile  (gene-io-temp-file-name "genegb")))

    (gene-message "Assembling GenBank file...")
    (gene-io-call-process-carefully gene-io-tabl-to-genbank-program
                  "--annotation-end" "12"
                  "--annotation-file" headerfile
                  "--sequence-end" "10"
                  "--sequence-file" sequencefile
                  "--out-file" genbankfile)
    (gene-message "Assembling GenBank file... done.")

    genbankfile))

(defun gene-io-make-genbank-annotation (short-id)
  "Return a GenBank annotation for a sequence named NAME."
  (format "LOCUS       %s\n\nORIGIN\n" short-id))



;;;; Calling EMBL<->tabl (e.g. Emacs buffer format) converters

(gene-format-define 'embl
  '((header-regexp . "^ID[ \t]")
    (filename-regexp . "\\.embl\\'")
    (to-tabl-method . gene-io-embl-to-tabl)
    (from-tabl-method . gene-io-tabl-to-embl)
    (make-annotation-method . gene-io-make-embl-annotation)))

(defvar gene-io-embl-to-tabl-program "embl-tabl"
  "*The name of the EMBL->tabl conversion program.")

(defvar gene-io-tabl-to-embl-program "tabl-embl"
  "*The name of the tabl->EMBL conversion program.")


(defun gene-io-embl-to-tabl (source &optional subset-op indices)
  "Extract loci, headers, and sequences from SOURCE, a file of EMBL entries.
SOURCE may also be a buffer (but not a buffer name, to avoid ambiguity).
This returns a vector [LOCUSFILE HEADERFILE SEQUENCEFILE],
where each element is the name of a file containing the obvious data.
Loci in LOCUSFILE are separated by newlines.
Headers in HEADERFILE are separated by form feeds (?\f).
Sequences in SEQUENCEFILE are separated by newlines.

Optional arguments SUBSET-OP and INDICES help extract only a subset of
the file.
INDICES should be a file/buffer of loci.
If SUBSET-OP is `include', extract only entries whose loci INDICES lists.
If SUBSET-OP is `exclude', extract only entries whose loci INDICES does not."

  (let ((sourcefile   (gene-io-coerce-to-file source))
	(locusfile    (gene-io-temp-file-name "genelo"))
	(headerfile   (gene-io-temp-file-name "genehe"))
	(sequencefile (gene-io-temp-file-name "genese"))
	(subset-args  (cond ((eq subset-op 'include)
			     (list "--include-keys"
				   (gene-io-coerce-to-file indices)))
			    ((eq subset-op 'exclude)
			     (list "--exclude-keys"
				   (gene-io-coerce-to-file indices)))
			    (t nil)))
	(subset-message (cond ((eq subset-op 'include) "selected ")
			      ((eq subset-op 'exclude) "unselected ")
			      (t ""))))
    (gene-message "extracting %sheaders and sequences..." subset-message)
    (apply 'gene-io-call-process-carefully
	   gene-io-embl-to-tabl-program
	   "--embl-file" sourcefile
	   "--entry-name-out" locusfile
	   "--annotation-out" headerfile
	   "--sequence-out" sequencefile
	   subset-args)
    (gene-message "extracting %sheaders and sequences... done."
		  subset-message)

    (vector locusfile headerfile sequencefile)))

(defun gene-io-tabl-to-embl (loci headers sequences)
  "Combine LOCI, HEADERS and SEQUENCES into an EMBL file, and return its name.
LOCI is ignored.
HEADERS should contain EMBL-format headers, separated by form feeds.
SEQUENCES should contain tabl-format sequences, separated by newlines."
  (let ((headerfile   (gene-io-coerce-to-file headers))
        (sequencefile (gene-io-coerce-to-file sequences))
        (emblfile  (gene-io-temp-file-name "geneem")))

    (gene-message "Assembling EMBL file...")
    (gene-io-call-process-carefully gene-io-tabl-to-embl-program
				    "--annotation-in" headerfile
				    "--sequence-in" sequencefile
				    "--embl-file" emblfile)
    (gene-message "Assembling EMBL file... done.")

    emblfile))

(defun gene-io-make-embl-annotation (short-id)
  "Return an EMBL annotation for a sequence named NAME."
  (format "ID   %s\nXX\nSQ   \n" short-id))


;;;; Calling FASTA<->tabl (e.g. Emacs buffer format) converters

(gene-format-define 'fasta
  '((header-regexp . "^>")
    (filename-regexp . "\\.fasta\\'")
    (to-tabl-method . gene-io-fasta-to-tabl)
    (from-tabl-method . gene-io-tabl-to-fasta)
    (make-annotation-method . gene-io-make-fasta-annotation)))

(defvar gene-io-fasta-to-tabl-program "fasta-tabl"
  "*The name of the FASTA->tabl conversion program.")

(defvar gene-io-tabl-to-fasta-program "tabl-fasta"
  "*The name of the tabl->FASTA conversion program.")


(defun gene-io-fasta-to-tabl (source &optional subset-op indices)
  "Extract loci, headers, and sequences from SOURCE, a file of FASTA entries.
SOURCE may also be a buffer (but not a buffer name, to avoid ambiguity).
This returns a vector [LOCUSFILE HEADERFILE SEQUENCEFILE],
where each element is the name of a file containing the obvious data.
Loci in LOCUSFILE are separated by newlines.
Headers in HEADERFILE are separated by form feeds (?\f).
Sequences in SEQUENCEFILE are separated by newlines.

Optional arguments SUBSET-OP and INDICES help extract only a subset of
the file.
INDICES should be a file/buffer of loci.
If SUBSET-OP is `include', extract only entries whose loci INDICES lists.
If SUBSET-OP is `exclude', extract only entries whose loci INDICES does not."

  (let ((sourcefile   (gene-io-coerce-to-file source))
        (locusfile    (gene-io-temp-file-name "genelo"))
        (headerfile   (gene-io-temp-file-name "genehe"))
        (sequencefile (gene-io-temp-file-name "genese"))
        (subset-args  (cond ((eq subset-op 'include)
                             (list "--include-keys"
                                   (gene-io-coerce-to-file indices)))
                            ((eq subset-op 'exclude)
                             (list "--exclude-keys"
                                   (gene-io-coerce-to-file indices)))
                            (t nil)))
        (subset-message (cond ((eq subset-op 'include) "selected ")
                              ((eq subset-op 'exclude) "unselected ")
                              (t ""))))
    (gene-message "extracting %sheaders and sequences..." subset-message)
    (apply 'gene-io-call-process-carefully
	   gene-io-fasta-to-tabl-program
	   "--fasta-file" sourcefile
	   "--entry-name-out" locusfile
	   "--annotation-out" headerfile
	   "--sequence-out" sequencefile
	   subset-args)
    (gene-message "extracting %sheaders and sequences... done."
                  subset-message)

    (vector locusfile headerfile sequencefile)))

(defun gene-io-tabl-to-fasta (loci headers sequences)
  "Combine LOCI, HEADERS and SEQUENCES into an FASTA file, and return its name.
LOCI is ignored.
HEADERS should contain FASTA-format headers, separated by form feeds.
SEQUENCES should contain tabl-format sequences, separated by newlines."
  (let ((headerfile   (gene-io-coerce-to-file headers))
        (sequencefile (gene-io-coerce-to-file sequences))
        (fastafile  (gene-io-temp-file-name "genefa")))

    (gene-message "Assembling FASTA file...")
    (gene-io-call-process-carefully gene-io-tabl-to-fasta-program
				    "--annotation-in" headerfile
				    "--sequence-in" sequencefile
				    "--fasta-file" fastafile)
    (gene-message "Assembling FASTA file... done.")

    fastafile))

(defun gene-io-make-fasta-annotation (short-id)
  "Return a FASTA annotation for a sequence named NAME."
  (format ">%s\n" short-id))



;;;; Calling Phylip<->tabl (e.g. Emacs buffer format) converters

(gene-format-define 'phylip
  '((header-regexp . "^[ \t]*[0-9]+[ \t]+[0-9]+")
    (filename-regexp . "\\.phy\\'")
    (to-tabl-method . gene-io-phylip-to-tabl)
    (from-tabl-method . gene-io-tabl-to-phylip)
    (make-annotation-method . gene-io-make-phylip-annotation)))

(defvar gene-io-phylip-converter-program "phylip-tabl"
  "*The name of the Phylip<->tabl conversion program.")


(defun gene-io-phylip-to-tabl (source &optional subset-op indices)
  "Extract loci, headers, and sequences from SOURCE, a Phylip file.
SOURCE may also be a buffer (but not a buffer name, to avoid ambiguity).
This returns a vector [LOCUSFILE HEADERFILE SEQUENCEFILE],
where each element is the name of a file containing the obvious data.
Loci in LOCUSFILE are separated by newlines.
Headers in HEADERFILE are separated by form feeds (?\f).
Sequences in SEQUENCEFILE are separated by newlines.

Optional arguments SUBSET-OP and INDICES help extract only a subset of
the file.
INDICES should be a file/buffer of loci.
If SUBSET-OP is `include', extract only entries whose loci INDICES lists.
If SUBSET-OP is `exclude', extract only entries whose loci INDICES does not."

  (let ((sourcefile   (gene-io-coerce-to-file source))
        (locusfile    (gene-io-temp-file-name "genelo"))
        (headerfile   (gene-io-temp-file-name "genehe"))
        (sequencefile (gene-io-temp-file-name "genese"))
        (subset-args  (cond ((eq subset-op 'include)
                             (list "--include"
                                   (gene-io-coerce-to-file indices)))
                            ((eq subset-op 'exclude)
                             (list "--exclude"
                                   (gene-io-coerce-to-file indices)))
                            (t nil)))
        (subset-message (cond ((eq subset-op 'include) "selected ")
                              ((eq subset-op 'exclude) "unselected ")
                              (t "")))

	;; I think it's cheezy to ask the user questions in this
	;; layer, but the Phylip format leaves us little choice.
	(format-trivialities
	 (let ((nice-name (abbreviate-file-name source)))
	   (cond
	    ((gene-y-or-n-p (format "Is `%s' intended for use with fastDNAml? "
			       nice-name))
	     '("--fastdnaml"))
	    ((gene-y-or-n-p (format "Is `%s' contain interleaved sequences? "
			       nice-name))
	     '("--interleaved"))
	    (t nil)))))
    (gene-message "extracting %sheaders and sequences..." subset-message)
    (apply 'gene-io-call-process-carefully
	   gene-io-phylip-converter-program
	   "--to-tabl"
	   sourcefile
	   "--names" locusfile
	   "--annotations" headerfile
	   "--sequences" sequencefile
	   (append format-trivialities subset-args))
    (gene-message "extracting %sheaders and sequences... done."
                  subset-message)

    (vector locusfile headerfile sequencefile)))

(defun gene-io-tabl-to-phylip (loci headers sequences &optional name)
  "Combine LOCI, HEADERS and SEQUENCES into an Phylip file; return its name.
LOCI should contain organism short ID's, separated by newlines.
HEADERS should contain headers of the sort generated by `phylip-tabl -t',
   separated by form feeds.
SEQUENCES should contain tabl-format sequences, separated by newlines.
Optional fourth argument NAME is the name of the tabl file."
  (let ((locusfile    (gene-io-coerce-to-file loci))
	(headerfile   (gene-io-coerce-to-file headers))
        (sequencefile (gene-io-coerce-to-file sequences))
        (phylipfile   (or name (gene-io-temp-file-name "geneph"))))

    (gene-message "Assembling Phylip file...")
    (gene-io-call-process-carefully gene-io-phylip-converter-program
				    "--to-phylip"
				    phylipfile
				    "--names" locusfile
				    "--annotations" headerfile
				    "--sequences" sequencefile
				    "--interleaved")
    (gene-message "Assembling Phylip file... done.")

    phylipfile))

(defun gene-io-make-phylip-annotation (short-id)
  "Return a Phylip annotation for a sequence named NAME."
  (format "\n" short-id))



;;;; Calling GDBM<->tabl (e.g. Emacs buffer format) inserters and extractors

(gene-format-define 'gdbm
  '((filename-regexp . "\\.gdbm\\'")

    ;; Match the magic number 0x13579ace, in either big-endian or
    ;; little-endian form.
    (header-regexp . "\\`\\(\x13\x57\x9a\xce\\|\xce\x9a\x57\x13\\)")

    (indexed-p . t)
    (to-tabl-method . gene-io-tabl-out-of-gdbm)
    (update-from-tabl-method . gene-io-tabl-into-gdbm)
    (list-ids-method . gene-io-gdbm-list-ids)
    ;; It doesn't really matter what annotation format we use in a
    ;; GDBM file, so we'll just use GenBank.
    (make-annotation-method . gene-io-make-genbank-annotation)))

(defvar gene-io-gdbm-program "gdbm"
  "*The name of the GDBM access program.")

(defun gene-io-tabl-out-of-gdbm (source &optional subset-op indices)
  "Extract sequence entries from SOURCE.
SOURCE should be the name of a GDBM file.

This returns a vector [LOCUSFILE HEADERFILE SEQUENCEFILE],
where each element is the name of a file containing the obvious data.
Loci in LOCUSFILE are separated by newlines.
Headers in HEADERFILE are separated by form feeds (?\f).
Sequences in SEQUENCEFILE are separated by newlines.

Second and third optional arguments SUBSET-OP and INDICES help extract
only a subset of the file.
INDICES should be a file/buffer of loci, separated by newlines.
If SUBSET-OP is `include', extract only entries whose loci INDICES lists.
Otherwise, signal an error."
  (let ((source       (expand-file-name source))
        (locusfile    (gene-io-temp-file-name "genelo"))
        (headerfile   (gene-io-temp-file-name "genehe"))
        (sequencefile (gene-io-temp-file-name "genese")))

    (gene-message "retrieving from database...")
    (apply 'gene-io-call-process-carefully
	   (append
	    (list gene-io-gdbm-program 
		  source)
	    (cond
	     ((eq subset-op 'include)
	      (list (concat "--choose="
			    (gene-io-coerce-to-file indices))))
	     ((null subset-op)
	      (list "--get"))
	     (t
	      (error 
	       "gene-io-tabl-out-of-gdbm: subset-op `%s' not supported."
	       subset-op)))
	    (list "--keys"        locusfile)
	    (list "--annotations" headerfile)
	    (list "--sequences"   sequencefile)))
    (gene-message "retrieving from database... done.")

    (vector locusfile headerfile sequencefile)))


(defun gene-io-tabl-into-gdbm (loci headers sequences dest)
  "Insert LOCI, HEADERS, and SEQUENCES into the gdbm file DEST.
Return nil.
LOCI should contain loci (or whatever the gdbm file is indexed by),
    separated by newlines.
HEADERS should contain headers (in any format --- these are not
    interpreted) separated by form feeds.
SEQUENCES should contain tabl-format sequences, separated by newlines."

  (let ((locusfile    (gene-io-coerce-to-file loci))
        (headerfile   (gene-io-coerce-to-file headers))
        (sequencefile (gene-io-coerce-to-file sequences))
        (dest         (expand-file-name dest)))
    
    (gene-message "storing in database...")
    (gene-io-call-process-carefully gene-io-gdbm-program
				    dest
				    "--put"
				    "--keys" locusfile
				    "--annotations" headerfile
				    "--sequences" sequencefile)
    (gene-message "storing in database... done.")

    nil))

(defun gene-io-gdbm-list-ids (gdbm-file)
  "List of the sequences in GDBM-FILE to a file, and return its name.
Loci in the file are separated by newlines."
  (setq gdbm-file (expand-file-name gdbm-file))
  (let ((locusfile (gene-io-temp-file-name "genelo")))
    (gene-message "listing sequences...")
    (gene-io-call-process-carefully gene-io-gdbm-program
				    gdbm-file
				    "--list"
				    "--keys" locusfile)
    (gene-message "listing sequences... done.")
    locusfile))



;;;; Generic format conversion functions

;;; "Coercing a file to a given format" means "taking whatever steps
;;; are necessary (perhaps none) to convert a file's contents to a
;;; given format."

(defun gene-io-coerce-to-tabl (file format &optional subset-op indices)
  "Convert FILE's contents to tabl format, assuming FILE uses FORMAT.
Run any necessary conversion programs.
This returns a vector [LOCUSFILE HEADERFILE SEQUENCEFILE],
where each element is the name of a file containing the obvious data.
Loci in LOCUSFILE are separated by newlines.
Headers in HEADERFILE are separated by form feeds (?\f).
Sequences in SEQUENCEFILE are separated by newlines.

Second and third optional arguments SUBSET-OP and INDICES help extract
only a subset of the file.
INDICES should be a file/buffer of loci.
If SUBSET-OP is `include', extract only entries whose loci INDICES lists.
If SUBSET-OP is `exclude', extract only entries whose loci INDICES does not."
  (funcall
   (or (gene-format-to-tabl-method format)
       (error "gene-io-coerce-to-tabl: unrecognized format: %s" format))
   file subset-op indices))


(defun gene-io-coerce-from-tabl (source-triplet format)
  "Write SOURCE-TRIPLET in file format FORMAT, and return the file name.
SOURCE-TRIPLET is a vector of the form
  [LOCI HEADERS SEQUENCES]
Each element is a file name or a buffer.
Loci in LOCI are separated by newlines.
Headers in HEADERS are separated by form feeds (?\f).
Sequences in SEQUENCES are separated by newlines.
FORMAT is the target format, for example `genbank', `embl', or `tabl'."
  (let ((loci      (elt source-triplet 0))
        (headers   (elt source-triplet 1))
        (sequences (elt source-triplet 2)))
    (let ((from-tabl (gene-format-from-tabl-method format)))
      (if from-tabl
	  (funcall from-tabl loci headers sequences)
	(let ((new-file (gene-io-temp-file-name "genedb")))
	  (funcall
	   (or (gene-format-update-from-tabl-method format)
	       (error
		"gene-io-coerce-from-tabl: don't know how write %s format"
		format))
	   loci headers sequences new-file)
	  new-file)))))


(defun gene-io-list-ids (file format)
  "Create a list of the sequences in FILE; FILE is in FORMAT.
Return the name of the file containing the list.
Loci in the file are separated by newlines.

For unindexed formats, this function extracts all the file's info, and
then deletes the sequence and annotation files.  Indexed formats like
GDBM can produce this information directly."
  (let ((list-ids (gene-format-list-ids-method format)))
    (if list-ids (funcall list-ids file)

      ;; This is dumb; we should just direct the annotations and
      ;; sequences to /dev/null, rather than writing them to disk and
      ;; then deleting them.  But that would require changing the
      ;; interface to the format-specific functions.
      (let ((triplet (gene-io-coerce-to-tabl file format)))
	(delete-file (elt triplet 1))
	(delete-file (elt triplet 2))
	(elt triplet 0)))))
  


;;;; Generic functions for making annotations.

(defun gene-io-make-annotation (name format)
  "Return an annotation for a sequence named NAME, in format FORMAT.
FORMAT should be an element of `gene-format-list'."
  (let ((method (gene-format-make-annotation-method format)))
    (unless method
      (error "gene-io-make-annotation: don't know about `%s' format!" format))
    (funcall method name)))


(defun gene-io-make-and-add-annotation (name format)
  "Return a new ann posn for a sequence named NAME, in format FORMAT.
This ruins any narrowing in the ann buffer.
FORMAT should be an element of `gene-format-list'."
  (let ((ann (gene-io-make-annotation name format)))
    (gene-save-excursion-in gene-ann-buffer
      (widen)
      (goto-char (point-max))
      (let ((opoint (point)))
        (insert ann "\f")
        (list (copy-marker opoint) (point-marker))))))



;;;; Triplet manipulation --- creating, reading and writing

;;; This page is supposed to contain generally useful I/O functions,
;;; suitable for use by any code which needs to read or write sections
;;; of the buffer.  The commands call these functions to do the real
;;; work.

;; Note that some functions use the word "triplet", and some use
;; "quadruplet".  This is because not everyone needs to know about or
;; deal with the status-buffer.  Since the information in the status
;; buffer is not read from or written to a file, those functions that
;; deal with file I/O are "triplet"-users.  Those that deal with
;; setting up all four buffers, a job that may have somewhat to do
;; with file contents and somewhat to do with run-time decisions, are
;; "quadruplet"-users.     -Karl

(defun gene-create-buffer-quadruplet (names)
  "Create a quadruplet of empty gene-mode buffers, with names taken from NAMES.
NAMES should be a vector like that returned by `gene-make-buffer-names'.
Return a quadruplet of buffers [ID-BUF ANN-BUF SEQ-BUF STATUS-BUF]."
  (let ((short-id-name   (elt names 0))
        (annotation-name (elt names 1))
        (sequence-name   (elt names 2))
        (status-name     (elt names 3)))
    (let ((short-id-buffer   (get-buffer-create short-id-name))
          (annotation-buffer (get-buffer-create annotation-name))
          (sequence-buffer   (get-buffer-create sequence-name))
          (status-buffer     (get-buffer-create status-name)))
      (gene-setup-buffer-pointers
       short-id-buffer annotation-buffer sequence-buffer status-buffer)
      (save-excursion (set-buffer short-id-buffer)
                      (gene-id-mode))
      (save-excursion (set-buffer annotation-buffer)
                      (gene-ann-mode))
      (save-excursion (set-buffer sequence-buffer)
                      (gene-seq-mode))
      (save-excursion (set-buffer status-buffer)
                      (gene-status-mode))
      (vector
       short-id-buffer annotation-buffer sequence-buffer status-buffer))))

(defun gene-setup-buffer-pointers (id-buf ann-buf seq-buf status-buf)
  "Initialize the `gene-X-buffer' variables."
  (setq gene-id-buffer       id-buf)
  (setq gene-ann-buffer      ann-buf)
  (setq gene-seq-buffer      seq-buf)
  (setq gene-status-buffer   status-buf)
  )



(defun gene-io-get-annotations-from-file-or-buffer (file-or-buf)
  "Insert annotations from FILE-OR-BUF into gene-ann-buffer.
String means FILE, buffer means BUF.
Check that the new annotations will not exceed the maximum
buffer-size.
This does not do a save-excursion, it just sets buffer to gene-ann-buffer.
It also widens the ann buffer."
  (set-buffer gene-ann-buffer)
  
  ;; Check that the new annotations will not exceed the
  ;; buffer-size limit:
  (let ((new-anns-size (if (stringp file-or-buf)
                           (nth 7 (file-attributes headerfile))
                         (save-excursion
                           (set-buffer file-or-buf)
                           (buffer-size))))
        (existing-anns-size (buffer-size)))
    (if (or
         (< new-anns-size 0)
         (< (+ new-anns-size existing-anns-size) 0))
        ;; If the sizes add to a negative number, then we've
        ;; exceeded the range of Lisp integers.  Oops!
        (gene-error "alignment annotations too large to edit")))
  
  (widen)
  (goto-char (point-max))
  (let ((buffer-undo-list t)
        (opoint (point)))
    (if (stringp file-or-buf)
        ;; Hmm, why doesn't insert-file-contents behave like any other
        ;; insertion function and move point along?
        (forward-char (nth 1 (insert-file-contents file-or-buf)))
      (insert-buffer-substring file-or-buf))
    (goto-char opoint)))


;; This function might at first glance look like it should deal with a
;; quadruplet, but in fact it is a triplet-o-phile.  This is because
;; it's only concerned with file contents.  The buttons and whatnot in
;; the status buffer are not related to file contents, so they get
;; dealt with separately.
(defun gene-io-insert-file-triplet (names source-file source-format)
  "Insert a tabl-format triplet of files into the current buffer triplet.
NAMES should be a vector of filenames [ID-FILE ANN-FILE SEQ-FILE].
SOURCE-FILE and SOURCE-FORMAT should be the file name and format to
place in the sequences' source structure.

Insert the sequences at user-friendly places in the buffer.  Leave
point in the sequence and id buffers at the top of the first sequence
inserted.

Note that this function undoes any restriction in the annotation
buffer.  Since there's no good way in Emacs to modify the text outside
a restriction, and then reinstate the restriction, I'm not going to
worry about this; the callers should take care to call
gene-ann-update-annotation."

  (message "Placing data...")
  (let (short-id-temp-buffer 
        sequence-temp-buffer
        
        ;; We need to preserve the modification state of the buffer.
        ;; If it was [un]modified before we inserted some new
        ;; sequences, it's certainly still [un]modified after the
        ;; virgin sequences are in it, since it's really modification
        ;; of sequence data -- not the buffer -- that we're interested in.
        ;; This var should be same as (buffer-modified-p): 
        (modified-p gene-modified-p)
        
        ;; Since we want to let gene-io-place-sequences move point in
        ;; the seq and id buffers, a full save-excursion isn't called
        ;; for.
        (original-buffer (current-buffer)))
    (unwind-protect
        (let ((locusfile    (elt names 0))
              (headerfile   (elt names 1))
              (sequencefile (elt names 2)))
          
          ;; Read in the short names --- they appear in the "locus file."
          (setq short-id-temp-buffer (generate-new-buffer "short-id-temp"))
          (set-buffer short-id-temp-buffer)
          (let ((buffer-undo-list t))
            (insert-file-contents locusfile))
          
          ;; Read in the sequences.
          (setq sequence-temp-buffer (generate-new-buffer "sequence-temp"))
          (set-buffer sequence-temp-buffer)
          (let ((buffer-undo-list t))
            (insert-file-contents sequencefile))
          
          ;; Check that the sequences will not exceed the
          ;; buffer-size limit:
          (let ((new-seq-data-size (buffer-size))
                (existing-seq-data-size (gene-save-excursion-in
                                            gene-seq-buffer
                                          (buffer-size))))
            (if (< (+ new-seq-data-size existing-seq-data-size) 0)
                ;; If the sizes add to a negative number, then we've
                ;; exceeded the range of Lisp integers.  Oops!
                (gene-error "alignment sequences too large to edit")))


          ;; Read in the annotations.  These can be placed
          ;; directly at the end of gene-ann-buffer, since it
          ;; doesn't matter what order that buffer is in.  Widen
          ;; the buffer while we do this.
          (gene-io-get-annotations-from-file-or-buffer headerfile)
          
          ;; Figure out where to put these sequences, and put them
          ;; there.
          (gene-io-place-sequences (vector short-id-temp-buffer
                                           gene-ann-buffer
                                           sequence-temp-buffer)
                                   source-file
                                   source-format))
      
      ;; Restore the modification state, whatever it was:
      (save-excursion
        (set-buffer gene-seq-buffer)
        (or 
         modified-p ; if it was modified before, we need do nothing
         (gene-mark-unmodified)))
      
      ;; Unwind: delete the temporary buffers.
      (if (bufferp sequence-temp-buffer)
          (kill-buffer sequence-temp-buffer))
      (if (bufferp short-id-temp-buffer)
          (kill-buffer short-id-temp-buffer))
      (set-buffer original-buffer)))
    (message "Placing data... done."))

;;; Since our buffers may contain sequences from many files, and any
;;; subset of the sequences in any given file, saving a buffer may
;;; entail replacing portions of several files, while leaving their
;;; other contents unchanged.
;;;
;;; Some file formats, like GDBM, support this sort of access very
;;; easily; the GDBM routines can efficiently extract or replace a
;;; section of a GDBM file.
;;;
;;; Other file formats, like GenBank, aren't arranged to easily
;;; support such access patterns.  In these cases, the best we can
;;; generally do is extract that portion of the file which we're not
;;; going to change into temporary files in tabl format, append our
;;; new, changed data to the end of them, and then convert the entire
;;; thing back into the orginal format.
;;;
;;; gene-io-update-sequences-indexed takes care of the first (GDBM-like)
;;; case.
;;;
;;; gene-io-update-sequences-unindexed takes care of the second
;;; (GenBank-like) case.


(defun gene-io-update-sequences-indexed (region-triplets file format)
  "Merge REGION-TRIPLETS into FILE.
FILE must be the name of a file in an indexed format, like GDBM;
FORMAT specifies its format.
REGION-TRIPLETS is a list of vectors of the form
  [(ID-START         ID-END)
   (ANNOTATION-START ANNOTATION-END)
   (SEQUENCE-START   SEQUENCE-END) ...]
giving the regions of gene-id-buffer, gene-ann-buffer, and
gene-seq-buffer to be merged with FILE.

By merging, we mean the following: if FILE contains sequences other
than those being written (i.e. sequences with different loci), they
are preserved in the result.  If FILE contains sequences with the same
loci as those being written, they are replaced in the result with the
new sequences.

FILE need not exist."
  (let ((update (gene-format-update-from-tabl-method format)))

    (unless update
      (error "gene-io-update-sequences-indexed: don't know how write %s format"
	     format))

    (let ((locusfile      (gene-io-temp-file-name "genelo"))
	  (annotationfile (gene-io-temp-file-name "genean"))
	  (sequencefile   (gene-io-temp-file-name "genesq")))
      
      ;; Write the loci out.
      (save-excursion
	(set-buffer gene-id-buffer)
	(gene-io-append-regions region-triplets 0 "\n" locusfile))
      
      ;; Write the headers out.
      (save-excursion
	(set-buffer gene-ann-buffer)
        (save-restriction
          (widen)
          (gene-io-append-regions region-triplets 1 "\f" annotationfile)))
      
      ;; Write the sequences out.
      (save-excursion
	(set-buffer gene-seq-buffer)
	(gene-io-append-regions region-triplets 2 "\n" sequencefile))

      (funcall update locusfile annotationfile sequencefile file))))
      

(defun gene-io-update-sequences-unindexed (region-triplets file format)
  "Merge REGION-TRIPLETS with FILE, yielding a new file.
Return the name of the resulting file.
FILE must be the name of a file in an unindexed format, like GenBank or EMBL;
FORMAT specifies its format.
REGION-TRIPLETS is a list of vectors of the form
  [(ID-START         ID-END)
   (ANNOTATION-START ANNOTATION-END)
   (SEQUENCE-START   SEQUENCE-END) ...]
giving the regions of gene-id-buffer, gene-ann-buffer, and
gene-seq-buffer to be merged with FILE.

By merging, we mean the following: if FILE contains sequences other
than those being written (i.e. sequences with different loci), they
are preserved in the result.  If FILE contains sequences with the same
loci as those being written, they are replaced in the result with the
new sequences.

FILE need not exist.

This may undo the current restriction on the annotation buffer.
Callers will just have to deal."

  (let ((file-triplet

	 (if (file-exists-p file)
	     ;; If FILE already exists, merging with it.  Extract all
	     ;; its sequences but those we're going to write.
	     (let ((exclude-file (gene-io-temp-file-name "geneex")))
	       (save-excursion
		 (set-buffer gene-id-buffer)
		 (gene-io-append-regions region-triplets 0 "\n" exclude-file)
		 (gene-io-coerce-to-tabl file format 'exclude exclude-file)))

	   ;; If FILE doesn't exist, we're writing a fresh file.  We
	   ;; just need to generate some new filenames.
	   (vector (gene-io-temp-file-name "genelo")
		   (gene-io-temp-file-name "genehe")
		   (gene-io-temp-file-name "genesq")))))

    ;; Write the loci out.
    (save-excursion
      (set-buffer gene-id-buffer)
      (gene-io-append-regions region-triplets 0 "\n" (elt file-triplet 0)))

    ;; Write the headers out.
    (save-excursion
      (set-buffer gene-ann-buffer)
      (save-restriction
        (widen)
        (gene-io-append-regions region-triplets 1 "\f" (elt file-triplet 1))))
      
    ;; Write the sequences out.
    (save-excursion
      (set-buffer gene-seq-buffer)
      (gene-io-append-regions region-triplets 2 "\n" (elt file-triplet 2)))

    ;; Convert it all back to the original format.
    (gene-io-coerce-from-tabl file-triplet format)))


(defun gene-io-append-regions (region-triplets index separator dest)
  "Append regions of the current buffer to disk.
REGION-TRIPLETS has the form ([(START END) ...] ...)
INDEX is a natnum.
SEPARATOR is a string, to be written between the regions.
DEST is a filename.
The regions of the current buffer to write are
    (mapcar (lambda (t) (elt t INDEX)) REGION-TRIPLETS)"

  (let ((source-buf (current-buffer)))
    (let (temp-buf)
      (unwind-protect
          (progn

            ;; The temporary buffer wouldn't be necessary if we
            ;; didn't need to put in SEPARATOR, but there's no other
            ;; way to deal with the sequence buffer; sometimes there
            ;; won't be a newline at the end of the sequence region.
            (save-excursion
              (set-buffer (setq temp-buf
                                (generate-new-buffer
                                 "gene-io-append-regions-temp")))
              (buffer-disable-undo)
              (dolist (triplet region-triplets)
                      (insert-buffer-substring source-buf
                                               (car (elt triplet index))
                                               (cadr (elt triplet index)))
                      (insert separator))
              (write-region (point-min) (point-max) dest t 'no-message)))

        ;; Unwind:
        (if (bufferp temp-buf) (kill-buffer temp-buf))))))
  

(defun gene-io-make-region-triplet ()
  "Return an id/ann/seq triplet for the sequence at point, and how far it goes.
The return value has the form
    (SOURCE END REGION-TRIPLET)
where REGION-TRIPLET has the form
     [(ID-START         ID-END)
      (ANNOTATION-START ANNOTATION-END)
      (SOURCE-START     SOURCE-END)]
where:
    SOURCE is the source at point,
    END is the next point with a different source,
    ID-START .. ID-END is the region of gene-id-buffer containing the
        short-id for this stretch of sequence,
    ANNOTATION-START .. ANNOTATION-END is the region of gene-ann-buffer
        containing the annotation for this stretch of sequence, and
    SOURCE-START .. SOURCE-END is the largest region of
        gene-seq-buffer starting at point which have the same short-id
        and annotation (in other words, the same source).

If there is no source at point, this returns (nil END nil)." 

  (let ((source (gene-source-at-point))
        (end    (gene-source-end)))
    (list source
          end
          (if source
              (let ((id-region
                     (let ((id-posn-triplet (gene-get-here-property
                                             'gene-id-posn)))
                       (list (car  id-posn-triplet)
                             (cadr id-posn-triplet))))
                    (ann-region
                     (let ((ann-pair (gene-source-annotation source)))

		       ;; The end marker for annotations doesn't float
		       ;; after the text when you insert text at the
		       ;; end of the annotation.  However, the start
		       ;; marker works fine, so we'll go to the start
		       ;; marker, and search for the end.
		       (save-excursion
			 (set-buffer gene-ann-buffer)
			 (goto-char (car ann-pair))
			 (list (point) (gene-end-of-page-point)))))
                    (seq-region
                     (list (point) end)))

                (vector id-region ann-region seq-region))))))


(defun gene-io-compile-sequence-files (&optional sequences)
  "Return a list of files visited in the current buffer, and their regions.
If optional argument SEQUENCES is non-nil, it is a list of sequences
to be considered.  If nil, all sequences are considered.

The return value has the form
  ((FILENAME FORMAT ([(ID-START         ID-END)
                      (ANNOTATION-START ANNOTATION-END)
                      (SOURCE-START     SOURCE-END)] ...)
   ...)
In other words, it is an alist mapping files onto their formats and
the regions of the short-id, annotation, and sequence buffers that
they contributed."

  (save-excursion
    (set-buffer gene-ann-buffer)
    (save-restriction
      (widen)
      
      (set-buffer gene-seq-buffer)
      (save-excursion

	;; We'll map over regions of the sequence buffer with different
	;; sources.  We'll register filenames in file-table (a hash
	;; table), and accumulate region lists as the values of the
	;; symbols we intern there.
	(let ((file-table (make-vector 101 0)))

	  (goto-char (point-min))

	  ;; Process one single-sourced region at a time.
	  (while (< (point) (point-max))
	    (let* ((source-end-region-triplet (gene-io-make-region-triplet))
		   (source         (elt source-end-region-triplet 0))
		   (end            (elt source-end-region-triplet 1))
		   (region-triplet (elt source-end-region-triplet 2)))
	      (and
	       source
	       (if sequences (memq (gene-get-here-sym) sequences) t)
	       (let ((filename-sym
		      (intern (gene-source-filename source) file-table)))
		 (set filename-sym
		      (cons region-triplet
			    (if (boundp filename-sym)
				(symbol-value filename-sym))))
		 (put filename-sym
		      'format
		      (gene-source-file-format source))))
	      (goto-char end)))
        
	  ;; Compute the final result.
	  (let ((file-alist nil))
	    (mapatoms (lambda (filename-sym)
			(setq file-alist
			      (cons (list (symbol-name filename-sym)
					  (get filename-sym 'format)
					  (nreverse 
					   (symbol-value filename-sym)))
				    file-alist)))
		      file-table)
	    file-alist))))))


(defun gene-io-write-regions (regions)
  "Save some sequences from the current buffer into some files.
REGIONS is an alist of the form computed by gene-io-compile-sequence-files;
the specified regions are written/updated into their corresponding files."
  
  ;; Write the files.
  (dolist (region regions)
    (let ((filename        (elt region 0))
          (file-format     (elt region 1))
          (region-triplets (elt region 2)))
      (gene-with-message-prefix
	  (format "Saving %s... " (file-name-nondirectory filename))
        (if (gene-format-indexed-p file-format)
            (gene-io-update-sequences-indexed region-triplets
                                              filename
                                              file-format)
          (let ((result-name
                 (gene-io-update-sequences-unindexed region-triplets
                                                     filename 
                                                     file-format)))
            (rename-file result-name filename 'okay-if-filename-exists)))))))



;;;; I/O functions for analysis packages.

;;; These can be a bit simpler than the other I/O functions, because
;;; they operate on whole sequences at a time, rather than on regions
;;; specified by gene-source properties.

(defun gene-io-write-ids (organisms file)
  "Write the ids of ORGANISMS to FILE.
ORGANISMS is a list of orgsyms."
  (gene-io-with-temp-buffer buf
    (set-buffer buf)
    (dolist (organism organisms)
      (let ((pos (gene-organism-id-posn organism)))
	;; Could we just insert gene-organism-name here? 
	(insert-buffer-substring gene-id-buffer (car pos) (cadr pos))
	(insert "\n")))
    (write-region (point-min) (point-max) file nil 'no-message)))

(defun gene-io-write-sequences (organisms file)
  "Write the sequences of ORGANISMS to FILE.
ORGANISMS is a list of orgsyms."
  (gene-io-with-temp-buffer buf
    (set-buffer buf)
    (dolist (organism organisms)
      (let ((pos (gene-organism-seq-posn organism)))
	(insert-buffer-substring gene-seq-buffer (car pos) (cadr pos))
	(insert "\n")))
    (write-region (point-min) (point-max) file nil 'no-message)))

(defun gene-io-write-annotations (organisms file)
  "Write the annotations of ORGANISMS to FILE.
ORGANISMS is a list of orgsyms."
  (save-excursion
    (gene-io-with-temp-buffer buf
      (set-buffer gene-ann-buffer)
      (save-restriction
        (widen)
        (set-buffer buf)
        (dolist (organism organisms)
          (let* ((posn  (gene-source-annotation
                         (gene-organism-default-source organism)))
                 (start (car posn))
                 (end   (cadr posn)))
            (insert-buffer-substring gene-ann-buffer start end)
            (insert "\f")))
        (write-region (point-min) (point-max) file nil 'no-message)))))
  
(defun gene-io-write-organisms-no-ann (organisms)
  "Write the ids and sequences of ORGANISMS to two files; return their names.
Don't write the sequences' annotations.
ORGANISMS is a list of orgsyms.
The return value has the format [IDFILE nil SEQFILE], where IDFILE and
SEQFILE are the names of the files to which we wrote the short-ids and
the sequences of ORGANISMS."
  (let ((idfile  (gene-io-temp-file-name "geneid"))
	(seqfile (gene-io-temp-file-name "genese")))
    (gene-io-write-ids organisms idfile)
    (gene-io-write-sequences organisms seqfile)
    (vector idfile nil seqfile)))
  
(defun gene-io-write-sequences-to-tabl-files (sequences prefix)
  "Write SEQUENCES to tabl files named according to PREFIX.
Returns the list of three files."
  (let ((id-file  (concat prefix ".ids"))
        (ann-file (concat prefix ".anns"))
        (seq-file (concat prefix ".seqs")))
    (gene-io-write-ids         sequences  id-file)
    (gene-io-write-annotations sequences ann-file)
    (gene-io-write-sequences   sequences seq-file)
    (list id-file ann-file seq-file)))


;;;; Inserting new sequences near close relatives

(defun gene-io-place-sequences
  (buffer-triplet &optional file format no-status-update)
  "Add new sequences and short ID's to editing buffers, placed cleverly.
BUFFER-TRIPLET has the form [LOCI HEADERS SEQUENCES], where each
element is a buffer containing the short-ID's, headers, or sequences
to add.

FILE is the name of the file the new data came from; FORMAT is its format.
They are passed directly to `gene-map-setup-properties'.  See that
function's documentation for more information.

Insert the contents of these buffers into `gene-id-buffer' and
`gene-seq-buffer', using heuristics to try insert organisms near close
relatives.  Leave point in the seq and id buffers on the top sequence
inserted.  Randomize point in the ann buffer.  Update the buttons in
the status buffer appropriately.

BUFFER-TRIPLET might really be a quadruplet.  You know how it is.

Optional fourth arg NO-STATUS-UPDATE means don't update the lock
status buffer; the caller will take care of that.

Grotesque hack alert: HEADERS can be gene-ann-buffer, with point
sitting just before the new headers.  Rationale: the annotation buffer
doesn't need to be ordered, so there's no point in wasting time and
space to bring it into a temp buffer and then copy it into the real
one.  However, the clipboard, for example, keeps annotations off in a
real tmp buffer \(i.e.: the clipboard\), so we don't want to assume
that HEADERS will *always* be gene-ann-buffer."

  ;; For now, we don't actually use any heuristics.  They'll come later.

  ;; Why use heuristics?  Are they necessarily the best thing to use?
  ;; What if we used Post-Modern Leftist French Political Theory to
  ;; determine where to place organisms?  Or how about using
  ;; Miracle-Gro?  What I'm trying to say is, let's keep an open mind
  ;; about this.  Heuristics have their place, but so does West
  ;; Virginia.

  ;; I did not write the above paragraph.  -JimB

  ;; He'd like to deny it, wouldn't he?    -Karl

  ;; Aye and merry, I have!  -JimB

  ;; I think the above should have been written in a more archaic
  ;; font, but he didn't do that, so his whole statement is
  ;; suspect.  -Karl

  ;; Yes, but he was a suspect to begin with.  -JimB

  ;; No.  Some people thought you were a suspect, but later they
  ;; realized they had been wrong the whole time.  -Karl

  ;; I did not write this either.  -JimB

  ;; Serious note (this function needs one badly): if we ever do use
  ;; heuristics, there will have to be a way to turn them off, because
  ;; now that this function does the dirty work of gene-open-sequence,
  ;; we'll just want it to place sequences at point in the seq buffer.
  ;; In sum, the clever sequence-placing code has to be isolated
  ;; somewhere and the caller of gene-io-place-sequences needs control
  ;; over whether it's used.

  (gene-save-and-restore-selected-group
   (let ((original-buffer (current-buffer))
         (empty-p (gene-alignment-empty-p))
         first-line
         id-start
         seq-start
         id-end
         seq-end
         final-sym)
     (unwind-protect
         (progn
           
           ;; Insert the sequences.
           (set-buffer gene-seq-buffer)
           (beginning-of-line)
           (setq seq-start (point))
           (setq first-line (gene-current-line))
           (setq id-start 
                 (if (< (point) (point-max))
                     (elt (gene-get-here-property 'gene-id-posn) 0)
                   'max))
           ;; now: set status-start/end here too.
           
           (let (buffer-read-only
                 (buffer-undo-list t))
             
             ;; We need to preserve the markers around the id after
             ;; point.  Since the only markers at bol are starting
             ;; markers for other sequences, if we insert a newline at
             ;; point before the markers, and then back up one, we'll be
             ;; on a line free of markers.
             (insert-before-markers "\n")
             (backward-char 1)
             (insert-buffer-substring (elt buffer-triplet 2))
             (or (bobp) (delete-char -1)))
           (setq seq-end (point))
           (goto-char seq-start) ; for mapping code, below.
           
           ;; Insert the ID's.
           (set-buffer gene-id-buffer)
           (goto-char (if (eq id-start 'max)
                          (point-max)
                        id-start))
           (beginning-of-line)
           (setq id-start (point))

           (save-excursion
             (let (buffer-read-only
                   (buffer-undo-list t)
                   (grp (if (bobp)
                            nil
                          (prog2
                              (forward-char -1)
                              (gene-sgroup-displayed-group-here)
                            (forward-char 1)))))
               (if grp (gene-sgroup-hide-group grp))
               (insert-before-markers "\n")
               (backward-char 1)
               (insert-buffer-substring (elt buffer-triplet 0))
               (if grp (gene-sgroup-display-group grp))
               (or (bobp) (delete-char -1)))
             (setq id-end (point)))
           (goto-char id-start) ; for mapping code, below.
           
           ;;  Insert the annotations... maybe.
           (if (eq (elt buffer-triplet 1) gene-ann-buffer)
               nil
             ;; Else the tmp buffer for headers is not
             ;; gene-ann-buffer, which probably means that we were
             ;; called by some clipboard-insertion function, rather
             ;; than by `gene-io-insert-file-triplet'... Anyway, the
             ;; important thing is that gene-ann-buffer doesn't
             ;; contain the headers, so we'll have to put them in
             ;; before we can go any farther.
             (gene-io-get-annotations-from-file-or-buffer
              (elt buffer-triplet 1)))
             
           ;; Prep everyone.
           (setq
            final-sym
            (gene-map-setup-properties
             id-start id-end seq-start seq-end file format empty-p))

           (set-buffer gene-seq-buffer)
           (goto-char seq-start)
           (set-buffer gene-id-buffer)
           (goto-char id-start)
           ;; Update the status buffer:
           (or no-status-update (gene-status-buffer-update 1))
           )
       
       ;; Unwind
       (set-buffer original-buffer)
       (gene-arrange-windows)
       (gene-ann-update-annotation))

     (if empty-p
         (gene-record-unchanged-state))
     final-sym
     )))


(defun gene-map-setup-properties
  (id-start id-end seq-start seq-end file format &optional no-undo)
  "Map over some contiguous sequences, setting up their properties.
The sequences' locations are given by the arguments
ID-START, ID-END, SEQ-START, and SEQ-END.

FILE is the file from which the sequences came.  Nil means they
already have source properties, so the file should be obtained
indivdually from those.

FORMAT is the FORMAT of the FILE from which the sequences came.  Nil
means they already have source properties, so the format should be
obtained indivdually from those.

Point is assumed to be at the correct location in the annotation
buffer; that is, in the first annotation.
Optional fifth arg NO-UNDO means the alignment was empty before
this, so it is not necessary to record undo."
  ;; If this is startup and/or we are about to insert
  ;; sequences into an empty alignment, then don't
  ;; record undo:
  (let ((noundo (or gene-startup-noundo no-undo))
        (region-quadruplet (make-vector 4 nil))
        (line (progn (set-buffer gene-id-buffer)
                     (gene-current-line)))
        (final-sym nil))
    
    ;; Initialize point in the ID and seq buffers:
    (set-buffer gene-id-buffer)
    (goto-char id-start)
    (set-buffer gene-seq-buffer)
    (goto-char seq-start)
    
    ;; Map over the new sequences setting up their properties:
    (while (progn (set-buffer gene-id-buffer)
                  (< (point) id-end))
      (aset region-quadruplet 0 (list (gene-bol-point)
                                      (gene-eol-point)
                                      line))
      (forward-line 1)
      
      (set-buffer gene-ann-buffer)
      (aset region-quadruplet 1
            (mapcar 'copy-marker 
                    (list (gene-beginning-of-page-point)
                          (gene-end-of-page-point))))
      (forward-page 1)
      
      (set-buffer gene-seq-buffer)
      (aset region-quadruplet 2 (list (gene-bol-point)
                                      (gene-eol-point)
                                      line))
      (forward-line 1)
      
      (let ((sym (gene-setup-plists region-quadruplet file format)))
        (or
         noundo
         (gene-undo-record-sequence-open sym))
        (setq final-sym sym))
      (incf line))
    final-sym))



;;;; Opening new sequences.
;;
;; When we start a new sequence, there are several bookkeeping details
;; that must be taken care of.  It needs an ID, an initial lock state
;; (usually unlocked), a `gene-source' property, and a `gene-ident'
;; property.

(defun gene-open-sequence (name data ann &optional file format state line)
  "Insert a new sequence into the alignment, return its org sym.
Its name is NAME, its sequence data is DATA \(which can be the empty
string\), its annotation is ANN.

Optional fourth arg FILE is the file to write this sequence to, and
fifth optional arg FORMAT is the format for that file.
Both default to the file and format of the sequence at point just
before the insertion of the new data.

The format will be deduced if not provided.  (now: We don't implement
this yet.)

Optional sixth arg STATE is an initial lock state; it defaults to
GapInOut.

Optional seventh arg LINE is the line at which to put the new
sequence; it defaults to the current line in the sequence buffer.

The `gene-source' property covers the whole sequence for now; we'll
allow fancier distributions later."
  (save-excursion
    (set-buffer gene-seq-buffer)
    (if line
        (or
         (gene-goto-line line)
         (goto-char (point-max))))

    (let ((id-temp-buffer  (generate-new-buffer "temp-id"))
          ;; gene-ann-buffer is its own temp buffer.
          (seq-temp-buffer (generate-new-buffer "temp-seq"))
          sym)

      (unwind-protect
          (progn
            ;; Fill the temp buffers with the good stuff:
            (set-buffer seq-temp-buffer)
            (erase-buffer) ; paranoia
            (insert data "\n")
            (set-buffer id-temp-buffer)
            (erase-buffer) ; paranoia
            (insert name "\n")
            ;; The ann buffer is its own temp-buffer.
            (set-buffer gene-ann-buffer)
            (widen)
            (goto-char (point-max))
            (let ((opoint (point))
                  (buffer-undo-list t))
              (insert ann "\f")
              (goto-char opoint))
            
            ;; Then just let gene-io-place-sequences do the dirty work.
            (setq
             sym
             (gene-io-place-sequences
              (vector id-temp-buffer gene-ann-buffer seq-temp-buffer)
              (or file
                  (save-excursion
                    (set-buffer gene-seq-buffer)
                    (gene-source-filename (gene-get-here-source))))
              (or format
                  (save-excursion
                    (set-buffer gene-seq-buffer)
                    (gene-source-file-format (gene-get-here-source))))
              ;; Pass the NO-STATUS-UPDATE flag
              t))
            
            (gene-locking-lock-sequence-to-state
             sym (or state gene-locking-gapinout-state))
            sym)
        
        ;; Unwind: kill the temp buffers.
        (kill-buffer id-temp-buffer)
        (kill-buffer seq-temp-buffer)
        (gene-ann-update-annotation)))))


;; Some of this code won't make sense unless I tell you an interesting
;; fact about the way groups work.  When a sequence SEQ is a member of
;; a group GRP, there are a couple of ways the outside world can know
;; about it.  We could get GRP from `gene-sgroup-list' and examine
;; GRP's cdr, which is a list of the sequences GRP contains.  Or we
;; could use `gene-sgroup-groups-containing-seq' to look at SEQ's
;; `gene-sgroups-here' property, which is a list of all the groups
;; containing SEQ.
;;
;; So you see how things are sort of cross-indexed: each group knows
;; what sequences it contains, and each sequence knows which groups
;; contain it.  Well and good.  However...
;;
;; When you cut SEQ from the alignment and put it in the clipboard, we
;; do something funny with that cross-indexing.  GRP is told that
;; sequence no longer exists (i.e.: SEQ is removed from GRP's cdr),
;; but SEQ is never told that it's not a member of GRP -- the property
;; is left on SEQ!  This allows us to restore SEQ to whatever groups
;; it belongs in when we paste SEQ back in.  This is the reason for
;; `gene-sgroup-delete-seq-from-group's optional LEAVE-PROPS argument.
;;
;; (If you cut SEQ, and it's a member of GRP, and then you maybe
;; delete GRP, SEQ never knows that GRP has been deleted.  But when
;; paste-back time comes, SEQ looks for a group with the same name as
;; GRP: if it finds one, it joins that group and changes its
;; `gene-sgroups-here' property accordingly (since the new group might
;; might not actually be the same as the old one, but just has the
;; same name), else the GRP property is removed from SEQ after pasting
;; is done, and GRP is never heard from again.  This process is gone
;; through for every group SEQ thinks it's in.)
;;
;; So, with that in mind, enjoy the show!

;; Attaching organism symbols and sources to new text
;; gene-setup-plists used to handle all the new sequences at once,
;; but it seemed silly and bug-prone to force it to rederive the
;; sequence/id pairings when its caller, who created the new
;; sequence, surely already knows.  So now it only operates on one
;; chunk of sequence, and should be called many times.

(defun gene-setup-plists (region-triplet &optional file format)
  "Set up text properties for some incoming sequence data.
Return the organism symbol created for the sequence.

REGION-TRIPLET should have the form [LOCUSPOS HEADERPOS SEQUENCEPOS],
where each element is a 2- or 3-list giving the starting and ending
position (and the line number, if known) of the new sequence's portion
of the appropriate buffer: gene-id-buffer, gene-ann-buffer, or
gene-seq-buffer.  The regions should not include any newlines, except
in the annotation buffer.

If any of the regions are nil, they are not used.
\(Why is this thus?  Because sometimes we only want to update some of
the properties, not everything.  Notably, when cutting and pasting
sequences, the SOURCE \(therefore HEADERPOS\) data does not change, but
position information in LOCUSPOS and SEQUENCEPOS do\).

We set the `gene-source' property of SEQUENCEPOS to SOURCE, and set the
`gene-ident' property of the text in SEQUENCEPOS and LOCUSPOS
to the appropriate organism symbol.

If FILE and FORMAT are passed in, a source structure is created and
given to the sequences.  Else the sequences are assumed to already
have sources, albeit obsolete ones, and the new sources are created
from those.

If the text in LOCUSPOS has no organism symbol, we create and initialize
one.  If the text in LOCUSPOS already has an organism symbol on its
`gene-ident' property, we use that organism symbol, on the
assumption that the new sequence is being added to the existing data for
that organism."
  (save-excursion

    (let* ((id-pos      (elt region-triplet 0))
           (ann-pos     (elt region-triplet 1))
           (seq-pos     (elt region-triplet 2))
           ;; Pick 'em apart:
           (id-start    (elt id-pos 0))
           (id-end      (elt id-pos 1))
           (ann-start   (elt ann-pos 0))
           (ann-end     (elt ann-pos 1))
           (seq-start   (elt seq-pos 0))
           (seq-end     (elt seq-pos 1))
           (source      (and
                         file
                         format
                         (make-gene-source
                          :filename file
                          :file-format format
                          :annotation ann-pos)))
           )
      
      (set-buffer gene-id-buffer)
      (let* ((here-sym     (gene-get-here-sym id-start nil t))
             (already-here (or
                            (text-property-any
                             (point-min) id-start 'gene-ident here-sym)
                            (text-property-any
                             id-end (point-max) 'gene-ident here-sym)))
             (organism-sym (if already-here nil here-sym))
             (organism-name (buffer-substring-no-properties id-start id-end))
             (here-source (or
                           source
                           (if here-sym
                               (let* ((this-source
                                       (gene-organism-default-source
                                        here-sym))
                                      (this-file
                                       (gene-source-filename this-source))
                                      (this-format
                                       (gene-source-file-format this-source)))
                                 (make-gene-source
                                  :filename    this-file
                                  :file-format this-format
                                  :annotation  ann-pos))))))
        ;; Prep the sym
        (setq organism-sym
              (gene-init-organism-sym
               organism-name id-pos seq-pos here-source organism-sym))
        (gene-sym-accum-sequence organism-sym seq-start seq-end)

        ;; Put the appropriate properties on the ID.
        (let (buffer-read-only
              (buffer-undo-list t))
          (put-text-property id-start id-end 'gene-ident organism-sym)
          (put-text-property id-start id-end 'mouse-face 'gene-mouse-face)
          (if already-here
              (gene-tweak-sequence-name organism-sym)))
        
        ;; Put the appropriate properties on the annotation.
        (set-buffer gene-ann-buffer)
        (let (buffer-read-only)
          (put-text-property ann-start ann-end 'gene-ident organism-sym))

        ;; Put the appropriate properties on the new sequence.
        (set-buffer gene-seq-buffer)
        (let (buffer-read-only
              (buffer-undo-list t))
          (put-text-property seq-start seq-end 'gene-ident organism-sym)
          (if gene-mouse-sensitive-sequences
              (put-text-property
               seq-start seq-end 'mouse-face 'gene-mouse-face))
          
          (put-text-property seq-start seq-end 'front-sticky t)
          (put-text-property seq-start seq-end 'rear-sticky t)
          (put-text-property seq-start seq-end 'gene-source here-source))

        ;; Return it to any groups of which it was a member.
        (gene-sgroup-reinstate-seq-into-groups organism-sym)

        ;; ... start monitoring it again it if it was being
        ;; monitored when cut ...
        (if (get organism-sym 'gene-posmon-monitored)
            (gene-posmon-add-sequence organism-sym))
        
        organism-sym))))


(defun gene-tweak-sequence-name (sym &optional newname)
  "Uniqify the ID name of sequence SYM.
Most probably, this means making the ID name be SYM's symbol-name,
since that's guaranteed to be unique from all other syms.
Or, optional second arg NEWNAME is the new name for the sequence.
This function will error if sequence SYM is not in the alignment."
  (gene-save-excursion-in gene-id-buffer
    (goto-char
     (text-property-any (point-min) (point-max) 'gene-ident sym))
    (let ((new-name (or newname (symbol-name sym)))
          (start (point-marker))
          end
          buffer-read-only)
      (insert new-name)
      (setq end (point-marker))
      (delete-region end (progn (end-of-line) (point)))
      (put-text-property start end 'gene-ident sym)
      (put sym 'gene-id-posn (list start end)))))

  

;;;; Reading and writing commands

(defun gene-switch-to-buffer (buffer)
  "Make gene-mode buffer BUFFER current, and display it in an appropriate frame.
BUFFER may be a buffer (or the name of a buffer) in Gene ID Mode, Gene
Sequence Mode, or Gene Annotation Mode, or Gene Status Mode.
Alternatively, BUFFER may be the root name for such a quadruplet.

If no buffer named BUFFER exists, a quadruplet using BUFFER as the root
name is created."
  (interactive
   (list
    (completing-read "Switch to gene buffer: "
                     (gene-mode-buffers-and-roots)
                     nil nil
                     (let ((other-buffer (gene-other-buffer (current-buffer))))
                       (if other-buffer (buffer-name other-buffer) nil)))))
  
  (let ((buffer

         ;; Did they ask for an extant gene-mode buffer by name?
         (if (and (get-buffer buffer)
                  ;; A root name may also be the name of an ordinary
                  ;; non-gene-mode buffer.  Check for that.
                  (save-excursion (set-buffer buffer)
                                  (memq major-mode gene-mode-list)))
             buffer

           ;; Did they ask for an extant buffer by its root?
           (let* ((names (gene-make-buffer-names buffer))
                  (short-id-name   (elt names 0))
                  (annotation-name (elt names 1))
                  (sequence-name   (elt names 2))
                  (status-name     (elt names 3)))
             ;; I guess right now ann and status buffers just aren't
             ;; checked in this:
             (or (get-buffer sequence-name)
                 (get-buffer short-id-name)

                 ;; They asked for some new buffers.  Let's create them.
                 (elt (gene-create-buffer-quadruplet names) 1))))))

    (set-buffer buffer)
    (gene-frame-show gene-seq-buffer (gene-frame-alignment-frame-params))
    (gene-frame-alignment-frame-other-params)
     
    ;; now: passing id-width arg for now.  What a kluge.
    (gene-display-all (if gene-seq-vertical-scroll-bars 11 10))
    ))
      

(defun gene-open-sequence-file (file)
  "Read the sequences in FILE into the editor, which must be empty."
  (interactive (list
                (if (gene-alignment-empty-p)
                    (read-file-name "Open sequence file: ")
                  (gene-error "Alignment not empty."))))
  (if (gene-alignment-empty-p)
      (gene-read-sequences file)
    (gene-error "Alignment not empty.")))


(defun gene-insert-sequence-file (file)
  "Insert the sequences in FILE into the editor at point."
  (interactive "fInsert sequence file: ")
  (gene-read-sequences file))


(defun gene-read-sequences (file)
  "Read the sequences in FILE into the sequence buffer."
  (interactive "fRead sequence file: ")
  
  ;; Make sure the file exists, is readable, and is in the same format
  ;; as gene-io-current-source-file; the conversion programs are
  ;; certainly capable of signalling the error, but it doesn't look as
  ;; good.
  (unless (file-exists-p file)
    (error "gene-read-sequences: file doesn't exist: %s" file))
  (unless (file-readable-p file)
    (error "gene-read-sequences: file not readable: %s:" file))
  (gene-io-check-format-conflict file)

  ;; If the alignment is empty, then we should set the master source
  ;; file now.  We don't want to save it on other reads, however.
  (if (gene-alignment-empty-p)
      ;; Don't call gene-io-set-source-file; it wants to map over the
      ;; sequences.
      (setq gene-io-current-source-file (expand-file-name file))
    (gene-mark-modified))

  (gene-with-watch-pointer
    (gene-io-with-temp-files
      (let ((format (gene-io-file-format file)))
	(gene-io-insert-file-triplet (gene-io-coerce-to-tabl file format)
				     file format)))
    (save-excursion
      (gene-switch-to-buffer gene-seq-buffer)
      (gene-ann-update-annotation)
      (gene-arrange-windows)
      (gene-handle-vertical-motion t)
      (gene-do-window-sync))))


(defun gene-read-selected-sequences (seq-file locus-file)
  "Read those sequences from SEQ-FILE whose loci appear in LOCUS-FILE.
LOCUS-FILE should be a newline-separated list of strings."
  (gene-io-with-temp-files
    (let ((format (gene-io-file-format seq-file)))
      ;; If the alignment is empty, then we should set the master source
      ;; file now.  We don't want to save it on other reads, however.
      (if (gene-alignment-empty-p)
          ;; Don't call gene-io-set-source-file; it wants to map over the
          ;; sequences.
          (setq gene-io-current-source-file (expand-file-name seq-file))
        (gene-mark-modified))
      (gene-io-insert-file-triplet
       (gene-io-coerce-to-tabl seq-file format 'include locus-file)
       seq-file format)))
  
  (save-excursion
    (set-buffer gene-seq-buffer)
    (gene-ann-update-annotation)
    (if gene-id-highlight-current-id
        (gene-seq-update-current-id-highlight))
    (gene-do-window-sync)))


;;;; Startup code

(defvar gene-default-alignment-name "Unnamed Alignment"
  "*The name to give the alignment when no data files are specified.")


(defvar gene-startup-hook nil
  "Hook run after the gene editor is invoked.
More precisely, it runs after all the sequence files have been loaded
and the windows arranged.
Even more precisely, this hook is the last thing run by
`gene-process-args'.")

(add-hook 'gene-startup-hook 'gene-err-display-accumulated-errors)


(defvar gene-startup-noundo nil
  "Non-nil means we're in the startup code, so don't record sequence
creation for undo.
Only `gene-io-place-sequences' examines this.")


(defun gene-process-args ()
  "Process command-line arguments; create and fill a gene-mode buffer triplet.
Assumes that the command-line arguments ended with 
\"-f gene-process-args SEQFILE-1 SEQFILE-2 ...\"."

  (let ((original-frame (selected-frame))
        (gene-startup-noundo t)
	alignment-files tree-file)

    ;; Build the faces we'll use.
    (gene-make-all-faces)

    ;; Sort out the files to read from the files to show phylo listings for.
    (while (consp command-line-args-left)
      (cond
       ((string= (car command-line-args-left) "--debug")
	(pop command-line-args-left)
	(setq debug-on-error t))
       ((string-match "\\.phylo\\'" (car command-line-args-left))
	(setq tree-file (pop command-line-args-left)))
       (t
	(push (pop command-line-args-left) alignment-files))))
    (setq alignment-files (nreverse alignment-files))

    ;; If the user specified a tree file, visit the files using the
    ;; phylo browser.
    (if tree-file 
	(while alignment-files
	  (gene-phylo-list (pop alignment-files) tree-file))

      ;; Otherwise, just bring them into the alignment buffer.
      (when alignment-files

	;; Name the alignment buffer and frame after the first
	;; alignment file.
	(gene-switch-to-buffer
	 (file-name-nondirectory (car alignment-files)))

	(while (and alignment-files
		    (condition-case dummy
			(progn
			  (gene-read-sequences (pop alignment-files))
			  t)
		      (error nil)))

	  ;; Later files should get appended to end of buffer.
	  (set-buffer gene-seq-buffer)
	  (goto-char (point-max)))))

    ;; We always need to have an alignment around.  Make sure we got
    ;; one.
    (unless (bufferp gene-seq-buffer)
	    (save-window-excursion
	      (gene-create-buffer-quadruplet
	       (gene-make-buffer-names gene-default-alignment-name))))

    ;; We always need a new frame, displaying a phylo browser or an
    ;; alignment.  If there's no extant phylo or alignment frame
    ;; around, create a frame looking at the new alignment.
    (when (not
	   (some (lambda (frame)
		   (let ((buf (window-buffer (frame-selected-window frame))))
		     (memq (save-excursion (set-buffer buf) major-mode)
			   '(gene-phylo-mode gene-seq-mode))))
		 (delq original-frame (frame-list))))
	  (gene-switch-to-buffer gene-default-alignment-name))

    (delete-frame original-frame)

    ;; Make sure any error frame produced by
    ;; gene-io-call-process-carefully is visible.
    (let ((frame (gene-frame-get-frame "*Errors*")))
      (when frame
	    (make-frame-visible frame)
	    (raise-frame frame)))

    ;; Position point correctly in alignment buffer.
    (set-buffer gene-seq-buffer)
    (set-window-point (gene-seq-buffer-window) (point-min))
    (set-window-start (gene-seq-buffer-window) (point-min))
    (gene-handle-vertical-motion)

    ;; Do anything else.
    (run-hooks 'gene-startup-hook 'gene-startup-hooks)

    ;; Paranoia.
    (gene-do-window-sync)
    ))


;;;; Karl's changes to implement the new file IO.
;; This is not necessarily the best grouping criterion, but it'll do
;; for now.

;; Notes:
;; To do this right, we'd probably want to change the fact that source
;; files are associated with subsequences, and simply have a current
;; "master file" to which things are saved.  This would mean big
;; changes to `gene-io-write-regions', I guess.

(defvar gene-io-current-source-file nil
  "The file to which sequences should be saved.")

(defun gene-io-current-source-file ()
  "The file to which sequences should be saved."
  (if (stringp gene-io-current-source-file)
      gene-io-current-source-file
    (setq
     gene-io-current-source-file 
     (expand-file-name (read-file-name "File to save in: ")))))


;; I think it was a premature optimization for us to support multiple
;; "sources" per "sequence" at this stage.  It means we have to do
;; things like physically map over the sequence buffer when changing
;; sources, instead of just dealing with orgsyms.  We don't have any
;; other support for it -- I mean, is there actually a way for us to
;; read in multi-sourced sequences right now?  Is it even on the list
;; for the near future?
;; -kff

(defun gene-io-set-source-file (file &optional format)
  "Change the filename and format of all sequences to FILE and FORMAT.
If FORMAT is nil or omitted, we will attempt to deduce it."
  (let ((format (or format (gene-io-file-format file))))
    (gene-save-excursion-in gene-seq-buffer
      (goto-char (point-min))
      (while (< (point) (point-max))
        (let ((source (gene-source-at-point)))
          (when source
            (setf (gene-source-filename source) file)
            (setf (gene-source-file-format source) format)))
        (goto-char (gene-source-end))))
    (setq gene-io-current-source-file (expand-file-name file))))


(defun gene-io-current-file-format ()
  "Return the current format of the alignment.
This is something like `genbank', `embl', `gdbm', etc.
If the alignment is empty, return nil; it has no format."
  (if (gene-alignment-empty-p)
      nil
    ;; This should never error, but if it does, then it should!
    (gene-save-excursion-in gene-seq-buffer
      (gene-source-file-format (gene-get-here-source)))))


(defun gene-io-check-format-conflict (new-file &optional no-error)
  "Error unless NEW-FILE and `gene-io-current-source-file' are same format.
Return non-nil if they do have the same format.
If optional arg NO-ERROR non-nil, then return nil instead of erroring."
  (let ((source-format (gene-io-current-file-format))
        (new-format (gene-io-file-format new-file no-error)))
    (if source-format
        (or
         (eq source-format new-format)
         (gene-error "Formats not identical: %S, %S"
                     source-format new-format))
      ;; Else there is no current source format, so we can't have a
      ;; conflict.  How nice.  Just return t.
      t)))

(defvar gene-save-hook nil
  "Hook to be run when sequences are saved.")

(defun gene-merge-sequences-internal (&optional sequences)
  "Save all sequences to their respective files.
NOTE: This does a \"merging\" save, not an overwrite.
Optional arg SEQUENCES, a list of syms, means only save those sequences."
  (interactive)
  (gene-with-watch-pointer
    (gene-io-with-temp-files
      (gene-io-write-regions (gene-io-compile-sequence-files sequences))
      (run-hooks 'gene-save-hook))))


(defun gene-save-sequences-as (file &optional sequences overwrite-no-ask)
  "Save the sequences in the current buffer in FILE.
Optional second arg SEQUENCES, a list of syms, means only save those
sequences.  If SEQUENCES is non-nil, then the buffer's master save
file, `gene-io-current-source-file', will not be reset.
Optional third arg OVERWRITE-NO-ASK means don't confirm overwrites."
  (interactive "FSave sequences to file: ")

  ;; We guarantee atomicity and avoid cheezy merging behavior with an
  ;; equally cheezy fix.
  ;; The code of gene-io-write-regions probably needs to be
  ;; substantially changed, but this will do for now.

  (let* ((fullpath (expand-file-name file))
         ;; If we're only saving a subset of the sequences, then we
         ;; don't want to reset the current file to the file they
         ;; happen to get saved to.  We only want to reset it when all
         ;; sequences are saved to a file.
         ;; The temp var below is both flag and data for this.
         (orig-source-file (if sequences (gene-io-current-source-file)))
         (already-exists (file-exists-p fullpath))
         (tmp-file (if already-exists
                       (expand-file-name
                        (make-temp-name "genesv")
                        gene-io-temp-file-dir)))
         (confirm  (or
                    overwrite-no-ask
                    (if already-exists
                        (or
                         (gene-y-or-n-p (format "Overwrite \"%s\"?" file))
                         (gene-error
                          "Overwrite of \"%s\" not confirmed." file))
                      t)))
	 (format (gene-io-current-file-format)))
    (condition-case
        data
        (gene-with-watch-pointer
	 (if already-exists (rename-file fullpath tmp-file))
	 (gene-io-set-source-file file format)
	 ;; Technically, we're doing a merging save, but since the
	 ;; file has been removed, it acts just like an overwrite
	 ;; would have:
	 (gene-merge-sequences-internal sequences)
	 (if orig-source-file
	     (gene-io-set-source-file
	      orig-source-file format))
	 (if already-exists (delete-file tmp-file)))
      (error
       ;; Maybe restore source file, why not?
       (if orig-source-file
           (gene-io-set-source-file orig-source-file format))
       ;; Restore the original contents of the file if there was an
       ;; error:
       (setq gene-error-data data)
       (if already-exists (rename-file tmp-file fullpath))
       (gene-error "%S" data)))))


(defun gene-merge-sequences-to-file (file &optional sequences)
  "Merge sequences into FILE, without destroying the other contents of FILE.
Optional arg SEQUENCES, a list of syms, means only save those sequences."
  ;; Kluge City: it wouldn't really help to modify
  ;; `gene-io-set-source-file' so that it operated on a subset of
  ;; the sequences.  We'd still have to map over them all asking
  ;; `gene-get-here-sym', so it doesn't cost much more to set the
  ;; source while we're there.
  ;;
  ;; Therefore, the plan here is to set
  ;; the source file to the file into which we're saving, do the
  ;; save, then reset the source.  Sigh.
  (let ((orig-source-file (gene-io-current-source-file)))
    (unwind-protect
        (progn
          (gene-io-set-source-file (expand-file-name file))
          (gene-merge-sequences-internal sequences))
      ;; Restore the old sources no matter what.
      (gene-io-set-source-file orig-source-file))))


(defun gene-save-sequences (&optional sequences)
  "Save all sequences to `gene-io-current-source-file'.
If the current file is a GDBM file, then do a merging save; else do an
overwriting save.
Optional arg SEQUENCES, a list of syms, means only save those sequences."
  (interactive)
  
  (let ((filename (gene-io-current-source-file)))

    ;; Make sure that the file is writable.
    (if (and (file-exists-p filename)
	     (not (file-writable-p filename)))
	(error "gene-save-sequences: %s not writable" filename))

    (if (eq (gene-io-file-format filename) 'gdbm)
	(gene-merge-sequences-to-file filename sequences)
      (gene-save-sequences-as filename sequences t))))


;; This probably shouldn't be called right now; I don't think it does
;; what we want it to.
(defun gene-save-sequences-to-original-files (&optional sequences)
  "Save all sequences to their respective files.
NOTE: this loses right now.  It can only obtain the original files
from the `gene-source' properties on the sequences, which means that
if you call it after having saved them all to some other file, the
original files have been lost.
Optional arg SEQUENCES, a list of syms, means only save those sequences."
  (interactive)
  (gene-merge-sequences-internal sequences))


;;;; Closing

(provide 'gene-io)
