;;;; gene-phylo.el --- phylogenetic browsing of organism lists, 2nd try
;;;; Copyright (C) 1994  Indiana University Foundation
;;;;
;;;; This software is copyrighted under the terms of the GNU General
;;;; Public License (see terms below).  The copyright covers this file,
;;;; specifically, and the program of which it is a part.
;;;; 
;;;;      This program is free software; you can redistribute it and/or modify
;;;;      it under the terms of the GNU General Public License as published by
;;;;      the Free Software Foundation; either version 1, or (at your option)
;;;;      any later version.
;;;; 
;;;;      This program is distributed in the hope that it will be useful,
;;;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;;      GNU General Public License for more details.
;;;; 
;;;;      You should have received a copy of the GNU General Public License
;;;;      along with this program; if not, write to the Free Software
;;;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;;;;
;;;; Author: Jim Blandy <jimb@cs.oberlin.edu>
;;;; Maintainer: Jim Blandy <jimb@cs.oberlin.edu>
;;;; Created: Apr 22 1994
;;;; Version: 0.0 (prototype)
;;;; Keywords: genes, ribosomes, sequences, molecular biology, milkshakes

;;; This code is supposed to help people choose the organisms whose
;;; genes they'd like to bring into the sequence buffer.  Given a file
;;; describing a phylogenetic tree and the organisms at its leaves,
;;; this code will build a buffer that allows users to select
;;; organisms or entire taxa, and hide taxa they're not interested in.
;;; When the user is done, we provide a function that returns a list
;;; of all the creatures selected by the user.

(require 'mouse)
(require 'cl)

(require 'gene-io)
(require 'gene-frame)


;;;; Phylo browsing mode

(defvar gene-phylo-mode-hooks nil
  "*Hooks to run when entering Phylogenetic Browsing mode.")

(defvar gene-phylo-sequence-file nil
  "Buffer-local; where to get the sequences mentioned in the present tree.")

(defconst gene-phylo-mode-map
  (let ((map (make-sparse-keymap "Phylogenetic Browsing")))

    ;; Get the global selection stuff out of the way.
    (define-key map [down-mouse-1] 'gene-phylo-dispatch-click)
    (define-key map [double-down-mouse-1] 'gene-phylo-dispatch-click)
    (define-key map [triple-down-mouse-1] 'gene-phylo-dispatch-click)
    (define-key map [S-down-mouse-1] 'gene-phylo-dispatch-click)

    (define-key map [down-mouse-2] 'gene-phylo-absorb-mouse-down)
    (define-key map [mouse-2] 'undefined)
    (define-key map [double-down-mouse-2] 'gene-phylo-absorb-mouse-down)
    (define-key map [double-mouse-2] 'undefined)
    (define-key map [triple-down-mouse-2] 'gene-phylo-absorb-mouse-down)
    (define-key map [triple-mouse-2] 'undefined)

    (define-key map [down-mouse-3] 'gene-phylo-pop-up-menu)

    map)
  "Keymap for Phylogenetic Browsing mode.")

(defvar gene-phylo-mode-buffers nil
  "A list of the buffers that have been in Phylogenetic Browsing mode.
This list is used to generate the phylo browser portion of the Views menu.
It's okay for it to contain deleted buffers, or buffers in other
modes;  We clean it up before generating the menu.")

;;;###autoload
(defun gene-phylo-mode ()
  "Major mode for browsing phylogenetic listings.

Click on an organism to select/unselect it for inclusion in the
sequence buffer.  Click on a taxon to include/exclude all the
organisms it contains.

You may click on a taxon or organism's name to select it, and click on
the `...' or `>>>' after a taxon's name to open or close it.

The right mouse button provides a menu offering commands to apply to
the taxon or organism clicked upon.

A full listing of the commands available:
\\<gene-phylo-mode-map>\\{gene-phylo-mode-map}"
  (interactive)
  (setq major-mode 'gene-phylo-mode)
  (setq mode-name "Phylogenetic Browsing")
  (use-local-map gene-phylo-mode-map)
  (buffer-disable-undo (current-buffer))
  (setq truncate-lines t)
  (setq mode-line-format '("---- Phylogenetic Listing of "
			   gene-phylo-sequence-file
			   " %-"))

  ;; The phylo browser should have the same pointer shape as the
  ;; alignment frame, I think.  -Karl
  (setq x-pointer-shape gene-seq-default-pointer-shape)
  (set-mouse-color
   (cdr (assq 'mouse-color (frame-parameters (selected-frame)))))


  ;; Make sure to do this after use-local-map.
  (gene-menu-bar-make-hide-button)

  (make-local-variable 'gene-phylo-server-buffer)
  (make-local-variable 'gene-phylo-sequence-file)

  ;; Arrange to kill the server when this buffer goes away.
  (add-hook 'kill-buffer-hook 'gene-phylo-kill-server)

  ;; Arrange for this buffer to appear on the Views menu.
  (push (current-buffer) gene-phylo-mode-buffers)

  (run-hooks 'gene-phylo-mode-hooks))

;;;###autoload
(put 'gene-phylo-mode 'mode-class 'special)

(defun gene-phylo-frame-parameters ()
  "Return a parameter list appropriate for a phylogenetic listing frame."
  (list (cons 'foreground-color     gene-foreground-color)
	(cons 'background-color     gene-background-color)
	(cons 'cursor-color         gene-cursor-color)
	(cons 'mouse-color          gene-foreground-color)
	(cons 'vertical-scroll-bars t)))

(defun gene-phylo-make-faces ()
  "Create the appropriate faces for the phylo browser.
`gene-make-all-faces' calls this.

Note that it's incorrect to make faces in top-level expressions, since
the construction of a face depends on run-time factors (like the fonts
available)."

  (copy-face 'default 'phylo-selected)
  (set-face-background
   'phylo-selected (gene-check-color-allowable "medium orchid"))

  (copy-face 'default 'phylo-mouse-unselected)
  (make-face-bold 'phylo-mouse-unselected nil t)

  (copy-face 'default 'phylo-mouse-selected)
  (make-face-bold 'phylo-mouse-selected nil t)
  (set-face-background
   'phylo-mouse-selected
   ;; redundant:
   (gene-check-color-allowable (face-background 'phylo-selected)))

  ;; This face should cover phylo-mouse-selected when applied to an
  ;; overlay.
  (let* ((parameters (gene-frame-alignment-frame-params))
         (foreground (cdr (assq 'foreground-color parameters)))
         (background (cdr (assq 'background-color parameters)))
         (font       (cdr (assq 'font             parameters))))
    (copy-face 'default 'phylo-cover-region)
    (set-face-foreground
     'phylo-cover-region (gene-check-color-allowable foreground))
    (set-face-background
     'phylo-cover-region (gene-check-color-allowable background))
    ;; This apparently got an error on some machines, sometimes, so it
    ;; was commented out.  But that was a long time ago; let's put it
    ;; back and see what happens:
    (set-face-font 'phylo-cover-region font)
    ))


;;;; Running and talking to the phylo server subprocess --- low-level stuff.

(defconst gene-phylo-server "phylo"
  "The program to run to do the heavy tree-processing work.")

(defconst gene-phylo-protocol-version 7
  "The version of the protocol this client knows how to use.")

(defvar gene-phylo-server-buffer nil
  "The buffer through which we communicate with the phylo server.")

(defvar gene-phylo-reply-start nil
  "Where the next unprocessed server reply starts.
This is a position (not a marker) in gene-phylo-server-buffer.")

(defvar gene-phylo-server-timeout 20
  "The number of seconds we'll wait for a server reply before giving up.")

(defun gene-phylo-start-server ()
  "Start the phylogenetic processing server."

  ;; Make sure the buffer exists.
  (unless (and (bufferp gene-phylo-server-buffer)
	       (buffer-name gene-phylo-server-buffer))
    (setq gene-phylo-server-buffer (generate-new-buffer "*phylo server*")))

  (save-excursion
    (let ((buf gene-phylo-server-buffer))
      (set-buffer buf)
      (make-local-variable 'gene-phylo-server-buffer)
      (setq gene-phylo-server-buffer buf))
    (erase-buffer)
    (make-local-variable 'gene-phylo-reply-start)
    (setq gene-phylo-reply-start (point))

    ;; Kill off any old servers.
    (let ((old-proc (get-buffer-process (current-buffer))))
      (when (processp old-proc)
	;; Don't stick a termination message in the buffer; that'll
	;; get in the way of reading replies.
	(set-process-sentinel old-proc
			      (lambda (process string) nil))
	(kill-process old-proc)))

    ;; Start a new server.
    (let ((server
	   (let ((process-connection-type nil))
	     (start-process "*phylo server*" (current-buffer)
			    gene-phylo-server))))
      (process-kill-without-query server))))

(defun gene-phylo-maybe-start-server ()
  "Start the phylogenetic processing server, if it's not already running.
If it is already running, leave it alone."
  (unless (and (bufferp gene-phylo-server-buffer)
	       (buffer-name gene-phylo-server-buffer)
	       (let ((process (get-buffer-process gene-phylo-server-buffer)))
		 (and (processp process)
		      (eq (process-status process) 'run))))
    (gene-phylo-start-server)))

(defun gene-phylo-check-server-health (process)
  "Make sure the phylo server is still running.
If it's not, signal an error."
  (unless (and (processp process)
	       (eq (process-status process) 'run))
    (error "gene-phylo-check-server-health: It's dead, Jim.")))

(defun gene-phylo-kill-server ()
  "Kill the phylo server associated with the current buffer."
  (and (eq major-mode 'gene-phylo-mode)
       (bufferp gene-phylo-server-buffer)
       (buffer-name gene-phylo-server-buffer)
       (kill-buffer gene-phylo-server-buffer)))

(defun phylo-send-request (request)
  "Send REQUEST (a string) to the phylo server.  Do not wait for a reply.
REQUEST must be terminated by a newline.
Signal an error if there are unread replies from previous requests.
If you're feeling asynchronous, REQUEST may actually be several
requests separated by newlines."
  (save-excursion
    (set-buffer gene-phylo-server-buffer)
    (let ((process (get-buffer-process (current-buffer))))
      (gene-phylo-check-server-health process)
      (goto-char gene-phylo-reply-start)
      (if (< (point) (point-max))
	  (error 
	   "phylo-send-request: But you haven't finished the last one!"))
      (insert request)
      (setq gene-phylo-reply-start (point))
      (set-marker (process-mark process) (point))
      (process-send-string process request))))

(defun gene-phylo-read-reply ()
  "Wait for a reply from the server, read it, and return it."
  (save-excursion
    (set-buffer gene-phylo-server-buffer)
    (let ((process (get-buffer-process (current-buffer))))
      (gene-phylo-check-server-health process)

      ;; Is there a newline string there, indicating a complete reply?
      (while (progn
	       (goto-char gene-phylo-reply-start)
	       (not (search-forward "\n" nil t)))

	;; Not yet; let's wait a bit.
	(unless (accept-process-output process gene-phylo-server-timeout)
	  (delete-process process)
	  (error "gene-phylo-read-reply: server seems to be hung")))

      (goto-char gene-phylo-reply-start)
      (prog1 (read (current-buffer))
	(unless (eolp)
	  (error "gene-phylo-read-reply: garbled reply"))
	(forward-line 1)
	(setq gene-phylo-reply-start (point))))))


;;;; Examining the tree text in the buffer

(defun gene-phylo-parse ()
  "Parse the line at point, and return a value describing its contents.
The return value takes one of the following forms:
\(interior NAME DESCRIPTION SELECTED-P OPEN-P\)
   meaning that the line is an interior node with the given properties, or
\(leaf     NAME DESCRIPTION SELECTED-P\)
   meaning that the line is a leaf node."
  (save-excursion
    (beginning-of-line)
    (cond
     ((looking-at " *\\([^ ]+\\) +\\(\\([^ ].*\\) +\\)?\\(\\.\\.\\.\\|>>>\\)$")
      (list 'interior
	    ;; name
	    (format "%s" (buffer-substring-no-properties (match-beginning 1) 
                            (match-end 1)))
	    ;; description
	    (if (match-beginning 3)
		(buffer-substring-no-properties (match-beginning 3) (match-end 3)))
	    ;; selected
	    (not (null (get-text-property (match-beginning 1) 'face)))
	    ;; open
	    (= (char-after (match-beginning 4)) ?>)))
     ((looking-at " *\\([^ ]+\\) +\\(.*\\)$")
      (list 'leaf
	    ;; name
	    (format "%s" (buffer-substring-no-properties (match-beginning 1) 
                           (match-end 1)))
	    ;; description
	    (buffer-substring-no-properties (match-beginning 2) (match-end 2))
	    ;; selected
	    (not (null (get-text-property (match-beginning 1) 'face)))))
     (t
      (error
       "gene-phylo-parse: unrecognized line.")))))


(defsubst gene-phylo-name (parsed)
  "Return the name of the node described by PARSED."
  (elt parsed 1))

(defsubst gene-phylo-description (parsed)
  "Return the description of the node described by PARSED."
  (elt parsed 2))

(defsubst gene-phylo-taxon-p (parsed)
  "Return non-nil iff the node described by PARSED has subnodes."
  (eq (elt parsed 0) 'interior))

(defsubst gene-phylo-selected-p (parsed)
  "Return non-nil iff the node described by PARSED is selected."
  (elt parsed 3))

(defsubst gene-phylo-open-p (parsed)
  "Return non-nil iff PARSED describes a taxon whose subnodes are displayed."
  (elt parsed 4))


(defun gene-phylo-node ()
  "Return the name of the node at point."
  (gene-phylo-name (gene-phylo-parse)))


(defun gene-phylo-in-ellipsis-p ()
  "Return non-nil iff point is in an ellipsis."
  (save-excursion
    (let ((point-in-question (point)))
      (end-of-line)
      (backward-char 3)
      (and (looking-at "\\(>>>\\|\\.\\.\\.\\)")
	   (<= (point) point-in-question)))))


(defun phylo-find-node-forward (node &optional no-error)
  "Search forward for the line for NODE; put point at beginning of line.
If optional argument NO-ERROR is non-nil, don't signal an error if you
can't find it."
  (let ((case-fold-search nil)
	pos)
    (while (and (< (point) (point-max))
		(search-forward node nil t))
      (when (string= (gene-phylo-node) node)
	(setq pos (point))
	(goto-char (point-max))))
    (cond
     (pos (goto-char pos)
	  (beginning-of-line))
     (no-error nil)
     (t
      (error "phylo-find-node-forward: couldn't find %S" node)))))

(defmacro phylo-save-excursion (&rest forms)
  "Evaluate FORMS like progn, but try to keep point on the same node."
  (list 'let
	'((phylo-save-excursion-node (if (eobp) 'eob (gene-phylo-node))))
	(list 'unwind-protect
	      (cons 'progn forms)
	      '(goto-char (point-min))
	      '(if (eq phylo-save-excursion-node 'eob)
		   (goto-char (point-max))
		 (phylo-find-node-forward phylo-save-excursion-node
					  'no-error)))))


;;;; Phylo server requests

(defun phylo-simple-request (request &rest args)
  "Send (format REQUEST ARGS ...) to the phylo server, and process replies."
  (phylo-send-request (apply 'format request args))
  (phylo-read-replies))

(defun phylo-request-check-version (version)
  "Verify that the server is speaking protocol version VERSION."
  (phylo-simple-request "check-version %s\n" version))

(defun phylo-request-use-tree (tree-file org-file)
  "Read and display the tree in TREE-FILE, showing organisms in ORG-FILE. 
TREE-FILE should contain a tree in the RDP `.phylo' format.
ORG-FILE should be a list of organism names, separated by newlines.
The tree is minimized to display those organisms."
  (phylo-simple-request "use-tree %S %S\n" 
			(expand-file-name tree-file)
			(expand-file-name org-file)))

(defun phylo-request-select (node)
  "Select or unselect the node called NODE, and redisplay."
  (phylo-simple-request "select %S\n" node))

(defun phylo-request-select-range (start-node end-node)
  "Select or unselect the region from START-NODE to END-NODE."
  (phylo-simple-request "select-range %S %S\n" start-node end-node))

(defun phylo-request-open (node)
  "Open or close the node called NODE, and redisplay."
  (phylo-simple-request "open %S\n" node))

(defun phylo-request-list-selected (filename)
  "List the selected nodes to FILENAME."
  (phylo-simple-request "list-selected %S\n" filename))

(defun phylo-request-unselect-all ()
  "Unselect all nodes."
  (phylo-simple-request "unselect-all\n"))

(defun phylo-request-open-all ()
  "Open all nodes."
  (phylo-simple-request "open-all\n"))

(defun phylo-request-open-groups ()
  "Close nodes which contain only leaves.  Open all others."
  (phylo-simple-request "open-groups\n"))

(defun phylo-request-open-selected ()
  "Open nodes which have any selected leaves.  Close all others."
  (phylo-simple-request "open-selected\n"))

(defun phylo-request-open-subtree (subtree)
  "Open all nodes in SUBTREE."
  (phylo-simple-request "open-subtree %S\n" subtree))


;;;; Phylo server replies

(defun phylo-read-replies ()
  "Read and process replies from the phylo server.
When the phylo server sends a `phylo-okay', we return.
We signal an error if the phylo server sends a `phylo-error' reply."
  (save-excursion
    (catch 'phylo-okay			; Thrown by phylo-okay.
      (while t (eval (gene-phylo-read-reply))))))

(defun phylo-display-all (text)
  "Replace the current buffer's contents with TEXT."
  (delete-region (point-min) (point-max))
  (insert text)
  (goto-char (point-min)))

(defun phylo-display-range (start end text)
  "Replace the region between nodes START and END with TEXT."
  (goto-char (point-min))
  (let ((start-posn (progn (phylo-find-node-forward start)
			   (point)))
	(end-posn (progn (phylo-find-node-forward end)
			 (forward-line 1)
			 (point))))
    (delete-region start-posn end-posn)
    (goto-char start-posn)
    (insert text)))

(defun phylo-error (string)
  "Signal an error, displaying STRING."
  (error "%s" string))

(defun phylo-message (string)
  "Display STRING in the minibuffer."
  (message string))

(defun phylo-okay ()
  "Stop reading replies."
  ;; Caught by phylo-read-replies.
  (throw 'phylo-okay nil))



;;;; Commands of phylo-mode

(defun gene-phylo-toggle-open ()
  "Toggle the openness of the node at point."
  (interactive)
  (phylo-save-excursion
   (phylo-request-open (gene-phylo-node))))

(defun gene-phylo-toggle-selected ()
  "Toggle the selectedness of the node at point."
  (interactive)
  (phylo-save-excursion
   (phylo-request-select (gene-phylo-node))))

(defun gene-phylo-select-nodes-in (start end)
  "Toggle the selectedness of the nodes between START and END."
  (phylo-save-excursion
   (phylo-request-select-range start end)))

(defun gene-phylo-show-all ()
  "Open all nodes."
  (interactive)
  (phylo-save-excursion
   (phylo-request-open-all)))

(defun gene-phylo-show-groups ()
  "Close nodes which contain only leaves.  Open all others."
  (interactive)
  (phylo-save-excursion
   (phylo-request-open-groups)))

(defun gene-phylo-show-selected ()
  "Open nodes which contain selected children.  Close all others."
  (interactive)
  (phylo-save-excursion
   (phylo-request-open-selected)))

(defun gene-phylo-open-subtree ()
  "Open all nodes in the subtree at point."
  (interactive)
  (phylo-save-excursion
   (phylo-request-open-subtree (gene-phylo-node))))

(defun gene-phylo-absorb-mouse-down ()
  "Ignore a mouse down event."
  (interactive))

(defun gene-phylo-dispatch-click (event)
  "Handle a mouse down-click in the phylogenetic browsing buffer.
This looks at the kind and portion of the node clicked on to decide
what command to invoke.

Emacs somehow doesn't seem to handle the `local-map' text property
correctly.  If we ever put our finger on what's going on, fixing
the problem might make this function unnecessary."
  (interactive "e")
  (mouse-set-point event)
  (if (eobp)
      (forward-line -1))
  (if (gene-phylo-in-ellipsis-p)
      (progn
	(gene-phylo-toggle-open)
	(read-event))
    (gene-phylo-mouse-select event)))

(defun gene-phylo-menu-read ()
  "Read the currently selected sequences into the sequence buffer."
  (interactive)
  (gene-with-watch-pointer
    (gene-io-with-temp-files
      (let ((phylo-buf (current-buffer))
	    (seq-file gene-phylo-sequence-file)
	    (locus-file (gene-io-temp-file-name "genelo")))
	(phylo-request-list-selected locus-file)

	;; Switch to the appropriate buffer triplet.
	(let ((other (gene-other-buffer phylo-buf)))
	  (gene-switch-to-buffer (or other gene-default-alignment-name)))

	(gene-read-selected-sequences seq-file locus-file)

	;; Unselect everything.  This way, the user doesn't have to be
	;; vigilant to avoid inserting the same sequences again the next
	;; time she uses the phylo browser.
	(save-excursion
	  (set-buffer phylo-buf)
	  (phylo-request-unselect-all))))))


;;;; Menus --- pop-up and menu bar

;;; Here's the menu hierarchy used for both the pop-up menus and the
;;; menu bar.  When invoked from the menu bar, context-sensitive items
;;; are based on the location of point.
;;;
;;;    File
;;; 	  Insert Selected Sequences
;;; 	  Open Phylogenetic Listing
;;; 	  Quit
;;;   
;;;    Edit
;;; 	  Select/Unselect <current node>
;;;   
;;; 	  Find Organism/Taxon
;;;   
;;;    Views
;;;       Show/Hide <taxon>
;;;       Show <taxon> Complete Contents
;;;       Show All
;;;       Show Only Groups
;;;       Show Only Selected
;;;   
;;;       (The following appear on the menu bar, but not on the pop-up.)
;;; 	  Show Group List
;;; 	  Show Lock Status
;;; 	  Show Annotations
;;; 	  Show Column Stats
;;; 	  Show <phylo list> ...


(defun gene-phylo-file-pane ()
  "Return the File menu/pane for the phylo browser frame."
  (list
   "-*- File -*-"
   ""
   '("Insert Selected Sequences" . gene-phylo-menu-read)
   '("Open Phylogenetic Listing" . gene-phylo-list)
   '("Close" . gene-phylo-close)
   ""
   (if gene-modified-p
       '("Quit (offers to save)" . gene-quit-maybe-offer-save)
     '("Quit" . gene-quit-maybe-offer-save))))


(defun gene-phylo-edit-pane ()
  "Return the Edit menu/pane for the phylo browser frame.
The menu is tailored for the organism at point."
  (let* ((parsed (gene-phylo-parse))
	 (taxon-p (gene-phylo-taxon-p parsed))
	 (selected-p (gene-phylo-selected-p parsed))
	 (title (if taxon-p
		    (or (gene-phylo-description parsed)
			(gene-phylo-name parsed))
		  (gene-phylo-name parsed))))

    (list
     "-*- Edit -*-"
     ""

     ;; Select/Unselect <current node>
     (let ((item (format "%s %s"
			 (if (eq selected-p t) "Unselect" "Select")
			 title)))
       (if (eq selected-p 'unavailable)
	   item
	 (cons item 'gene-phylo-toggle-selected)))

     ""
     ;; Not implemented yet:
     "Search for Organism/Taxon")))


(defun gene-phylo-views-subpane ()
  "Return the items common to the phylo's pop-up and menu bar Views menus.
The items are tailored for the organism at point."
  (let* ((parsed (gene-phylo-parse))
	 (taxon-p (gene-phylo-taxon-p parsed))
	 (title (if taxon-p
		    (or (gene-phylo-description parsed)
			(gene-phylo-name parsed))
		  (gene-phylo-name parsed))))
    (list

     ;; Show/Hide <taxon> Contents
     (let ((item (format "%s %s Contents"
			 (if (and taxon-p (gene-phylo-open-p parsed))
			     "Hide" "Show")
			 title)))
       (if taxon-p
	   (cons item 'gene-phylo-toggle-open)
	 item))

     ;; Show <taxon> Complete Contents
     (let ((item (format "Show %s Complete Contents" title)))
       (if taxon-p
	   (cons item 'gene-phylo-open-subtree)
	 item))

     '("Show All" . gene-phylo-show-all)
     '("Show Only Groups" . gene-phylo-show-groups)
     '("Show Only Selected" . gene-phylo-show-selected))))


(defun gene-phylo-views-menubar-pane ()
  (append '("-*- Views -*-"
	    "")
	  (gene-phylo-views-subpane)
	  '("")
	  (gene-phylo-common-views)))


(defun gene-phylo-views-popup-pane ()
  (list* "-*- Views -*-"
	 ""
	 (gene-phylo-views-subpane)))


(defun gene-phylo-common-views ()
  "Construct a list of hide/show items for phylo frames.
These items appear in the Views menu of both the phylo and alignment frames.
This works by consulting `gene-phylo-mode-buffers'."

  (save-excursion
    (let* ((head (cons 'dummy gene-phylo-mode-buffers))
	   (elt head)
	   (items nil))

      (while (consp (cdr elt))
	(let ((buf (cadr elt)))
	  
	  ;; If buf hasn't been deleted, and it's in gene-phylo-mode,
	  ;; make a menu item for it.
	  (if (and (buffer-name buf)
		   (progn (set-buffer buf)
			  (eq major-mode 'gene-phylo-mode)))
	      (let* ((window (get-buffer-window buf 0))
		     (frame (and window
				 (window-live-p window)
				 (window-frame window)))
		     (visible-p (if frame (eq t (frame-visible-p frame))))
		     (title (format "%s Phylo Listing Of %s"
				    (if visible-p "Hide" "Show")
				    (file-name-nondirectory 
				     gene-phylo-sequence-file)))
		     (closure
		      (list 'lambda '() '(interactive)
			    (if visible-p
				(list 'delete-frame frame)
			      (list 'gene-frame-show
				    buf 
				    '(gene-phylo-frame-parameters))))))
		(push (cons title closure) items)
		(setq elt (cdr elt)))

	    ;; The buf has been deleted, or it's no longer in
	    ;; gene-phylo-mode.  Remove it from the list.
	    (setcdr elt (cddr elt)))))

      ;; Put the purified list back in gene-phylo-mode-buffers.
      (setq gene-phylo-mode-buffers (cdr head))

      (sort items
	    (lambda (item1 item2)
	      (string-lessp (car item1) (car item2)))))))


(defun gene-phylo-pop-up-menu (event)
  "Pop up the phylogenetic browsing menu, and run the user's choice."
  (interactive "e")
  (mouse-set-point event)
  (if (eobp) (forward-line -1))
  (let ((choice (x-popup-menu event
			      (list "-*- Phylo Pop-up -*-"
				    (gene-phylo-views-popup-pane)
				    (gene-phylo-file-pane)
				    (gene-phylo-edit-pane)))))
    (if choice
	(call-interactively choice))))


(defun gene-phylo-menubar (pane)
  (let ((choice (x-popup-menu t (list "untitled" pane))))
    (if choice (call-interactively choice))))


(defun gene-phylo-file-menubar ()
  (interactive)
  (gene-phylo-menubar (gene-phylo-file-pane)))


(defun gene-phylo-edit-menubar ()
  (interactive)
  (gene-phylo-menubar (gene-phylo-edit-pane)))


(defun gene-phylo-views-menubar ()
  (interactive)
  (gene-phylo-menubar (gene-phylo-views-menubar-pane)))


;;; Put them all together on the menu.  These entries should appear in
;;; right-to-left order.
(define-key gene-phylo-mode-map [menu-bar gene-views]
  (cons "Views" 'gene-phylo-views-menubar))
(define-key gene-phylo-mode-map [menu-bar phylo-select]
  (cons "Edit" 'gene-phylo-edit-menubar))
(define-key gene-phylo-mode-map [menu-bar phylo-file]
  (cons "File" 'gene-phylo-file-menubar))



;;;; Mouse dragging

;;; Much of this code comes from Mike William's mouse-sel package,
;;; which is available under the GPL.  The full copyright notice is:

;;; ;;; mouse-sel.el --- Multi-click selection support for Emacs 19
;;; 
;;; ;; Copyright (C) Mike Williams 1993
;;; 
;;; ;; Author: Mike Williams <mikew@gopher.dosli.govt.nz>
;;; ;; Keywords: mouse
;;; ;; Version: Revision: 2.1
;;; 
;;; ;; This file is not (yet) part of GNU Emacs, but is made available under
;;; ;; the same conditions.
;;; ;;
;;; ;; GNU Emacs is free software; you can redistribute it and/or modify
;;; ;; it under the terms of the GNU General Public License as published by
;;; ;; the Free Software Foundation; either version 2, or (at your option)
;;; ;; any later version.
;;; ;;
;;; ;; GNU Emacs is distributed in the hope that it will be useful,
;;; ;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; ;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; ;; GNU General Public License for more details.

(defun gene-phylo-mouse-select (event)
  "Select taxa using the mouse.

On click, select the taxon clicked on.
Dragging selects all taxa included in the drag.

This should be bound to a down-mouse event."
  (interactive "e")
  (mouse-set-point event)
  (beginning-of-line)
  (let* ((node (if (eobp) nil (gene-phylo-parse)))
         (new-selected-p (and node (not (gene-phylo-selected-p node))))
         (original-face (overlay-get mouse-drag-overlay 'face))
         (original-mouse-face (overlay-get mouse-drag-overlay 'mouse-face))
         (new-face (if new-selected-p 'phylo-selected 'phylo-cover-region)))
    (unwind-protect
        (progn
          (overlay-put mouse-drag-overlay 'face new-face)
	  (overlay-put mouse-drag-overlay 'mouse-face new-face)

	  (save-restriction
	    (narrow-to-region (point) (point)))

	  (let* ((selection-bounds
		  (gene-phylo-mouse-drag (posn-window (event-end event))
					 (posn-point (event-end event))))
		 (start-node (progn (goto-char (elt selection-bounds 0))
				    (gene-phylo-node)))
		 (end-node (progn (goto-char (elt selection-bounds 1))
				  (gene-phylo-node))))
	    (gene-phylo-select-nodes-in start-node end-node)))
      (overlay-put mouse-drag-overlay 'face original-face)
      (overlay-put mouse-drag-overlay 'mouse-face original-mouse-face))))


(defsubst gene-phylo-drag-overlay (overlay root-left root-right)
  "Move OVERLAY to enclose point, given an anchor from ROOT-LEFT to ROOT-RIGHT.
OVERLAY is moved to enclose the line containing point, and the line
delimited by ROOT-LEFT and ROOT-RIGHT.
OVERLAY is assumed to be in the correct buffer already."

  (cond

   ((< (point) root-left)
    (beginning-of-line)
    (move-overlay overlay (point) root-right))

   ((> (point) root-right)
    (forward-line 1)
    (move-overlay overlay root-left (point))
    (forward-line -1))

   (t (move-overlay overlay root-left root-right))))


(defsubst gene-phylo-event-row (posn line-height frame)
  "Return the row number of the mouse position POSN, or nil if we can't tell.
Assume the frame's lines are line-height pixels tall."
  (let ((window (posn-window posn))
	(y (cdr (posn-x-y posn))))
    (cond
     ((and (framep window)
	   (eq window frame))
      (/ y line-height))
     ((and (windowp window)
	   (eq (window-frame window) frame))
      (+ (/ y line-height)
	 (elt (window-edges window) 1)))
     (t nil))))


(defun gene-phylo-mouse-drag (window root-posn)
  "Process a mouse drag in the phylo browser buffer.
ROOT-POSN specifies the drag's rooted end; WINDOW says which window to drag in.
Let the mouse vary the other end of the selection.

When the user releases the mouse button, return a list (START END),
giving the positions of the first and last lines in the selection.
START is the position of the beginning of the line where the drag
started; END is the position of the beginning of the line where the
drag was released.  NOTE THAT THIS IS NOT THE USUAL REPRESENTATION FOR
REGIONS.  If the selection includes only one line, START and END will
be equal.  There is no representation for an empty selection.  This is
a bit weird, but it is simpler for the callers to work with.

If the mouse goes below or above the window during a drag, scroll the
window to show more text in the direction of the mouse."

  (let* ((frame (window-frame window))
	 (line-height (/ (frame-pixel-height frame)
			 (frame-height frame)))
	 (buffer (window-buffer window))
	 (top (elt (window-edges window) 1))
	 (bottom (- (elt (window-edges window) 3) 2))
	 (root-left (progn (goto-char root-posn)
			   (beginning-of-line)
			   (point)))
	 (root-right (progn (forward-line 1)
			    (point)))
	 (orig-cursor-type (list (assq 'cursor-type (frame-parameters frame))))
	 event event-type scrolling)

    (select-window window)

    ;; Point is the "growth tip" of the selection.  It always sits at
    ;; the beginning of the line containing the mouse cursor, when the
    ;; mouse is in the window.
    (goto-char root-posn)

    ;; Put the mouse overlay in the right buffer.
    (move-overlay mouse-drag-overlay root-left root-right buffer)

    (unwind-protect
	(track-mouse

	  ;; Use the bar cursor while dragging --- it's less confusing.
	  (modify-frame-parameters frame '((cursor-type . bar)))

	  ;; Handle dragging.
	  (while (progn

		   (gene-phylo-drag-overlay 
		    mouse-drag-overlay root-left root-right)

		   ;; Read an event.  Ignore switch-frame events.  If
		   ;; we're scrolling, time out after 0.2 seconds.
		   (if (and scrolling (sit-for 0.2))
		       (setq event-type 'auto-scroll)
		     (while (eq (setq event-type
				      (event-basic-type 
				       (setq event 
					     (read-event))))
				'switch-frame)))

		   (message "%S" event-type)

		   ;; Continue the loop if this expression is true.
		   (memq event-type '(auto-scroll mouse-movement)))

	    ;; Position point according to the mouse.
	    (let* ((posn (event-start event))
		   (posn-window (posn-window posn))
		   (posn-point (posn-point posn)))
	      
	      ;; If we're in the text area of the original window,
	      ;; we can do this pretty cheaply.
	      (if (and (eq posn-window window)
		       (numberp posn-point))
		  (progn
		    (goto-char posn-point)
		    (beginning-of-line)
		    (setq scrolling nil))

		;; Where is the mouse, vertically?
		(let ((row (gene-phylo-event-row
			    posn line-height frame)))

		  ;; Position point according to the mouse.
		  (setq scrolling
			(cond
			 ((null row) nil)
			 ((< row top)
			  (goto-char (window-start window))
			  (vertical-motion -2 window)
			  (beginning-of-line)
			  (set-window-start window (point))
			  t)
			 ((> row bottom)
			  (goto-char (window-start window))
			  (vertical-motion 2 window)
			  (beginning-of-line)
			  (set-window-start window (point))
			  (vertical-motion (- bottom top 1))
			  t)
			 (t
			  (goto-char (window-start window))
			  (vertical-motion (- row top))
			  (beginning-of-line)
			  nil))))))))
		
      ;; end track-mouse

      ;; unwinding clauses:

      ;; Restore cursor
      (modify-frame-parameters (selected-frame) orig-cursor-type)

      ;; Remove overlay
      (delete-overlay mouse-drag-overlay))

    ;; Return a list (START END), giving the beginnings of the lines
    ;; where the drag started and ended.
    (list root-left (point))))


;;;; Indexing a sequence file

(defun gene-phylo-list-buffer-name (filename)
  "Return the name of the buffer used to store the phylo list for FILENAME."
  (format "Phylogenetic listing for %s" filename))

(defun gene-phylo-list-noselect (seq-file tree-file)
  "Return a phylogenetic listing buffer for SEQ-FILE and TREE-FILE.
SEQ-FILE should contain an alignment.
TREE-FILE should be a phylogenetic tree."
  (save-excursion
    (let ((name (gene-phylo-list-buffer-name seq-file)))
      (set-buffer (get-buffer-create name))
      (erase-buffer)
      (gene-phylo-mode)
      (setq gene-phylo-sequence-file seq-file)

      (gene-phylo-start-server)
      (phylo-request-check-version gene-phylo-protocol-version)
      (gene-io-with-temp-files
	(let ((ids (gene-io-list-ids seq-file
				     (gene-io-file-format seq-file))))
	  (phylo-request-use-tree tree-file ids)))

      ;; Put point somewhere helpful.
      (goto-char (point-min))
      (forward-line 1)
      (current-buffer))))


(defun gene-phylo-list (seq-file tree-file)
  "Present a phylogenetic listing for SEQ-FILE, using the tree in TREE-FILE."
  (interactive
   (let* ((sequence-file
	   (read-file-name "File to list phylogenetically: " nil nil t))
	  (default-phylo-file
	    (if (string-match "\\`.*\\(\\.[^.]*\\'\\)" sequence-file)
		(concat (substring sequence-file 0 (match-beginning 1))
			".phylo")
	      (concat sequence-file ".phylo")))
	  (default-phylo-file (if (file-exists-p default-phylo-file)
				  default-phylo-file))
	  (prompt (format "Tree file: %s"
			  (if default-phylo-file
			      (format "(default %s) " default-phylo-file)
			    "")))
	  (phylo-name (read-file-name prompt nil default-phylo-file t nil)))
     (list sequence-file phylo-name)))
  (gene-frame-show (gene-phylo-list-noselect (expand-file-name seq-file)
					     (expand-file-name tree-file))
		   (gene-phylo-frame-parameters)))


(defun gene-phylo-close ()
  (interactive)
  (unless (eq major-mode 'gene-phylo-mode)
    (error "gene-phylo-close: applied to non-phylo buffer"))
  (let* ((window (get-buffer-window (current-buffer) 0))
	 (frame (window-frame window)))
    (gene-phylo-kill-server)
    (kill-buffer (current-buffer))
    (delete-frame frame)))


;;;; The End.
;;;; "The killer arose at dawn.  He put his boots on..."
(provide 'gene-phylo)
