;; gene-posmon.el (part of ale).
;; Copyright (C) 1993, 1994  Karl Fogel <kfogel@darwin.life.uiuc.edu>
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;;
;; Search for instances of "todo:" to find notes about future plans.


;;;; Position monitoring in current sequence and reference sequences.
(require 'gene-mode-tools)

;; The Plan:
;; We have a separate frame, which shows a bunch of sequences and the
;; current position in each (base & column).  The current sequence
;; is the first of those; the rest of them are known as "reference
;; sequences".  When you're in a reference sequence, its stats just
;; show twice, that's all.
;;
;; The Implementation:
;;
;; The prefix for this feature is "gene-posmon", for "position
;; monitoring".
;;
;; `gene-posmon-monitoring-on' non-nil means do the monitoring.
;; It is only accessible through the `gene-posmon-monitoring-on-off'
;; function, which takes positive arg to toggle on (& create the
;; frame, etc), and anything else to turn off.
;;
;; The reference sequences are kept in a list:
;; `gene-posmon-ref-sequences'.
;;
;; After every command, the following things happen:
;; The current seq and each ref sequence are visited in turn, the
;; current column and base-pos number are determined, and the list of
;; seqs is updated in a special frame.
;;
;; There are several ways that updating can happen:  
;;
;; The slow way is to recalculate from the 0th column of each
;; sequence.  Because this slow way turns out not to be so slow, it's
;; the way we're doing it right now.  Performance is acceptable with
;; 5-10 reference sequences averaging 2000 chars each.
;;
;; The fast way is to use a cache.  If you can remember where you
;; *last* were in each sequence, and only count the bases from there.
;; Since you probably only moved a small distance, this will result in
;; a much more efficient update most of the time.  If you moved a
;; large distance, then one can always toss the information and start
;; over (since bol is cheap w/ the newline cache running).  A plan for
;; doing it this way is described below.
;;
;; THE IMPROVEMENT PLAN:
;;
;; gene-posmon-ref-sequences is going to be a list like this:
;;
;;      ((seq last-point-in-seq num-bases last-base-pos ...) ...)
;;
;; or something like that.  There will also be a var,
;; `gene-posmon-last-column', kept up-to-date by someone.  If the
;; column doesn't change, there's no need to update any of the ref
;; sequences; only the current sequence needs to be redone.
;;
;; The current sequence will also have corresponding vars, either in
;; the same list, or out in global space: gene-posmon-last-point,
;; gene-posmon-last-base-pos, and anything else that might be implied
;; in the "...".
;;
;; Updating will work like this:
;;
;; 1. discover the current column.
;;    a) Don't bother to figure out whether we're in the same sequence
;;       or not; just go to bol and count forward again.
;;
;; 2. update the current sequence
;;    a) count bases between gene-posmon-last-point and point, then
;;       add or subtract depending on whether you've moved forward or
;;       back.  Watch out for the obvious o-b-o-es.
;;
;; 3. update each of the ref sequences.
;;    a) goto last-point-in-seq.
;;    b) count the number of bases between num-bases and current
;;       point, add or subtract depending on whether cur-point is
;;       before or after last-point.
;;    c) make sure to update last-point, num-bases, and anything else.
;;
;; Question: how is last-base-pos used, is it necessary?
;;
;; 
;; INVALIDATING THE CACHE:
;; There's a problem, though.  Operations which change the text of the
;; buffer can also affect the cached information.  I guess this is the
;; classic problem with caches -- how to invalidate them correctly.
;;
;; Conservatively, the cache needs to be invalidated after any change
;; to the buffer.  If we had a function F, which returned t iff the
;; buffer had changed since the last time F was called, we'd be
;; winning in the invalidation department.  Thanks to
;; buffer-modified-tick, we can implement such a function easily and
;; efficiently: see `gene-posmon-changed-since-last-call', hereafter
;; referred to as c-s-l-c.  It is already implemented.
;;
;; Each update of the position-monitoring buffer will first call
;; c-s-l-c; if it returns t, then the cache is not to be trusted.
;;
;; A further optimization, to make invalidation happen less often,
;; would be to see if the user is only changing the current sequence.
;; If she is,  then the ref sequences' cache entries remain as valid
;; after the change as they were before.
;; One way to do this is to have gene-seq-insert-char and
;; gene-seq-delete-char set gene-posmon-last-buffer-modified-tick if
;; they are not acting on a group.
;;
;;
;; GOING OVERBOARD:
;; We could be super-efficient and try to notice when the cached
;; information is less useful than normal newline info (for example,
;; moving all the way from one side of the alignment to another).  But
;; this is probably overkill; things like C-e and C-a can be slow
;; because the user will have to pause for a moment afterwards to get
;; her bearings anyway.  The vast majority of movements -- and, more
;; importantly, the ones which a user makes without thinking -- are
;; small, and those are what the cache will help.


(defvar gene-posmon-last-buffer-modified-tick -1
  "Checked by `gene-posmon-changed-since-last-call'.")

(defun gene-posmon-changed-since-last-call ()
  "Return non-nil iff buffer changed since the last time we were called.
Well, also return t if this is the first call."
  (let ((modified-tick (buffer-modified-tick gene-seq-buffer)))
    (prog1
        (/= gene-posmon-last-buffer-modified-tick modified-tick)
      (setq gene-posmon-last-buffer-modified-tick modified-tick))))


(defvar gene-posmon-buffer-name "Position Numbers"
  "The name of the buffer in which position statistics are displayed.")


(defun gene-posmon-init-buffer ()
  "Initialize the position monitoring buffer."
  (gene-save-excursion-in
      gene-posmon-buffer
    (let (buffer-read-only)
      (goto-char (point-min))
      (if (looking-at "^Sequence")
          nil ; already initialized
        (insert "Sequence		Seq Position\n") ; two tabs separate
        (insert "--------		------------\n") ; two tabs separate
        (insert "\n" "\n")))))


(defvar gene-posmon-buffer (get-buffer-create gene-posmon-buffer-name)
  "The buffer in which position statistics are displayed.")

;; Prepare the buffer.
(gene-posmon-init-buffer)


(defvar gene-posmon-monitoring-on nil
  "Non-nil means position monitoring is on.
This variable should never be set directly, but should be accessed
through the function `gene-posmon-monitoring-on-off'.")


(defvar gene-posmon-ref-sequences nil
  "The list of reference sequences against whom the current position
should be displayed.")


(defun gene-posmon-toggle-monitoring ()
  "Turn monitoring on if it's off, off if it's on."
  (interactive)
  (gene-posmon-monitoring-on-off
   (if gene-posmon-monitoring-on -1 1)))


(defun gene-posmon-monitoring-on-off (flag)
  "Turn position monitoring on iff FLAG is positive, off otherwise.
Turning it on includes raising the monitoring frame, turning it off
hides the frame."
  (interactive "p")
  (if (> flag 0)
      (gene-posmon-monitoring-on)
    (gene-posmon-monitoring-off)))


(defun gene-posmon-monitoring-on ()
  "Turn position monitoring on.
Turning it on includes raising the monitoring frame.
This should only be called if it's off."
  (interactive)
  (gene-posmon-update)
  (gene-frame-show gene-posmon-buffer
                   `((height . 15)
                     (width  . 62)
                     (background-color . ,gene-background-color)
                     (foreground-color . ,gene-foreground-color)
                     (cursor-color     . ,gene-cursor-color)))
  (gene-save-excursion-in
      gene-posmon-buffer
    (toggle-read-only 1)
    (gene-menu-bar-make-hide-button 'gene-posmon-monitoring-off)
    ;; Annoying; why do I have to deal with down-mouse-1 when all I
    ;; want to bind is mouse-1?
    (local-unset-key [down-mouse-1])
    (local-unset-key [drag-mouse-1])
    (local-set-key [mouse-1] 'gene-absorb-mouse-event)
    (local-set-key [down-mouse-1] 'gene-posmon-mouse-button-command)
    (setq mode-line-format 'gene-posmon-modeline-string))
  (add-hook 'post-command-hook 'gene-posmon-hook-function)
  (setq gene-posmon-monitoring-on t))


(defun gene-posmon-monitoring-off ()
  "Turn position monitoring off.
Turning it off includes hiding the monitoring frame.
This should only be called if it's on."
  (interactive)
  ;; todo: this might have to clear some data structures, like the
  ;; cached position list, which doesn't exist yet...
  (gene-frame-hide gene-posmon-buffer)
  (remove-hook 'post-command-hook 'gene-posmon-hook-function)
  (setq gene-posmon-monitoring-on nil))


(defun gene-posmon-add-sequence-interactive (sym)
  "Add SEQUENCE to the list of reference sequences and display a message."
  (interactive (list (gene-get-here-sym)))
  (let ((name (gene-organism-name sym)))
    (gene-posmon-add-sequence sym)
    (gene-message "%s added to reference sequences." name)))


(defun gene-posmon-add-sequence (sym)
  "Add sequence SYM to the list of reference sequences, update the stats.
If it's already there, do nothing."
  (if (memq sym gene-posmon-ref-sequences)
      nil ; do nothing
    
    ;;  Else put it in the list, propify it, and put it in the posmon
    ;;  buffer.
    (setq gene-posmon-ref-sequences
          (nconc gene-posmon-ref-sequences (list sym)))
    (put sym 'gene-posmon-monitored t)
    (gene-save-excursion-in
        gene-posmon-buffer
      (goto-char (point-max))
      (let* (buffer-read-only
             (name (gene-organism-name sym))
             (len  (length name))
             (col  (gene-save-excursion-in
                       gene-seq-buffer
                     (gene-current-column)))
             (seq-pos  (gene-posmon-base-position sym col))
             (pos-str  (format "%d" seq-pos))
             )
        
        ;; Make sure this line knows who it is.
        (put-text-property 0 (length name) 'gene-org-short-name sym name)
        ;; Make sure we can t-prop search for the pos number later.
        (put-text-property 0 (length pos-str) 'seq-pos t pos-str)
        
        ;; Insert the data
        (insert name)
        (insert-char ?  (- 24 len))
        (insert pos-str "		")
        (insert gene-posmon-sequence-removal-string "\n")))))


(defun gene-posmon-remove-sequence (sym &optional in-cut)
  "Remove sequence SYM from the list of reference sequences, update the stats.
If it's not there, do nothing.
If optional second arg IN-CUT is non-nil, then don't change the
organism's `gene-posmon-monitored' property, so that when we paste
back in, we know to start monitoring it again."
  (if (not (memq sym gene-posmon-ref-sequences))
      nil ; do nothing
    
    ;; Else remove it from the list, de-propify it, and remove it from
    ;; the posmon buffer.
    (setq gene-posmon-ref-sequences (delq sym gene-posmon-ref-sequences))
    (put sym 'gene-posmon-monitored in-cut)
    (let ((name (gene-organism-name sym)))
      (gene-save-excursion-in
          gene-posmon-buffer
        ;; Skip past first sequence, it might also be a ref sequence,
        ;; who knows.
        (goto-char (point-min))
        (forward-line 3)
        ;; Now we're searching only in ref sequences...
        (goto-char
         (text-property-any (point) (point-max) 'gene-org-short-name sym))
        (beginning-of-line)
        (let (buffer-read-only)
          (gene-delete-whole-line))))))
  
;; The posmon frame will look like this:
;;
;; Sequence   Position
;; --------   --------
;; cur.seq    1453             <add this seq>
;;
;; E.coli     xxx              <remove>
;; foo.bar    xxx              <remove>
;; baz.bim    xxx              <remove>
;;
;; --- (column X) -------------------------- [this is the modeline]

(defun gene-posmon-update ()
  "Update the statistics in the position monitoring buffer."
  (let ((cur-seq (gene-get-here-sym))
        (cur-col (gene-current-column))
        (ref-seqs gene-posmon-ref-sequences))

    (gene-save-excursion-in
        gene-posmon-buffer
      (let (buffer-read-only)
        (goto-char (point-min))
        
        (forward-line 2)
        (gene-delete-whole-line)
        
        ;; Insert stats for the current sequence, passing a non-nil
        ;; CURRENT-SEQ flag.
        (gene-posmon-update-sequence-stats cur-seq cur-col t)
        
        (if gene-posmon-ref-sequences
            ;; Insert stats for each reference sequence.
            (while ref-seqs
              (gene-posmon-update-sequence-stats (car ref-seqs) cur-col)
              (setq ref-seqs (cdr ref-seqs)))
          ;; Else make sure everything is cleaned up, can't hurt.
          ;; (Though shouldn't be necessary...)
          (goto-char (point-min))
          (forward-line 3) ; at least 3 newlines are sure to be there.
          (delete-region (point) (point-max))
          (insert "\n")
          )))))


;; This function could use cached information to avoid the
;; gene-seq-goto-organism, which is probably costing us a lot.
;; Only a temporary function; will go away soon.
(defun gene-posmon-base-position (sym col)
  "Return the base-position of column COL in sequence SYM."
  (gene-save-excursion-in
      gene-seq-buffer
    (gene-seq-goto-organism sym col t) ; t for non-interactive
    (gene-seq-base-number)))


(defun gene-posmon-mouse-button-command (evt)
  "Get text-prop `command' at point clicked on and run it."
  (interactive "e")
  (let* ((pt (posn-point (event-start evt)))
         (command (get-text-property pt 'command)))
    (if command
        (progn
          (funcall command pt)))))


(defun gene-posmon-mouse-remove-sequence (pt)
  "Remove the sequence whose remove button was clicked on.
This is called with PT, a location, as argument; don't bind it to a
mouse event."
  (interactive)
  (gene-save-excursion-in
      gene-posmon-buffer
    (goto-char pt)
    (let ((sym (get-text-property (gene-bol-point) 'gene-org-short-name)))
      (gene-posmon-remove-sequence sym))))


(defun gene-posmon-mouse-add-current-sequence (pt)
  "Remove the sequence whose remove button was clicked on.
This is called with PT, a location, as argument; don't bind it to a
mouse event."
  (interactive)
  (gene-save-excursion-in
      gene-seq-buffer
    (let ((sym (gene-get-here-sym)))
      (gene-posmon-add-sequence sym))))


(defvar gene-posmon-sequence-removal-string
  (let ((raw-str "<remove>"))
    (put-text-property 0 8 'mouse-face 'gene-mouse-face raw-str)
    (put-text-property
     0 8 'command 'gene-posmon-mouse-remove-sequence raw-str)
    raw-str)
  "The text-prop-ified string used as a removal button in the
position-monitoring buffer.")


(defvar gene-posmon-sequence-add-string
  (let ((raw-str "<add this seq>"))
    (put-text-property 0 14 'mouse-face 'gene-mouse-face raw-str)
    (put-text-property
     0 14 'command 'gene-posmon-mouse-add-current-sequence raw-str)
    raw-str)
  "The text-prop-ified string used as a removal button in the
position-monitoring buffer.")


(defvar gene-posmon-modeline-string "  (Position Numbers)"
  "String displayed in position monitoring buffer's modeline.
It is updated to include the column when monitoring is on.")


(defun gene-posmon-update-sequence-stats (sym col &optional current-flag)
  "Insert position-monitoring stats for sequence SYM at column COL.
Optional second arg CURRENT-FLAG non-nil means this is the current
sequence, so insert a column figure instead of a \"<remove>\" button.

This function assumes that point is already at the correct place in
the posmon buffer, and advances point to the beginning of the next
line."
  (if current-flag
      (let* ((name (gene-organism-name sym))
             (len  (length name)))
        (setq gene-posmon-modeline-string
              (format "  (column %d)" col))
        (insert
         (concat name
                 ;; 24 is the seq-pos number column
                 ;; todo: this may GC too much, tho.
                 (make-string (- 24 len) ? )
                 (format "%d" (gene-posmon-base-position sym col))
                 "		" ; two TABs
                 ;; (format "(col: %d)" col)
                 ;; "   "
                 gene-posmon-sequence-add-string
                 "\n")))
    
    ;; Else this is a ref sequence stat we're updating.
    ;; First get ahead of top sequence, since it's already done:
    (goto-char (point-min)) (forward-line 3)
    ;; Now it's safe to do property searches without interference from
    ;; the current sequence.
    (goto-char
     (text-property-any (point) (point-max) 'gene-org-short-name sym))
    (goto-char
     (text-property-any (point) (point-max) 'seq-pos t))
    (gene-delete-word 1)
    (let ((str (format "%d" (gene-posmon-base-position sym col))))
      (put-text-property 0 (length str) 'seq-pos t str)
      (insert str))))


(defun gene-posmon-hook-function ()
  "Update the position monitoring buffer; put this in post-command-hook."
  (and
   ;; gene-posmon-monitoring-on
   ;; We don't check `gene-posmon-monitoring-on' because we won't even
   ;; be in post-command-hook unless it is on.
   (frame-live-p    (gene-frame-get-frame gene-posmon-buffer-name))
   (frame-visible-p (gene-frame-get-frame gene-posmon-buffer-name))
   (eq (current-buffer) gene-seq-buffer)
   (progn
     (gene-posmon-update)
     (set-buffer gene-posmon-buffer)
     (goto-char (point-max))
     (set-window-point (get-buffer-window gene-posmon-buffer t) (point-max))
     )))



;;;
(provide 'gene-posmon)
