;; gene-selection.el (part of ale).
;; Copyright (C) 1994  Karl Fogel <kfogel@darwin.life.uiuc.edu>
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;;
;; Search for instances of "todo:" to find notes about future plans.



;;; Code for dealing with "selections" (rectangles, essentially) in
;;; the sequence buffer.  Because dragging a base-chunk is really
;;; dragging a temporary selection, that code is in here too.

;;; Code:
(require 'gene-frame)
(require 'gene-mode-tools)
(require 'gene-id-mode)
(require 'gene-seq-mode)
(require 'gene-locking)


(defvar gene-sel-stretch-selections t
  "Non-nil means char insertion can stretch a selection, instead of
making holes in it.")



;;; Faces and colors:

;; todo: what other colors do we want?  What different kinds of
;; selections are there?  Use a color ring?  How?  RGB values for
;; gradation, or just pick 100 discrete colors and bury the most
;; recently selected one?

;; todo: we'll have a better color generation scheme later.  For now,
;; this will do.

;;; This number is determined by the X cursor font.
(defconst gene-pointer-horiz-double-arrow 108)

(defvar gene-seq-dragging-pointer-shape gene-pointer-horiz-double-arrow
  "*Shape of the mouse pointer when dragging something.")

(defvar gene-sel-startpos-color "red"
  "The color for the selection starting-corner marker.")

(defconst gene-sel-left-side-color "#ac8a00" ; dark brown
  "Color for temporary left-side selections.")

(defconst gene-sel-right-side-color "#009098" ; blue green
  "Color for temporary right-side selections.")

;; make a circular color list
(defconst gene-sel-default-colors
  (list
   "#a60282" ; green
   "#ba161f" ; brick
   "#fc6900" ; red
   "#e933c2" ; dark pink ?
   "#b349cc" ; purple ?
   "#a09aef" ; light slate blue ?
   "#7854f6" ; blue ?
   "#2827c1" ; darker blue ?
   "#00a0da" ; light blue, dark cyan ?
   "#97b4a4" ; pastel olive
   gene-sel-right-side-color
   "#e98200" ; orange
   "#d6a000" ; light brown
   gene-sel-left-side-color
   "#89b600" ; green
   "#229700" ; purple
   )
  "List of colors to be used for selections.")


(defvar gene-sel-colors nil
  "*List of colors used by sequence selections.")


(defmacro gene-define-selection-color (color)
  "Define a new color to be used for selections."  
  (list 'gene-define-selection-color-internal
	(if (symbolp color) (symbol-name color) color)))


(defun gene-define-selection-color-internal (color)
  "Define a new color to be used for selections."
  (setq color (gene-check-color-allowable color))
  (if gene-sel-colors
      ;; We have some colors already, so add this one:
        (setq gene-sel-colors (cons color gene-sel-colors))
    ;; Else create the color list.
    (setq gene-sel-colors (list color))))


(defvar gene-sel-faces nil
  "The circular list of selection faces.
Initialized by gene-selection-make-faces.")

(defun gene-selection-make-faces ()
  "Create the appropriate faces for selections.
`gene-make-all-faces' calls this.
Note that it's incorrect to make faces in top-level expressions, since
the construction of a face depends on run-time factors \(like the fonts
available\)."

  (setq gene-sel-faces
        (mapcar
         (lambda (color)
           (let ((face (gensym "gene-sel-face-")))
             (copy-face 'default face)
             (set-face-background
              face
              (gene-check-color-allowable color))
             face))
         (or gene-sel-colors
             gene-sel-default-colors)))
  (nconc gene-sel-faces gene-sel-faces)

  (copy-face 'default 'gene-sel-left-side-face)
  (copy-face 'default 'gene-sel-right-side-face)
  (set-face-background
   'gene-sel-left-side-face
   (gene-check-color-allowable gene-sel-left-side-color))
  (set-face-background
   'gene-sel-right-side-face
   (gene-check-color-allowable gene-sel-right-side-color))

  (copy-face 'default 'gene-sel-start-pos)
  (set-face-background
   'gene-sel-start-pos
   (gene-check-color-allowable gene-sel-startpos-color)))


(defun gene-sel-get-next-face ()
  "Get the next color from gene-sel-faces, a circular face list."
  (prog1
      (car gene-sel-faces)
    (setq gene-sel-faces (cdr gene-sel-faces))))



;;; misc selection creation and maintenance stuff ;;;

;; If a selection is created and later we want to add to it, do it
;; like this: pretend you're creating a new selection, but then when
;; it's time to add it to the selection-ring, just merge it with the
;; appropriate selection in the Right Way.
;;
;; A selection looks like this:
;; ###selection-dependency:
;; (selection ; symbol used by `selection-p', says this is a selection
;;  (seg-list
;;  face
;;  read-only
;;  (attributes) ??
;;  (undo-info)  ??
;;  (misc-info)  ??
;;  ...
;; )
;;
;; Each element of the seg-list looks like this:
;; (start end ldiff rdiff lgap rgap...)
;; start and end are markers.  ldiff and rdiff are resynced when a
;; drag starts to happen, so that it's not necessary to search for the
;; drag boundaries more than once; they are also used in uncollapsing
;; fluid drags.  lgap and rgap tell record which gap to insert on that
;; side when the selection is being dragged.
;;
;; face is the face being used for the selection (probably being
;; implemented with text properties).
;; read-only says whether the selection is read-only or not.  What
;; type shall it be?
;;
(defsubst gene-selection-p (selection-perhaps)
  "Return non-nil if SELECTION-PERHAPS is really a selection.
This doesn't do anything fancy; any list whose car is `selection' will pass."
  (eq (car selection-perhaps) 'selection))


;; (defun gene-sel-merge-selections (sel1 sel2 ...) ?
;;   (...))
;; Should this return a selection that is the merging of all the other
;; selections?  Or merge them all into the first one?  That's probably
;; better.  Think about this.


(defun gene-sel-set-seg-list (selection seg-list)
  "Set the segment list of SELECTION to SEG-LIST."
  ;; todo: should we error instead?
  (if (gene-selection-p selection)
      (setcar (nthcdr 1 selection) seg-list)))


(defun gene-sel-get-seg-list (selection &optional no-verify)
  "Return the segment list of SELECTION, or nil if none or not a SELECTION.
This sometimes side-effects the seg-list, see below.

If optional second arg NO-VERIFY is non-nil, then the selection's
segment list will not be verified by a text-property search in the
alignment buffer; otherwise it will, and the seg list will be reset
according to the results of that search.

WARNING: this function returns the actual seg list, not a copy of the
seg list.  This has tripped me up a number of times.  If you want a
copy and not the actual object, then use gene-sel-generate-seg-list."
  (if (gene-selection-p selection)
      (progn
        ;; This sets the selection's seg-list...
        (or no-verify (gene-sel-generate-seg-list selection t))
        ;; ... so it's safe to do this now:
        (car (cdr selection)))))


(defun gene-sel-generate-seg-list (selection &optional reset)
  "Return SELECTION's segment list.
This is accomplished by text-property searches in the alignment
buffer.
If optional-arg RESET is non-nil, then reset it in the selection as well."
  (message "Generating selection's segment list...")

  ;; The format of selection segment lists is:
  ;; ((start end ldiff rdiff lgap rgap ...) ...)
  ;; However, we ignore the diff fields, because
  ;; gene-sel-make-selection-distances deals with 
  ;; those when it's important to do so.  Here, we 
  ;; just fill them in with nil.
  (gene-with-watch-pointer
    (gene-save-excursion-in
        gene-seq-buffer
      (goto-char (point-min))
      (let ((start (point))
	    end
	    (pmax (point-max))		;hah!
	    new-seg-list
	    next-chunk-start)
	(while (setq start
		     (text-property-any
		      start pmax 'gene-selection selection))
	  (setq end
		(text-property-not-all
		 start pmax 'gene-selection selection))
	  (setq new-seg-list
		(cons (list (copy-marker start)
			    (copy-marker end)
			    0 0 
			    (gene-seq-gap-before start) 
			    (gene-seq-gap-after end))
		      new-seg-list))
	  (setq start end))
            
	(message "Generating selection's segment list... done.")
	;; Reverse the seg-list, since it's nice to have it in order:
	(let ((return-list (nreverse new-seg-list)))
	  (if reset
	      (gene-sel-set-seg-list selection return-list)
	    return-list))))))


(defun gene-sel-get-face (selection)
  "Return SELECTION's face."
  ;; todo: why do some check selection-p and others don't?  Is there a
  ;; reason?
  (car (cdr (cdr selection))))


(defvar gene-sel-column-model 'fluid
  "*Default model for assignment of selection columns.
Choices are:
  - `fluid', meaning conform to residues,
  - `hard', meaning straight edged, ignoring residues,
  - `left', meaning hard on the left, fluid on the right,
  - `right', opposite of above.")


(defvar gene-sel-current-start-pos nil
  "Internal var used by selection-creation routines.")


(defun gene-sel-cancel-selection (selection &optional nomsg)
  "Delete SELECTION.
If interactive, uses the selection at point; if none, nothing happens.
If optional second arg NOMSG is non-nil, then don't display a message.
If not in a selection, a message to that effect will be displayed no
matter what the value of NOMSG."
  (interactive (list (gene-sel-selection-at-pos)))
  (gene-save-excursion-in
      gene-seq-buffer
    (gene-dont-modify-buffer
     (if selection
         (let (buffer-read-only
               (seglist (gene-sel-get-seg-list selection t)))
           (mapcar
            (lambda (segment)
              ;; Don't claim that there's a selection here anymore:
              (put-text-property
               (car segment)
               (car (cdr segment))
               'gene-selection
               nil)
              ;; Don't leave the selection highlighted:
              (put-text-property
               (car segment)
               (car (cdr segment))
               'face
               nil))
            seglist)
           
           ;; There might be stragglers -- parts of the alignment that
           ;; still claim to be in this selection even though the
           ;; selection doesn't record them.  It's an error if so, but we
           ;; might as well be extra safe and take care of them too.
           (let ((pos (point-min))
                 next)
             (while (setq next
                          (text-property-any pos
                                             (point-max)
                                             'gene-selection
                                             selection))
               (let ((end (text-property-not-all next
                                                 (point-max)
                                                 'gene-selection
                                                 selection)))
                 (put-text-property next end 'gene-selection nil)
                 (put-text-property next end 'face nil)
                 (setq pos next))))
           (or
            nomsg
            (gene-message "Unselected."))
           )
       ;; Else this wasn't in a selection in the first place.
       (gene-message "Not in a selection.")))))


;; The NAPOLEON arg is necessary because when called from the
;; Selection Menu, this function will automatically be passed a
;; selection as argument.  This just happens to be the one function on
;; that menu that doesn't need to know which selection it's operating
;; on, but the menu-handling code doesn't know that and passes it a
;; selection anyway.
(defun gene-sel-cancel-all-selections (&optional napoleon)
  "Delete all selections.
This is not the same as deleting all their contents!
Optional arg NAPOLEON has no effect whatsoever."
  (interactive)
  (gene-sel-cancel-selections-in-range (point-min) (point-max))
  (message "All selections deleted."))


(defun gene-sel-cancel-selections-in-range (b e)
  "Cancel all selections which contain chars in the range B to E.
This is not the same as deleting all their contents!"
  (let (next)
    (while (setq next (text-property-not-all b e 'gene-selection nil))
      (gene-sel-cancel-selection
       (get-text-property (setq b next) 'gene-selection) t)
      )))


(defun gene-sel-delete-selection-indels (selection)
  "Delete all the indels in SELECTION."
  (interactive (list (gene-sel-selection-at-pos)))

  ;; Check that the selection's locking can handle this:
  (gene-save-excursion-in
      gene-seq-buffer

    (gene-sel-map-over-selection
     selection
     (lambda (start end)
       (goto-char start)
       (gene-locking-sequence-state-p nil gene-locking-gapinout-state t)))

    (let (buffer-read-only)
      (gene-sel-map-over-selection
       selection
       (lambda (start end)
         (goto-char start)
         (while (<= (point) end)
           (if (gene-gap-p (char-after (point)))
               (progn
                 (delete-char 1)
                 (setq end (1- end)))
             (forward-char 1)))))
      (message "Selection's indels deleted."))))


(defun gene-sel-delete-selection-contents (selection)
  "Delete the contents of SELECTION, and then delete SELECTION itself."
  (interactive (list (gene-sel-selection-at-pos)))

  ;; Check that the selection's locking can handle this:
  (gene-save-excursion-in
      gene-seq-buffer

    (if (catch 'bad-locking
          (gene-sel-map-over-selection
           selection
           (lambda (start end)
             (if (string-match gene-base-regexp (buffer-substring-no-properties start end))
                 (or
                  (gene-locking-sequence-state-p
                   (gene-get-here-sym start) gene-locking-unlocked-state)
                  (throw 'bad-locking t))
               (or
                (gene-locking-sequence-state-p
                 (gene-get-here-sym start) gene-locking-gapinout-state)
                (throw 'bad-locking t))))))
        (gene-error
         "Locking does not permit deletion of this selection's contents."))
    
    (let (buffer-read-only
          (seg-list (gene-sel-get-seg-list selection)))
      
      (gene-sel-cancel-selection selection)
      ;; Now that it's deleted, we have the only remaining handle on the
      ;; selection.  Oh, how the garbage collector must envy us!
      
      (mapcar
       (lambda (seg)
         (delete-region (car seg) (car (cdr seg))))
       seg-list))
    (message "Selection contents deleted.")))


;; todo: this needs to scroll someday.
(defun gene-sel-drag-select (event &optional col-model two-sided)
  "Create a selection by dragging.
The first argument is a mouse-down EVENT.
From Lisp, optional second argument COL-MODEL is a selection column
model, `hard' or `fluid'.
Optional third argument TWO-SIDED non-nil means this is a columnar
selection."
  (interactive "e")
  (condition-case
      nil
      (let (buffer-read-only
            (start (posn-point (event-start event))))
        (cond

         ((not (eq (current-buffer) gene-seq-buffer))
          ;; Whoa!  Just restore the buffer.
          (mouse-set-point event))

         ;; Shift-clicks should destroy selections
         ((and two-sided (gene-sel-selection-at-pos start))
          (let ((sel (gene-sel-selection-at-pos start)))
            (gene-sel-cancel-selection sel)))

         ;; If there's a selection here, don't make another one
         ;; overlapping it, just go to the spot.
         ((gene-sel-selection-at-pos start)
          (mouse-set-point event)
          (gene-handle-vertical-motion))

         ;; If this is the finish-click after a S-mouse-1
         ;; 'gene-sel-start-click-select', then just finish it up,
         ;; don't drag.
         ;; fooo: will need col-model and range
         (gene-sel-current-start-pos
          (gene-sel-finish-click-select event col-model two-sided))

         ;; Else drag to create a selection:
         (t
          (track-mouse
            (let* ((start-window     (selected-window))
                   (half-w-height    (/ (window-height start-window) 2))
                   (quarter-w-height (/ half-w-height 2))
                   (eighth-w-height  (/ quarter-w-height 2))
                   (top              (nth 1 (window-edges start-window)))
                   (bottom           (nth 3 (window-edges start-window)))
                   (evt              (read-event))
                   (posn             (event-start evt))
                   (latest           (posn-point posn))
                   did-it            )
              
              (while (mouse-movement-p evt)
                (cond
                 ;; Just sitting on one spot, so do nothing
                 ((and (numberp latest) (= start latest))
                  nil)
                 
                 ;; Ignore any movement outside the frame
                 ((eq (car-safe evt) 'switch-frame)
                  (gene-handle-vertical-motion t t))
                 
                 ;; On the mode line, so scroll buffer up
                 ((eq latest 'mode-line)
                  (gene-seq-scroll-buffer-internal
                   'scroll-up eighth-w-height))
                 
                 ((eq latest 'vertical-line)
                  ;; Do nothing much if on the window divider
                  (gene-handle-vertical-motion t t))

                 ((not (eq (posn-window posn) start-window))
                  ;; Different window, same frame
                  (let ((end-row (cdr (cdr (mouse-position)))))
                    (cond
                     ((and end-row
                           (not (bobp))
                           (< end-row top))
                      (gene-seq-scroll-buffer-internal
                       'scroll-down quarter-w-height))
                     ((and end-row
                           (not (eobp))
                           (>= end-row bottom))
                      (gene-seq-scroll-buffer-internal
                       'scroll-up quarter-w-height))
                     (t
                      (gene-handle-vertical-motion t t)))))
                 (t
                  ;; Else we're still in the same window, ahh, relief.
                  ;; (gene-handle-vertical-motion t t)
                  (or did-it
                      (progn
                        (setq did-it t)
                        ;; mark the first corner with a red splotch:
                        (gene-sel-mark-unmark-start-pos start)))))

                (setq
                 evt (read-event)
                 posn (event-start evt)
                 latest (posn-point posn))
                ) ; matches `while'
              
              (setq posn (event-end evt) latest (posn-point posn))
              
              (if (= start latest) 
                  ;; The two corners are at the same spot.  This
                  ;; probably means we'll just go there, instead of
                  ;; creating a selection; however, `two-sided' might
                  ;; imply that we're creating a tall, thin selection,
                  ;; one char wide and as tall as the alignment.
                  (if two-sided
                      (gene-sel-start-click-select event)
                    (mouse-set-point event)
                    (gene-handle-vertical-motion))
                
                ;; Else the two corners are not at the same spot.
                ;; This is normally the case.
                (if did-it
                    ;; Make the selection now that we know the corners
                    (let ((selection
                           (gene-sel-make-selection
                            start latest nil col-model two-sided)))
                      (if (gene-sel-get-seg-list selection) ; if not empty
                          (progn
                            (gene-message "Selection created.")
                            (gene-sel-mark-unmark-start-pos
                             gene-sel-current-start-pos)
                            ;; redundant:
                            (setq gene-sel-current-start-pos nil))
                        ;; else it's empty, so forget it
                        (gene-sel-mark-unmark-start-pos
                         gene-sel-current-start-pos)
                        (setq gene-sel-current-start-pos nil) ; redundant
                        (gene-message "No selection created.")))
                  
                  ;; Else don't build selection, just go to click
                  (gene-sel-finish-click-select event))))))))
  
    ;; Handler to cleanup in case of error:
    (error
     (if gene-sel-current-start-pos
         (gene-sel-mark-unmark-start-pos
          gene-sel-current-start-pos))
     (error "Cleaning up after error in `gene-sel-drag-select'."))))


(defvar gene-sel-start-pos-overlay nil
  "Overlay used by `gene-sel-mark-unmark-start-pos'.")
  
(defun gene-sel-mark-unmark-start-pos (pt)
  "Toggle on/off the marked selection start-corner at PT.
This is used by selection creation routines, you probably don't need
to know about it."
  (gene-save-excursion-in
      gene-seq-buffer
    (gene-dont-modify-buffer
     (let (buffer-read-only)
       (if gene-sel-current-start-pos
           (progn
             (setq gene-sel-current-start-pos nil)
             (delete-overlay gene-sel-start-pos-overlay))
         (if pt
             (progn
               (setq gene-sel-start-pos-overlay (make-overlay pt (1+ pt)))
               (overlay-put gene-sel-start-pos-overlay 
                            'face
                            'gene-sel-start-pos)
               (setq gene-sel-current-start-pos pt))))))))


(defun gene-sel-start-click-select (event)
  "Start creating a selection by marking one corner with the mouse.
However, if called inside a selection, this deletes the selection.
See also `gene-sel-finish-click-select'."
  (interactive "e")
  (gene-save-excursion-in
      gene-seq-buffer
    (let* (buffer-read-only
           (posn (posn-point (event-start event)))
           (already-selection-here (gene-sel-selection-at-pos posn)))
      (if already-selection-here
          (gene-sel-cancel-selection already-selection-here)
        ;; else mark spot as starting corner for new selection, or
        ;; complete the selection if one was already in progress.
        (if gene-sel-current-start-pos    ; one already in progress
            (gene-sel-two-sided-select event)
          (gene-sel-mark-unmark-start-pos posn))))))


(defun gene-sel-finish-click-select (event &optional col-model two-sided)
  "Finish creating a selection with  the mouse.
Optional second argument COL-MODEL is a selection column model.
Optional third argument TWO-SIDED non-nil means this is a two-sided
selection.
See also `gene-sel-start-click-select'."
  (interactive "e")
  (if (not (eq (selected-window) (posn-window (event-start event))))
      ;; The following should guarantee that the seq window ends up
      ;; selected, since this won't be bound in other windows.
      ;; However, I can't see that the later code (in the
      ;; condition-case) which makes sure that the current window is
      ;; the seq window is doing any harm, so we'll leave it in.
      (progn (mouse-set-point event)
             (gene-handle-vertical-motion))

    ;; Else we were in the seq window already, so proceed normally.
    (condition-case
        nil
        (if (eq (current-buffer) gene-seq-buffer)
            (let ((pt (posn-point (event-start event))))
              (gene-sel-finish-select-internal pt col-model two-sided))
          (or
           (eq (selected-window) (gene-seq-buffer-window))
           (select-window (gene-seq-buffer-window)))
          (gene-switch-to-click event)
          (let ((pt (posn-point (event-start event))))
            (gene-sel-finish-select-internal pt col-model two-sided)))
      
      ;; Handle errors:
      (error
       (if gene-sel-current-start-pos
           (gene-sel-mark-unmark-start-pos
            gene-sel-current-start-pos))
       (error "Cleaning up after error in `gene-sel-mouse-select'.")))))
    

(defun gene-sel-finish-select-internal (pt &optional col-model two-sided)
  "Internal function used by gene-sel-finish-click-select and
gene-sel-two-sided-select.
It isolates functionality common to both; basically, you just vary the
value of PT -- it's the other corner of the selection.
Optional arg COL-MODEL is the selection column model.
Optional arg TWO-SIDED should be non-nil if this is a two-sided selection."
    (if (and gene-sel-current-start-pos
             (not (gene-sel-selection-at-pos pt)) ; paranoia
             (not (= gene-sel-current-start-pos pt)))
        (gene-save-excursion-in
            gene-seq-buffer
          (let ((opoint (point))
                (selection
                 (gene-sel-make-selection
                  (if two-sided
                      (save-excursion
                        (goto-char gene-sel-current-start-pos)
                        (gene-seq-this-col-last-seq))
                    gene-sel-current-start-pos)
                  (if two-sided
                      (save-excursion
                        (goto-char pt)
                        (gene-seq-this-col-first-seq))
                    pt)
                  nil col-model)))
            (progn
              (save-excursion (goto-char opoint) (sit-for 0))
              (if (gene-sel-get-seg-list selection)
                  (progn
                    (gene-sel-mark-unmark-start-pos
                     gene-sel-current-start-pos))
                ;; else it's empty, so forget it
                (gene-sel-mark-unmark-start-pos gene-sel-current-start-pos)
                (gene-message "No selection created")))))

      ;; else clear the air and return to normal operation
      (gene-sel-mark-unmark-start-pos gene-sel-current-start-pos)
      (goto-char pt)
      ;; might not have been line-motion, but could have been.
      (setq this-command 'gene-seq-line-motion)
      (setq gene-seq-preserved-column
            (gene-current-column))
      ;; todo: should point be moved into the selection as Niels
      ;; suggested?  Probably.  Figure out the nicest place to put it,
      ;; though.
      (gene-handle-vertical-motion t t)))


(defun gene-sel-two-sided-select (event &optional col-model)
  "Select the other side of a two-sided selection.
The first side was selected with \\[gene-sel-start-click-select].

Optional arg COL-MODEL is the selection column model."
  (interactive "e")
  (gene-save-excursion-in
      gene-seq-buffer
    (let ((pt (gene-seq-this-col-first-seq
               (save-excursion 
                 (goto-char (posn-point (event-start event)))
                 (gene-current-column)))))
      ;; finish-select-internal will take care of the other side by
      ;; calling gene-seq-this-col-last-seq with
      ;; `gene-sel-current-start-pos' as its argument.
      ;; The final argument t below indicates that this is a two sided
      ;; selection going on.
      (gene-sel-finish-select-internal pt col-model t))))


(defun gene-sel-make-selection (start end &optional face col-model two-sided)
  "Wrapper function for gene-sel-make-selection-internal.
Args are the same, except that optional fifth argument TWO-SIDED means
this is a two-sided \(i.e.: columnar\) selection, so adjust first two
arguments, START and END, accordingly, before calling
gene-sel-make-selection-internal.

This function must return the selection."
  (gene-save-excursion-in
      gene-seq-buffer
    (gene-dont-modify-buffer
     (if two-sided
         (let ((new-start
                (gene-seq-this-col-first-seq
                 (save-excursion
                   (goto-char start)
                   (gene-current-column))))
               (new-end 
                (gene-seq-this-col-last-seq
                 (save-excursion
                   (goto-char end)
                   (gene-current-column)))))
           (gene-sel-make-selection-internal
            new-start new-end face col-model two-sided))
       
       ;; Else just pass on the args unmodified:
       (gene-sel-make-selection-internal start end face col-model)))))
  

(defun gene-sel-make-selection-internal
  (start end &optional face col-model two-sided)
  "Make and return selection with corners at START and END.
Optional arg FACE is a face for the selection.
Optional arg COL-MODEL is the selection column model.
Optional arg TWO-SIDED signals that this is a two-sided selection, for
display purposes only."

  ;; `start' and `end' are the start and end points of a mouse drag,
  ;; maybe.  We must canonicalize them into the upper-left and
  ;; lower-right corners of the rectangle that they delimit.

  (let* ((gc-cons-threshold gene-really-high-number)
         buffer-read-only
         ;; Note that a seg-list can become out of date.  Normally, it
         ;; is reset by `gene-sel-generate-seg-list' before it is
         ;; returned; however, certain functions need it fast and
         ;; unreset, so they pass the NO-VERIFY flag to
         ;; `gene-sel-get-seg-list'.
         (seg-list (prog2
                       (message "Making selection's segment list...")
                       (gene-sel-make-seg-list start end col-model)
                     (message "Making selection's segment list... done.")
                     ))
         (face (or face (gene-sel-get-next-face)))
         (selection
          (list
           'selection
           seg-list
           face
           ;; other stuff, unimplemented right now:
           ;; ( ... attributes ... )
           ;; ( ... undo-info ... )
           ;; ( ... misc-info ... )
           ;; ( ... links ... )
           )))
    
    ;; Highlight the selection and set up its other text properties:
    (message "Highlighting and putting text properties...")
    (mapcar
     (lambda (cell)
       (put-text-property
        (car cell) (nth 1 cell) 'gene-selection selection)
       ;; Set the face (normally a background) for the selected area:
       (put-text-property
        (car cell) (nth 1 cell) 'face face))
     seg-list)
  
    (message "Highlighting and putting text properties... done.")

    ;; Return the selection itself.
    selection))


(defun gene-sel-selection-at-pos (&optional pt)
  "Return the selection around PT, or nil if none.  PT defaults to point."
  (get-text-property (or pt (point)) 'gene-selection))


;; This returns a list:
;; ((start end 0 0 lgap rgap...) (start end 0 0 lgap rgap...) ...)
;;
;; The idea is to concentrate complexity in this function, which
;; other, wrapper functions will use shamelessly.  It is the place to
;; change the configuring of selections based on given rectangle
;; corners.

(defun gene-sel-make-seg-list (start end &optional col-model)
  "Make a selection's segment list, based on corners at START and END.
Optional arg COL-MODEL is the selection column model; it defaults to `fluid'."
  (if (> start end) ; First, swap 'em if necessary
      (setq start (prog1 end (setq end start))))
  (gene-save-excursion-in
      gene-seq-buffer
    (let* ((col-model (or col-model gene-sel-column-model))
           (endcol (progn (goto-char end)
                          (gene-current-column)))
           (startcol (progn (goto-char start)
                            (gene-current-column)))
           seg-list)
      (if (> startcol endcol) ; swap the columns if they're reversed
          (setq startcol (prog1 endcol (setq endcol startcol))))
      (setq endcol (1+ endcol)) ; compensate for o-b-o-e in right column
      (while (<= (point) end)
        (let* ((bol (gene-bol-point))
               (eol (gene-eol-point))
               (startcol-pos (min eol (+ bol startcol)))
               (endcol-pos (min eol (+ bol endcol)))
               ;; todo: can we just use copy-marker for these later?
               (start-marker (make-marker))
               (end-marker (make-marker))
               (lgap 0)
               (rgap 0)
               cell)
          (goto-char startcol-pos)
          (if (eq col-model 'hard)           ; todo: deal w/ hard outside loop
              (setq cell (list (set-marker start-marker startcol-pos)
                               (set-marker end-marker endcol-pos)
                               ;; Don't forget the ldiff and rdiff
                               ;; fields, which
                               ;; make-selection-distances will fill
                               ;; in later:
                               0 0
                               ;; And gaps for use in drags
                               (gene-seq-gap-before start-marker)
                               (gene-seq-gap-after end-marker)
                               ))
            ;; The commented-out skip-chars used to make selections
            ;; "spread out" to cover contiguous bases on both sides.
            (or (eq col-model 'right)
                (> (skip-chars-forward gene-skip-gaps-regexp endcol-pos) 0)
                )
            ;; Record the beginning of this segment
            (setq startcol-pos (point)
                  ;; and start filling in the cell as a whole
                  cell (list (set-marker start-marker startcol-pos)))
            ;; Record the gap-char that we found here
            (setq lgap (gene-seq-gap-before))

            ;; go to right col
            (goto-char endcol-pos)
            (or (eq col-model 'left)
                (< (skip-chars-backward gene-skip-gaps-regexp startcol-pos) 0)
                )
            ;; Record the end of the segment
            (setq endcol-pos (point))
            ;; Record the gap-char on this side
            (setq rgap (gene-seq-gap-after))

            ;; The final two 0's are the ldiff and rdiff fields, to be
            ;; filled in by make-selection-distances later.  `cell'
            ;; already contains start-marker, remember.
            (nconc cell (list (set-marker end-marker endcol-pos)
                              0 0
                              lgap rgap)))

          (if (= startcol-pos endcol-pos)
              ;; who needs 0-sized segs?
              nil
            ;; If there's another selection containing this segment,
            ;; delete it.  Selections are ephemeral critters anyway,
            ;; the one the user is asking to create right now should
            ;; always take precedence.
            (gene-sel-cancel-selections-in-range (nth 0 cell) (nth 1 cell))
            (setq seg-list (cons cell seg-list)))

          ;; Move to the next line.
	  ;; If there isn't a next line, we're done.  Since end could
	  ;; be at (point-max), trash it to make sure the loop exits.
	  (or (zerop (forward-line 1))
	      (setq end (1- (point))))))
      ;; return the list of columns, in order
      (nreverse seg-list))))


;; The `col-model' parameter is 'fluid or 'hard.  'hard  means that the
;; selection will only move until a collision in some line.  Otherwise
;; it will move each line until that line collides (unless it doesn't
;; collide within `count' distance).
;;
;; Adjusts the diff fields of the segment-list in the selection as
;; necessary.
;;
;; Returns a list of ((start end ldiff rdiff lgap rgap ...) ...)

(defun gene-sel-make-selection-distances (selection count &optional col-model)
  "Make a list of safe movement distances for each segment in SELECTION.
COUNT is the desired distance that the selection wants to move, it can
be negative or positive.
If COL-MODEL is `hard' then all segments will be limited to moving the
shortest permissible distance among them.  Otherwise, the selection
slides fluidly, and its segments can move in and out of register with
each other.

This function does *not* check text properties when getting a
selection's segment list, so you must reset it by hand before calling
this.

Returns a list of ((start end dist ...) (start end dist ...) ...).

If COUNT is nil, then this function is used for resetting certain
fields in SELECTION, the fields that tell each segment how much gappy
space it has on each side.  So this function is called once at the
beginning of a selection drag, with COUNT nil, to set all these
distance fields correctly.  Then it's called during the drag to adjust
those differences and return a list (per mouse-event) indicating how
far each segment can go.
If any of the segments are not in a locking state that permits drags,
an error will be signalled."

  ;; If `count' is nil, it means that this call is being used to reset
  ;; the diffs in a seg-list.  Remember: distances on the left side
  ;; are given negatively, whereas ones on the right are positive.  If
  ;; you find that it's reversed for one side, it means that that side
  ;; has hit a wall but the user kept dragging.  They should never be
  ;; reversed on both sides.
  (gene-save-excursion-in
      gene-seq-buffer
    ;; Remember to pass NO-VERIFY argument to get-seg-list, because we
    ;; need to preserve ldiffs and rdiffs across calls while dragging
    ;; the selection around.
    (let ((seg-list (gene-sel-get-seg-list selection t)))
      (let (distances)
        (while seg-list
          (let* ((segment (car seg-list))
                 (lcolpnt (nth 0 segment))
                 (rcolpnt (nth 1 segment))
                 (ldiff-cdr (nthcdr 2 segment))
                 (rdiff-cdr (nthcdr 3 segment))
                 ;; Get lgap or rgap, depending on direction.
                 (gap-char (if (> count 0)
                               (nth 4 segment)
                             (nth 5 segment)))
                 (safedist 0))
            
            ;; If count is positive, we're moving right
            (if (> count 0)
                
                ;; If ldiff is positive, then we have some
                ;; decrementing to do before anything actually
                ;; moves, because someone tried to drag past a
                ;; wall previously.
                (if (> (car ldiff-cdr) 0)
                    ;; ldiff can absorb all of count
                    (if (>= (- (car ldiff-cdr) count) 0)
                        ;; zero safedist and adjust ldiff
                        ;; accordingly, but don't touch rdiff
                        (progn
                          (setq safedist 0)
                          (setcar ldiff-cdr (- (car ldiff-cdr) count)))
                      ;; else ldiff can't absorb all of count
                      (let ((newcount (- count (car ldiff-cdr)))
                            (rdist (car rdiff-cdr))) ; save old val
                        (setcar rdiff-cdr (- (car rdiff-cdr) newcount))
                        ;; still room left after adjusting diffs
                        (if (> (car rdiff-cdr) 0)
                            (progn
                              (setcar ldiff-cdr (- (car ldiff-cdr) count))
                              (setq safedist newcount))
                          ;; else collision in other direction too! ***
                          (setcar ldiff-cdr (- rdist))
                          (setq safedist (max 0 rdist))))) ; paranoia
                  ;; else ldiff already zero or negative, so just
                  ;; move as far as can and set diffs accordingly
                  (let ((rdist (car rdiff-cdr)))
                    (setcar rdiff-cdr (- (car rdiff-cdr) count))
                    (if (>= (car rdiff-cdr) 0)
                        (progn
                          (setcar ldiff-cdr (- (car ldiff-cdr) count))
                          (setq safedist count))
                      ;; else right collision happened
                      (setcar ldiff-cdr
                              (- (car ldiff-cdr) (max 0 rdist)))
                      (setq safedist (max 0 rdist)))))
              
              ;; Else count is negative, so we're moving left:
              
              ;; rdiff is negative, so adjust absorb some
              ;; movement first
              (if (< (car rdiff-cdr) 0)
                  ;; rdiff can absorb all of count
                  (if (<= (- (car rdiff-cdr) count) 0)
                      ;; zero safedist and adjust rdiff
                      ;; accordingly, but don't touch ldiff 
                      (progn
                        (setq safedist 0)
                        (setcar rdiff-cdr (- (car rdiff-cdr) count)))
                    ;; else rdiff can't absorb all of count
                    (let ((newcount (- count (car rdiff-cdr)))
                          (ldist (car ldiff-cdr))) ; save old val
                      (setcar ldiff-cdr (- (car ldiff-cdr) newcount))
                      ;; still room left after adjusting diffs
                      (if (< (car ldiff-cdr) 0)
                          (progn
                            (setcar rdiff-cdr (- (car rdiff-cdr) count))
                            (setq safedist newcount))
                        ;; else collision in other direction too! ***
                        (setcar rdiff-cdr (- ldist))
                        (setq safedist (min 0 ldist))))) ; paranoia
                ;; else rdiff already zero or positive, so just
                ;; move and set diffs accordingly
                (let ((ldist (car ldiff-cdr)))
                  (setcar ldiff-cdr (- (car ldiff-cdr) count))
                  (if (<= (car ldiff-cdr) 0)
                      (progn
                        (setcar rdiff-cdr (- (car rdiff-cdr) count))
                        (setq safedist count))
                    ;; else left collision happened
                    (setcar rdiff-cdr
                            (- (car rdiff-cdr) (min 0 ldist)))
                    (setq safedist (min 0 ldist))))))
            
            ;; Uncomment this for debugging:
            ;; (message "l: %d, r: %d" (car ldiff-cdr) (car rdiff-cdr))
            ;; (sit-for 0)
            
            ;; A simple paranoia check here might be in order.
            ;; One nice fact is that the gappy runs on either
            ;; side of a moving base run always sum to the same
            ;; number, according to this formula:
            ;; N = max (0, abs(ldiff)) + max(0, rdiff)
            ;; The absolute value of ldiff is taken because ldiff
            ;; is usually given as a negative number of gaps.
            ;; N stays constant for the duration of the drag; it
            ;; need only be determined once when the mouse is
            ;; first clicked, and recorded somewhere (with the
            ;; diffs, probably), and we can check it with the
            ;; formula each this function is called with a non-nil
            ;; `count'.  If it doesn't pass the check, just error
            ;; with a message saying that there's a bug here, and
            ;; maybe record some useful data abou the context.
            
            ;; Now that the diffs have been adjusted correctly and
            ;; the allowable distances each line can move
            ;; calculated, record the distances for later return:
            (setq distances
                  (cons (list lcolpnt rcolpnt safedist gap-char)
                        distances)))
          ;; Go to next segment and run the same calculations on
          ;; it:
          (setq seg-list (cdr seg-list)))
        
        ;; all can only move as far as the least distance
        (if (eq col-model 'hard)
            (let* ((distlist distances)
                   (maxdist (nth 2 (car distlist))))
              (if (>= count 0)
                  (while distlist
                    (setq maxdist
                          (min maxdist (nth 2 (car distlist)))
                          distlist (cdr distlist)))
                (while distlist
                  (setq maxdist (max maxdist (nth 2 (car distlist)))
                        distlist (cdr distlist))))
              ;; ...then set all distances to it
              (mapcar (lambda (lst) (setcar (cdr (cdr lst)) maxdist))
                      distances)))
        ;; return the distances in the right order
        (nreverse distances)))))


;; From: kfogel@phylo.life.uiuc.edu (Karl Fogel)
;; To: jimb@totoro
;; Subject: Re: gene-sel-make-selection-distances
;; Date: Mon, 13 Jun 1994 11:10:43 -0500
;;
;; Jim Blandy sez:
;; >Karl, that's a monstrosity.  Explain it to me.
;; 
;;         I'm sorry.  It's really two functions, and I'm going to split
;; it up into them later (didn't want to do something like that right
;; before Niels uses the editor).
;; 
;;         If you pass it nil `count', then it is the function that
;; computes how much side a selection has on each side, and fills in the
;; "ldiff" and "rdiff" fields of the selection's segments (a "segment" is
;; one line of the selection, don'cha know).  These tell each segment how
;; much gappy space it has to the right and left.  The rightward number
;; is positive, leftward is negative.
;; 
;;         Once the setting of these distances is done, moving the
;; selection back and forth is a matter of adding/subtracting from these
;; distances, and watching when they hit 0.  When they cross 0, it means
;; that that segment has hit a base, so don't go any farther.
;; 
;;         This adding and subtracting is what the monstrous first part
;; of gene-sel-make-selection-distances is about.  It doesn't actually do
;; the moving of the segments, it just computes how far each one can
;; move, and records (in the diff field) the difference between the
;; desired distance and the actual allowable distance.  On the right
;; side, this means the number crosses 0 and goes negative, on the left
;; it crosses 0 and goes positive.  (Thus, there should never be a
;; situation where the right is negative AND the left is positive.)
;; 
;;         One of the things it does is take into account past
;; discrepancies between desired and actual distances (as recorded in the
;; diff field), so that sometimes the new distances include compensation
;; for old collisions.  This is how fluid dragging works: the relative
;; positions of the segments to each other are not being recorded, it's
;; just that the amounts by which segments might have failed to move in
;; the past (due to collision) are exactly the amounts by which they must
;; now absorb requested movement in order to come back in line with each
;; other.
;; 
;;         Having computed these things, it returns a list of distances
;; to move everyone (if you passed the `hard' col-model, then that list has
;; been minimized so no one moves farther than the segment moving the
;; least).  The list looks like, uh
;; 
;;         ((start end allowable-distance) ; numbers for first segment
;;          (start end allowable-distance) ; numbers for second segment
;;          ...)                           ; etc
;; 
;;         This list is used by gene-sel-slide-selection, which basically
;; `apply's gene-shift-region to each element of it (sticking in the
;; GAP-CHAR fourth arg to gene-shift-region, though).
;; gene-shift-region's arguments are, recall, START END DISTANCE
;; FILL-CHAR.
;; 
;;         Yes, make-distances should be computing the gap-char from
;; context and including it, instead of gene-sel-slide-selection having
;; to put in in.  That's on the todo list. :-)
;;
;;    [This is done now. -Karl]
;; 
;;         There are two things going on here: the list with allowable
;; distances is returned, but the diff fields of the selection are
;; actually side-effected too, because they are used for recording
;; information that might be important the next time make-distances is
;; called.
;; 
;;         One thing that might seem like a problem is that the
;; equalizing of distances in the case of `hard' is done *after* all the
;; diff fields have been side-effected.  The reason things work out is
;; that make-distances knows that the next time it's called, it will
;; either be during the same hard drag (because the user hasn't let up
;; the mouse button), so that the distances will just get equalized again
;; next time around and the discrepant adjustments resulting from the
;; diffs having been taken into account will come out in the wash anyway,
;; or it will be an entirely new drag, in which case the diffs have been
;; reset anyway.
;; 
;;         Whew.  Do things make any more sense now?
;; 
;;         I haven't been looking in make-distances for the bug, because
;; even if it has a bug, I can't see how it's causing this particular
;; problem.  No matter how bogus the information it returns is, as long
;; as `lcolpnt' and `rcolpnt' (start and end) don't get messed up, then
;; gene-shift-region can't be affecting anything inside the region being
;; shifted.  Or so I reason.  Not that I couldn't be way wrong, or that
;; start and end couldn't be being messed up, therefore I thank you very
;; much and applaud your courage at venturing into make-distances, the
;; Bermuda Triangle of the selection code.
;;
;;   [Bug now fixed; it had to do with relocating buffers.  -Karl]


(defun gene-sel-reset-selection-distances (selection)
  "Reset the `diff' fields of SELECTION, to prepare for sliding."
  ;; Check the locking states on the segments, too, and error if
  ;; anyone can't be dragged.
  (gene-save-excursion-in
      gene-seq-buffer
    (let ((seg-list (gene-sel-get-seg-list selection t)))
      (while seg-list
        (let* ((segment (car seg-list))
               (lcolpnt (nth 0 segment))
               (rcolpnt (nth 1 segment))
               (safedist 0))
          
          ;; Do the left side:
          (goto-char lcolpnt)
          ;; Check lock state before anything else.  This will error
          ;; if the locking is too strict.  It's strictly a paranoia
          ;; check, because the selection has already been checked
          ;; as a whole.  However, this costs little, so for right
          ;; now I'll leave it in:
          (gene-locking-sequence-state-p nil gene-locking-gapshift-state t)
          ;; Safe to go on:
          (let* ((bol (gene-bol-point))
                 (next-selection
                  ;; Sadly, can't use `text-property-not-all' here
                  ;; because it only searches from left to right.
                  (let ((prev
                         (previous-single-property-change
                          lcolpnt 'gene-selection nil bol)))
                    (if (get-text-property prev 'gene-selection)
                        ;; We border another selection, so we can't
                        ;; move to the left at all.
                        lcolpnt
                      ;; Else find any preceding selections.
                      (previous-single-property-change
                       lcolpnt 'gene-selection nil bol))))
                 (barrier (max (- next-selection lcolpnt)
                               (skip-chars-backward
                                gene-skip-gaps-regexp bol))))
            ;; set the `ldiff' field to the max safedist:
            (setcar (nthcdr 2 segment) barrier))
          
          ;; Now do the right side:
          (goto-char rcolpnt)
          (let* ((eol (gene-eol-point))
                 (next-selection
                  (or
                   (text-property-not-all rcolpnt eol 'gene-selection nil)
                   eol))
                 (barrier (min (- next-selection rcolpnt)
                               (skip-chars-forward
                                gene-skip-gaps-regexp eol))))
            (setcar (nthcdr 3 segment) barrier))
          (setq seg-list (cdr seg-list)))))
    nil)) ; return nil


(defun gene-sel-right-throw-selection (selection &optional col-model)
  "Throw SELECTION as far to the right as it can go.
Optional second arg COL-MODEL is `fluid' or `hard'; nil means `hard'.
Note: this default is the opposite of gene-sel-slide-selection's
default."
  (let ((old-seg-list (gene-sel-generate-seg-list selection)))
    (gene-sel-generate-seg-list selection t)
    (gene-sel-reset-selection-distances selection)
    (gene-sel-slide-selection selection
                              gene-really-high-number
                              (or col-model 'hard))
    (gene-undo-record-selection-drag
     old-seg-list
     (gene-sel-get-seg-list selection t))))


(defun gene-sel-left-throw-selection (selection &optional col-model)
  "Throw SELECTION as far to the left as it can go.
Optional second arg COL-MODEL is `fluid' or `hard'; nil means `hard'.
Note: this default is the opposite of gene-sel-slide-selection's
default."
  (let ((old-seg-list (gene-sel-generate-seg-list selection)))
    (gene-sel-generate-seg-list selection t)
    (gene-sel-reset-selection-distances selection)
    (gene-sel-slide-selection selection
                              (- 0 gene-really-high-number)
                              (or col-model 'hard))
    (gene-undo-record-selection-drag
     old-seg-list
     (gene-sel-get-seg-list selection t))))


(defun gene-sel-slide-selection (selection count &optional col-model)
  "Slide current SELECTION COUNT places, or to the point of residue
collision in each line.  That is, this function is \"safe\" -- it
performs bounds checking before the slide is done.

COUNT can be positive \(slide to the right\), or negative \(to the left\).
Inserts gap characters to fill in vacated space.
Optional third arg COL-MODEL is nil (`fluid') or `hard'.

This function records no undo.

The selection's internal distances must usually be reset with
`gene-sel-reset-selection-distances' before calling this function.
However, because certain dragging functions take advantage of those
internal distances, this function does not do that resetting itself --
you must take care of it in the caller, if it needs to be done."
  (save-excursion
    (let (buffer-read-only
          (buffer-undo-list t)
          (distlist
           (gene-sel-make-selection-distances selection count col-model))
          (gc-cons-threshold gene-really-high-number))
      (while distlist
        (apply 'gene-shift-region (car distlist))
        (setq distlist (cdr distlist))))))



;;; end selection specs ;;;


;; dragging selections and base-groups

(defun gene-sel-drag-selection-or-base-group (start-event flag)
  "Drag a selection with mouse.
This should be bound to a button-down mouse event.
A white-tie mouse event would be ideal."
  (interactive "e")
  (gene-save-excursion-in
      gene-seq-buffer
    (gene-with-pointer-shape
        gene-pointer-horiz-double-arrow
      (let* ((start-posn (event-start start-event))
             (start-col (car (posn-col-row start-posn)))
             (start-point (posn-point start-posn))
             (start-window (posn-window start-posn))
             (selection (gene-sel-selection-at-pos start-point))
             ;; This gets us a new copy without affecting the one in
             ;; the selection.  We'll reset that separately later.
             old-seg-list)
        
        (if selection
            (progn
              ;; reset the selection's ldiffs and rdiffs, and check its
              ;; locking state:
              (gene-locking-selection-check-state
               selection gene-locking-gapshift-state)
              
              ;; The locking check above reset the selection's seg
              ;; list, so we can avoid regenerating and just copy the
              ;; list.
              (setq old-seg-list
                    (copy-sequence (gene-sel-get-seg-list selection t)))
              
              ;; make-distances won't do this, so we have to.
              (gene-sel-generate-seg-list selection t)
              
              (gene-sel-reset-selection-distances selection)
              
              (let (event end end-col)
                (track-mouse
                  (while (progn
                           (setq event (read-event))
                           (or (mouse-movement-p event)
                               (eq (car-safe event) 'switch-frame)))
                    (if (eq (car-safe event) 'switch-frame)
                        nil                 ; do nothing (see below)
                      (setq end (event-end event)
                            end-col (car (posn-col-row end)))
                      (if (eq (posn-window end) start-window)
                          ;; we are moving within the original window
                          (progn
                            (gene-sel-slide-selection selection
                                                      (- end-col start-col)
                                                      flag)
                            (setq start-col end-col))
                        ;; do nothing for now, have something clever later
                        )))))
              
              (gene-undo-record-selection-drag
               old-seg-list
               (gene-sel-get-seg-list selection t))
              )
          ;; Else there is no selection here, so make base groups
          ;; and drag those:
          (gene-sel-drag-base-group start-event))))))
    
  
(defun gene-sel-drag-base-group (start-event)
  "Drag a contiguous stretch of residues with the mouse."
  (interactive "e")
  ;; save-excursion is overkill, since we just need to move point to
  ;; bind `beg' and `end' correctly.
  (gene-save-excursion-in
      gene-seq-buffer
    (let* ((orig-point (point))
           (mid-posn (event-start start-event))
           (orig-col (car (posn-col-row mid-posn)))
           (start-window (posn-window mid-posn))
           (mid-point (posn-point mid-posn))
           (beg
            (progn
              (mouse-set-point start-event)
              (skip-chars-backward gene-skip-bases-regexp (gene-bol-point))
              (point)))
           (end
            (prog2 (skip-chars-forward gene-skip-bases-regexp (gene-eol-point))
                (point)
              (goto-char orig-point)))
           (selection-left (gene-sel-make-selection
                            beg
                            (max beg (1- mid-point))
                            'gene-sel-left-side-face))
           (selection-right (gene-sel-make-selection
                             mid-point
                             (max mid-point end)
                             'gene-sel-right-side-face))
           (old-left-seg-list
            (if selection-left
                (gene-sel-generate-seg-list selection-left)))
           (old-right-seg-list
            (if selection-right
                (gene-sel-generate-seg-list selection-right))))
      
      ;; This will error if locking does not permit gap shifting:
      (gene-locking-sequence-state-p
       (gene-get-here-sym mid-point) gene-locking-gapshift-state
       t)
      
      (progn
        
        ;; Since selection creation is inclusive on the two corners, we
        ;; have to explicitly nullify any selection created between two
        ;; equal points.
        (if (= beg mid-point)
            (gene-sel-set-seg-list selection-left nil))
        (if (= mid-point end)
            (gene-sel-set-seg-list selection-right nil))
        
        ;; Have to do this by hand here, since make-distances won't want
        ;; to later:
        (gene-sel-reset-selection-distances selection-left)
        (gene-sel-reset-selection-distances selection-right)
        
        (let (event end end-col start-col)
          (setq start-col orig-col)
          (track-mouse
            (while (progn
                     (setq event (read-event))
                     (or (mouse-movement-p event)
                         (eq (car-safe event) 'switch-frame)))
              (if (eq (car-safe event) 'switch-frame)
                  nil                 ; do nothing (see below)
                (setq end (event-end event)
                      end-col (car (posn-col-row end)))
                (if (eq (posn-window end) start-window)
                    ;; we are moving within the original window
                    (progn
                      
                      (gene-sel-slide-selection selection-left
                                                (- end-col start-col))
                      (gene-sel-slide-selection selection-right
                                                (- end-col start-col))
                      
                      (setq start-col end-col))
                  ;; do nothing for now, have something clever later
                  )))))
        
        (if old-left-seg-list
            (gene-undo-record-selection-drag
             old-left-seg-list
             (gene-sel-get-seg-list selection-left t)))
        (if old-right-seg-list
            (gene-undo-record-selection-drag
             old-right-seg-list
             (gene-sel-get-seg-list selection-right t)))
        
        (undo-boundary)
        
        (gene-sel-cancel-selection selection-left)
        (gene-sel-cancel-selection selection-right)
        ))))
  

;; Adding to / Removing from a pre-existing selection:
;;
;; todo: how to handle adding?

(defun gene-sel-unselect-chunk (event)
  "Remove a chunk from the clicked-on selection."
  (interactive "e")
  (let* ((pt (posn-point (event-start event)))
         (selection (gene-sel-selection-at-pos pt)))
    (if selection
        (gene-sel-do-unselect-chunk selection pt))))


(defun gene-sel-remove-seq-from-selection (sym selection)
  "Remove sequence SYM from SELECTION."
  (gene-save-excursion-in
      gene-seq-buffer
    (gene-seq-goto-organism sym)
    (beginning-of-line)

    (let ((eol (gene-eol-point))
          next-chunk-start)
      (while (setq next-chunk-start
                   (text-property-any (point)
                                      eol
                                      'gene-selection
                                      selection))
        (goto-char next-chunk-start)
        (gene-sel-do-unselect-chunk (gene-sel-selection-at-pos))))))


(defun gene-sel-remove-seq-from-all-selections (sym)
  "Remove sequence SYM from any selections of which it is a member.
This function does not record any undo.  It is used mainly by the cut
and paste code to take sequences out of selections before cutting
them."
  (gene-save-excursion-in
      gene-seq-buffer
    (gene-seq-goto-organism sym)
    (beginning-of-line)
    (gene-sel-remove-selection-fragments-to-eol)))


(defun gene-sel-remove-selection-fragments-to-eol ()
  "Remove all traces of selectionness from the sequence following point.
You probably want to make sure point is at the beginning of the
sequence when this is called."
  (save-excursion
    (let ((eol (gene-eol-point))
          (pt (point))
          (buffer-undo-list t)
          (sel (gene-sel-selection-at-pos))
          next-selection-start)
      
      ;; next-prop-change won't catch a selection that starts at the
      ;; beginning of the line, so we have to watch out for that case
      ;; specially right here, before going any further.
      (if sel (gene-sel-do-unselect-chunk sel))
      
      ;; Now go looking for other selections downstream.
      (while (let ((next-next
                    (next-single-property-change pt
                                                 'gene-selection
                                                 nil
                                                 eol)))
               (if (or (= next-next pt) (= next-next eol))
                   nil
                 (setq next-selection-start next-next)))
        
        (goto-char (setq pt next-selection-start))
        (gene-sel-do-unselect-chunk (gene-sel-selection-at-pos))
        ))))


(defun gene-sel-do-unselect-chunk (selection &optional spot)
  "Remove from SELECTION the chunk around SPOT.
SPOT defaults to point."
  (or spot (setq spot (point)))
  (gene-save-excursion-in
      gene-seq-buffer
    (gene-dont-modify-buffer
     (goto-char spot)
     
     ;; para-para noia-noia check-check-check:
     (or
      (eq (gene-sel-selection-at-pos) selection)
      (gene-error "Attempt to delete chunk from a selection it's not in."))
     
     (let* (buffer-read-only
            (start (or 
                    (if (eolp) (point)) ; selection chunks don't cross lines
                    ;; next/prev-prop-change fails if we're at the edge
                    ;; of a chunk.
                    (if (not (eq
                              (get-text-property (1- (point)) 'gene-selection)
                              selection))
                        (point))
                    (previous-single-property-change spot
                                                     'gene-selection
                                                     nil
                                                     (gene-bol-point))
                    ;; If no prop change, then we're on the first
                    ;; char in the chunk.
                    (point)))
            (end (or
                  (if (eolp) (point))
                  ;; next/prev-prop-change fails if we're at the edge
                  ;; of a chunk.
                  (if (not (eq
                            (get-text-property (1+ (point)) 'gene-selection)
                            selection))
                      (1+ (point)))
                  (next-single-property-change spot
                                               'gene-selection
                                               nil
                                               (gene-eol-point))
                  ;; If no prop change, then we're on the last
                  ;; char in the chunk.  If this is the case, the
                  ;; property search above could not have returned
                  ;; nil.
                  (point))))
       
       ;; Delete this line from the selection
       (put-text-property start end 'gene-selection nil)
       (put-text-property start end 'face nil)
       
       (or
        (text-property-any
         (point-min) (point-max) 'gene-selection selection)
        ;; Else the selection is empty, so remove it completely.
        (gene-sel-cancel-selection selection))))))

  

(defun gene-sel-center-justify (selection)
  "Center the residues on each line of SELECTION.
Interactively, SELECTION is the selection at point.
If perfect centering is impossible, the bases will tend toward the
left.
Programmers with a sense of justice may wish to avoid this function.
In order to center-justify, it left-justifies and then moves
everything halfway to the right.  Ah hahahahaha."
  (interactive (list (gene-sel-selection-at-pos)))
  (if (gene-selection-p selection)
      (gene-save-excursion-in
          gene-seq-buffer
        (message "Center justifying selection...")
        (gene-locking-selection-check-state
         selection gene-locking-gapshift-state)
        
        (let* (buffer-read-only
               ;; These below must be in this order:
               (seglist (gene-sel-get-seg-list selection t))
               (dummy   (gene-undo-record-selection-justification seglist))
               (buffer-undo-list t))
          
          ;; Dang it, my time is worth more than the CPU's:
          (gene-sel-left-justify selection t)
          
          ;; Okay, now that it's justified left, just transpose half of the
          ;; gaps back to the other side:
          (while seglist
            (let* ((segment (car seglist))
                   (start (nth 0 segment))
                   (end (nth 1 segment)))
              
              (goto-char start)
              (if (looking-at gene-gap-regexp)
                  nil ; nothing to transpose on this line
                (let* ((gappy-start
                        (progn
                          (skip-chars-forward gene-skip-bases-regexp end)
                          (point)))
                       (gappy-len (- end gappy-start))
                       (halfway (floor (/ gappy-len 2)))
                       (right-edge (1- (+ gappy-start halfway)))
                       (char-pt    (1- gappy-start)))
                  (if (> gappy-len 1)
                      (progn
                        (while (>= char-pt start)
                          (gene-translate-point-to-char
                           right-edge (char-after char-pt))
                          (setq right-edge (1- right-edge)
                                char-pt (1- char-pt)))
                        (gene-translate-region-to-char
                         start (1+ right-edge) (gene-seq-gap-before start)))
                    )
                  )))
            (setq seglist (cdr seglist))))
        (message "Center justifying selection... done."))))


(defun gene-sel-left-justify (selection &optional noundo)
  "Left-justify all residues in SELECTION.
Interactively, SELECTION is the selection at point.

Optional arg NOUNDO means don't record undo.
Normally, only gene-sel-center-justify passes this flag."
  (interactive (list (gene-sel-selection-at-pos)))
  (if (gene-selection-p selection)
      (gene-save-excursion-in
          gene-seq-buffer
        (gene-locking-selection-check-state
         selection gene-locking-gapshift-state)
        (let* (buffer-read-only
               (line 0)
               ;; These next three must happen in this order:
               (seglist (gene-sel-get-seg-list selection t))
               (dummy (or
                       noundo
                       (gene-undo-record-selection-justification seglist)))
               (buffer-undo-list t))
          (while seglist
            (let* ((segment (car seglist))
                   (start (nth 0 segment))
                   (end (nth 1 segment))
                   did-something
                   (char-pt 0)
                   (gappy-start 0))

              (goto-char start)

              (message "Justifying line %d..." (setq line (1+ line)))
              (if (gene-seq-base-after-p)
                  ;; We're on a base, so forward until gaps
                  (skip-chars-forward gene-skip-bases-regexp end))

              (setq gappy-start (point))
              (if (< (point) end)
                  (progn
                    (skip-chars-forward gene-skip-gaps-regexp end)
                    (setq char-pt (point))))
              
              (while (< (point) end)
                
                (while (and
                        (< char-pt end)
                        (gene-base-p (char-after char-pt)))
                  (gene-translate-point-to-char
                   gappy-start (char-after char-pt))
                  (setq gappy-start (1+ gappy-start)
                        char-pt (1+ char-pt)
                        did-something t))
              
                (goto-char char-pt)
                (skip-chars-forward gene-skip-gaps-regexp end)
                (setq char-pt (point)))
              
              ;; Now gappify the end region

              (and
               did-something ; do nothing if no bases in this line
               (< gappy-start end)
               (let ((gap
                      (save-excursion
                        (goto-char end)
                        (skip-chars-backward
                         gene-skip-gaps-regexp gappy-start)
                        (gene-seq-gap-after))))
                 (gene-translate-region-to-char
                  gappy-start end gap))))

            (setq seglist (cdr seglist))))
        (message "Left-justifying selection... done."))))


(defun gene-sel-right-justify (selection)
  "Right-justify all residues in SELECTION.
Interactively, SELECTION is the selection at point."
  (interactive (list (gene-sel-selection-at-pos)))
  (if (gene-selection-p selection)
      (gene-save-excursion-in
          gene-seq-buffer
        (gene-locking-selection-check-state
         selection gene-locking-gapshift-state)
        (let* (buffer-read-only
               (line 0)
               ;; These next three must happen in this order:
               (seglist (gene-sel-get-seg-list selection t))
               (dummy (gene-undo-record-selection-justification seglist))
               (buffer-undo-list t))
          (while seglist
            (let* ((segment (car seglist))
                   (start (nth 0 segment)) ; "start" always refers to left
                   (end (1- (nth 1 segment))) ; "end" always to right
                   did-something
                   (char-pt 0)
                   (gappy-end 0))
              
              (goto-char end)
              
              (message "Justifying line %d..." (setq line (1+ line)))
              (if (gene-seq-base-after-p)
                  ;; We're on a base, so backward until gaps.
                  (progn ; note that this stops "on" a base, though.
                    (skip-chars-backward gene-skip-bases-regexp start)
                    (forward-char -1)))

              (setq gappy-end (point))
              (if (>= (point) start)
                  (progn
                    (skip-chars-backward gene-skip-gaps-regexp start) 
                    (forward-char -1) (setq char-pt (point))))
              
              (while (>= (point) start)
                
                (while (and
                        (>= char-pt start)
                        (gene-base-p (char-after char-pt)))
                  (gene-translate-point-to-char
                   gappy-end (char-after char-pt))
                  (setq gappy-end (1- gappy-end)
                        char-pt (1- char-pt)
                        did-something t))
                
                (goto-char char-pt)
                (skip-chars-backward gene-skip-gaps-regexp start)
                (forward-char -1) (setq char-pt (point)))
              
              ;; Now gappify the end region
              (and
               did-something
               (>= gappy-end start)
               (let ((gap
                      (save-excursion
                        (goto-char start)
                        (skip-chars-forward
                         gene-skip-gaps-regexp gappy-end)
                        (gene-seq-gap-before))))
                 (gene-translate-region-to-char
                  start (1+ gappy-end) gap))))
            
            (setq seglist (cdr seglist))))
        (message "Right-justifying selection... done."))))


;; We can use native commands `upcase-region' and `downcase-region'
;; for this, because they are fast subrs.
(defun gene-sel-upcase-selection (selection)
  "Convert all alphabetic characters in SELECTION to upper case.
Interactively, SELECTION is the selection at point."
  (interactive (list (gene-sel-selection-at-pos)))
  ;; We decided not to check locking for case ops.
  ;; (gene-locking-selection-check-state
  ;;  selection gene-locking-unlocked-state)
  (gene-save-excursion-in
      gene-seq-buffer
    (gene-undo-record-selection-case selection)
    (let ((buffer-undo-list t))
      (gene-sel-map-over-selection
       selection
       (symbol-function 'upcase-region)))))


(defun gene-sel-downcase-selection (selection)
  "Convert all alphabetic characters in SELECTION to lower case.
Interactively, SELECTION is the selection at point."
  (interactive (list (gene-sel-selection-at-pos)))
  ;; We decided not to check locking for case ops.
  ;; (gene-locking-selection-check-state
  ;;  selection gene-locking-unlocked-state)
  (gene-save-excursion-in
      gene-seq-buffer
    (gene-undo-record-selection-case selection)
    (let ((buffer-undo-list t))
      (gene-sel-map-over-selection
       selection
       (symbol-function 'downcase-region)))))


(defun gene-sel-flipcase-selection (selection)
  "Toggle the case of all alphabetic characters in SELECTION.
Interactively, SELECTION is the selection at point."
  (interactive (list (gene-sel-selection-at-pos)))
  ;; We decided not to check locking for case ops.
  ;; (gene-locking-selection-check-state
  ;;  selection gene-locking-unlocked-state)
  (gene-save-excursion-in
      gene-seq-buffer
    (gene-undo-record-selection-case selection)
    (let ((buffer-undo-list t))
      (gene-sel-map-over-selection
       selection
       (symbol-function 'gene-flipcase-region)))))


(defun gene-sel-subst-char-in-selection (selection &optional fromchar tochar)
  "In SELECTION, substitute all occurences of FROMCHAR with TOCHAR.
If FROMCHAR and/or TOCHAR is nil or omitted, then the user will be
prompted interactively for them."
  (interactive (list (gene-sel-selection-at-pos)))
  ;; We decided not to check locking for this too.
  ;; (gene-locking-selection-check-state
  ;;  selection gene-locking-unlocked-state)
  (or
   fromchar
   (setq fromchar (gene-read-char "Char to replace: " t)))
  (or
   tochar
   (setq tochar (gene-read-char (format "Replace \"%c\" with: " fromchar) t)))

  ;; Check for non-printables
  (if (or (< tochar ?!) (> tochar ?~))
      (gene-error "\"%c\" is not a regular printing char!" tochar))

  (message "Replacing \"%c\" with \"%c\"..." fromchar tochar)
  (gene-sel-map-over-selection selection 'subst-char-in-region fromchar tochar)
  (message "Replacing \"%c\" with \"%c\"... done." fromchar tochar))


;; In order to use the above, we need a convenient way to change the
;; lock status of an entire selection at once.  Groups don't quite cut
;; it for this.
(defun gene-sel-next-lock-status (selection)
  "Toggle the lock status of every sequence in SELECTION.
They all toggle according to the status of the first sequence in the
selection; naturally, after one iteration, they're all synchronized."
  (interactive (list (gene-sel-selection-at-pos)))
  (let* ((sym (gene-get-here-sym ; first seq in the selection
               (car (car (gene-sel-get-seg-list selection)))))
         (new-state (gene-locking-next-state-from-this-one
                      (gene-locking-get-sequence-state sym))))
    (message "Locking selection to state \"%s\"..."
             (gene-locking-state-string new-state))
    (gene-sel-map-over-selection selection
                             (lambda (start end state)
                               (gene-locking-lock-sequence-to-state
                                (gene-get-here-sym start)
                                ;; Yes, that's right, ignore `end'
                                state))
                             new-state)
    (message "Locking selection to state \"%s\"... done."
             (gene-locking-state-string new-state))))


(defun gene-sel-map-over-selection (selection func &rest args)
  "Map over SELECTION's segments, calling function FUNC.
FUNC's first two args will be the start and end of that segment.
Remaining ARGS will be also be passed to func after them.  That is,
the first ARG willl be the third arg of FUNC, second ARG would be
fourth arg of FUNC, and so on..."
  (gene-save-excursion-in
      gene-seq-buffer
    (let (buffer-read-only
          (seglist (gene-sel-get-seg-list selection)))
      (while seglist
        (let* ((cell (car seglist))
               (start (car cell))
               (end (car (cdr cell))))
          (apply func start end args)
          (setq seglist (cdr seglist)))))))



(defun gene-sel-selection-to-string (selection &optional no-pad)
  ;; The alignment in the docstring below is correct; the skewed
  ;; appearance is due to backslash-quoting.
  "Return SELECTION as a single string, suitable for insertion into text.
The selection will be padded so that it remains in alignment; each
segments will be separated with newlines.
Optional argument NO-PAD means don't pad ragged left edges with
indels.  An unpadded selection may not appear to be in alignment.
For example,

                 A                  \"-----A                   \"A     
this           AAU   becomes this    ---AAU   but this         AAU   
selection      AAC   when padded     ---AAC   when unpadded    AAC   
------>     AAUCAC   ------->        AAUCAC   ------->         AAUCAC
            AAA-UC                   AAA-UC\"                   AAA-UC\"
"
  (gene-save-excursion-in
      gene-seq-buffer
    (let (segment-strings
          cols)
      (gene-sel-map-over-selection
       selection
       (lambda (start end)
         (goto-char start)
         ;; This will no longer work when disjoint selections are legal:
         (or no-pad (setq cols (cons (gene-current-column) cols)))
         (setq segment-strings
               (cons (buffer-substring-no-properties start end) segment-strings)
               segment-strings
               (cons "\n" segment-strings))))

      ;; Put them back in "natural" order
      (setq cols (nreverse cols) segment-strings (nreverse segment-strings))

      (if no-pad
          ;; No padding, so return segment strings separated by newlines
          (let ((final-str (apply 'concat segment-strings)))
            (set-text-properties 0 (length final-str) nil final-str)
            final-str)

        ;; Else we need to splice in the correct number of padding
        ;; gaps at the beginning of each segment string.
        (let ((min-col (apply 'min cols))
              (tmp-cols cols)
              (tmp-strs segment-strings))
          (while tmp-strs
            (let* ((this-col    (car tmp-cols))
                   (padding-amt (- this-col min-col))
                   (padding     (if (> padding-amt 0)
                                    (make-string padding-amt ?-))))
              (if padding
                  (setcar tmp-strs (concat padding (car tmp-strs)))))
            (setq tmp-cols (cdr tmp-cols)
                  tmp-strs (cdr (cdr tmp-strs)))))

        ;; Now the final string can be created and returned:
        (let ((final-str (apply 'concat segment-strings)))
          (set-text-properties 0 (length final-str) nil final-str)
          final-str)
        ))))


(defun gene-sel-copy-selection-for-interprogram-paste (selection)
  "Copy the text of SELECTION into the windowing system's paste buffer.
Don't be confused by the fact that under X Windows, these clipboard
entries are also called selections, so that what this function does is
take an Ale selection and make it into an X selection."
  (interactive (list (gene-sel-selection-at-pos)))
  (gene-message "Copying selection to interprogram paste buffer...")
  (x-set-selection nil (gene-sel-selection-to-string selection))
  (gene-message
   "Copying selection to interprogram paste buffer... done."))


(defun gene-sel-save-selection-to-file (selection &optional file preface)
  "Save SELECTION in FILE, in the following format:

START-POSITION  ID  FRAGMENT
START-POSITION  ID  FRAGMENT
START-POSITION  ID  FRAGMENT
START-POSITION  ID  FRAGMENT
...

The START-RESIDUE is the sequence position of the first columns in
SELECTION.  Following it is the short ID for that sequence, then the
FRAGMENT of sequence itself.  The FRAGMENTS will be padded with indels
on the left to preserve their relative alignment.

If FILE already exists, its will be appended to, not destroyed.
If FILE is nil or omitted, the user will be prompted for a file, even
if this function was not called interactively.

Optional third arg PREFACE is text to insert before SELECTION; for
example, a note about the relevance of SELECTION."
  (interactive
   (list
    (gene-sel-selection-at-pos)
    (expand-file-name (read-file-name "Save selection to file: "))))

  (or ; only set FILE if necessary -- we might be non-interactive
   file
   (setq file (expand-file-name (read-file-name "Save selection to file: "))))

  (message "Writing selection...")
  
  (gene-save-excursion-in
      gene-seq-buffer
    (let ((sel-string (gene-sel-selection-to-string selection))
          (buf (find-file-noselect file))
          opoint)
      (set-buffer buf)
      (goto-char (point-max))
      (unless (= (point) (point-min))
        (insert "\n")) ; just to be safe
      (if preface (insert preface "\n"))
      
      ;; Put in the names, make everything fasta format
      
      (setq opoint (point))
      (insert sel-string)
      (goto-char opoint)
      (set-buffer gene-seq-buffer)
      (gene-sel-map-over-selection
       selection
       (lambda (start end)
         (save-excursion
           (goto-char start)
           (let ((name (gene-organism-name)) ; defaults to seq@point
                 (start-pos (gene-seq-base-number)))
             (set-buffer buf)
             (insert ">" name ?\t (number-to-string start-pos) "\n")
             (forward-line 1)))))
      (set-buffer buf)
      (save-buffer)
      (kill-buffer buf))
    
    (message "Writing selection... done.")))



(provide 'gene-selection)

;; An idea for simplifying undo *and* making selections behave a bit
;; more intuitively:
;;
;; Drag is to the right, so `count' is positive.  The top two chunks
;; lead the bottom one by 6 when in unsquished formation.
;;
;;            |
;;        aaa | +1
;;        aaa | +1
;;  aaa       | +7
;;            |
;;
;; Top two collide with wall first, but bottom one still drags:
;;
;;            |
;;         aaa| -3
;;         aaa| -3
;;      aaa   | +3
;;            |
;;
;; So in this case, with `count' positive, minimizing diffs means
;; don't do anything.  Why?  Because we have diffs of both kinds.  If
;; we had all diffs of one kind (i.e.: all positive or all negative),
;; then it would make sense to minimize them so that the next drag
;; (manual or from a program, say undo) took effect immediately, but
;; was still able to undo *relative* differences between segments.
;; The trick is to record this in the selection, so that drags always
;; do it right!  make-sel-distances is going to have to behave
;; differently when COUNT arg is nil: it will have to check for an old
;; fluidity record and compensate if it can, being careful not to
;; violate the text of the buffer -- which might have changed since
;; the drag was done, thus impossibilizing a full recovery of the
;; original formation.  Of course, in undo, we are guaranteed that the
;; original formation is possible, and make-sel-distances must be able
;; to effect a perfect recovery in that case.


;; The Plan, then, is to make selection dragging a bit smarter.  This
;; will be done by recording a diff list in the selection itself,
;; after the drag is over.  When make-sel-distances is called for
;; resetting, it will attempt to use this diff-list, *even* if the
;; drag is not explicitly requested to be fluid.
;;
;; Another possibility is to only use it during undo; this makes the
;; job of using the diff-list easier, because we don't have to check
;; for changes in the buffer around us.
