;; gene-sgroup.el (part of ale).
;; Copyright (C) 1994  Karl Fogel <kfogel@darwin.life.uiuc.edu>
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Code for manipulating sequence groups.

(require 'gene-mode-tools)


;;; Keybindings:

;; (define-key gene-id-mode-map [S-mouse-1]
;;   'gene-sgroup-start-click-select)
;; (define-key gene-seq-id-map [down-mouse-1]   'gene-sgroup-select-unselect)

(define-key gene-id-mode-map [mouse-1]      'gene-absorb-mouse-event)
(define-key gene-id-mode-map [down-mouse-1] 'gene-sgroup-drag-select)

(define-key gene-id-mode-map [double-mouse-1]
  'gene-absorb-mouse-event)
(define-key gene-id-mode-map [triple-mouse-1]
  'gene-absorb-mouse-event)
(define-key gene-id-mode-map [double-down-mouse-1]
  'gene-absorb-mouse-event)
(define-key gene-id-mode-map [triple-down-mouse-1]
  'gene-absorb-mouse-event)
(define-key gene-id-mode-map [drag-down-mouse-1]
  'gene-absorb-mouse-event)

(define-key gene-id-mode-map [double-mouse-2]
  'gene-absorb-mouse-event)
(define-key gene-id-mode-map [triple-mouse-2]
  'gene-absorb-mouse-event)
(define-key gene-id-mode-map [mouse-2]
  'gene-absorb-mouse-event)
(define-key gene-id-mode-map [drag-mouse-2]
  'gene-absorb-mouse-event)
(define-key gene-id-mode-map [down-mouse-2]
  'gene-absorb-mouse-event)
(define-key gene-id-mode-map [double-down-mouse-2]
  'gene-absorb-mouse-event)
(define-key gene-id-mode-map [triple-down-mouse-2]
  'gene-absorb-mouse-event)


;;; Code.

;; Vocabulary:
;; A "displayed" group is group that's currently visible to the user.
;; More than one group can be displayed at once, so long as none of
;; the displayed groups have sequences in common.
;;
;; The "selected" group is the one to which sequences will get added
;; if you click on their ID's.  It should be brighter than the rest of
;; the groups, although we don't really know yet how we're going to
;; choose it and change it.
;;
;; The selected group is also the one that the sequence buffer pays
;; attention to; merely displayed groups don't affect editing.
;;
;; mouse-1 on a sequence should make it a member of the current group
;; if that sequence is not part of a displayed group already.  If it
;; is, mouse-1 removes it from that group.  IDs will be
;; mouse-sensitive, to make all this easier to navigate.
;;
;; S-mouse-1 should delete a group.  FOO-mouse-1 should hide it (we
;; don't really have a corresponding operation on selections -- should
;; we?).  Hmm, give the binding some more thought, talk to Jim about
;; it...
;;
;; When a group function takes a "sequence" as an argument, what it
;; really takes is the orgsym of that sequence.  Any information about
;; the sequence can be gotten from there, and any bookkeeping will be
;; done there anyway.
;;
;; Each orgsym has two properties: `gene-sgroup-displayed' and
;; `gene-sgroups-here'.  The first is the displayed sgroup of this
;; sequence, or nil if none right now; the second is a list of all
;; groups that contain this sequence as a member.

;; The group list looks like:
;; ((("group-name" displayed-p face ...) seglist) ; now: `undisp-face'?
;;  (("group-name" displayed-p face ...) seglist)
;;  ...)
;;
;; and therefore each group looks like:
;; (("group-name" displayed-p face ...) seglist)
;; 
;; where each `seglist' is of the form:
;; ((orgsym id-overlay) (orgsym id-overlay) ...)
;;
;; So:
;; The little list that is the car of each group just contains
;; miscellaneous information about the group, such as its name,
;; whether it's being displayed at the moment, its face, whatever.
;;
;; The actual sequences (and their corresponding overlays) of the
;; group are contained in the cdr of the group.
;;
;; See also the long explanatory comment near the top of
;; gene-clipboard.el, the one beginning "Some of this code won't make
;; sense unless...".  It confides some useful secrets about how groups
;; work.



;;; Common objects and constants

(defvar gene-sgroup-list ()
  "The list of sequence groups.")

(defvar gene-sgroup-selected-group nil
  "Selected sequence group.  Chosen sequences are added to this group.
Nil means no group is currently selected.")

(defvar gene-sgroup-unnamed-count 1
  "Extension to put in the names of auto-generated groups.")

(defvar gene-sgroup-displayed-list ()
  "The list of currently displayed sequence groups.")

(defconst gene-sgroup-non-name "" ; just the empty string for now
  "*What to call an unnamed sequence group by default.")


;;; Helper functions

(defun gene-sgroup-selected-group ()
  "Returns the selected group if there is one, else makes a new one."
  (or
   gene-sgroup-selected-group
   (prog1
       ;; This selects it, too:
       (gene-sgroup-make-new-group
        (concat "Unnamed " (number-to-string gene-sgroup-unnamed-count)))
     (setq gene-sgroup-unnamed-count
           (1+ gene-sgroup-unnamed-count)))))


(defun gene-sgroup-rename-group (&optional sgroup)
  "Rename SGROUP interactively, prompting for a new name.
If SGROUP is nil, then prompt for its old name, too.
If you want to change a group's name from Lisp, then just use
`gene-sgroup-set-name'."
  (interactive)
  (or sgroup (setq sgroup (gene-sgroup-get-group-by-name
                           nil
                           "Change name of group: ")))
  (gene-sgroup-set-name sgroup
                        (read-string "New name: "))
  ;; Show the change:
  (gene-sgroup-frame-update))


(defun gene-sgroup-group-unnamed-p (sgroup)
  "Return t iff SGROUP is unnamed.
See also variable `gene-sgroup-non-name'."
  (string-equal gene-sgroup-non-name (gene-sgroup-get-name sgroup)))


(defun gene-sgroup-get-name (sgroup)
  "Return SGROUP's name (a string)."
  (car (car sgroup)))


(defun gene-sgroup-set-name (sgroup name)
  "Set SGROUP's name to NAME."
  (setcar (car sgroup) name))


(defun gene-sgroup-get-face (sgroup)
  "Return SGROUP's face."
  (nth 2 (car sgroup)))


(defun gene-sgroup-set-face (sgroup face)
  "Set SGROUP's face to FACE."
  (setcar (nthcdr 2 (car sgroup)) face))


;; Returns a list of the sequences in the group.
(defalias 'gene-sgroup-seglist (symbol-function 'cdr))


(defun gene-sgroup-sequences (sgroup)
  "Return a list of the sequences in sgroup."
  (mapcar 'car (gene-sgroup-seglist sgroup)))


(defun gene-sgroup-get-mouse-face (sgroup)
  "Return SGROUP'S mouse-face."
  (nth 3 (car sgroup)))


(defun gene-sgroup-set-mouse-face (sgroup face)
  "Set SGROUP'S mouse-face to FACE."
  (setcar (nthcdr 3 (car sgroup)) face))


(defun gene-sgroup-get-displayed-p (sgroup)
  "Return t iff SGROUP is currently displayed."
  ;; Just return the displayed-p flag in the group's car
  (nth 1 (car sgroup)))


(defun gene-sgroup-set-displayed-p (sgroup flag)
  "Set SGROUP's displayed-p flag to FLAG."
  ;; Just return the displayed-p flag in the group's car
  (setcar (nthcdr 1 (car sgroup)) flag))


(defun gene-sgroup-displayed-group (sym)
  "Return the displayed sgroup containing sequence SYM, or nil if none."
  (get sym 'gene-sgroup-displayed))


(defun gene-sgroup-displayed-group-here ()
  "Return displayed sgroup containing current sequence, or nil if none."
  (gene-sgroup-displayed-group (gene-get-here-sym)))


(defun gene-sgroup-selected-group-at-seq (sym)
  "Return selected group if sequence SYM is in it, or else nil."
  (get sym 'gene-sgroup-selected))


(defun gene-sgroup-selected-here ()
  "Return selected group if current sequence is in it, or else nil."
  (gene-sgroup-selected-group-at-seq (gene-get-here-sym)))


(defun gene-sgroup-groups-containing-seq (sym)
  "Return list of sequence groups containing sequence SYM.
This examines a property of SYM.  Thus, if SYM is absent from the
alignment \(say, in the clipboard\), this function will return the
groups SYM *thinks* it's a member of."
  (get sym 'gene-sgroups-here))


(defun gene-sgroup-put-groups-in-seq (sym grouplist)
  "Tell sequence SYM that it's contained by the groups in GROUPLIST.
SYM is an organism symbol, GROUPLIST is a list of sequence groups."
  (put sym 'gene-sgroups-here grouplist))


(defun gene-sgroup-delete-group-from-plists (sgroup)
  "Remove SGROUP from the plists of its sequences' organism symbols.
This is usually only done in the process of deleting a group."
  (mapcar
   (lambda (seg)
     (gene-sgroup-delete-group-from-sym sgroup (car seg)))
   (gene-sgroup-seglist sgroup)))


(defun gene-sgroup-add-group-to-sym (sgroup sym)
  "Add SGROUP to the list of groups containing sequence SYM.
This is done by modifying the value of the `gene-sgroups-here'
property in SYM's plist.
If SYM thinks it's already a member of SGROUP, then do nothing."
  (let ((groups (gene-sgroup-groups-containing-seq sym)))
    (or
     ;; If it's already in the group, then do nothing.
     (memq sgroup groups)
     (progn
       ;; Add the sgroup, then re-put the list.
       (setq groups (cons sgroup groups))
       (gene-sgroup-put-groups-in-seq sym groups)))))


(defun gene-sgroup-delete-group-from-sym (sgroup sym)
  "Remove SGROUP from the sequences containing SYM.
This is done by modifying the value of the `gene-sgroups-here'
property in SYM's plist."
  (let ((groups (gene-sgroup-groups-containing-seq sym)))
    ;; Get the list, delq the sgroup, then re-put the list
    (setq groups (delq sgroup groups))
    (gene-sgroup-put-groups-in-seq sym groups)))


(defun gene-sgroup-group-empty-p (sgroup)
  "Return t iff SGROUP contains no sequences."
  (not (gene-sgroup-seglist sgroup)))


(defun gene-sgroup-seq-in-group-p (sym sgroup)
  "Return t iff the sequence SYM is in SGROUP.
This examines both the property on SYM and the syms stored in SGROUP."
  (and
   (memq sgroup (gene-sgroup-groups-containing-seq sym))
   (assq sym (gene-sgroup-seglist sgroup))))


(defun gene-sgroup-build-sgroup-completion-table ()
  "Build and return a completion table for sequence group names."
  (mapcar
   (lambda (sgroup)
     (list (gene-sgroup-get-name sgroup) sgroup))
   gene-sgroup-list))


(defun gene-sgroup-get-group-by-name (&optional name prompt)
  "Return the sgroup named NAME, or nil if none.
If NAME is nil or omitted, prompt user for NAME, completing on groups.
From Lisp, second optional arg PROMPT is a prompt to use when
reading a name from the user.  It is possible to pass both a NAME and a
PROMPT string; in that case, the PROMPT will never be used."
  (or name
      (setq name
	    (completing-read
	     (or prompt "Enter a group name: ")
	     (gene-sgroup-build-sgroup-completion-table)
	     nil t)))
  (catch 'the-group
    (mapcar
     (lambda (sgroup)
       (if (string-equal name (gene-sgroup-get-name sgroup))
           (throw 'the-group sgroup)))
     gene-sgroup-list)
    (throw 'the-group nil)))



;;; Group color and face stuff.

(defconst gene-sgroup-default-colors 
  (let ((lst (list 
              "LightBlue"
              "yellow3"
              "pink2"
              "cyan3"
              "gold2"
              "pink"
              "coral"
              "wheat"
              "light grey"
              "green2"
              )))
    (nconc lst lst))
  "Cdr-circular list of colors to be used for sequence groups.")


(defvar gene-sgroup-colors nil
  "*List of colors used by sequence groups.
This list is circular; only access it with `gene-define-group-color'
or `gene-define-group-color-internal'.")


(defmacro gene-define-group-color (color)
  "Define a new color to be used for sequence groups."  
  (list 'gene-define-group-color-internal
	(if (symbolp color) (symbol-name color) color)))


(defun gene-define-group-color-internal (color)
  "Define a new color to be used for sequence groups."
  (setq color (gene-check-color-allowable color))
  (if gene-sgroup-colors
      ;; We have some colors already, so add this one:
      (progn
        (setcdr gene-sgroup-colors (cons color (cdr gene-sgroup-colors)))
        ;; Rotate the new color to the front:
        (setq gene-sgroup-colors (cdr gene-sgroup-colors)))
    ;; Else create the color list.
    (setq gene-sgroup-colors (list color))
    (nconc gene-sgroup-colors gene-sgroup-colors)))


(defun gene-sgroup-color ()
  "Returns the \"next\" sequence group color."
  ;; It's a circular list, that's why this works:
  (car (if gene-sgroup-colors
           (setq gene-sgroup-colors (cdr gene-sgroup-colors))
         (setq gene-sgroup-default-colors (cdr gene-sgroup-default-colors)))))


(defun gene-sgroup-generate-face (color)
  "Return a face whose foreground is set to COLOR.
Use optional arg COLOR instead, if non-nil.
The foreground being set makes this group appropriate for displayhood
but not selectedness.  Selecting it will swap the foreground and
background colors."
  (let ((face (make-face (gensym "gene-sgroup-face-"))))
    (set-face-foreground face (gene-check-color-allowable color))
    face))


(defun gene-sgroup-generate-mouse-face (color)
  "Return a bold face whose foreground is COLOR.
The foreground being set makes this group appropriate for displayhood
but not selectedness.  Selecting it will swap the foreground and
background colors."
  (let ((face (make-face (gensym "gene-sgroup-mouse-face-"))))
    (set-face-foreground face (gene-check-color-allowable color))
    (make-face-bold face nil t)
    face))


;;; Core functions

(defun gene-sgroup-generate-blank-group ()
  "Return a new, empty, unnamed sequence group.
Only the face is defined, and the group is undisplayed by default."
  ;; List of the form ((name displayed-flag face) ...)
  (let ((color (gene-sgroup-color)))
    (list
     (list
      ""
      nil
      (gene-sgroup-generate-face color)
      (gene-sgroup-generate-mouse-face color)
      ))))


(defvar gene-sgroup-making-group nil
  "Non-nil while a group creation is in progress.
For internal use only; don't depend on this var or set it.")


(defun gene-sgroup-make-new-group (&optional name)
  "Make and select a new sequence group named NAME.
Returns the group.
If NAME is nil or omitted, then the user will be prompted.
If a group name NAME already exists, an error will be signalled."
  (interactive)
  (if gene-sgroup-making-group
      ;; don't allow recursive invocations
      nil
    ;; Else we are not already in the process of making a group, so go
    ;; ahead and make one.
    (let ((gene-sgroup-making-group t))
      ;; Later on, this could use a dialog box or some other gimmick.
      (gene-sgroup-make-new-group-internal
       (or
        name
        (let ((cursor-in-echo-area t))
          (read-string "Name for new group: ")))))))


(defun gene-sgroup-make-new-group-internal (name)
  "Internal helper function -- makes a new sequence group named NAME.
If a group name NAME already exists, an error will be signalled."
  (let ((grp (gene-sgroup-generate-blank-group)))
    (if (gene-sgroup-get-group-by-name name) ; enforce unique names
        (gene-error "A group named %s already exists." name)
      (gene-sgroup-set-name grp name))
    (setq gene-sgroup-list (cons grp gene-sgroup-list))
    ;; The new group is always selected too:
    (gene-sgroup-select-group grp)
    grp))


(defun gene-sgroup-just-display-group (sgroup)
  "See to it that SGROUP is displayed but not selected."
  (gene-sgroup-hide-group sgroup)
  (gene-sgroup-display-group sgroup))


(defun gene-sgroup-select-group (sgroup)
  "Make SGROUP the selected group.
If SGROUP is nil, unselect the current selected group if there is one."
  (cond
   ;; Do nothing if SGROUP is already selected, or if it's nil and so
   ;; is gene-sgroup-selected-group.
   ((eq gene-sgroup-selected-group sgroup)
    nil)

   ;; If the arg is nil, then unselect the selected group.  It will
   ;; become just another displayed group.
   ((null sgroup)
    (invert-face (gene-sgroup-get-face       gene-sgroup-selected-group))
    (invert-face (gene-sgroup-get-mouse-face gene-sgroup-selected-group))
    ;; Tell the seqs they're no longer in the selected group:
    (mapcar
     (lambda (seg)
       (let ((sym (car seg)))
         (put sym 'gene-sgroup-selected nil)))
     (gene-sgroup-seglist gene-sgroup-selected-group))
    (setq gene-sgroup-selected-group nil)
    (gene-sgroup-frame-update sgroup))

   ;; Else select sgroup!
   (t
    (if gene-sgroup-selected-group
        ;; Make a recursive call to unselect the old group.  We
        ;; already know that the current group is not the selected
        ;; one, or the first condition in this cond would have
        ;; returned true, or at least pepperoni.
        (gene-sgroup-select-group nil))
    (gene-sgroup-display-group sgroup) ; a no-op if already displayed
    (invert-face (gene-sgroup-get-face sgroup)) ; show its selectedness
    (invert-face (gene-sgroup-get-mouse-face sgroup))
    (setq gene-sgroup-selected-group sgroup) ; record its selectedness
    (gene-sgroup-frame-update sgroup) ; tell the frame about it too

    ;; Give each sequence a quick way to say it's in the selected
    ;; group:
    (mapcar
     (lambda (seg)
       (let* ((sym (car seg)))
              (put sym 'gene-sgroup-selected sgroup)))
     (gene-sgroup-seglist sgroup))

    (if gene-sgroup-follow-selected-group
        (gene-sgroup-point-to-group gene-sgroup-selected-group))
    )))


(defun gene-sgroup-display-group (sgroup)
  "Display the sequence sequence group SGROUP.
SGROUP is a sequence group, but if called interactively, you will be
prompted for the name of a sequence group."
  (interactive (list (gene-sgroup-get-group-by-name)))

  ;; If this encounters any conflicts, it should make sure to hide
  ;; the old conflicting group.  The idea is that it's okay to hide a
  ;; group automagically, just not to display one in a surprising way
  ;; or at a surprising time.
  (if (gene-sgroup-get-displayed-p sgroup)
      nil ; do nothing if displayed
    ;; If any of its sequences are already in displayed groups, then
    ;; undisplay those groups before going on.
    (mapcar (lambda (seq)
              (let ((dgroup (gene-sgroup-displayed-group (car seq))))
                (if dgroup
                    (gene-sgroup-hide-group dgroup))))
            (gene-sgroup-seglist sgroup))
    
    ;; Then just display the group.  This takes care of setting any
    ;; flags on the group or its syms.
    (gene-sgroup-display-group-internal sgroup)
    ;; Now it's safe to go ahead and put this group in the list of
    ;; displayed groups
    (setq gene-sgroup-displayed-list
          (cons sgroup gene-sgroup-displayed-list))
    (gene-sgroup-frame-update sgroup)))


(defun gene-sgroup-display-group-internal (sgroup)
  "Internal function for displaying groups.
Makes SGROUP visible by putting its overlays in the right places.
Probably the only caller of this should be `gene-sgroup-display-group'."
  (mapcar
   (lambda (seg)
     (let* ((sym (car seg))
            (id-pos-list (get sym 'gene-id-posn))
            (id-buf (marker-buffer (car id-pos-list)))
            (start (car id-pos-list))
            (end (nth 1 id-pos-list)))
       ;; We use end+1 because that makes the highlighting go all the
       ;; way to the end of the line.
       (move-overlay (nth 1 seg) start (1+ end) id-buf)
       (put sym 'gene-sgroup-displayed sgroup)
       ))
   (gene-sgroup-seglist sgroup))
  ;; Then set the display-p flag for the group as a whole
  (gene-sgroup-set-displayed-p sgroup t))


(defun gene-sgroup-hide-group (sgroup)
  "Hide (i.e.: undisplay, and unselect if necessary) SGROUP."
  (interactive (list (gene-sgroup-get-group-by-name)))
  ;; Only hide it if it's displayed in the first place
  (if (gene-sgroup-get-displayed-p sgroup)
      (progn
        ;; If this was the selected group, then there won't be any more
        ;; selected group right now, because hiding it unselects it.
        (if (eq gene-sgroup-selected-group sgroup)
            (gene-sgroup-select-group nil))
        
        ;; Remove the overlays and the `group-displayed-here' prop from the
        ;; sequence's symbol:
        (mapcar
         (lambda (seg)
           (let ((sym (car seg)))
             (delete-overlay (nth 1 seg))
             ;; Don't change the sym's props unless this sgroup is the group
             ;; designated as being displayed here
             (if (eq sgroup (gene-sgroup-displayed-group sym))
                 (put sym 'gene-sgroup-displayed nil))))
         (gene-sgroup-seglist sgroup))

        ;; Remove this group from the list of displayed groups:
        (setq gene-sgroup-displayed-list
              (delq sgroup gene-sgroup-displayed-list))     

        ;; Tell the group itself that it's not displayed:
        (gene-sgroup-set-displayed-p sgroup nil)
        (gene-sgroup-frame-update sgroup))))


(defun gene-sgroup-toggle-visibility (sgroup)
  "If SGROUP is displayed, the hide it, else display it."
  (if (gene-sgroup-get-displayed-p sgroup)
      (gene-sgroup-hide-group sgroup)
    (gene-sgroup-display-group sgroup)))


(defun gene-sgroup-delete-group (sgroup)
  "Delete SGROUP.  This removes it from the universe."
  (interactive (list (gene-sgroup-get-group-by-name)))
  (gene-sgroup-hide-group sgroup)
  (mapcar (lambda (seg)
            (gene-sgroup-delete-seq-from-group sgroup (car seg)))
          (gene-sgroup-seglist sgroup))
  (setq gene-sgroup-list (delq sgroup gene-sgroup-list))
  (gene-sgroup-frame-update))


(defun gene-sgroup-add-seq-to-group (sgroup sym)
  "Add sequence SYM (second arg) to SGROUP (first arg).
Interactively, prompts for the group and uses the sequence at point."
  (interactive (list (gene-sgroup-get-group-by-name) (gene-get-here-sym)))
  (or
   ;; Don't put it in if it's already there.
   (gene-sgroup-seq-in-group-p sym sgroup)

   ;; Okay, it's not a member, so make it one:
   (let* ((id-pos-list (get sym 'gene-id-posn))
          (start (car id-pos-list))
          (end (nth 1 id-pos-list))
          (ov (make-overlay start (1+ end) gene-id-buffer)))
     (and (gene-sgroup-displayed-group sym)
          ;; below useful?  Guess not...
          ;; (gene-sgroup-seq-in-group-p sym sgroup)
          (gene-sgroup-hide-group (gene-sgroup-displayed-group sym)))
     (overlay-put ov 'face (gene-sgroup-get-face sgroup))
     (overlay-put ov 'mouse-face (gene-sgroup-get-mouse-face sgroup))

     ;; Put it in the right place.
     (gene-sgroup-splice-seq-into-sgroup sgroup sym ov)

     (gene-sgroup-add-group-to-sym sgroup sym)
     (if (eq sgroup gene-sgroup-selected-group)
         (put sym 'gene-sgroup-selected sgroup))
     ;; paranoia -- can't imagine when we'd be adding to a hidden
     ;; group, or even to anything other than the selected group, for
     ;; that matter.  But it would not be proper for this function to
     ;; make that assumption.
     (if (gene-sgroup-get-displayed-p sgroup)
         (put sym 'gene-sgroup-displayed sgroup)
       (delete-overlay ov)))))


(defun gene-sgroup-splice-seq-into-sgroup (sgroup sym overlay)
  "Internal function called by gene-sgroup-add-seq-to-group.
Wants the SGROUP, the sequence SYM, and the OVERLAY.
You really shouldn't need to call this function from anywhere else.

All it does is make sure that when a sequence is added to a group,
it's inserted into the right place in the list that is the group,
based on line-numbers.
Maintaining this order is useful in several places, and it's most
convenient to make sure it happens right when the sequence is entered
into the group, since the cut and paste functions also use
gene-sgroup-add-seq-to-group."

  (let* ((seq-pos (car (get sym 'gene-seq-posn)))
         (seglist (gene-sgroup-seglist sgroup))
         (sym1 (car (car seglist)))
         (seq-pos1 (car (get sym1 'gene-seq-posn))))

    ;; No two line numbers should ever be the same; should we include
    ;; a sanity check here?
    (if (or (not sym1) (< seq-pos seq-pos1)) ; don't cmp if seglist empty

        ;; SYM should be the first sequence in this group
        (setcdr sgroup (cons (list sym overlay) seglist))
      
      ;; Else it belongs somewhere inside or at the end of the group.
      (let ((seglist-cdr (cdr seglist)))
        (catch 'gronk
          (while seglist-cdr
            (let* ((symN (car (car seglist-cdr)))
                   (seq-posN (car (get symN 'gene-seq-pos))))
              (if (and seq-posN (< seq-pos seq-posN))
                  ;; Splice here.
                  (progn
                    (setcdr seglist (cons (list sym overlay) seglist-cdr))
                    (throw 'gronk nil))
                ;; Else try the next spot in the list.
                (setq seglist seglist-cdr
                      seglist-cdr (cdr seglist-cdr)))))
          ;; Else this sequence belongs at the end.  Seglist is quite
          ;; short now, so this is even cheaper than usual, whoopee.
          (nconc seglist (list (list sym overlay))))))))


;; See the long explanatory comment near the top of gene-clipboard.el,
;; the one beginning "Some of this code won't make sense unless..."
(defun gene-sgroup-delete-seq-from-group (sgroup sym &optional leave-props)
  "Delete sequence SYM (second arg) from SGROUP (first arg).
Interactively, prompts for the group and uses the sequence at point.
If optional third arg LEAVE-PROPS is non-nil, then don't remove the
group properties from the sequence SYM.  This is used when cutting a
sequence for later pasting."
  (interactive (list (gene-sgroup-get-group-by-name) (gene-get-here-sym)))
  (let ((seg (assq sym (gene-sgroup-seglist sgroup))))
    (if seg ; if the sym is really in this group in the first place
        (progn
          (delete-overlay (nth 1 seg))
          (delq seg sgroup))) ; delq is enough -- `seg' is not the car
    (if leave-props
        nil
      (gene-sgroup-delete-group-from-sym sgroup sym)
      (if (eq (gene-sgroup-selected-group-at-seq sym) sgroup)
          (put sym 'gene-sgroup-selected nil))
      (if (eq (gene-sgroup-displayed-group sym) sgroup)
          (put sym 'gene-sgroup-displayed nil)))))


(defun gene-sgroup-add-delete-seq-to-group (sgroup sym)
  "Put SGROUP in sequence SYM if it's not there, else remove it."
  ;; Essentially, it's a toggle between the two above functions.
  (if (memq sgroup (gene-sgroup-groups-containing-seq sym))
      (gene-sgroup-delete-seq-from-group sgroup sym)
    (gene-sgroup-add-seq-to-group sgroup sym)))


;;; Interface functions

;; The behavior is that when you click on an ID, a few different
;; things might happen:
;;
;; 1) If the ID is in the selected group, then remove it from the
;;    group.
;;
;; 2) If the ID is not in the selected group, but there is a selected
;;    group, then add it to the selected group.
;;
;; 3) If the ID is in a displayed group, then select that group.

;;; Mouse clicking and dragging commands

;;; Much of this code comes from Mike William's mouse-sel package,
;;; which is available under the GPL.  The full copyright notice is:

;;; ;;; mouse-sel.el --- Multi-click selection support for Emacs 19
;;; 
;;; ;; Copyright (C) Mike Williams 1993
;;; 
;;; ;; Author: Mike Williams <mikew@gopher.dosli.govt.nz>
;;; ;; Keywords: mouse
;;; ;; Version: \$\Revision: 1.55 \$\
;;; 
;;; ;; This file is not (yet) part of GNU Emacs, but is made available under
;;; ;; the same conditions.
;;; ;;
;;; ;; GNU Emacs is free software; you can redistribute it and/or modify
;;; ;; it under the terms of the GNU General Public License as published by
;;; ;; the Free Software Foundation; either version 2, or (at your option)
;;; ;; any later version.
;;; ;;
;;; ;; GNU Emacs is distributed in the hope that it will be useful,
;;; ;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; ;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; ;; GNU General Public License for more details.

(defun gene-sgroup-range-with-group (func sgroup start end)
  "Use FUNC to add to or delete from SGROUP the sequences from START to END.
START and END is taken in the ID buffer, of course.
FUNC is typically `gene-sgroup-add-seq-to-group' or
`gene-sgroup-delete-seq-from-group'."
  ;; you *could* call this the "swap space"... but I wouldn't
  ;; recommend it:
  (if (> start end)
      (setq start (prog1 end (setq end start))))
  (save-excursion
    (set-buffer gene-id-buffer)
    (goto-char start)
    (while (<= (point) end)
      (funcall func sgroup (gene-get-here-sym))
      (forward-line 1))))


(defun gene-sgroup-point-to-group (sgroup)
  "Move point to a sequence in SGROUP.
Try to preserve column and orientation with respect to the group.

If the group has no sequences (i.e.: is empty), then do nothing.

TODO: CONDITION BELOW IS NOT IMPLEMENTED YET, BUT SHOULD BE SOMEDAY:
If possible, stay on the visible screen."
  ;; Find the line number in SGROUP closest to current-line, and move
  ;; there.
  (set-buffer gene-seq-buffer)
  (if (gene-sgroup-seglist sgroup)
      (or
       ;; Don't move unless necessary:
       (memq sgroup (gene-sgroup-groups-containing-seq
                     (gene-get-here-sym)))
       ;; Okay, necessary to move:
       (let* ((pt (point))
              (syms (gene-sgroup-sequences sgroup))
              (pos-list
               (sort
                (mapcar
                 (lambda (sym)
                   (car (get sym 'gene-seq-posn)))
                 syms)
                '<))
              prev-cons)

         (while (and pos-list (> pt (car pos-list)))
           (setq prev-cons pos-list pos-list (cdr pos-list)))

         ;; Now (car prev-cons) is the last pos less than point,
         ;; and (car pos-list) is the first one greater than point.
         (let ((prev-pos (car prev-cons))
               (next-pos (car pos-list)))
           (gene-seq-goto-organism
            (get-text-property (or next-pos prev-pos) 'gene-ident)))))))


(defun gene-sgroup-make-faces ()
  "Create the appropriate faces for sequence groups.
`gene-made-all-faces' calls this.

Note that it's incorrect to make faces in top-level expressions, since
the construction of a face depends on run-time factors (like the fonts
available)."

  ;; for drag-selecting sequence groups
  (copy-face 'default 'gene-sgroup-background-face)
  (set-face-background
   'gene-sgroup-background-face
   (gene-check-color-allowable gene-background-color))

  (copy-face 'default 'gene-sgroup-background-mouse-face)
  (set-face-background
   'gene-sgroup-background-mouse-face
   (gene-check-color-allowable gene-background-color))
  )


(defun gene-sgroup-drag-select (start-event)
  "Select IDs to be put into the current sequence group.
This should be bound to a down-mouse event.

If the down-click happens on a member of the selected group, then the
dragging has the effect of removing sequences from the selected group
if they're in it; otherwise, they're left alone.

Else it adds sequences to the selected group, unless they're already
in it, in which case they're left alone."
  (interactive "e")
  (let* ((seq-opoint (save-excursion
                       (set-buffer gene-seq-buffer)
                       (point)))
         (start-posn (event-start start-event)) ; a posn thingy
         (start-window (posn-window start-posn))
         (start-pt (posn-point (event-start start-event)))
	 start-bol
	 (start-eol (save-excursion
		      (set-buffer (window-buffer start-window))
		      (goto-char start-pt)
		      (beginning-of-line)
		      (setq start-bol (point))
		      (forward-line 1)
		      (point)))
         (olay (make-overlay start-bol start-eol gene-id-buffer))
         (start-frame (window-frame start-window))
         (half-w-height (/ (window-height start-window) 2))
         end-event         ; will be used later
         end-posn          ; will be used later
         end-pt            ; will be used later
         first-seq-added   ; will be used later
         (top (nth 1 (window-edges start-window)))
         (bottom (nth 3 (window-edges start-window)))
         direction
         ;; Don't follow while dragging:
         gene-sgroup-follow-selected-group
         ;; The sequence we first clicked on.
         (sym (gene-get-here-sym start-pt gene-id-buffer))
         ;; The disp-group-here and sel-group-here might be the same.
         ;; If not, then sel-group-here is nil, and maybe
         ;; disp-group-here is too.
         (disp-group-here (gene-sgroup-displayed-group sym))
         (sel-group-here 
          (if gene-sgroup-selected-group
              ;; If there is a selected group, then we want to add or
              ;; delete sequences from it.  If none, then we want to
              ;; create one and add sequences to it.  So, if there is
              ;; a selected group, just see if we clicked inside it,
              ;; since we're deciding between the two alternatives of
              ;; the first sentence.
              (gene-sgroup-selected-group-at-seq sym)
            ;; Else create the selected group but don't set
            ;; `sel-group-here' to it -- this way we'll be
            ;; automatically adding to the new selected group.
            ;; Of course, if there's a displayed group here, then
            ;; don't do it, because the end result of this call should
            ;; just be to select that group.
            (if disp-group-here
                nil
              (gene-sgroup-selected-group)
              nil)))
         ) ; end of `let*' bindings

    (condition-case
        the-error
        (progn
          (if (and disp-group-here (not sel-group-here))
              ;; Mouse-1 on a merely displayed group selects it and nothing
              ;; more:
              (progn
                ;; Select the merely-displayed group that was clicked on
                (gene-sgroup-select-group disp-group-here)
                ;; And go to it, now that it's selected
                (gene-sgroup-point-to-group gene-sgroup-selected-group))
            
            ;; Else we drag to include or exclude sequences from the
            ;; selected group.  We decide whether we're adding or deleting
            ;; group members by the location of the first click: if it's in
            ;; the selected group, then we're deleting; else we're adding. 
            
            ;; Do some housekeeping:
            (select-window (gene-id-buffer-window))
            (set-buffer gene-id-buffer) ; redundant but harmless
            
            (overlay-put olay 'priority 7) ; 7 is > nil, right?
            
            ;; Get overlay's face background right:
            (if sel-group-here
                ;; If the selected group is under the click, then our job is
                ;; to unselect anything we happen to get dragged over. 
                ;; Sequences that aren't selected in the first place are
                ;; unaffected by this.
                (progn
                  (overlay-put
                   olay 'face 'gene-sgroup-background-face)
                  (overlay-put
                   olay 'mouse-face 'gene-sgroup-background-mouse-face))
              ;; Else there wasn't a selected sequence under the click, so
              ;; select it, and do the same to any unselected sequences we get
              ;; dragged over.
              (overlay-put olay 'face (gene-sgroup-get-face
                                       gene-sgroup-selected-group))
              (overlay-put olay 'mouse-face (gene-sgroup-get-mouse-face
                                             gene-sgroup-selected-group)))
            
            ;; While we're actually tracking the mouse, we'll just stretch
            ;; the overlay so it appears as though the sequences are getting
            ;; selected.  In reality, they won't actually be added to the
            ;; selected group until after the drag is over.
            
            (track-mouse
              (while (progn
                       ;; Gulp up start-event, then use read-event
                       ;; ever after:
                       (setq end-event (if start-event
                                           (prog1
                                               start-event
                                             (setq start-event nil))
                                         (read-event)))
                       
                       (and
                        (consp end-event)
                        (memq
                         (car end-event)
                         '(down-mouse-1 mouse-movement switch-frame))))
                
                (setq end-posn (event-end end-event) ; a posn thingy
                      end-pt   (posn-point end-posn))
                
                (cond
                 
                 ;; Just sitting on one spot, so do nothing
                 ((and (numberp end-pt) (= start-pt end-pt))
                  nil)
                 
                 ;; Ignore any movement outside the frame
                 ((eq (car-safe end-event) 'switch-frame)
                  (gene-handle-vertical-motion t t))
                 
                 ;; On the mode line, so scroll buffer up
                 ((eq end-pt 'mode-line)
                  (gene-seq-scroll-buffer-internal 'gene-mouse-scroll-subr
                                                   half-w-height
                                                   olay
                                                   start-pt))
                 
                 ((eq end-pt 'vertical-line)
                  ;; Do nothing much if on the window divider
                  (gene-handle-vertical-motion t t))

                 ((not (eq (posn-window end-posn) start-window))
                  ;; Different window, same frame
                  (let ((end-row (cdr (cdr (mouse-position)))))
                    (cond
                     ((and end-row
                           (not (bobp))
                           (< end-row top))
                      (gene-seq-scroll-buffer-internal 'gene-mouse-scroll-subr
                                                       (- half-w-height)
                                                       olay
                                                       start-pt))
                     ((and end-row
                           (not (eobp))
                           (>= end-row bottom))
                      (gene-seq-scroll-buffer-internal 'gene-mouse-scroll-subr
                                                       half-w-height
                                                       olay
                                                       start-pt))
                     (t
                      (gene-handle-vertical-motion t t)))))
                 
                 (t
                  ;; Else we're still in the same window, ahh, relief.
                  (gene-handle-vertical-motion t t)
		  (save-excursion
		    (if (< start-pt end-pt)
			(progn
			  (goto-char end-pt)
			  (forward-line 1)
			  (move-overlay olay start-bol (point)))
		      (progn
			(goto-char end-pt)
			(beginning-of-line)
			(move-overlay olay start-eol (point)))))))))
            
            ;; Look at the overlay to determine which sequences to add or
            ;; delete from the selected group:
            (let* ((gc-cons-threshold gene-really-high-number)
                   (start (overlay-start olay))
                   (end   (1- (overlay-end olay)))
                   (func  (if sel-group-here
                              'gene-sgroup-delete-seq-from-group
                            
                            ;; Else if we're adding seqs to the group,
                            ;; then we need to record which one was the
                            ;; first one added, so we can go to that
                            ;; organism after all is said and done.
                            (lambda (sgroup seq)
                              (if first-seq-added
                                  (gene-sgroup-add-seq-to-group sgroup seq)
                                (setq first-seq-added seq)
                                (gene-sgroup-add-seq-to-group sgroup seq))))))
              
              (gene-sgroup-range-with-group
               func gene-sgroup-selected-group start end))
            
            ;; ... and then delete the overlay, now that it's no longer
            ;; needed:
            (delete-overlay olay))
          
          ;; Heavy-handed kluge to put us back where we belong:
          (select-window (gene-seq-buffer-window t))
          (set-buffer gene-seq-buffer) ; redundant, probably
          (goto-char seq-opoint)
          (if first-seq-added
              (gene-seq-goto-organism first-seq-added))
          (gene-handle-vertical-motion))
    
      ;; Handle clean-up in case of error:
      (error
       (set-buffer gene-id-buffer)
       (delete-overlay olay)
       (select-window (gene-seq-buffer-window t))
       (set-buffer gene-seq-buffer)
       (gene-handle-vertical-motion t)
       (setq gene-error-data the-error)
       (error
        "Error encountered in group drag... cleaning up after myself.")))))
    
  
(defvar gene-sgroup-follow-selected-group t
  "*Non-nil means always move point into selected group.")



(defun gene-sgroup-canonicalize-group (name &optional create)
  "Return the GROUP named NAME, error if none.
If NAME appears to be a sequence group, then just return it.
Optional arg CREATE means create a group named NAME if one doesn't
exist, adding numeric suffices if necessary."
  (if (stringp name)
      (setq name
            (or
             (gene-sgroup-get-group-by-name name)
             (and
              create
              (gene-sgroup-make-group-from-name-unconditionally name)))))
  (unless name (gene-error "Not a valid group: %S" name))
  name)


;; For the results of searching functions.
(defun gene-sgroup-make-group-from-name-unconditionally (name)
  "Make an sgroup whose name starts with the string NAME.
\(Add numbered suffixes until finding a free name, basically\)."
  (if (not (gene-sgroup-get-group-by-name name))
      (gene-sgroup-make-new-group name)
    (let ((count 1))
      (while (gene-sgroup-get-group-by-name
              (format "%s (%d)" name (setq count (1+ count)))))
      (gene-sgroup-make-new-group (format "%s (%d)" name count)))))


(defun gene-sgroup-make-group-from-sequences (sequences sgroup)
  "Put all SEQUENCES into SGROUP.
SEQUENCES is a list of sequences.
SGROUP is a group to use, or a group name.  If it's a name, the group
will be created if it doesn't already exist, and the new group's name
will at least contain the requested name as a prefix, if there are any
conflicts."
  (let ((group (gene-sgroup-canonicalize-group sgroup)))
    (mapcar
     (lambda (sym)
       (gene-sgroup-add-seq-to-group group sym))
     sequences)))


(defun gene-sgroup-all-sequences ()
  "Create a group containing all the sequences in the editor.
The group's name will be \"ALL\", or at least will start with
\"ALL\"."
  (interactive)
  (let ((seqs (gene-list-sequences)))
    (gene-sgroup-make-group-from-sequences seqs "ALL")))


(defun gene-sgroup-reinstate-seq-into-groups (sym)
  "If SYM thinks it's in some groups, make sure it's right.
For use mainly by cut/copy/paste."
  (mapcar
   (lambda (sgroup)
     ;; If there is a group with the same name as an old
     ;; group SEQ thinks it's in, then add SEQ to the new
     ;; group.  The "new" group might well be the same as
     ;; the old one, or it might not; the point is to
     ;; associate by name not by actual group objects.
     (let ((real-group (gene-sgroup-get-group-by-name
                        (gene-sgroup-get-name sgroup))))
       (if real-group
           (progn
             ;; There is a group with this name.
             ;; First, just remove the possibly obsolete
             ;; "ghost" group from SEQ:
             (gene-sgroup-delete-seq-from-group sgroup sym)
             ;; Then add SEQ to the real group, which might
             ;; be the same as the old one.  Hide it while
             ;; doing this, if necessary, so that everything
             ;; is displayed correctly after the
             ;; restoration:
             (let ((displayed-p (gene-sgroup-get-displayed-p
                                 real-group)))
               (if displayed-p
                   (gene-sgroup-hide-group real-group))
               (gene-sgroup-add-seq-to-group real-group sym)
               (if displayed-p
                   (gene-sgroup-display-group real-group)))
             )
         ;; Else disabuse this sequence of the notion that it
         ;; has anything to do with this ghost group anymore.
         (gene-sgroup-delete-seq-from-group sgroup sym))))
   (gene-sgroup-groups-containing-seq sym)))



;;; Group Frame stuff

(defconst gene-sgroup-frame-mode-map (make-sparse-keymap))
(define-key gene-sgroup-frame-mode-map [mouse-1]
  'gene-absorb-mouse-event)
(define-key gene-sgroup-frame-mode-map [down-mouse-1]
  'gene-sgroup-frame-dispatch-click)
(define-key gene-sgroup-frame-mode-map [double-down-mouse-1]
  'gene-absorb-mouse-event)
(define-key gene-sgroup-frame-mode-map [triple-down-mouse-1]
  'gene-absorb-mouse-event)
(define-key gene-sgroup-frame-mode-map [double-mouse-1]
  'gene-absorb-mouse-event)
(define-key gene-sgroup-frame-mode-map [triple-mouse-1]
  'gene-absorb-mouse-event)

(define-key gene-sgroup-frame-mode-map [double-mouse-2]
  'gene-absorb-mouse-event)
(define-key gene-sgroup-frame-mode-map [triple-mouse-2]
  'gene-absorb-mouse-event)
(define-key gene-sgroup-frame-mode-map [mouse-2]
  'gene-absorb-mouse-event)
(define-key gene-sgroup-frame-mode-map [drag-mouse-2]
  'gene-absorb-mouse-event)
(define-key gene-sgroup-frame-mode-map [down-mouse-2]
  'gene-absorb-mouse-event)
(define-key gene-sgroup-frame-mode-map [double-down-mouse-2]
  'gene-absorb-mouse-event)
(define-key gene-sgroup-frame-mode-map [triple-down-mouse-2]
  'gene-absorb-mouse-event)

(define-key gene-sgroup-frame-mode-map [double-mouse-3]
  'gene-absorb-mouse-event)
(define-key gene-sgroup-frame-mode-map [triple-mouse-3]
  'gene-absorb-mouse-event)
(define-key gene-sgroup-frame-mode-map [mouse-3]
  'gene-absorb-mouse-event)
(define-key gene-sgroup-frame-mode-map [drag-mouse-3]
  'gene-absorb-mouse-event)
(define-key gene-sgroup-frame-mode-map [double-down-mouse-3]
  'gene-absorb-mouse-event)
(define-key gene-sgroup-frame-mode-map [triple-down-mouse-3]
  'gene-absorb-mouse-event)


;; I wonder if this is working.  Hmm.  Well, at least it doesn't seem
;; to be hurting anything.
;; Don't let outside Emacs bindings show through:
;; (nconc
;;  gene-sgroup-frame-mode-map
;;  (list
;;   (cons t
;;         (lambda ()
;;           (interactive)
;;           (gene-error
;;        "%s has been unbound here.  Please check the number and dial again."
;;            (this-command-keys)
;;            )))))

(defvar gene-sgroup-frame nil
  "The Group Frame.")

(defvar gene-sgroup-frame-buffer (get-buffer-create "Groups")
  "Return the buffer displayed in the Group Frame.")


(defun gene-sgroup-frame-show-frame ()
  "Show the Group Frame.  This may include de-iconifying and raising it."
  (interactive)
  ;; If the frame is dead or never existed, then (re)suscitate it:
  (or
   (frame-live-p gene-sgroup-frame)
   (save-excursion
     (set-buffer gene-sgroup-frame-buffer)
     (setq gene-sgroup-frame
           (gene-frame-show
            gene-sgroup-frame-buffer
            `((width            . 70)
              (height           . 15)
              (name             . "Groups")
              (menu-bar-lines   . 1)
              (unsplittable     . t)
              (background-color . ,gene-background-color)
              (foreground-color . ,gene-foreground-color)
              (cursor-color     . ,gene-cursor-color)
              )))
     (gene-sgroup-frame-mode)
     (set-window-dedicated-p
      (frame-selected-window gene-sgroup-frame) t)
     ;; A save-frame-excursion would be cute...
     (let ((oframe (selected-frame)))
       (select-frame gene-sgroup-frame)
       (toggle-scroll-bar 1)
       (select-frame oframe))))
  
  (make-frame-visible gene-sgroup-frame)
  ;; Make sure it's up-to-date
  (gene-sgroup-frame-update)
  (raise-frame gene-sgroup-frame))


(defun gene-sgroup-frame-hide-frame ()
  "Hide the Group Frame."
  (interactive)
  (make-frame-invisible gene-sgroup-frame))


(defun gene-sgroup-frame-update (&optional sgroup)
  "Update the Group Frame with SGROUP.
If SGROUP is nil, then a full update will be done.  This is,
therefore, the function that displays the Group Frame in the first
place, when there are no groups."
  (when (and (frame-live-p gene-sgroup-frame)
             (frame-visible-p gene-sgroup-frame))

    ;; Remove this when the THEN-clause below is working!
    (setq sgroup nil)
    ;; Hmm, things seem fast enough this way that maybe we'll just never
    ;; remove it.  Hee hee.  Think of all those poor callers, imagining
    ;; that they're getting some careful update by passing an sgroup to
    ;; this function, when they could have just saved themselves the
    ;; trouble.  Ahh, Lisp.  Bring me another martini, Jenks, would you
    ;; be so kind?
    
    (if sgroup
        (save-excursion
          ;; do an amazingly efficient single-group update... someday.
          )
      
      ;; Else do a full update.  This is always a "superset" of a
      ;; single-group update; in other words, it's always okay to do a
      ;; full update instead of a single update, just slower.
      (save-excursion
        (set-buffer gene-sgroup-frame-buffer)
        (let (buffer-read-only
              opoint)
          (delete-region (point-min) (point-max))
          
          ;; Insert the groups and their standard command buttons:
          (mapcar
           (lambda (grp)
             (gene-sgroup-frame-insert-buttons grp)
             (insert "\n"))
           gene-sgroup-list)
          
          (setq opoint (progn (beginning-of-line) ; paranoia
                              (point)))
          
          ;; Insert the Make-New-Group button here
          (insert "\n")
          (insert "<Create New Group>")
          (put-text-property opoint
                             (progn (end-of-line) (point))
                             'func
                             'gene-sgroup-make-new-group)
          (put-text-property opoint
                             (progn (end-of-line) (point)) ; a no-op now
                             'mouse-face
                             'gene-mouse-face)
          
          ;; Insert the All Sequences button here:
          (insert "    ")
          (setq opoint (point))
          (insert "<All Sequences>")
          (put-text-property opoint
                             (progn (end-of-line) (point))
                             'func
                             'gene-sgroup-all-sequences)
          (put-text-property opoint
                             (progn (end-of-line) (point))
                             'mouse-face
                             'gene-mouse-face)
          
          ;; resize the frame to be as large as needed
          (gene-sgroup-frame-resizify)
          )))))

  
(defun gene-sgroup-frame-resizify ()
  "Resize the Group Frame to fit nicely the number of groups in it.
Unimplemented right now, though I believe it does have some callers."
  ;; Make the frame fit the groups.
  ;; This does nothing right now, since we have scrollbars and the
  ;; window can be resized anyway.  In the future, it may do
  ;; something, or maybe not.
  )


;; The groups are always displayed at the right, becauce their names
;; are of varying length and we'd like to avoid calculating the
;; padding for them.  We can control the length of buttons and
;; therefore precalculate their padding, however.
(defun gene-sgroup-frame-insert-buttons (grp)
  "Insert the correct buttons for GRP in the Group Frame, at point.
This function must be called from within the group buffer."
  (let (buffer-read-only
        (start-pt (point))
        (small-spacer " ") ; goes between some buttons
        (bigger-spacer "  ") ; goes between other buttons
        (face (if (gene-sgroup-get-displayed-p grp)
                  (gene-sgroup-get-face grp)
                ;; now: some var orig-face???
                'default))
        (mouse-face (if (gene-sgroup-get-displayed-p grp)
                        (gene-sgroup-get-mouse-face grp)
                      'gene-mouse-face)))
    
    ;; Select-group button:
    (insert "<select>")
    (put-text-property start-pt (point) 'func 'gene-sgroup-select-group)
    (put-text-property start-pt (point) 'group grp)
    (put-text-property start-pt (point) 'face face)
    (put-text-property start-pt (point) 'mouse-face mouse-face)
    (insert bigger-spacer)
    (setq start-pt (point))

    ;; Rename-group button:
    (insert "<rename>")
    (put-text-property start-pt (point) 'func 'gene-sgroup-rename-group)
    (put-text-property start-pt (point) 'group grp)
    (put-text-property start-pt (point) 'face face)
    (put-text-property start-pt (point) 'mouse-face mouse-face)
    (insert small-spacer)
    (setq start-pt (point))
    
    ;; Delete-group button:
    (insert "<cancel>")
    (put-text-property start-pt (point) 'func 'gene-sgroup-delete-group)
    (put-text-property start-pt (point) 'group grp)
    (put-text-property start-pt (point) 'face face)
    (put-text-property start-pt (point) 'mouse-face mouse-face)
    (insert small-spacer)
    (setq start-pt (point))

    ;; Toggle-visibility button:
    (insert (if (gene-sgroup-get-displayed-p grp) "<hide>" "<show>"))
    (put-text-property start-pt (point) 'func 'gene-sgroup-toggle-visibility)
    (put-text-property start-pt (point) 'group grp)
    (put-text-property start-pt (point) 'face face)
    (put-text-property start-pt (point) 'mouse-face mouse-face)
    (insert bigger-spacer)
    (setq start-pt (point))
    
    ;; The group itself:
    (insert (gene-sgroup-get-name grp))
    (put-text-property start-pt (point) 'func 'gene-sgroup-select-group)
    (put-text-property start-pt (point) 'group grp)
    (put-text-property start-pt (point) 'face face)
    (put-text-property start-pt (point) 'mouse-face mouse-face)))


(defun gene-sgroup-frame-mode ()
  "Major mode for the Group Frame buffer."
  (interactive)
  (setq mark-active nil)
  (transient-mark-mode -1)
  (toggle-read-only 1)

  (setq mode-line-format "Sequence Groups")

  (put 'gene-sgroup-frame-mode 'mode-class 'special)
  (setq major-mode 'gene-sgroup-frame-list-mode)
  (setq mode-name "Group List")
  (use-local-map gene-sgroup-frame-mode-map)
  (gene-menu-bar-make-hide-button
   'gene-sgroup-frame-hide-frame))


(defun gene-sgroup-frame-dispatch-click (event)
  "Handle a mouse click in the Group Frame."
  (interactive "e")
  (gene-switch-to-click event)
  (let ((pt (posn-point (event-start event))))
    ;; Possible commands are gene-toggle-group-visible, gene-select-group,
    ;; and gene-make-new-group.
    (let* ((command (get-text-property (point) 'func))
           (grp (get-text-property (point) 'group)))
      (if command
          (if grp
              (funcall command grp)
            (funcall command))))))


(defun gene-group-menu (event)
  "Popup a sequence group menu."
  (interactive "e")
  (gene-id-menu event)) ;;; ;cheap
                        ;;;
                        ;;; This comment was too expensive, so I
                        ;;; commented it out for now.
                        ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun gene-id-menu (evt)
  "Popup a menu for the id buffer... someday."
  (interactive "e")
  (message "Not yet implemented."))


;;; Operations done on Groups but not about Groups:
(defun gene-sgroup-upcase-group (sgroup)
  "Convert the sequences in SGROUP to upper-case."
  (interactive (list (gene-sgroup-selected-group)))
  (mapcar
   (lambda (seg)
     (gene-seq-upcase-sequence (car seg)))
   (gene-sgroup-seglist sgroup)))


(defun gene-sgroup-downcase-group (sgroup)
  "Convert the sequences in SGROUP to lower-case."
  (interactive (list (gene-sgroup-selected-group)))
  (mapcar
   (lambda (seg)
     (gene-seq-downcase-sequence (car seg)))
   (gene-sgroup-seglist sgroup)))


(defun gene-sgroup-flipcase-group (sgroup)
  "Toggle the case of all sequences in SGROUP."
  (interactive (list (gene-sgroup-selected-group)))
  (mapcar
   (lambda (seg)
     (gene-seq-flipcase-sequence (car seg)))
   (gene-sgroup-seglist sgroup)))


(defun gene-sgroup-subst-char-in-group (sgroup &optional fromchar tochar)
  "In SGROUP, substitute all occurences of FROMCHAR with TOCHAR.
If FROMCHAR and/or TOCHAR is nil or omitted, then the user will be
prompted interactively for them."
  (interactive (list (gene-sgroup-selected-here)))
  ;; We decided not to check locking for this too.
  ;; (gene-locking-sgroup-check-state selection gene-locking-unlocked-state)
  (or
   fromchar
   (setq fromchar (gene-read-char "Char to replace: " t)))
  (or
   tochar
   (setq tochar (gene-read-char (format "Replace \"%c\" with: " fromchar) t)))
  (message "Replacing \"%c\" with \"%c\"..." fromchar tochar)
  (mapcar
   (lambda (seg)
     (gene-seq-subst-char-in-sequence (car seg) fromchar tochar t))
   (gene-sgroup-seglist sgroup))
  (message "Replacing \"%c\" with \"%c\"... done." fromchar tochar))



;;; Saving/Writing groups.

(defun gene-sgroup-write-sgroup-to-tabl-file (sgroup file)
  "Write SGROUP to tabl files with prefix FILE.
In other words, it will create FILE.ids FILE.anns, and FILE.seqs.
If they exist, they will be overwritten.

This may be used as input to filter programs, which convert the
sequences into a form suitable for distance matrix and treeing
programs, for example."
  (gene-io-write-sequences-to-tabl-files
   (gene-sgroup-sequences sgroup) file))


(defun gene-sgroup-write-sgroup-to-phylip-file (sgroup &optional file)
  "Write SGROUP to a phylip FILE.
If FILE is nil or omitted, then the user will be prompted for its name."
  (setq file (expand-file-name
              (or file (read-file-name "Write to phylip file: "))))
  (let ((tabl-files (gene-io-write-sequences-to-tabl-files
                     (gene-sgroup-sequences sgroup) file)))
    (apply
     'gene-io-tabl-to-phylip
     (append tabl-files (list file)))
    (mapcar 'delete-file tabl-files)))


(defun gene-sgroup-merge-group-to-file (sgroup &optional file)
  "Merge the sequences in SGROUP to a file."
  (interactive (list
                gene-sgroup-selected-group
                (read-file-name "Save group to file: ")))
  (let* ((name (gene-sgroup-get-name sgroup))
         (prompt-str (format "Save group %s to file: " name)))
    (unless file
      (setq file (expand-file-name (read-file-name prompt-str))))
    (gene-merge-sequences-to-file file (gene-sgroup-sequences sgroup))))
                        

(defun gene-sgroup-save-group-to-file (sgroup &optional file)
  "Save the sequences in SGROUP to a new file.
If second arg FILE is not provided, the user will be prompted."
  (interactive (list
                gene-sgroup-selected-group
                (read-file-name "Save group to file: ")))
  (let* ((name (gene-sgroup-get-name sgroup))
         (prompt-str (format "Save group %s to file: " name)))
    (unless file
      (setq file (expand-file-name (read-file-name prompt-str))))
    (gene-save-sequences-as file (gene-sgroup-sequences sgroup))))
                        


(provide 'gene-sgroup)
