;; gene-translate.el (part of ale).
;; Copyright (C) 1994  Ribosomal Database Project (RDP).
;;                     Karl Fogel <kfogel@darwin.life.uiuc.edu>,
;;                     
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;;
;; Search for instances of "todo:" to find notes about future plans.


(require 'gene-mode-tools)

;;; Code:

;;; Translating ribosomal RNA sequences to proteins, and displaying them.

;; This is the genetic code:

;; UUU = F  CUU = L  AUU = I  GUU = V    UCU = S  CCU = P  ACU = T  GCU = A
;; UUC = F  CUC = L  AUC = I  GUC = V    UCC = S  CCC = P  ACC = T  GCC = A
;; UUA = L  CUA = L  AUA = I  GUA = V    UCA = S  CCA = P  ACA = T  GCA = A
;; UUG = L  CUG = L  AUG = M  GUG = V    UCG = S  CCG = P  ACG = T  GCG = A

;; UAU = Y  CAU = H  AAU = N  GAU = D    UGU = C  CGU = R  AGU = S  GGU = G
;; UAC = Y  CAC = H  AAC = N  GAC = D    UGC = C  CGC = R  AGC = S  GGC = G
;; UAA = *  CAA = Q  AAA = K  GAA = E    UGA = *  CGA = R  AGA = R  GGA = G
;; UAG = *  CAG = Q  AAG = K  GAG = E    UGG = W  CGG = R  AGG = R  GGG = G

;; Note that "T" may be substituted for any "U" on the nucleotide
;; side.  Asterisk is the end-of-chain marker.

;;; The Plan:
;;
;; According to Carl, translations could be usefully displayed like
;; this:
;;
;; Take the original sequence.  It's sitting in the buffer -- we'll
;; just display the amino translations right beneath it.  We need to
;; show all three reading frames (i.e.: offset by 0, 1, and 2, since
;; these are triplets).  Thus,
;;
;; AAUCUGUUU-GAU-CCUGGCAGAGGCU-ACUGC...  /*  This is the sequence. */
;; N--L--F---D---P--G--R--G--Y---C--...  /*  First reading frame.  */
;; -I--C--L---I---L--A--E--A---T--.....  /*  Second reading frame. */
;; --S--V--*---S---W--Q--R--L---L......  /*  Third reading frame.  */
;;
;; Then, immediately below that, have the translations for the
;; "reverse complement" of the sequence, in all three reading frames
;; of course.  The reverse complement is just what you think it is:
;; the sequence backwards and "complemented" (so that A becomes U/T
;; and C becomes G).
;; 
;;; The Implementation:
;;
;; First insert the six new sequences, as all gaps.  Make sure they
;; have names that reflect the original sequence and the reading
;; frame.  Go over them, translating with
;; gene-translate-point-to-char, turning them into aminos.  Question:
;; how well does the editor deal with "unusual" sequences in the
;; buffer?  That is, ones that don't really have any file, or format,
;; or annotation associated with them?  I guess it should be okay; if
;; saving functions complain, we'll just modify them so that they
;; don't mind anymore.


;; There are more efficient ways of doing this, but who cares?
(defconst gene-trans-triplet-table
  (list
   ;; On each line/col, the pattern runs UCAG, with U translated to T.

   (list ; leftmost U
    ?T (list ?T (cons ?T ?F) (cons ?C ?F) (cons ?A ?L) (cons ?G ?L))   ; mid U
       (list ?C (cons ?T ?S) (cons ?C ?S) (cons ?A ?S) (cons ?G ?S))   ; mid C
       (list ?A (cons ?T ?Y) (cons ?C ?Y) (cons ?A ?*) (cons ?G ?*))   ; mid A
       (list ?G (cons ?T ?C) (cons ?C ?C) (cons ?A ?*) (cons ?G ?W)))  ; mid G
   (list ; leftmost C
    ?C (list ?T (cons ?T ?L) (cons ?C ?L) (cons ?A ?L) (cons ?G ?L))   ; mid U
       (list ?C (cons ?T ?P) (cons ?C ?P) (cons ?A ?P) (cons ?G ?P))   ; mid C
       (list ?A (cons ?T ?H) (cons ?C ?H) (cons ?A ?Q) (cons ?G ?Q))   ; mid A
       (list ?G (cons ?T ?R) (cons ?C ?R) (cons ?A ?R) (cons ?G ?R)))  ; mid G
   (list ; leftmost A
    ?A (list ?T (cons ?T ?I) (cons ?C ?I) (cons ?A ?I) (cons ?G ?M))   ; mid U
       (list ?C (cons ?T ?T) (cons ?C ?T) (cons ?A ?T) (cons ?G ?T))   ; mid C
       (list ?A (cons ?T ?N) (cons ?C ?N) (cons ?A ?K) (cons ?G ?K))   ; mid A
       (list ?G (cons ?T ?S) (cons ?C ?S) (cons ?A ?R) (cons ?G ?R)))  ; mid G
   (list ; leftmost G
    ?G (list ?T (cons ?T ?V) (cons ?C ?V) (cons ?A ?V) (cons ?G ?V))   ; mid U
       (list ?C (cons ?T ?A) (cons ?C ?A) (cons ?A ?A) (cons ?G ?A))   ; mid C
       (list ?A (cons ?T ?D) (cons ?C ?D) (cons ?A ?E) (cons ?G ?E))   ; mid A
       (list ?G (cons ?T ?G) (cons ?C ?G) (cons ?A ?G) (cons ?G ?G)))) ; mid G
  "Table for translating nucleotide triplets into amino acids.")


(defun gene-trans-triplet-to-amino (ch1 ch2 ch3)
  "Translate the triplet CH1 CH2 CH3 to an amino acid character.
Case-insensitive.  U or T may be used interchangably.
If the triplet does not code for anything, return `X'.
The input characters are three separate arguments, not a list."
  (or
   (gene-trans-triplet-to-amino-internal
    (min (upcase ch1) ?T) (min (upcase ch2) ?T) (min (upcase ch3) ?T))
   ?X))


(defun gene-trans-triplet-to-amino-internal (ch1 ch2 ch3)
  "Translate the triplet CH1 CH2 CH3 to an amino acid character.
Case-sensitive, and only handles T, not U.
Returns nil if no such triplet code.
You probably don't want to use this function.  Take a look at
`gene-trans-triplet-to-amino' instead."
  (cdr (assoc
        ch3
        (cdr (assoc
              ch2
              (cdr (assoc ch1 gene-trans-triplet-table)))))))



(defun gene-translate-sequence-to-amino (sym)
  "Translate sequence SYM to amino acid.
Place results in three dummy sequences right below sequence SYM."
  (interactive (list (gene-get-here-sym)))
  (let* ((symlist (progn
                    (message "Creating amino acids...")
                    (gene-trans-open-blank-aminos sym)))
         (sym1    (nth 0 symlist))   ; for original, reading-frame 1
         (sym2    (nth 1 symlist))   ; for original, reading-frame 2
         (sym3    (nth 2 symlist))   ; for original, reading-frame 3
         (sym4    (nth 3 symlist))   ; for rev comp, reading-frame 1
         (sym5    (nth 4 symlist))   ; for rev comp, reading-frame 2
         (sym6    (nth 5 symlist)))  ; for rev comp, reading-frame 3

    ;; do the sequence
    (message "Translating in reading frame 1...")
    (gene-translate-sequence-to-amino-in-frame sym sym1 1)
    (message "Translating in reading frame 2...")
    (gene-translate-sequence-to-amino-in-frame sym sym2 2)
    (message "Translating in reading frame 3...")
    (gene-translate-sequence-to-amino-in-frame sym sym3 3)

    ;; do its reverse complement
    (message "Translating reverse complement in reading frame 1...")
    (gene-translate-reverse-complement-to-amino-in-frame sym sym4 1)
    (message "Translating reverse complement in reading frame 2...")
    (gene-translate-reverse-complement-to-amino-in-frame sym sym5 2)
    (message "Translating reverse complement in reading frame 3...")
    (gene-translate-reverse-complement-to-amino-in-frame sym sym6 3)

    (gene-arrange-windows)
    (message "Done translating.")
    ))


(defun gene-trans-open-derived-sequence-internal
  (sym name suffix length-or-data line ann file format lockstate)
  "Internal function for use by the `open-derived-sequence' family.
If LENGTH-OR-DATA is a negative number, then derive a length and
insert that many indels.  Else if a string, use it as sequence data;
else if nil, use SYM's sequence data."
  (or sym (setq sym (gene-get-here-sym)))
  (unless sym
    (gene-error "No sequence to derive from."))
  (gene-save-excursion-in gene-seq-buffer
    (gene-seq-goto-organism sym nil t)
    (let ((name   (concat (or name (gene-organism-name sym)) (or suffix "#")))
          (data   (if (numberp length-or-data)
                      ;; Make a gappy string
                      (make-string
                       (if (< length-or-data 0)
                           (- (gene-eol-point) (gene-bol-point))
                         length-or-data)
                       gene-default-gap)
                    ;; Else init to SYM's sequence data
                    (gene-organism-seq-data sym)))
          (line   (or line (1+ (gene-current-line))))
          (ann    (or ann (gene-io-make-annotation name 'genbank))))
      (prog1
          ;; We must return the sym:
          (gene-open-sequence
           name data ann file format lockstate line)
        (gene-handle-vertical-motion)
        (gene-arrange-windows)))))
  

(defun gene-trans-open-derived-sequence
  (&optional sym name suffix line ann file format lockstate)
  "Open a blank \(i.e.: all-indel\) sequence derived from sequence SYM.
SYM defaults to the organism at point.
The new sequence will be adjacent to SYM, and will have the same data.
Return the new organism symbol.
Other optional arguments are:
NAME, SUFFIX, LINE, ANN, FILE, FORMAT, and LOCKSTATE.
NAME defaults to SYM's name, SUFFIX to \"#\".  The new sequence's name
will be NAMESUFFIX.
LENGTH is the length of the new sequence; it defaults to the length of
SYM's sequence data.
LINE defaults to SYM's line.
ANN, FILE, and FORMAT are made up in the obvious ways if not provided.
LOCKSTATE defaults to GapInOut, as usual."
  (gene-trans-open-derived-sequence-internal
   sym name suffix nil line ann file format lockstate))


(defun gene-trans-open-blank-derived-sequence 
  (&optional sym name suffix length line ann file format lockstate)
  "Open a blank \(i.e.: all-indel\) sequence derived from sequence SYM.
SYM defaults to the organism at point.
The new sequence will be adjacent to SYM, and will consist of indels.
Return the new organism symbol.
Other optional arguments are:
NAME, SUFFIX, LENGTH, LINE, ANN, FILE, FORMAT, and LOCKSTATE.
NAME defaults to SYM's name, SUFFIX to \"#\".  The new sequence's name
will be NAMESUFFIX.
LENGTH is the length of the new sequence; it defaults to the length of
SYM's sequence data.
LINE defaults to SYM's line.
ANN, FILE, and FORMAT are made up in the obvious ways if not provided.
LOCKSTATE defaults to GapInOut, as usual."
  (gene-trans-open-derived-sequence-internal
   sym name suffix (or length -1) line ann file format lockstate))



(defun gene-trans-open-blank-aminos (&optional sym)
  "Open blank lines to translate SYM's sequence to amino-acid.
SYM defaults to the organism of the current line.
Returns the list of org syms, from top to bottom."
  (or sym (setq sym (gene-get-here-sym)))
  (gene-save-excursion-in
      gene-seq-buffer
    (gene-seq-goto-organism sym nil t)
    (let* ((len  (- (gene-eol-point) (gene-bol-point)))
           (line (gene-current-line))
           (sym1 (gene-trans-open-blank-derived-sequence
                  sym nil "+1" len (setq line (1+ line))))
           (sym2 (gene-trans-open-blank-derived-sequence
                  sym nil "+2" len (setq line (1+ line))))
           (sym3 (gene-trans-open-blank-derived-sequence
                  sym nil "+3" len (setq line (1+ line))))
           (sym4 (gene-trans-open-blank-derived-sequence
                  sym nil "-1" len (setq line (1+ line))))
           (sym5 (gene-trans-open-blank-derived-sequence
                  sym nil "-2" len (setq line (1+ line))))
           (sym6 (gene-trans-open-blank-derived-sequence
                  sym nil "-3" len (setq line (1+ line)))))
      (list sym1 sym2 sym3 sym4 sym5 sym6))))
  

(defun gene-translate-sequence-to-amino-in-frame (source dest frame)
  "Translate sequence SOURCE to amino sequence DEST in reading frame FRAME.
SOURCE and DEST are organism symbols.  SOURCE's text is not affected,
but DEST's will be overwritten with the new amino text \(it is assumed
that DEST is a line of indels, probably hyphens and very likely
produced by gene-trans-open-blank-aminos\).
FRAME is 1, 2, or 3."
  (gene-save-excursion-in
      gene-seq-buffer
    (gene-seq-goto-organism source 0 t)
    (if (re-search-forward gene-base-regexp (gene-eol-point) t frame)
        (let* ((buffer-undo-list t)
               buffer-read-only
               base1   ; first base in current triplet
               base2   ; second base in current triplet
               base3   ; home plate^H^H^H^H third base in current triplet
               (source-bol  (gene-bol-point))
               (source-eol  (gene-eol-point))
               (source-pnt  (progn (forward-char -1) ; be ON the base
                                   (point)))
               (source-opnt source-pnt)
               (source-col  (gene-current-column))
               (dest-pnt    (progn
                              (gene-seq-goto-organism dest source-col t)
                              (point)))
               (dest-opnt dest-pnt))
          (while
              (and
               (progn
                 (goto-char source-pnt) 
                 (skip-chars-forward gene-skip-gaps-regexp source-eol)
                 (setq base1 (if (eolp) nil (char-after (point)))))
               (progn
                 (forward-char 1)
                 (skip-chars-forward gene-skip-gaps-regexp source-eol)
                 (setq dest-pnt (+ dest-opnt (- (point) source-opnt)))
                 (setq base2 (if (eolp) nil (char-after (point)))))
               (progn
                 (forward-char 1)
                 (skip-chars-forward gene-skip-gaps-regexp source-eol)
                 (setq base3 (if (eolp) nil (char-after (point)))))
               (progn
                 (forward-char 1)
                 (setq source-pnt (point))
                 (gene-translate-point-to-char
                  dest-pnt
                  (gene-trans-triplet-to-amino base1 base2 base3))))))

      ;; Else the re-search-forward found no bases.
      (gene-error
       "Not enough residues for translation in %s."
       (gene-organism-name)))))


(defun gene-translate-reverse-complement-to-amino-in-frame
  (source dest frame)
  "Translate the reverse complement of sequence SOURCE to amino
sequence DEST in reading frame FRAME.
SOURCE and DEST are organism symbols.  SOURCE's text is not affected,
but DEST's will be overwritten with the new amino text \(it is assumed
that DEST is a line of indels, probably hyphens and very likely
produced by gene-trans-open-blank-aminos\).
FRAME is 1, 2, or 3."
  (gene-save-excursion-in
      gene-seq-buffer
    (gene-seq-goto-organism source 0 t)
    (end-of-line)
    (if (re-search-backward gene-base-regexp (gene-bol-point) t frame)
        (let* ((buffer-undo-list t)
               buffer-read-only
               base1   ; first base in current triplet
               base2   ; second base in current triplet
               base3   ; home plate^H^H^H^H third base in current triplet
               (source-eol  (gene-eol-point))
               (source-bol  (gene-bol-point))
               ;; compensate for the first re-search:
               (source-pnt  (1+ (point)))
               ;; compensate for above compensation:
               (source-opnt (1- source-pnt))
               (source-col  (1- (gene-current-column)))
               (dest-pnt    (progn
                              (gene-seq-goto-organism dest source-col t)
                              (point)))
               (dest-opnt dest-pnt))

          (while
              (and
               (progn
                 (goto-char source-pnt) 
                 
                 (let ((opt (point)))
                   (re-search-backward gene-base-regexp source-bol t)
                   (if (= opt (point)) ; we haven't moved
                       (setq base1 nil)
                     (setq base1 (if (bolp) nil (char-after (point)))))))
               (progn
                 (let ((opt (point)))
                   (re-search-backward gene-base-regexp source-bol t)
                   (if (= opt (point)) ; we haven't moved
                       (setq base2 nil)
                     ;; "-1" below compensates for oboe
                     (setq dest-pnt (- dest-opnt -1 (- source-opnt (point))))
                     (setq base2 (if (bolp) nil (char-after (point)))))))
               (progn
                 (let ((opt (point)))
                   (re-search-backward gene-base-regexp source-bol t)
                   (if (= opt (point)) ; we haven't moved
                       (setq base3 nil)
                     (setq base3 (if (bolp) nil (char-after (point)))))))
               (progn
                 (setq source-pnt (point))
                 (gene-translate-point-to-char
                  dest-pnt
                  (gene-trans-triplet-to-amino
                   (gene-complement-char base1)
                   (gene-complement-char base2)
                   (gene-complement-char base3)))
                 ))))
      
      ;; Else the re-search-forward found no bases.
      (gene-error
       "Not enough residues for translation in %s."
       (gene-organism-name)))))


;;;; Reversing, Complementing, and Reverse-Complementing
;;; (as wfisher wanted)

;;; reversing

(defun gene-trans-reverse-string (string)
  "Return a copy of STRING, backwards."
  (let ((len (length string))
        (idx 0))
    (if (<= len 0)
        string ; empty string, or worse...
      (let ((new-string (make-string len 0)))
        (setq len (1- len))
        (while (<= idx len)
          (aset new-string idx (aref string (- len idx)))
          (setq idx (1+ idx)))
        new-string))))


(defun gene-trans-open-reverse-sequence (sym &optional suffix)
  "Open a sequence which is the reverse of sequence SYM.
Return the new organism symbol."
  (gene-message "Reversing %s..." (gene-organism-name sym))
  (let* ((sufx (or suffix "<"))
         (nsym (gene-trans-open-blank-derived-sequence sym nil sufx)))
    (gene-save-excursion-in gene-seq-buffer
      (let* ((o-data (gene-organism-seq-data-unpadded sym))
             (n-data (gene-trans-reverse-string o-data))
             (buffer-undo-list t))
        (gene-seq-goto-organism sym 0)
        (skip-chars-forward gene-skip-gaps-regexp (gene-eol-point))
        (let (buffer-read-only
              (col (gene-current-column)))
          (gene-seq-goto-organism nsym col)
          (gene-pave-with-string n-data)))
      (gene-message "Reversing %s... done." (gene-organism-name sym))
      nsym)))


;;; complementing
;; todo: what about the fact that we default to U, not T?

(defun gene-trans-complementify-sequence (sym)
  "Translate sequence SYM to its complement."
  (gene-message "Complementifying %s..." (gene-organism-name sym))
  (gene-save-excursion-in gene-seq-buffer
    (gene-seq-goto-organism sym 0)
    (let (buffer-read-only
          (eol (gene-eol-point))
          (pt  (point)))
    (while (<= pt eol)
      (gene-translate-point-to-char
       pt (gene-complement-char (char-after pt)))
      (setq pt (1+ pt)))))
  (gene-message "Complementifying %s... done." (gene-organism-name sym)))


(defun gene-trans-open-complement-sequence (sym)
  "Open a derived sequence near SYM that will be SYM's complement."
  (let ((nsym (gene-trans-open-derived-sequence sym nil "!"))
        (buffer-undo-list t))
    (gene-trans-complementify-sequence nsym)))


(defun gene-trans-open-reverse-complement-sequence (sym)
  "Open a derived sequence near SYM that will be SYM's reverse complement."
  (let ((nsym (gene-trans-open-reverse-sequence sym "!<"))
        (buffer-undo-list t))
  (gene-trans-complementify-sequence nsym)))



(provide 'gene-translate)
