;; gene-undo.el (part of ale).
;; Copyright (C) 1994  Karl Fogel <kfogel@darwin.life.uiuc.edu>
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;;
;; Search for instances of "todo:" to find notes about future plans.


;;;; Things in ale that need special undo:
;; - selection creation, dragging, case-fiddling
;; - sequence case-fiddling
;; - group case-fiddling and char-substituting
;; - cut-&-paste !!!
;; - what else?


(require 'gene-mode-tools)


;; I sent in a patch that implements custom undo records.  However, at
;; this time, I don't know if the patch will be accepted for the
;; distribution.  Therefore, I'll implement the patch in Lisp here,
;; and if it's accepted, what's below can be changed to a simple call
;; to native undo.

(defvar gene-undo-message nil
  "Undo function may set this to a string for control over undo messages.
`gene-undo' will check if it's a string, display it in place of the
standard undo message if so, and then set it to nil.")


(defun gene-undo ()
  "Make the last change not have happened.
Time will run backwards -- briefly."
  (interactive)
  ;; We don't want to use a save-excursion, because point has to move
  ;; to where the undo happened.
  (set-buffer gene-seq-buffer)

  (gene-message "Undoing...")

  (let ((max-lisp-eval-depth gene-really-high-number))
    
    (while (and
            buffer-undo-list
            (eq (car buffer-undo-list) nil))
      (setq buffer-undo-list (cdr buffer-undo-list)))
    
    (if buffer-undo-list
        (progn
          (let (buffer-read-only
                (max-specpdl-size 1048577)
                (this-undo-record (car buffer-undo-list)))
            
            (if (not (eq (car this-undo-record) 'gene-undo))
                (let ((lst buffer-undo-list))
                  (let ((buffer-undo-list t))
                    (setq lst (primitive-undo 1 lst)))
                  (setq buffer-undo-list lst)
                  (gene-message "Undoing... done."))
              ;; Else we're looking at a custom record:
              (let ((form (cdr this-undo-record)))
                (cond
                 ((eq (car form) 'gene-unmodified-marker)
                  (ding)
                  (if gene-saved-flag
                      (gene-mark-unmodified))
                  (gene-message "No more changes to undo."))
                 ((eq (car form) 'gene-saved-marker)
                  (ding)
                  ;; this flag non-nil means most recent save has not
                  ;; yet been reached in undo.
                  (if gene-saved-flag
                      (gene-mark-unmodified))
                  (setq gene-saved-flag nil)
                  (gene-message "No changes since last save.")
                  (setq buffer-undo-list (cdr buffer-undo-list)))

                 ;; Else it's a real custom record, not a no-changes
                 ;; marker, so run the record, and don't record undo
                 ;; while running it!
                 (t
                  (let ((buffer-undo-list t))
                    (apply 'funcall form))
                  (setq buffer-undo-list (cdr buffer-undo-list))
                  (if (car buffer-undo-list)
                      ;; We haven't reached an undo boundary yet
                      (gene-undo))
                  (if gene-lock-status-update-needed
                      (gene-status-buffer-update 1))
                  (if gene-undo-message
                      (progn
                        (gene-message gene-undo-message)
                        (setq gene-undo-message nil))
                    (gene-message "Undoing... done."))
                  ))))))
      
      ;; Else no more undo information.
      (progn
        (ding)
        (gene-message "No more undo information available.")
        ))
    (gene-handle-vertical-motion)))



(defun gene-record-unchanged-state ()
  "Record the unmodified state of the buffer.

The earliest undo event is the inverse of marking the buffer modified
after the first change.  That way, when the undo list reaches this
element, it marks the alignment as unmodified; however, if the undo
list ever gets truncated, this event is lost, so we won't be lying to
the user about the modification state of their alignment."
  (gene-undo-register-method 'gene-unmodified-marker)
  (undo-boundary))


(defvar gene-saved-flag t
  "Flag var telling undo if a save-event is the most recent one or not.")


(defun gene-record-saved-state ()
  "Record the unmodified state of the buffer after a save."
  (setq gene-saved-flag t)
  (gene-undo-register-method 'gene-saved-marker)
  (undo-boundary))


(defvar gene-modified-p nil
  "Non-nil means that the alignment buffer has been modified since the
last save.")


(defun gene-mark-modified ()
  "Set the buffer-modified-p and the modeline to indicate changes."
  (gene-save-excursion-in gene-seq-buffer
    (set-buffer-modified-p t)
    (setq gene-modified-p t)
    (force-mode-line-update)))


(defun gene-mark-unmodified ()
  "Set the alignment buffer's modeline to indicate no modifications.
This is meant to be put in `gene-save-hook'."
  (gene-save-excursion-in
      gene-seq-buffer
    (set-buffer-modified-p nil)
    (setq gene-modified-p nil)
    (force-mode-line-update))
  (gene-save-excursion-in
      gene-ann-buffer
    (set-buffer-modified-p nil)))


(defun gene-mark-save ()
  "Set the alignment buffer's modeline to indicate no modifications.
This is meant to be put in `gene-save-hook'."
  (gene-save-excursion-in
      gene-seq-buffer
    (gene-record-saved-state)
    (set-buffer-modified-p nil)
    (setq gene-modified-p nil)
    (force-mode-line-update))
  (gene-save-excursion-in
      gene-ann-buffer
    (set-buffer-modified-p nil)))


(defun gene-undo-register-method (&rest method)
  "Place a method for undoing a change into the buffer's undo list.
A method is a function followed by the arguments to pass to that
function.  For example:
   \(record-undo-method 'put-text-property 34 54 'face nil\)

or perhaps...

   \(record-undo-method \(lambda \(\) \(ding\) insert \"foo\"\)\)

This works by adding a new element to undo list.  The element looks
like `\(apply 'funcall \(func ...\)\).  See documentation for variable
`buffer-undo-list' for more details."
  (save-excursion
    (set-buffer gene-seq-buffer)
    (or
     (eq buffer-undo-list t)
     (setq buffer-undo-list
           (cons (cons 'gene-undo method) buffer-undo-list)))))


;; I sure hope this was worth it.  I think sequence data is generally
;; good for run-length encoding, but maybe should examine stats
;; someday.
(defun gene-undo-make-case-record (start end)
  "Returns a list indicating case of characters between START and END.
The list looks like:

         \(\(FLAG . NUMBER\) \(FLAG . NUMBER\) ...\)

and uses run-length encoding to record the case of characters in a region.
FLAG is t if all the chars in the run are upper-case, and NUMBER is
the length of the run, starting from the left.

START and END are not recorded anywhere in the result, although they do
affect the result, of course."
  (gene-save-excursion-in
      gene-seq-buffer
    (let ((flag (gene-upper-case-p (char-after (setq end (1- end)))))
          (run-length 1)
          lst)
      (while (> end start)
        ;; Ahhh.  These `if's could be nested either way, or a `cond'
        ;; used instead.  The `cond' might be more readable.  Think
        ;; about it.
        (if flag
            ;; We're in an upper-case run
            (if (gene-upper-case-p (char-after (setq end (1- end))))
                ;; And the next char was upper case too
                (setq run-length (1+ run-length))
              ;; else next char is lower case, and it broke the run
              (setq lst (cons (cons flag run-length) lst)
                    flag nil
                    run-length 1))
          ;; Else we're in a lower case run
          (if (gene-upper-case-p (char-after (setq end (1- end))))
              ;; Next char is upper case, and it broke the run
              (setq lst (cons (cons flag run-length) lst)
                    flag t
                    run-length 1)
            ;; else the next char was lower case too
            (setq run-length (1+ run-length)))))
      (cons (cons flag run-length) lst))))
  
  
(defun gene-undo-restore-region-case (start end record)
  "Restore the case from START to END according to RECORD.
START and END are positions in a buffer, RECORD is of the form
returned by `gene-undo-make-case-record'.

Do the restoration without disturbing any markers, text properties, or
point.  In other words, be as non-destructive as the original
case-change operation which necessitates this restore.

This function is normally only called by a special undo record."
  
  (set-buffer gene-seq-buffer)

  (while record
    (let* ((this-run (car record))
           (flag (car this-run))
           (run-end (+ start (cdr this-run))))
      (if (> run-end end) ; paranoia
          (gene-error "Bad region-case encoding!"))
      
      (if flag
          (upcase-region start run-end)
        (downcase-region start run-end))
      (setq start run-end
            record (cdr record))))
  
  (goto-char start))
  

;; "record" is a *verb* in the function name, a verb the first time in
;; the doc string, and a noun the second time in the doc string.  I
;; hate English.
(defun gene-undo-record-region-case (start end)
  "Record region START to END for case-change undo.
This conses a new element onto `buffer-undo-list'.  The element looks
like this:

        \(eval gene-undo-restore-region-case start end record...\)

The format of the record is the same as that returned by
`gene-undo-make-case-record'."
  (let ((record (gene-undo-make-case-record start end)))
    (gene-undo-register-method
     'gene-undo-restore-region-case start end record)))


(defun gene-undo-record-selection-case (selection)
  "Record SELECTION for case-change undo.
This affects buffer-undo-list; see function `gene-undo-register-method'."
  ;; this save-excursion not strictly necessary, but will make the one
  ;; called by `gene-undo-make-case-record' be a lot cheaper.
  (gene-save-excursion-in
      gene-seq-buffer
    (let* ((seg-list (gene-sel-get-seg-list selection)))
      (mapcar
       (lambda (seg)
         (let ((start (marker-position (car seg)))
               (end (marker-position (car (cdr seg)))))
           (gene-undo-register-method
            'gene-undo-restore-region-case
            start end (gene-undo-make-case-record start end))))
       seg-list))))


(defun gene-undo-restore-selection-drag (shifts)
  "Restore the positions recorded in SHIFTS.
SHIFTS is of the type returned by `gene-undo-record-selection-drag'."
  (set-buffer gene-seq-buffer)
  (mapcar
   (lambda (record)
     (apply 'gene-shift-region record))
   shifts)
  (let ((place-it-ended-up (car (car shifts)))
        (offset            (car (cdr (cdr (car shifts))))))
    (if place-it-ended-up
        (goto-char (+ place-it-ended-up offset)))))


(defun gene-undo-record-selection-drag (old-seg-list new-seg-list)
  "Record a selection's drag for later undo.
Takes as arguments OLD-SEG-LIST, the selection's segment list just
before the drag, and NEW-SEG-LIST, the same right after the drag.
When recording seg lists, it is best to obtain the old seg list with
gene-sel-generate-seg-list, not gene-sel-get-seg-list, to make sure
that none of the data in the lists is shared.

The record produced by this function looks like:

\(\(NEW-START NEW-END COUNT GAP\) ...\)

and is stored in a custom record in buffer-undo-list."

  (if (/= (length old-seg-list) (length new-seg-list)) ; impossible
      (gene-error ; paranoia
       "Selection undo error: seg lists' lengths don't match.")
    ;; Okay, we're safe.  Doesn't matter which one we test:
    (let (record)
      (while old-seg-list
        (let* ((old-seg     (car old-seg-list))
               (new-seg     (car new-seg-list))
               (old-start   (marker-position (nth 0 old-seg)))
               (new-start   (marker-position (nth 0 new-seg)))
               (old-end     (marker-position (nth 1 old-seg)))
               (new-end     (marker-position (nth 1 new-seg)))
               (old-lgap    (nth 4 old-seg))
               (old-rgap    (nth 5 old-seg))
               (count       (- old-start new-start))
               )
          
          ;; We must manufacture a bunch of gene-shift-region calls from
          ;; the above information.  Form a list of lists, each one
          ;; wants gene-shift-region applied to it.  
          
          (if (/= count (- old-end new-end)) ; also impossible
              ;; Avoid bad information
              (gene-error ; paranoia
               "Selection undo error: start and end counts don't match.") 

            ;; Else it's safe, so go ahead and record
            (setq record
                  (cons (list new-start
                              new-end
                              count
                              (if (>= count 0)
                                  old-lgap
                                old-rgap))
                        record))
            (setq
             old-seg-list (cdr old-seg-list)
             new-seg-list (cdr new-seg-list)))))

      (gene-undo-register-method 'gene-undo-restore-selection-drag record))))


;;; Recording selection justification for undo.
;;
;; The basic plan is to record a list that looks like this:
;;
;;    ((START END NUM-BASES NUM-GAPS NUM-BASES NUM-GAPS ...) ...)
;;
;; Each element of the list encodes one segment of the selection.
;; START and END mark the boundaries of the segment being undone.
;; Then come a series of numbers.  The first is the number of bases
;; (counting from the right) of the segment before the justification
;; took place.  The next number is the number of gaps following it,
;; etc.  They alternate until the end of the segment.  So for the
;; following selection:
;;
;; AAAA----AA-UU
;; GCCG----AAA--
;; -CCG----AAA--
;;
;; the encoding might be: 
;;
;; ((2311 2324 4 4 2 1 2)
;;  (3312 3325 4 4 3 2)
;;  (4313 4326 0 1 3 4 3 2)
;;
;; When the time comes to undo the change, we know that all the bases
;; are still in the selection, they've just been shifted around.  The
;; encoding is used to shift them back, by inserting and deleting gaps
;; at the right places.
;;
;; Now, this scheme has a flaw.  It doesn't record the *type* of the
;; gap characters -- they're all assumed to be "-", the default gap.
;; Not only does justification mutate gap characters (which might
;; itself be a bug), but the undo recording for justifications throws
;; away the information about what the actual characters were.  The
;; result is that there's no way to restore the original gap
;; characters, even though we can restore all their *positions*
;; correctly.
;;
;; The fix: each of the recorded gappy regions, "NUM-GAPS" in the
;; format above, can be a string.  If it is just a number, that is
;; equivalent to saying "NUM-GAPS worth of `-' were here".  However,
;; if it's a string, then it is an exact copy of the gappy region, and
;; can be inserted as-is.  The string's length speaks for itself, of
;; course.
;;
;; This will require two changes to the current undo-code (as of
;; gene-undo.el:1.24).  `gene-undo-record-selection-justification'
;; will have to record the new information, and
;; `gene-undo-unjustify-selection' will have to handle it.
;;

(defun gene-undo-record-selection-justification (seglist)
  "Record that the selection containing SEGLIST was justified.
This creates and undo record of the form

\(\(START END BASE-LENGTH GAP-LENGTH BASE-LENGTH GAP-LENGTH ...\) ...\)

and stores it in buffer-undo-list as a custom undo record.

START and END mark the boundaries of a selection segment.  Then the
number of bases in a row, starting from the left, then number of gaps
in a row, and so forth, until the right edge of that segment is
reached.  Then the next segment is done, and so on until the end of
the selection." 
  (message "Recording justification for undo...")
  (gene-save-excursion-in
      gene-seq-buffer
    (let* ((non-defgap-re
            (concat "[^" (char-to-string gene-default-gap) "]"))
           (record
            (mapcar
             (lambda (segment)
               (let* ((record-elt ())
                      (count      0) ; sometimes a string, see explanation
                      (gapping    nil)
                      (start      (marker-position (car segment)))
                      (end        (marker-position (car (cdr segment)))))
                 
                 (goto-char start)
                 
                 (while (< (point) end)
                   (setq count

                         (if gapping

                             ;; We're recording a gappy run.
                             (let ((opoint (point))
                                   (ocount
                                    (skip-chars-forward
                                     gene-skip-gaps-regexp end)))
                               ;; Determine if this gappy run is all
                               ;; default gap, or  if it's contaminated
                               ;; with other gaps.
                               (if (save-excursion
                                     (re-search-backward
                                      non-defgap-re opoint t))
                                   ;; There are gaps which aren't the
                                   ;; default gap in the run, so we
                                   ;; just record it as a string.
                                   (buffer-substring-no-properties opoint (point))
                                 ;; Else it's all default gaps, so
                                 ;; just record how many
                                 ocount))

                           ;; Else we're recording a base run.
                           (skip-chars-forward gene-skip-bases-regexp end)))

                   (setq record-elt (cons count record-elt)
                         gapping (not gapping)))
                 
                 (nconc (list start end) (nreverse record-elt))))
             seglist)))
      
      (gene-undo-register-method 'gene-undo-unjustify-selection record)))
  (message "Recording justification for undo... done."))


(defun gene-undo-unjustify-selection (record)
  "Undo the justification of a selection, as recorded by RECORD..
RECORD is of the form created by `gene-undo-record-selection-justification'
and stored in the undo list."
  (gene-save-excursion-in
      gene-seq-buffer
    (while record
      (let* ((record-elt (car record))
             (start      (car record-elt))
             ;; `end' has to be a marker so that it floats as we do
             ;; our insertions and deletions: 
             (end        (copy-marker (car (cdr record-elt))))
             (gapping    nil)
             all-gaps
             (run-list   (cdr (cdr record-elt)))
             (gap-insertion-count 0))

        (goto-char start)
        
        ;; If there's a run of gaps here, then this selection was
        ;; probably right-justified, or center-justified.  Delete the
        ;; run; we'll restore according to the recorded run-list only.
        (if (looking-at gene-gap-regexp)
            (let ((len (save-excursion
                         (skip-chars-forward gene-skip-gaps-regexp end)
                         )))
              (if (= (+ (point) len) end)
                  (setq all-gaps t)
                (delete-char len))))
        
        ;; This depends on the fact that the bases of each segment are
        ;; guaranteed to be contiguous after a justification.
        (if all-gaps
            nil ; The justification did nothing, so we won't either
          (while run-list
            (if gapping
                (progn
                  (insert-before-markers ?*)
                  (forward-char -1)
                  (if (stringp (car run-list))
                      (insert-and-inherit (car run-list))
                    (insert-char gene-default-gap (car run-list) t))
                  (delete-char 1))
              ;; Else this is a base length, so just skip forward.
              (forward-char (car run-list)))
            (setq run-list (cdr run-list)
                  gapping  (not gapping)))
        
          ;; We'd better be at the end now, so anything between here and
          ;; the recorded end is junk.  Delete it.
          (delete-region (point) end)))
    
      (setq record (cdr record)))))



;; Undo borrows clipboard code to implement sequence-deletion undo,
;; because in a way undoing that's just the same as doing a "paste".

(defconst gene-undo-sequence-clipboard
  (list
   (get-buffer-create "*Gene Undo ID Clipboard*")
   (get-buffer-create "*Gene Undo Ann Clipboard*")
   (get-buffer-create "*Gene Undo Seq Clipboard*")
   )
  "The undo clipboard buffers.")
(mapcar 'buffer-disable-undo gene-undo-sequence-clipboard)


(defun gene-undo-record-sequence-delete (sym &optional line)
  "Record the deletion of sequence SYM for undo.
IMPORTANT: make sure you call this before the sequence has actually
been deleted -- otherwise, it won't be able to find the data to
record."
  ;; Pass REVERSE-FLAG argument to `gene-cb-add-sequence-data' so
  ;; sequences are added at the beginning of the clipboard instead of
  ;; the end. 
  ;; This is because order is important in the undo clipboard: we have
  ;; to always find the most-recently-deleted sym first when searching
  ;; forward with property-searching primitives.
  ;; Other clipboards don't care about order, so we simply insert at
  ;; the end for cleanliness' sake.
  (gene-cb-add-sequence-data sym gene-undo-sequence-clipboard)
  (gene-undo-register-method
   'gene-cb-insert-then-remove-sequence
   sym
   gene-undo-sequence-clipboard
   (or line (gene-seq-line-number sym))))


(defun gene-undo-record-sequence-open (sym)
  "Record that sequence SYM was opened.
SYM defaults to the organism symbol at point.

Implementer's bow to insecurity: 
There is no need to record any of the initial sequence data.  By the
time we are undoing this, we know that the only remaining material
related to this sequence is that initial data -- thus, we can just
remove it all and wash our hands of the matter."
  (or sym (setq sym (gene-get-here-sym)))
  (gene-undo-register-method 'gene-seq-delete-sequence sym t))


        
(provide 'gene-undo)
